/* $Id: main.c,v 1.6 2002/10/02 03:18:08 cvs Exp $ */
/*-----------------------------------------------------------------------

	DST data generator...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include "main.h"

/* Module constants ---------------------------------------------------*/
#define TWO_PI	6.283185307

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL stop = FALSE;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL MakeFrame(DST_FRAME * frame, DATA * data);
static BOOL GenerateFunctions(MAIN_ARGS * args, DATA * data);
static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args);
static VOID CmdLineHelp(VOID);
static BOOL SetSignalHandler(VOID);
static VOID CatchSignal(int sig);

/*---------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	CHAR string[128];
	MAIN_ARGS args;
	DATA data;
	DST_FRAME frame;
	UINT8 buffer[DST_FRAME_SIZE];
	size_t n;
	USTIME time;

	printf("dstgen version %s (%s %s), pid:%d\n\n", VERSION, __DATE__, __TIME__, getpid());

	if (!SetSignalHandler())
		exit(1);

	if (!ParseCommandLine(argc, argv, &args))
		exit(1);

	if (!GenerateFunctions(&args, &data))
		exit(1);

	printf("Opening device: %s\n", AsyncFormatParams(string, 128, &args.dst_args.params));
	if ((args.dst_args.handle = AsyncOpen(&args.dst_args.params)) == ASYNC_INVALID) {
		printf("ERROR: AsyncOpen: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(args.dst_args.handle)));
		exit(1);
	}

	printf("Function: amplitude = %.0lf counts, frequency = %.2lf Hz\n", args.amplitude, args.frequency);
	printf("Sampling rate: %.2lf samples/second\n", args.rate);

	data.time = SystemUSTime() + UST_SECOND;
	data.time -= data.time % UST_SECOND;

	while (!stop) {
		/* Make up and encode a frame... */
		MakeFrame(&frame, &data);

		/* Wait until it's time to transmit */
		time = SystemUSTime();
		if (time < data.time)
			USSleep(data.time - time);
		data.time += data.period;

		if (data.time % UST_SECOND == 0) {
			printf("\r%s", FormatUSTime(string, 128, UST_DATE_TIME_MSEC, data.time));
			frame.aux = 0x80;
		}
		else
			frame.aux = 0x00;

		n = DST_FRAME_SIZE;
		DSTEncodeFrame(&frame, DST_BO_MSBLSB, buffer, &n);

		if (!AsyncWrite(args.dst_args.handle, buffer, &n)) {
			printf("ERROR: AsyncWrite: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(args.dst_args.handle)));
			break;
		}
	}

	if (!AsyncClose(args.dst_args.handle)) {
		printf("ERROR: AsyncClose: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(args.dst_args.handle)));
		exit(1);
	}

	return 0;
}

/*--------------------------------------------------------------------- */
static BOOL MakeFrame(DST_FRAME * frame, DATA * data)
{

	ASSERT(frame != NULL);
	ASSERT(data != NULL);

	frame->sync = DST_DFL_SYNC;

	frame->channel[0] = data->x[data->i];
	frame->channel[1] = data->y[data->i];
	frame->channel[2] = data->z[data->i];

	frame->aux = 0;

	data->i++;
	if (data->i >= data->n)
		data->i = 0;

	//printf("%6d %6d %6d\n", frame->channel[0], frame->channel[1], frame->channel[2]);

	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL GenerateFunctions(MAIN_ARGS * args, DATA * data)
{
	size_t i, n4;
	REAL64 n2;

	ASSERT(args != NULL);
	ASSERT(data != NULL);

	data->n = (size_t)(args->rate * (1.0 / args->frequency));
	n2 = (REAL64)data->n / 2.0;
	n4 = data->n / 4;

	data->i = 0;
	data->period = (USTIME)((1.0 / args->rate) * (REAL64)UST_SECOND);

	if (!AllocateMemory((VOID **)&data->x, data->n * sizeof(INT16)))
		return FALSE;
	if (!AllocateMemory((VOID **)&data->y, data->n * sizeof(INT16)))
		return FALSE;
	if (!AllocateMemory((VOID **)&data->z, data->n * sizeof(INT16)))
		return FALSE;

	/* x is a sine wave... */
	for (i = 0; i < data->n; i++)
		data->x[i] = (INT16)(args->amplitude * sin((REAL64)i * (TWO_PI / args->rate) * args->frequency));

	/* y is a triangle wave... */
	for (i = 0; i < data->n; i++)
		data->y[i] = (INT16)(args->amplitude * (1.0 - fabs(((REAL64)((i + n4) % data->n) - n2) / n2) * 2.0));

	/* z is a square wave */
	for (i = 0; i < data->n; i++)
		data->z[i] = (INT16)(args->amplitude * (1.0 - (data->y[i] < 0 ? 1.0 : 0.0) * 2.0));

	return TRUE;
}

/*---------------------------------------------------------------------*/
#include "getopt.c"
static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args)
{
	int i, option;

	ASSERT(args != NULL);
	ASSERT(argv != NULL);

	/* Set defaults... */
	DSTSetDefaults(&args->dst_args);
	args->amplitude = DFL_AMP;
	args->frequency = DFL_FREQ;
	args->rate = DFL_RATE;

	/* Don't write errors to stderr */
	my_opterr = 0;

	/* Options... */
	while ((option = my_getopt(argc, argv, "ha:f:r:")) != -1) {
		switch (option) {
		case 'h':
			CmdLineHelp();
			break;
		case 'a':
			args->amplitude = atof(my_optarg);
			break;
		case 'f':
			args->frequency = atof(my_optarg);
			break;
		case 'r':
			args->rate = atof(my_optarg);
			break;
		case '?':
			if (isprint(my_optopt))
				fprintf(stderr, "Unknown option `-%c'.\n", my_optopt);
			else
				fprintf(stderr, "Unknown option character `\\x%x'.\n", my_optopt);
			CmdLineHelp();
			return FALSE;
		default:
			abort();
		}
	}

	/* Arguments... */
	args->dst_args.params.device[0] = '\0';
	for (i = my_optind; i < argc; i++) {
		if (args->dst_args.params.device[0] == '\0')
			AsyncParseParams(argv[i], &args->dst_args.params);
		else {
			fprintf(stderr, "unrecognized argument: `%s'\n", argv[i]);
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CmdLineHelp(VOID)
{
	printf("Usage: dstgen [-a amplitude] [-f frequency] [-r rate] device[,speed,parity,databits,stopbits]\n");
	printf("\n");
	printf("   -h           Help display.\n");
	printf("   -a           Function amplitude in counts. (%.0lf)\n", DFL_AMP);
	printf("   -f           Function frequency in hertz. (%.2lf)\n", DFL_FREQ);
	printf("   -r           Sampling rate in hertz. (%.2lf)\n", DFL_RATE);
	printf("\n");
	printf("The port parameters default to '9600,n,8,1'\n");
	printf("\n");
    printf("[] = optional, () = default, | = mutually exclusive.\n");

	exit(0);
}

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{
	switch (sig) {
	case SIGTERM:
		fprintf(stderr, "Caught SIGTERM! \n");
		stop = TRUE;
		break;
	case SIGINT:
		fprintf(stderr, "Caught SIGINT! \n");
		stop = TRUE;
		break;
	default:
		fprintf(stderr, "Caught unexpected signal %d, ignored! \n", sig);
		break;
	}

	signal(sig, CatchSignal);
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandler(VOID)
{
	if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
		perror("signal(SIGTERM)");
		return FALSE;
	}
	if (signal(SIGINT, CatchSignal) == SIG_ERR) {
		perror("signal(SIGINT)");
		return FALSE;
	}

	return TRUE;
}
