/* $Id: main.c,v 1.7 2002/10/02 03:18:09 cvs Exp $ */
/*-----------------------------------------------------------------------

	GRF packet checker utility main module...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include "main.h"
#include "log.h"
#include "mem.h"

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL quit = FALSE;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL GRFCheck(MAIN_ARGS * args);
static BOOL OpenOutput(GRF_UNIT *unit, MAIN_ARGS *args);
static BOOL CloseOutput(GRF_UNIT *unit, MAIN_ARGS *args);
static CHAR *OutputFilename(GRF_UNIT *unit, MAIN_ARGS *args, CHAR *filename);
static BOOL ReadPacket(MAIN_ARGS * args, GRF_PACKET * packet);
static BOOL OpenInput(MAIN_ARGS * args);
static VOID CloseInput(MAIN_ARGS * args);
static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args);
static VOID CommandLineHelp(VOID);
static BOOL SetSignalHandlers(VOID);
static VOID CatchSignal(int sig);

/*---------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	MAIN_ARGS args;

	if (!SetSignalHandlers())
		abort();

	if (!SocketsInit())
		abort();

	LogMessage(LOG_NOTICE, "GRFCheck version %s (%s %s), pid:%d\n", VERSION, __DATE__, __TIME__, getpid());

	if (!ParseCommandLine(argc, argv, &args))
		exit(1);

	if (!LogOpen(argv[0], NULL, NULL))
		exit(1);
    if (args.verbose)
	    LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_INFO);
    else
	    LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_NOTICE);

	if (args.input_spec[0] == '\0') {
		LogMessage(LOG_ERR, "ERROR: No input source specified!\n");
		CommandLineHelp();
		LogClose();
		exit(1);
	}

    LogMessage(LOG_NOTICE, "Output path: %s\n", args.output_path);
    LogMessage(LOG_NOTICE, "Output record length: %.1lf seconds, %lu packets\n", 
            (REAL64)args.length / (REAL64)UST_SECOND, args.n);
    
    if (!OpenInput(&args)) {
		LogClose();
		exit(1);
	}

	if (!GRFCheck(&args)) {
		CloseInput(&args);
		LogClose();
		exit(1);
	}

	CloseInput(&args);
	LogClose();

	return 0;
}

/*--------------------------------------------------------------------- */
static BOOL GRFCheck(MAIN_ARGS * args)
{
	BOOL ok;
    CHAR string[32];
	UINT32 t_packets, t_seq_errors, t_time_tears;
	GRF_PACKET *packet;
	GRF_UNIT *list = NULL, *unit;
	REAL64 elapsed;
	USTIMER timer;
	REAL64 rate;
	USTIME time, period;

	if (!AllocateMemory((VOID **)&packet, GRF_MAX_PACKET_LEN)) {
		LogMessage(LOG_ERR, "ERROR: GRFLog: AllocateMemory failed!\n");
		return FALSE;
	}

	StartUSTimer(&timer, 0);
	time = 0;
	ok = TRUE;
	while (!QuitRequested()) {
		if (ReadPacket(args, packet)) {
			if (packet->hdr.length == 0)
				continue;

	        if (packet->hdr.type == GRF_TYPE_INFO) {
		        if (packet->hdr.unit != 0)
			        LogMessage(LOG_INFO, "Unit %d: %s\n", packet->hdr.unit, packet->serial.info.message);
		        else
			        LogMessage(LOG_INFO, "%s\n", packet->serial.info.message);
	        }

            /* Find or create a unit record for this unit */
			for (unit = list; unit != NULL; unit = unit->next) {
				if (packet->hdr.unit == unit->unit_id)
					break;
			}
			if (unit == NULL) {
				if (!AllocateMemory((VOID **)&unit, sizeof(GRF_UNIT))) {
					LogMessage(LOG_ERR, "ERROR: Unable to allocate memory!\n");
					ok = FALSE;
					break;
				}
				unit->unit_id = packet->hdr.unit;
				unit->channel = 0;
				unit->sequence = packet->hdr.sequence;
				unit->count = 0;
				unit->seq_errors = 0;
				unit->time_tears = 0;
				unit->expect_time = 0;
                
                unit->fileptr = NULL;
                unit->filename[0] = '\0';
                unit->start_time = 0;
                unit->length = 0;

				unit->next = list;
				list = unit;
			}

			/* Check sequence numbers */
			if (packet->hdr.sequence != unit->sequence) {
				LogMessage(LOG_ERR, "ERROR: sequence break on unit %u, got:%u, expected: %u\n",
					unit->unit_id, packet->hdr.sequence, unit->sequence);
				unit->seq_errors++;
                CloseOutput(unit, args);
			}
			unit->sequence = packet->hdr.sequence + 1;
			unit->count++;

            if (args->n > 0 && unit->count > args->n) {
                unit->count = 1;
                CloseOutput(unit, args);
            }

            /* Check timing */
            if (packet->hdr.type == GRF_TYPE_DATA) {
	            time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
	            rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;
                if (rate > 0.0) {
		            period = (USTIME)((1.0 / rate) * (REAL64)UST_SECOND);
                } else {
                    period = 1;
                }

		        LogMessage(LOG_INFO, "%s:%s:%s %u:%u:%05u %s (%s) %.6lf (%s) %+.6lf\n",
			        packet->payload_hdr.data.name.network,
			        packet->payload_hdr.data.name.station,
			        packet->payload_hdr.data.name.component,
			        packet->hdr.unit,
			        packet->payload_hdr.data.channel,
			        packet->hdr.sequence,
			        FormatUSTime(string, 32, UST_DATE_TIME_USEC, time),
			        GRFDescribeQuality(packet->payload_hdr.data.timing.time_quality),
			        rate,
			        GRFDescribeQuality(packet->payload_hdr.data.timing.rate_quality),
			        (REAL64)(time - unit->expect_time) / (REAL64)UST_SECOND);

                if (unit->channel == 0) {
                    unit->channel = packet->payload_hdr.data.channel;
                }

	            if (unit->channel == packet->payload_hdr.data.channel) {
                    if (unit->expect_time != 0 ) { 
		                if (time < unit->expect_time - period || time > unit->expect_time + period) {
			                unit->time_tears++;
			                LogMessage(LOG_ERR, "ERROR: time tear on unit %u, %+.6lf\n",
				                packet->hdr.unit, (REAL64)(time - unit->expect_time) / (REAL64)UST_SECOND);
                            CloseOutput(unit, args);
		                }
                    }
    	            unit->expect_time = time + (USTIME)(((REAL64)packet->payload_hdr.data.n / rate) * (REAL64)UST_SECOND);
	            }
            }

            if (!unit->fileptr) {
                if (!OpenOutput(unit, args)) {
					ok = FALSE;
					break;
				}
            }
            if (!GRFWriteFile(unit->fileptr, packet)) {
				ok = FALSE;
				break;
			}
            if (packet->hdr.type == GRF_TYPE_DATA) {
                if (unit->fileptr) {
                    if (unit->start_time == 0) {
                        unit->start_time = time;
                        unit->length = 0;
                    } else {
                        unit->length = time - unit->start_time;
                        if (unit->length > args->length) 
                            CloseOutput(unit, args);
                    }
                }
            }
		}
		else {
			break;
		}
	}

	elapsed = (REAL64)USTimerElapsed(&timer) / (REAL64)UST_SECOND;
	ReleaseMemory(packet);

	/* Release sequence check list */
	t_packets = t_seq_errors = t_time_tears = 0;
	while (list != NULL) {
		unit = list;
		list = list->next;
        CloseOutput(unit, args);
		LogMessage(LOG_NOTICE, "Unit: %u packets: %u, %.2lf/sec, seq errors: %u, time tears: %u\n",
			unit->unit_id, unit->count, (REAL64)unit->count / elapsed, unit->seq_errors, unit->time_tears);
		t_packets += unit->count;
		t_seq_errors += unit->seq_errors;
		t_time_tears += unit->time_tears;
		ReleaseMemory(unit);
	}
	LogMessage(LOG_NOTICE, "Total packets: %u, %.2lf/sec, seq errors: %u, time tears: %u\n",
		t_packets, (REAL64)t_packets / elapsed, t_seq_errors, t_time_tears);

	DumpMemoryInfo(LOG_NOTICE);

	return ok;
}


/*---------------------------------------------------------------------*/
static BOOL CloseOutput(GRF_UNIT *unit, MAIN_ARGS *args) 
{
    CHAR filename[MAX_PATH_LEN + 1];
    
    ASSERT(unit != NULL);
    ASSERT(args != NULL);

    if (unit->fileptr) {
        fclose(unit->fileptr);
        unit->fileptr = NULL;

        /* If no data is in the file, set timestamp to system time */
        if (unit->start_time == 0)
            unit->start_time = SystemUSTime();
        OutputFilename(unit, args, filename);

        LogMessage(LOG_NOTICE, "Closing: %s\n", filename);

        if (strcmp(unit->filename, filename) != 0) {
            rename(unit->filename, filename);
        }

        unit->filename[0] = '\0';
        unit->start_time = 0;
        unit->length = 0;
    }

    return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL OpenOutput(GRF_UNIT *unit, MAIN_ARGS *args) 
{

    ASSERT(unit != NULL);
    ASSERT(args != NULL);

    if (unit->fileptr) 
        CloseOutput(unit, args);

    sprintf(unit->filename, "%sunit%u.grf", args->output_path, unit->unit_id);

	if ((unit->fileptr = fopen(unit->filename, "w+b")) == NULL) {
		LogMessage(LOG_ERR, "ERROR: OpenOutput: fopen(%s) failed: %s\n", 
                unit->filename, strerror(errno));
		return FALSE;
	}

    unit->start_time = 0;
    unit->length = 0;

    return TRUE;
}

/*---------------------------------------------------------------------*/
/*
Filename format specifiers:

    %Y  - four digit year
    %y  - two digit year
    %M	- two digit month
    %D  - two digit day of month
    %d  - three digit day of year
    %h	- two digit hour of day
    %m  - two digit minute of hour
    %s  - two digit integer second of minute
    %ms - three digit integer millisecond of second
    %us - six digit integer microsecond of second

    %l  - record length as integer seconds (variable length)
    %lm - three digit integer millisecond of record length
    %lu - six digit integer microsecond of record length

    %u  - GRF unit number (variable length)

    %%	- literal '%'
*/
static CHAR *OutputFilename(GRF_UNIT *unit, MAIN_ARGS *args, CHAR *filename)
{
    BOOL escape;
    CHAR *ptr;
    INT32 year, month, day, doy, hour, minute;
    REAL64 second;
    size_t i;

    ASSERT(unit != NULL);
    ASSERT(args != NULL);
    ASSERT(filename != NULL);

    DecodeUSTime(unit->start_time, &year, &doy, &hour, &minute, &second);
    MonthDay(&month, &day, year, doy);

    ptr = args->name_format;
    i = sprintf(filename, "%s", args->output_path);
    escape = FALSE;

    while (*ptr) {

        if (*ptr == '%') {
            if (escape) {
                escape = FALSE;
            }
            else {
                escape = TRUE;
                ptr++;
                continue;
            }
        }

        if (escape) {
            if (*ptr == 'Y')
                i += sprintf(&filename[i], "%04d", year);
            else if (*ptr == 'y')
                i += sprintf(&filename[i], "%02d", year % 100);
            else if (*ptr == 'M')
                i += sprintf(&filename[i], "%02d", month);
            else if (*ptr == 'D')
                i += sprintf(&filename[i], "%02d", day);
            else if (*ptr == 'd')
                i += sprintf(&filename[i], "%03d", doy);
            else if (*ptr == 'h')
                i += sprintf(&filename[i], "%02d", hour);
            else if (*ptr == 'm') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&filename[i], "%03d", (INT32) ((second - (INT32) second) * 1000));
                }
                else {
                    i += sprintf(&filename[i], "%02d", minute);
                }
            }
            else if (*ptr == 's')
                i += sprintf(&filename[i], "%02d", (INT32) second);
            else if (*ptr == 'u') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&filename[i], "%06d", (INT32) ((second - (INT32) second) * 1000000));
                }
                else {
                    i += sprintf(&filename[i], "%d", unit->unit_id);
                }
            }
            else if (*ptr == 'l') {
                if (*(ptr + 1) == 'm') {
                    ptr++;
                    i += sprintf(&filename[i], "%03d", (UINT32) ((unit->length % UST_SECOND) / UST_MSECOND));
                }
                else if (*(ptr + 1) == 'u') {
                    ptr++;
                    i += sprintf(&filename[i], "%06d", (UINT32) (unit->length % UST_SECOND));
                }
                else {
                    i += sprintf(&filename[i], "%u", (UINT32) (unit->length / UST_SECOND));
                }
            }
            else
                i += sprintf(&filename[i], "%c", *ptr);
        }
        else {
            i += sprintf(&filename[i], "%c", *ptr);
        }

        ptr++;
        escape = FALSE;
    }

    return filename;
}

/*--------------------------------------------------------------------- */
static BOOL ReadPacket(MAIN_ARGS * args, GRF_PACKET * packet)
{

	ASSERT(args != NULL);
	ASSERT(packet != NULL);

	if (!GRFReadFile(args->file, packet))
		return FALSE;
	/* End-of-file? */
	if (packet->hdr.type == GRF_TYPE_NONE && packet->hdr.length == 0) {
		RequestQuit(TRUE);
		return FALSE;
	}

	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL OpenInput(MAIN_ARGS * args)
{
	ASSERT(args != NULL);

	if ((args->file = fopen(args->input_spec, "rb")) == NULL) {
		LogMessage(LOG_ERR, "ERROR: OpenInput: fopen(%s) failed: %s\n", args->input_spec, strerror(errno));
		return FALSE;
	}

	return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID CloseInput(MAIN_ARGS * args)
{
	ASSERT(args != NULL);

    fclose(args->file);

	return;
}

/*--------------------------------------------------------------------- */
#include "getopt.c"

static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args)
{
	CHAR *ptr;
    int i, option;

	ASSERT(args != NULL);
	ASSERT(argv != NULL);

	/* Set defaults */
	strcpy(args->output_path, DFL_PATH);
    strcpy(args->name_format, DFL_NAME_FORMAT);
    args->length = DFL_LENGTH;
    args->n = DFL_N;
    args->verbose = FALSE;

	args->input_spec[0] = '\0';
	args->file = NULL;
	args->output = NULL;

	/* Don't write errors to stderr */
	my_opterr = 0;

	/* Options... */
	while ((option = my_getopt(argc, argv, "hvl:o:f:n:")) != -1) {
		switch (option) {
		  case 'h':
			CommandLineHelp();
			break;
		  case 'v':
			args->verbose = TRUE;
			break;
          case 'l':
            if (my_optarg)
                args->length = (USTIME) strtoul(my_optarg, NULL, 0) * UST_SECOND;
            break;
          case 'n':
            if (my_optarg)
                args->n = strtoul(my_optarg, NULL, 0);
            break;
		  case 'f':
			if (my_optarg)
				strncpy(args->name_format, my_optarg, MAX_PATH_LEN);
			else {
				LogMessage(LOG_ERR, "ERROR: No name format specified: '-f'\n");
				return FALSE;
			}
			break;
		  case 'o':
            if (my_optarg) {
				strncpy(args->output_path, my_optarg, MAX_PATH_LEN);
                ptr = &args->output_path[strlen(args->output_path) - 1];
                if (ptr == NULL) 
                    strcpy(args->output_path, DFL_PATH);
                else if (*ptr != '\\' && *ptr != '/') 
                    strcat(args->output_path, "/");
            } else {
				LogMessage(LOG_ERR, "ERROR: No output path specified: '-o'\n");
				return FALSE;
			}
			break;
		  case '?':
			if (isprint(my_optopt))
				fprintf(stderr, "Unknown option '-%c'.\n", my_optopt);
			else
				fprintf(stderr, "Unknown option character '\\x%x'.\n", my_optopt);
			return FALSE;
		  default:
			abort();
		}
	}

	/* Arguments... */
    args->input_spec[0] = '\0';
	for (i = my_optind; i < argc; i++) {
        if (args->input_spec[0] == '\0') {
			strncpy(args->input_spec, argv[i], MAX_PATH_LEN);
        } else {
    		fprintf(stderr, "unrecognized argument: `%s'\n", argv[i]);
	    	return FALSE;
        }
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CommandLineHelp(VOID)
{
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "Usage: grfcheck [-hv] [-l length] [-f filename_format]\n"); 
    LogMessage(LOG_NOTICE, "       [-o output_path] input_file\n");
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "   -h  Help display.\n");
	LogMessage(LOG_NOTICE, "   -v  Verbose logging. (Off)\n");
	LogMessage(LOG_NOTICE, "   -f  Output filename format. (\"%s\")\n", 
            DFL_NAME_FORMAT);
	LogMessage(LOG_NOTICE, "   -l  Max output record length in seconds. (%lu)\n",
            (UINT32)(DFL_LENGTH / UST_SECOND));
	LogMessage(LOG_NOTICE, "   -n  Max output record length in packets. (Unlimited)\n");
	LogMessage(LOG_NOTICE, "   -o  Output path. (%s)\n", DFL_PATH);
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "[] = optional, () = default, | = mutually exclusive.\n");

	exit(0);
}

/*--------------------------------------------------------------------- */
BOOL QuitRequested(VOID)
{
	return quit;
}

/*--------------------------------------------------------------------- */
VOID RequestQuit(BOOL request)
{
	quit = request;
	return;
}

/*--------------------------------------------------------------------- */
VOID HexDump(UINT32 loglevel, VOID *ptr, size_t length)
{
	CHAR string[128];
	UINT8 *data;
	size_t i, j, n;

	ASSERT(ptr != NULL);

	if (length == 0)
		return;

/*                        0123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789 */
	LogMessage(loglevel, "%08X  00 01 02 03  04 05 06 07  08 09 0A 0B  0C 0D 0E 0F  0123456789ABCDEF\n", length);
	LogMessage(loglevel, "--------  --------------------------------------------------  ----------------\n");
	data = (UINT8 *)ptr;
	for (n = 0; n < length; n += 16) {
		j = sprintf(string, "%08x ", n);
		for (i = n; i < n + 16; i++) {
			if (i % 4 == 0)
				j += sprintf(string + j, " ");
			if (i < length)
				j += sprintf(string + j, "%02x ", data[i]);

			else
				j += sprintf(string + j, "   ");
		}
		j += sprintf(string + j, " ");
		for (i = n; i < n + 16; i++) {
			if (i < length) {
				if (isprint(data[i]))
					j += sprintf(string + j, "%c", data[i]);

				else
					j += sprintf(string + j, ".");
			}
		}
		LogMessage(loglevel, "%s\n", string);
	}
	return;
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandlers(VOID)
{
	if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
		perror("signal(SIGTERM)");
		return FALSE;
	}
	if (signal(SIGINT, CatchSignal) == SIG_ERR) {
		perror("signal(SIGINT)");
		return FALSE;
	}
#if !defined WIN32
	/* Ignore the SIGPIPE signal */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal(SIGPIPE)");
		return FALSE;
	}
#endif

	return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{
	switch (sig) {
	case SIGTERM:
		//fprintf(stderr, "\rCaught SIGTERM! \n");
		RequestQuit(TRUE);
		break;
	case SIGINT:
		//fprintf(stderr, "\rCaught SIGINT! \n");
		RequestQuit(TRUE);
		break;
	default:
		fprintf(stderr, "\rCaught unexpected signal %d, ignored! \n", sig);
		break;
	}

	signal(sig, CatchSignal);
}
