/* $Id: gse.c,v 1.4 2003/04/03 20:54:45 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	GSE2.0 waveform segment output.  These routines are dependent
    upon the ASCII character set.

    As per: GSE Conference Room Paper 243 (GSE/CRP/243), Annex 3 -
        Formats and Protocols for Data Exchange, Section 4.5 - GSE2.0
        Waveform Segments - Sub-format CM6.

-----------------------------------------------------------------------*/

#include "gse.h"

/* Module constants ---------------------------------------------------*/
#define MAX_LINE_LEN        1024
#define CM6_MODULO          100000000L

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static CHAR name_format[MAX_PATH_LEN + 1];
static CHAR line[MAX_LINE_LEN + 1];

static const UINT32 powers_of_two[32] = {
	0x00000001, 0x00000002, 0x00000004, 0x00000008,
	0x00000010, 0x00000020, 0x00000040, 0x00000080,
	0x00000100, 0x00000200, 0x00000400, 0x00000800,
	0x00001000, 0x00002000, 0x00004000, 0x00008000,
	0x00010000, 0x00020000, 0x00040000, 0x00080000,
	0x00100000, 0x00200000, 0x00400000, 0x00800000,
	0x01000000, 0x02000000, 0x04000000, 0x08000000,
	0x10000000, 0x20000000, 0x40000000, 0x80000000
};

static const CHAR code_chars[64] = {
	'+', '-', '0', '1', '2', '3', '4', '5',
	'6', '7', '8', '9', 'A', 'B', 'C', 'D',
	'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
	'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
	'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b',
	'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
	'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',
	's', 't', 'u', 'v', 'w', 'x', 'y', 'z'
};

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL WriteWID2(FILE *file, RECORD * record);
static BOOL WriteDAT2(FILE *file, RECORD * record);
static size_t CM6EncodeSample(INT32 sample, CHAR *codes);
static UINT32 PowerOfTwo(UINT32 exponent);

/*---------------------------------------------------------------------*/
VOID GSENameFormat(CHAR *format)
{
	strcpy(name_format, format);
	return;
}

/*---------------------------------------------------------------------*/
BOOL GSEConvert(RECORD * head, CHAR *net_name)
{
	CHAR filename[MAX_PATH_LEN + 1];
	FILE *file;
	RECORD *record;

	ASSERT(head != NULL);

	OutputFilename(head, net_name, name_format, filename);
	LogMessage(LOG_INFO, "%s\n", filename);
	if ((file = fopen(filename, "w+t")) == NULL) {
		LogMessage(LOG_ERR, "ERROR: OutputGSE: fopen(%s) failed: %s\n", filename, strerror(errno));
		return FALSE;
	}

	/* Output an GSE waveform segments */
	for (record = head; record != NULL; record = record->next) {
		if (net_name != NULL) {
			if (strcmp(record->name.network, net_name) != 0)
				continue;
		}
		if (!WriteWID2(file, record))
			break;
		if (!WriteDAT2(file, record))
			break;
	}

	fclose(file);

	AppendFilenameToList(filename);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL WriteWID2(FILE *file, RECORD * record)
{
	INT32 year, doy, month, day, hour, minute;
	REAL64 second;

	ASSERT(file != NULL);
	ASSERT(record != NULL);

	DecodeUSTime(record->ist, &year, &doy, &hour, &minute, &second);
	MonthDay(&month, &day, year, doy);

/* WID2 1996/06/06 06:47:46.639 FOO   S Z      CM6     1202    4.590642   0.10E+01   1.000         -1.0  0.0 */

/*                 12345    a    3    6    9    2       9      5      9    4   8    7      9      h     8      5     1      */
/*                      6789   12   45   78   b1  345678  c1234    678 d123 567 ..56 ..5678 ..6789 ..567 9i1234 6789j 2345  */
/*                      year  mon  day   hr  min  second    stn   comp  aux CM6 leng   rate  calib calpr   type  azim incid */
	sprintf(line, "WID2 %04d/%02d/%02d %02d:%02d:%06.3lf %-5.5s %-3.3s      CM6 %8ld %11.6f %10.2E %7.3f        %5.1f %4.1f",
		year, month, day, hour, minute, second,
		record->name.station,
		record->name.component, 
		record->n, 
		record->rate, 
		(REAL32)1.0, 
		(REAL32)1.0, 
		(REAL32)-1.0, 
		(REAL32)0.0
	);

	fprintf(file, "%s\n", line);
	if (ferror(file)) {
		LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL WriteDAT2(FILE *file, RECORD * record)
{
	CHAR codes[6];
	INT32 *sample, diff, checksum, xim1, xim2;
	size_t i, j, n, length;

	ASSERT(file != NULL);
	ASSERT(record != NULL);

	fprintf(file, "DAT2\n");
	if (ferror(file)) {
		LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		return FALSE;
	}

	sample = record->samples;
	checksum = xim1 = xim2 = 0;
	length = 0;
	for (i = 0; i < record->n; i++) {
		/* Compute checksum */
		if (abs(sample[i]) >= CM6_MODULO)
			checksum += sample[i] - ((sample[i] / CM6_MODULO) * CM6_MODULO);
		else
			checksum += sample[i];
		/* Apply modulo division to the checksum */
		if (labs(checksum) >= CM6_MODULO)
			checksum -= ((checksum / CM6_MODULO) * CM6_MODULO);

		/* Compute second difference */
		diff = sample[i] - (2 * xim1) + xim2;
		xim2 = xim1;
		xim1 = sample[i];

		/* Encode the difference */
		n = CM6EncodeSample(diff, codes);
		for (j = 0; j < n; j++) {
			line[length++] = codes[j];
			if (length >= 80) {
				line[length] = '\0';
				fprintf(file, "%s\n", line);
				if (ferror(file)) {
					LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
					return FALSE;
				}
				length = 0;
			}
		}
	}

	/* Fill out the last line with spaces (end-of-record)... */
	while (length < 80)
		line[length++] = ' ';
	line[length] = '\0';
	fprintf(file, "%s\n", line);
	if (ferror(file)) {
		LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		return FALSE;
	}

	/* Output the checksum */
	/*             12345         */
	/*                  6789a123 */
	fprintf(file, "CHK2 %11ld \n", abs(checksum));
	if (ferror(file)) {
		LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static size_t CM6EncodeSample(INT32 sample, CHAR *codes)
{
	size_t i, n_codes;
	UINT32 sign, magnitude, control, code_index, mask, shift;

	/* codes must point to at least six bytes of memory! */
	ASSERT(codes != NULL);

	/* Save sign in bit 4 and convert to unsigned magnitude */
	if (sample < 0) {
		sign = PowerOfTwo(4);
		magnitude = (UINT32)(-sample);
	}
	else {
		sign = 0;
		magnitude = (UINT32)sample;
	}

	/* Compute starting point and number of codes */
	i = 4;
	shift = 0;
	control = 0;
	while (magnitude >= PowerOfTwo(i) && i < 32) {
		i += 5;
		shift += 5;
		control = PowerOfTwo(5);
	}

	/* Encode most significant 4 data bits with control bit (bit 5) set if
	   more to come and saved sign bit (bit 4) */
	n_codes = 0;
	mask = PowerOfTwo(4) - 1;
	code_index = ((magnitude >> shift) & mask) | sign | control;
	codes[n_codes++] = code_chars[code_index];

	/* Encode remaining 5 bit segments with set control bit (bit 5) */
	mask = PowerOfTwo(5) - 1;
	control = PowerOfTwo(5);
	while (shift > 0) {
		shift -= 5;
		/* If this is the last segment, clear the control bit */
		if (shift == 0)
			control = 0;
		code_index = ((magnitude >> shift) & mask) | control;
		codes[n_codes++] = code_chars[code_index];
	}

	/* Return number of codes */
	return n_codes;
}

/*---------------------------------------------------------------------*/
static UINT32 PowerOfTwo(UINT32 exponent)
{
	if (exponent < 32)
		return powers_of_two[exponent];

	return 0;
}
