/* $Id: main.c,v 1.18 2003/04/03 20:54:45 cvs Exp $ */
/*-----------------------------------------------------------------------

	GRFConvert main module...

	Copyright (C) 2000-2003 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include "main.h"

#include "ascii.h"
#include "gse.h"
#include "sac.h"
#include "segy.h"
#include "miniseed.h"
#include "suds.h"
#include "matlab.h"
#include "stats.h"

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL shut_down = FALSE;
static UINT32 log_level = LOG_INFO;
static CHAR output_path[MAX_PATH_LEN + 1];
static CHAR list_file[MAX_PATH_LEN + 1];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/

static BOOL(*Convert) (RECORD * record, CHAR * net_name);   /* Function pointer set by SelectOutputFormat() */

static BOOL GRFConvert(ARGS * args);
static BOOL HandleGRFPacket(ARGS * args, STATE * state, GRF_PACKET * packet);
static BOOL CopyData(RECORD * record, GRF_PACKET * packet);
static VOID ClearState(STATE * state, CHAR * net_name);
static BOOL ReadGRFPacket(ARGS * args, GRF_PACKET * packet);
static BOOL SelectOutputFormat(ARGS * args);
static BOOL OpenGRFSource(ARGS * args);
static VOID CloseGRFSource(ARGS * args);

static BOOL SetSignalHandlers(VOID);

#if !defined WIN32
static VOID CatchSignal(int sig);

#else
static BOOL CtrlHandler(UINT32 ctrl_type);

#endif

/*---------------------------------------------------------------------*/
#if 1
int main(int argc, char *argv[])
{
    ARGS args;

    if (!SetSignalHandlers())
        exit(1);

    if (!SocketsInit())
        exit(1);

    LogMessage(LOG_NOTICE, "GRFConvert version %s (%s %s), pid:%d\n\n", 
			VERSION, __DATE__, __TIME__, getpid());

    SetDefaults(&args);
    if (!ParseCommandLine(argc, argv, &args))
        exit(1);

    DumpConfiguration(&args);

    if (!GRFConvert(&args))
        exit(1);

    return 0;
}

#endif

/*---------------------------------------------------------------------*/
static BOOL GRFConvert(ARGS * args)
{
    GRF_PACKET *packet;
    STATE state;
    GRF_UNIT *list = NULL, *unit;

    ASSERT(args != NULL);

    if (!SelectOutputFormat(args))
        return FALSE;

    while (!shut_down) {
        if (OpenGRFSource(args))
            break;
        USSleep(UST_SECOND * 5);
    }

    if (!AllocateMemory((VOID **) & packet, sizeof(GRF_PACKET))) {
        LogMessage(LOG_ERR, "ERROR: GRFConvert: unable to allocate memory!\n");
        SOCK_CLOSE(args->src.des.sock.fd);
        return FALSE;
    }

    strcpy(output_path, args->output_path);
    strcpy(list_file, args->list_file);
    state.earliest = state.latest = VOID_USTIME;
    state.record = NULL;

    while (!shut_down) {
        /* Read a packet... */
        if (ReadGRFPacket(args, packet)) {
            /* Did we get a packet? */
            if (packet->hdr.length == 0)
                continue;

            /* Check sequence numbers */
            for (unit = list; unit != NULL; unit = unit->next) {
                if (packet->hdr.unit == unit->unit_id)
                    break;
            }
            if (unit == NULL) {
                if (!AllocateMemory((VOID **) & unit, sizeof(GRF_UNIT))) {
                    LogMessage(LOG_ERR, "ERROR: Unable to allocate memory!\n");
                    break;
                }
                unit->unit_id = packet->hdr.unit;
                unit->sequence = packet->hdr.sequence;
                unit->next = list;
                list = unit;
            }
            if (packet->hdr.sequence != unit->sequence) {
                LogMessage(LOG_ERR, "Sequence break on unit %u, got:%u, expected:%u\n",
                    unit->unit_id, packet->hdr.sequence, unit->sequence);
				if (state.earliest != VOID_USTIME) {
					if (!Convert(state.record, NULL)) {
						ClearState(&state, NULL);
						return FALSE;
					}
					ClearState(&state, NULL);
				}
            }
            unit->sequence = packet->hdr.sequence + 1;

            if (!HandleGRFPacket(args, &state, packet))
                break;
        }
        /* If we're reading a socket... */
        else if (args->src.type == SOURCE_SOCKET) {
            /* Try to reconnect forever... */
            while (!shut_down) {
                CloseGRFSource(args);
                if (OpenGRFSource(args))
                    break;
                USSleep(UST_SECOND * 5);
            }
        }
        else { 
            shut_down = TRUE; 
        }
    }

    /* Flush out any pending data... */
    if (state.earliest != VOID_USTIME) {
        if (!Convert(state.record, NULL)) {
            ClearState(&state, NULL);
            return FALSE;
        }
        ClearState(&state, NULL);
    }

    ReleaseMemory((VOID *) packet);

    /* Release sequence check list */
    while (list != NULL) {
        unit = list;
        list = list->next;
        ReleaseMemory(unit);
    }

    CloseGRFSource(args);

#if defined _DEBUG
    DumpMemoryInfo(LOG_NOTICE);
#endif

    return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL HandleGRFPacket(ARGS * args, STATE * state, GRF_PACKET * packet)
{
    CHAR string[32], net_name[GRF_MAX_NET_LEN + 1], *ptr;
    REAL64 rate;
    USTIME time, margin;
    RECORD *record, *node, *prev;

    ASSERT(args != NULL);
    ASSERT(state != NULL);
    ASSERT(packet != NULL);

    if (packet->hdr.type == GRF_TYPE_INFO) {
        if (packet->hdr.unit == 0)
            LogMessage(LOG_INFO, "%s\n", packet->serial.info.message);
        else
            LogMessage(LOG_INFO, "Unit %u: %s\n", packet->hdr.unit, packet->serial.info.message);

        /* Is this a network detrigger message? */
        if (strncasecmp(packet->serial.info.message, "Network detrigger", 17) == 0) {
            ptr = net_name;
            memset(ptr, 0, GRF_MAX_NET_LEN + 1);
            strncpy(ptr, &packet->serial.info.message[19], GRF_MAX_NET_LEN);
            while (*ptr != '\0') {
                if (*ptr == ' ') {
                    *ptr = '\0';
                    break;
                }
                ptr++;
            }

            /* Output data... */
            if (!Convert(state->record, net_name)) {
                ClearState(state, NULL);
                return FALSE;
            }
            ClearState(state, net_name);
        }
        return TRUE;
    }
    else if (packet->hdr.type != GRF_TYPE_DATA)
        return TRUE;

	/* Drop unknown or bad time quality packets */
	if (args->drop_bad_times) {
		if (packet->payload_hdr.data.timing.time_quality <= GRF_TIME_BAD) {
            LogMessage(LOG_ERR, "WARNING: dropping packet with '%s' time quality from %s:%s:%s (%u:%u)\n", 
				GRFDescribeQuality(packet->payload_hdr.data.timing.time_quality),
				packet->payload_hdr.data.name.network,
				packet->payload_hdr.data.name.station,
				packet->payload_hdr.data.name.component,
				packet->hdr.unit, packet->payload_hdr.data.channel);
			return TRUE;
		}
	}

    time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
    rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;
    margin = (USTIME) (((1.0 / rate) * (REAL64) UST_SECOND) * 1.0);

    /* If starting a new record... */
    if (state->earliest == VOID_USTIME) {
        state->earliest = time;
        state->latest = time;
    }

    /* Time to go? */
    if (time > state->earliest) {
        if (time - state->earliest >= args->length) {
            /* Output data... */
            if (!Convert(state->record, NULL)) {
                ClearState(state, NULL);
                return FALSE;
            }
            ClearState(state, NULL);
        }
    }

    /* Lookup this channel record... */
    for (record = state->record; record != NULL; record = record->next) {
        ASSERT(IsValidBasePointer(record));
        if (record->unit == packet->hdr.unit && record->channel == packet->payload_hdr.data.channel)
            break;
    }
    /* If we didn't find one... */
    if (record == NULL) {
        /* Create one */
        if (!AllocateMemory((VOID **) & record, sizeof(RECORD))) {
            LogMessage(LOG_ERR, "ERROR: HandleGRFPacket: unable to allocate memory!\n");
            ClearState(state, NULL);
            return FALSE;
        }

        /* Find last node... */
        prev = NULL;
        for (node = state->record; node != NULL; node = node->next)
            prev = node;
        if (prev != NULL) {
            record->next = prev->next;
            prev->next = record;
        }
        else {
            record->next = state->record;
            state->record = record;
        }
        /* Initialize this record */
        record->unit = packet->hdr.unit;
        record->channel = packet->payload_hdr.data.channel;
        record->ist = time;
        record->lst = time;
        record->rate = rate;
        record->counts_per_volt = packet->payload_hdr.data.counts_per_volt;
        record->n = 0;
        memcpy(&record->name, &packet->payload_hdr.data.name, sizeof(GRF_NAME));
        memcpy(&record->timing, &packet->payload_hdr.data.timing, sizeof(GRF_TIMING));
        record->samples = NULL;
    }

    if (record->lst < time - margin || record->lst > time + margin) {
        LogMessage(LOG_WARNING, "WARNING: time tear on %s:%s:%s at %s, %.6lf\n",
            packet->payload_hdr.data.name.network,
            packet->payload_hdr.data.name.station,
            packet->payload_hdr.data.name.component,
            FormatUSTime(string, 32, UST_DATE_TIME_MSEC, time), (REAL64) (time - record->lst) / (REAL64) UST_SECOND);
    }

    /* Add this packet... */
    if (!CopyData(record, packet)) {
        ClearState(state, NULL);
        return FALSE;
    }
    record->lst = time + (USTIME) (((REAL64) packet->payload_hdr.data.n / rate) * (REAL64) UST_SECOND);

    /* Update state... */
    if (state->earliest > record->ist)
        state->earliest = record->ist;
    if (state->latest < record->lst)
        state->latest = record->lst;

    return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL CopyData(RECORD * record, GRF_PACKET * packet)
{
    INT32 *sample;

    ASSERT(record != NULL);
    ASSERT(packet != NULL);

    /* Make room for new sample data */
    if (record->n == 0) {
        if (!AllocateMemory((VOID **) & record->samples, sizeof(INT32) * packet->payload_hdr.data.n)) {
            LogMessage(LOG_ERR, "ERROR: CopyData: unable to allocate memory!\n");
            return FALSE;
        }
    }
    else {
        ASSERT(IsValidBasePointer(record->samples));
        if (!ResizeMemory((VOID **) & record->samples, sizeof(INT32) * (record->n + packet->payload_hdr.data.n))) {
            LogMessage(LOG_ERR, "ERROR: CopyData: unable to resize memory!\n");
            return FALSE;
        }
    }
    sample = record->samples + record->n;
    memset(sample, 0, sizeof(INT32) * packet->payload_hdr.data.n);
    record->n += packet->payload_hdr.data.n;

    /* Unwind and copy the sample data */
    GRFDecodeSamples(packet, sample, packet->payload_hdr.data.n);

    return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID ClearState(STATE * state, CHAR * net_name)
{
    RECORD *record, *ptr, *prev;

    ASSERT(state != NULL);

    if (net_name != NULL) {
        prev = NULL;
        record = state->record;
        while (record != NULL) {
            if (strcmp(record->name.network, net_name) == 0) {
                ptr = record;
                record = record->next;
                if (ptr->samples != NULL)
                    ReleaseMemory(ptr->samples);
                ReleaseMemory(ptr);
                if (prev == NULL)
                    state->record = record;
                else
                    prev->next = record;
            }
            else {
                prev = record;
                record = record->next;
            }
        }
        state->earliest = state->latest = VOID_USTIME;
        for (record = state->record; record != NULL; record = record->next) {
            if (state->earliest == VOID_USTIME || record->ist < state->earliest)
                state->earliest = record->ist;
            if (state->latest == VOID_USTIME || record->lst > state->latest)
                state->latest = record->lst;
        }
    }
    else {
        record = state->record;
        while (record != NULL) {
            ptr = record;
            record = record->next;
            if (ptr->samples != NULL)
                ReleaseMemory(ptr->samples);
            ReleaseMemory(ptr);
        }

        state->earliest = state->latest = VOID_USTIME;
        state->record = record;
    }

    return;
}

/*--------------------------------------------------------------------- */
static BOOL ReadGRFPacket(ARGS * args, GRF_PACKET * packet)
{

    ASSERT(args != NULL);
    ASSERT(packet != NULL);

    /* Read from file? */
    if (args->src.type == SOURCE_FILE) {
        if (!GRFReadFile(args->src.des.file, packet))
            return FALSE;
        /* End-of-file? */
        if (packet->hdr.type == GRF_TYPE_NONE && packet->hdr.length == 0) {
            shut_down = TRUE;
            return FALSE;
        }
    }
    else {
        if (!GRFReadSocket(args->src.des.sock.fd, packet, args->timeout))
            return FALSE;
    }

    return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL SelectOutputFormat(ARGS * args)
{

    ASSERT(args != NULL);

    switch (args->format) {
      case FMT_ASCII:
        Convert = ASCIIConvert;
        if (args->name_format[0] == '\0')
            ASCIINameFormat(DFL_ASCII_NAME_FMT);
        else
            ASCIINameFormat(args->name_format);
        return TRUE;
      case FMT_GSE:
        Convert = GSEConvert;
        if (args->name_format[0] == '\0')
            GSENameFormat(DFL_GSE_NAME_FMT);
        else
            GSENameFormat(args->name_format);
        return TRUE;
      case FMT_SAC:
        Convert = SACConvert;
        if (args->name_format[0] == '\0')
            SACNameFormat(DFL_SAC_NAME_FMT);
        else
            SACNameFormat(args->name_format);
        return TRUE;
      case FMT_SEGY:
        Convert = SEGYConvert;
        if (args->name_format[0] == '\0')
            SEGYNameFormat(DFL_SEGY_NAME_FMT);
        else
            SEGYNameFormat(args->name_format);
        return TRUE;
      case FMT_SEED:
        Convert = MiniSEEDConvert;
        if (args->name_format[0] == '\0')
            MiniSEEDNameFormat(DFL_MINISEED_NAME_FMT);
        else
            MiniSEEDNameFormat(args->name_format);
        return TRUE;
      case FMT_SUDS:
        Convert = SUDSConvert;
        if (args->name_format[0] == '\0')
            SUDSNameFormat(DFL_SUDS_NAME_FMT);
        else
            SUDSNameFormat(args->name_format);
        return TRUE;
      case FMT_MATLAB:
        Convert = MatLabConvert;
        if (args->name_format[0] == '\0')
            MatLabNameFormat(DFL_MATLAB_NAME_FMT);
        else
            MatLabNameFormat(args->name_format);
        return TRUE;
      case FMT_STATS:
        Convert = ComputeStats;
        return TRUE;
    }

    LogMessage(LOG_ERR, "ERROR: Unknown output data format: %u\n", args->format);
    return FALSE;
}

/*---------------------------------------------------------------------*/
static BOOL OpenGRFSource(ARGS * args)
{
    BOOL result;
    CHAR string[128];
    GRF_PACKET *packet;

    ASSERT(args != NULL);

    if (args->src.type == SOURCE_SOCKET) {
        if ((args->src.des.sock.fd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
            LogMessage(LOG_ERR, "ERROR: OpenGRFSource: socket: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
            return FALSE;
        }
        args->src.des.sock.endpoint.sin_family = AF_INET;
        if (connect(args->src.des.sock.fd, (struct sockaddr *) & args->src.des.sock.endpoint, sizeof(ENDPOINT)) != 0) {
            LogMessage(LOG_ERR, "ERROR: OpenGRFSource: connect(%s): %s\n",
                FormatEndpoint(&args->src.des.sock.endpoint, string), SOCK_ERRSTRING(SOCK_ERRNO()));
            SOCK_CLOSE(args->src.des.sock.fd);
            args->src.des.sock.fd = INVALID_SOCKET;
            return FALSE;
        }
        if (!AllocateMemory((VOID **) & packet, sizeof(GRF_PACKET))) {
            LogMessage(LOG_ERR, "ERROR: OpenGRFSource: unable to allocate memory!\n");
            SOCK_CLOSE(args->src.des.sock.fd);
            args->src.des.sock.fd = INVALID_SOCKET;
            return FALSE;
        }
        if ((result = GRFConnect(args->src.des.sock.fd, &args->src.des.sock.endpoint, GRF_ATTRIB_DATA,
                    GRF_DFL_TIMEOUT, packet)) == FALSE) {
            SOCK_CLOSE(args->src.des.sock.fd);
            args->src.des.sock.fd = INVALID_SOCKET;
        }
        ReleaseMemory((VOID *) packet);
        return result;
    }
    else {
        if ((args->src.des.file = fopen(args->src.name, "rb")) == NULL) {
            LogMessage(LOG_ERR, "ERROR: OpenGRFSource: fopen(%s) failed: %s\n", args->src.name, strerror(errno));
            return FALSE;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CloseGRFSource(ARGS * args)
{
    static GRF_PACKET packet;

    ASSERT(args != NULL);

    if (args->src.type == SOURCE_SOCKET && args->src.des.sock.fd != INVALID_SOCKET) {
        GRFDisconnect(args->src.des.sock.fd, &packet);
        SOCK_CLOSE(args->src.des.sock.fd);
        args->src.des.sock.fd = INVALID_SOCKET;
    }
    else if (args->src.type == SOURCE_FILE) {
        if (args->src.des.file != NULL) {
            fclose(args->src.des.file);
            args->src.des.file = NULL;
        }
    }

    return;
}

/*---------------------------------------------------------------------*/
/*
Filename format specifiers:

    %Y  - four digit year
    %y  - two digit year
    %M	- two digit month
    %D  - two digit day of month
    %d  - three digit day of year
    %h	- two digit hour of day
    %m  - two digit minute of hour
    %s  - two digit integer second of minute
    %ms - three digit integer millisecond of second
    %us - six digit integer microsecond of second

    %l  - record length as integer seconds (variable length)
    %lm - three digit integer millisecond of record length
    %lu - six digit integer microsecond of record length

    %u  - GRF unit number (variable length)
    %c  - GRF channel number (variable length)

    %N  - GRF network identifier (<= 8 characters)
    %S  - GRF station identifier (<= 16 characters)
    %C  - GRF component identifier (<= 8 characters)

    %%	- literal '%'
*/
CHAR *OutputFilename(RECORD * record, CHAR * net, CHAR * format, CHAR * filename)
{
    BOOL escape;
    CHAR *ptr;
    INT32 year, month, day, doy, hour, minute;
    REAL64 second;
    size_t i;

    ASSERT(record != NULL);
    ASSERT(format != NULL);
    ASSERT(filename != NULL);

    DecodeUSTime(record->ist, &year, &doy, &hour, &minute, &second);
    MonthDay(&month, &day, year, doy);

    ptr = format;
    i = sprintf(filename, "%s", output_path);
    escape = FALSE;

    while (*ptr) {

        if (*ptr == '%') {
            if (escape) {
                escape = FALSE;
            }
            else {
                escape = TRUE;
                ptr++;
                continue;
            }
        }

        if (escape) {
            if (*ptr == 'Y')
                i += sprintf(&filename[i], "%04d", year);
            else if (*ptr == 'y')
                i += sprintf(&filename[i], "%02d", year % 100);
            else if (*ptr == 'M')
                i += sprintf(&filename[i], "%02d", month);
            else if (*ptr == 'D')
                i += sprintf(&filename[i], "%02d", day);
            else if (*ptr == 'd')
                i += sprintf(&filename[i], "%03d", doy);
            else if (*ptr == 'h')
                i += sprintf(&filename[i], "%02d", hour);
            else if (*ptr == 'm') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&filename[i], "%03d", (INT32) ((second - (INT32) second) * 1000));
                }
                else {
                    i += sprintf(&filename[i], "%02d", minute);
                }
            }
            else if (*ptr == 's')
                i += sprintf(&filename[i], "%02d", (INT32) second);
            else if (*ptr == 'u') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&filename[i], "%06d", (INT32) ((second - (INT32) second) * 1000000));
                }
                else {
                    i += sprintf(&filename[i], "%d", record->unit);
                }
            }
            else if (*ptr == 'c')
                i += sprintf(&filename[i], "%d", record->channel);
            else if (*ptr == 'l') {
                if (*(ptr + 1) == 'm') {
                    ptr++;
                    i += sprintf(&filename[i], "%03d", (UINT32) (((record->lst - record->ist) % UST_SECOND) / UST_MSECOND));
                }
                else if (*(ptr + 1) == 'u') {
                    ptr++;
                    i += sprintf(&filename[i], "%06d", (UINT32) ((record->lst - record->ist) % UST_SECOND));
                }
                else {
                    i += sprintf(&filename[i], "%u", (UINT32) ((record->lst - record->ist) / UST_SECOND));
                }
            }
            else if (*ptr == 'N') {
                if (net == NULL)
                    i += sprintf(&filename[i], "%s", record->name.network);
                else
                    i += sprintf(&filename[i], "%s", net);
            }
            else if (*ptr == 'S')
                i += sprintf(&filename[i], "%s", record->name.station);
            else if (*ptr == 'C')
                i += sprintf(&filename[i], "%s", record->name.component);
            else
                i += sprintf(&filename[i], "%c", *ptr);
        }
        else {
            i += sprintf(&filename[i], "%c", *ptr);
        }

        ptr++;
        escape = FALSE;
    }

    return filename;
}

/*---------------------------------------------------------------------*/
BOOL AppendFilenameToList(CHAR *filespec)
{
	CHAR *p;
	FILE *file;

	ASSERT(filespec != NULL);

	if (list_file[0] == '\0')
		return TRUE;

	if ((file = fopen(list_file, "at")) == NULL) {
		LogMessage(LOG_ERR, "ERROR: AppendFilenameToList(): fopen(%s) failed: %s\n", 
				list_file, strerror(errno));
		list_file[0] = '\0';
		return FALSE;
	}

	if ((p = strrchr(filespec, '/')) != NULL) 
		fprintf(file, "%s\n", ++p);
	else if ((p = strrchr(filespec, '\\')) != NULL) 
		fprintf(file, "%s\n", ++p);
	else 
		fprintf(file, "%s\n", filespec);

	fclose(file);

	return TRUE;
}

/*---------------------------------------------------------------------*/
VOID SetLogLevel(UINT32 level)
{

    log_level = level;
    return;
}

/*---------------------------------------------------------------------*/
VOID LogMessage(UINT32 level, CHAR * format,...)
{
    va_list marker;

    ASSERT(format != NULL);

    if (level <= log_level) {
        va_start(marker, format);
        vprintf(format, marker);
        va_end(marker);
    }

    return;
}

/*---------------------------------------------------------------------*/
VOID HexDump(UINT32 loglevel, VOID * ptr, size_t length)
{
    CHAR string[128];
    UINT8 *data;
    size_t i, j, n;

    if (ptr == NULL)
        return;

/*                        0123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789 */
    LogMessage(loglevel, "%08X  00 01 02 03  04 05 06 07  08 09 0A 0B  0C 0D 0E 0F  0123456789ABCDEF\n", length);
    LogMessage(loglevel, "--------  --------------------------------------------------  ----------------\n");
    data = (UINT8 *) ptr;
    for (n = 0; n < length; n += 16) {
        j = sprintf(string, "%08x ", n);
        for (i = n; i < n + 16; i++) {
            if (i % 4 == 0)
                j += sprintf(string + j, " ");
            if (i < length)
                j += sprintf(string + j, "%02x ", data[i]);

            else
                j += sprintf(string + j, "   ");
        }
        j += sprintf(string + j, " ");
        for (i = n; i < n + 16; i++) {
            if (i < length) {
                if (isprint(data[i]))
                    j += sprintf(string + j, "%c", data[i]);

                else
                    j += sprintf(string + j, ".");
            }
        }
        LogMessage(loglevel, "%s\n", string);
    }
    return;
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandlers(VOID)
{

#if !defined WIN32
    if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
        perror("signal(SIGTERM)");
        return FALSE;
    }
    if (signal(SIGINT, CatchSignal) == SIG_ERR) {
        perror("signal(SIGINT)");
        return FALSE;
    }
    if (signal(SIGHUP, CatchSignal) == SIG_ERR) {
        perror("signal(SIGHUP)");
        return FALSE;
    }
    /* Ignore the SIGPIPE signal */
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
        perror("signal(SIGPIPE)");
        return FALSE;
    }

#else
    return (SetConsoleCtrlHandler((PHANDLER_ROUTINE) CtrlHandler, TRUE));
#endif

    return TRUE;
}

#if defined WIN32
/*--------------------------------------------------------------------- */
static BOOL CtrlHandler(UINT32 ctrl_type)
{

    switch (ctrl_type) {
 case CTRL_C_EVENT:
        fprintf(stderr, "\rControl-c event! \n");
        shut_down = TRUE;
        break;
      case CTRL_CLOSE_EVENT:
        fprintf(stderr, "\rControl-close event! \n");
        shut_down = TRUE;
        break;
      case CTRL_BREAK_EVENT:
        fprintf(stderr, "\rControl-break event! \n");
        shut_down = TRUE;
        break;
      case CTRL_LOGOFF_EVENT:
        fprintf(stderr, "\rControl-logoff event! \n");
        shut_down = TRUE;
        break;
      case CTRL_SHUTDOWN_EVENT:
        fprintf(stderr, "\rControl-shut_down event! \n");
        shut_down = TRUE;
        break;
      default:
        fprintf(stderr, "\rUnknown control type received: %u \n", ctrl_type);
        return (FALSE);
    }

    return TRUE;
}

#else

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{

    switch (sig) {
 case SIGTERM:
        // fprintf(stderr, "\rCaught SIGTERM! \n");
        shut_down = TRUE;
        break;
      case SIGINT:
        // fprintf(stderr, "\rCaught SIGINT! \n");
        shut_down = TRUE;
        break;
      case SIGHUP:
        // fprintf(stderr, "\rCaught SIGHUP! \n");
        shut_down = TRUE;
        break;
      default:
        fprintf(stderr, "\rCaught unexpected signal %d, ignored! \n", sig);
        break;
    }

    signal(sig, CatchSignal);
}

#endif
