/* $Id: miniseed.c,v 1.5 2003/03/22 21:16:35 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2001 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Data only SEED volume (Mini-SEED) output as peer 'SEED Reference 
    Manual, Appendix G'.

-----------------------------------------------------------------------*/

#include "miniseed.h"

/* Module constants ---------------------------------------------------*/
#define SEED_REC_SIZE       4096


/* Module types -------------------------------------------------------*/
typedef struct TAG_BLK_1000 {
    UINT16 type;
    UINT16 next;
    UINT8 encoding;
    UINT8 order;
    UINT8 length;
    UINT8 reserved;
} BLK_1000;

typedef struct TAG_SEED_REC {
    DATA_HEADER hdr;
    BLK_1000 blk_1000;
    UINT8 fill[8];
    FRAME frame[63];
} SEED_REC;

/* Module globals -----------------------------------------------------*/
static CHAR name_format[MAX_PATH_LEN + 1];
static SEED_REC seed;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static VOID ConvertToBTime(BTIME *btime, USTIME ustime);
static VOID DumpSEEDRecord(SEED_REC *seed);

/*---------------------------------------------------------------------*/
VOID MiniSEEDNameFormat(CHAR *format)
{
	strcpy(name_format, format);
	return;
}

/*---------------------------------------------------------------------*/
BOOL MiniSEEDConvert(RECORD * head, CHAR *net_name)
{
	CHAR filename[MAX_PATH_LEN + 1], string[32];
	FILE *file;
	RECORD *record;
    size_t size;
    INT16 n;
    UINT32 i, sequence;
    INT32 *samples;
    USTIME ist;

	ASSERT(head != NULL);

    if(sizeof(SEED_REC) != SEED_REC_SIZE) {
        LogMessage(LOG_ERR, "ERROR: size of SEED_REC is %u, not %u, contact Banfill Software Engineering!\n", 
            sizeof(SEED_REC), SEED_REC_SIZE);
    }

	/* Output SEED records */
	for (record = head; record != NULL; record = record->next) {
		if (net_name != NULL) {
			if (strcmp(record->name.network, net_name) != 0)
				continue;
		}

	    OutputFilename(record, net_name, name_format, filename);
	    LogMessage(LOG_INFO, "%s\n", filename);
	    if ((file = fopen(filename, "w+b")) == NULL) {
		    LogMessage(LOG_ERR, "ERROR: MiniSEEDConvert: fopen(%s) failed: %s\n", filename, strerror(errno));
		    return FALSE;
	    }
        memset(&seed, 0, sizeof(SEED_REC));

        /* SEED data header values that only need to be set once */
        seed.hdr.type = 'D';
        seed.hdr.align1 = ' ';
    
        size = strlen(record->name.station);
        memset(seed.hdr.stn, ' ', 5);
        memcpy(seed.hdr.stn, record->name.station, (size > 5 ? 5 : size));

        memcpy(seed.hdr.loc, "  ", 2);

        size = strlen(record->name.component);
        memset(seed.hdr.chn, ' ', 3);
        memcpy(seed.hdr.chn, record->name.component, (size > 3 ? 3 : size));

        size = strlen(record->name.network);
        memset(seed.hdr.net, ' ', 2);
        memcpy(seed.hdr.net, record->name.network, (size > 2 ? 2 : size));

        if(record->rate < 32.0) {
            seed.hdr.rate_fac = SwapINT16((INT16)((record->rate * 1000.0) + 0.5));
            seed.hdr.rate_mul = SwapINT16(-1000);
        }
        else if(record->rate < 320.0) {
            seed.hdr.rate_fac = SwapINT16((INT16)((record->rate * 100.0) + 0.5));
            seed.hdr.rate_mul = SwapINT16(-100);
        }
        else if(record->rate < 3200.0) {
            seed.hdr.rate_fac = SwapINT16((INT16)((record->rate * 10.0) + 0.5));
            seed.hdr.rate_mul = SwapINT16(-10);
        }
        else {
            seed.hdr.rate_fac = SwapINT16((INT16)(record->rate + 0.5));
            seed.hdr.rate_mul = SwapINT16(1);
        }

        seed.hdr.act_flags = 0;
        seed.hdr.io_flags = 0;
        seed.hdr.qual_flags = 0;

        seed.hdr.total_blks = 0;

        seed.hdr.time_correct = SwapINT32((INT32)(record->timing.time_correction / 10.0));

        seed.hdr.data_start = SwapINT16(0x40);
        seed.hdr.first_blk = SwapINT16(0x30);

        /* Add Blockette 1000 */
        seed.blk_1000.type = SwapINT16(1000);
        seed.blk_1000.next = 0;
        seed.blk_1000.encoding = 10;    /* Steim 1*/
        seed.blk_1000.order = (GetByteOrder() == ENDIAN_BIG ? 1 : 0);
        seed.blk_1000.length = 12;      /* 2^12 = 4096 */
        seed.blk_1000.reserved = 0;

        sequence = 1;
        ist = record->timing.time;
        samples = record->samples;
        i = 0;

        while(i < record->n) {

            n = EncodeSteim1(samples, (INT16)(record->n - i > 4096 ? 4096 : record->n - i), &seed);

            sprintf(string, "%06lu", sequence);
            sequence++;
            if(sequence > 999999)
                sequence = 0;
            memcpy(seed.hdr.seq, string, 6);

            ConvertToBTime(&seed.hdr.start_time, ist);

            DumpSEEDRecord(&seed);

            if(fwrite(&seed, SEED_REC_SIZE, 1, file) != 1) {
		        LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		        return FALSE;
	        }

            samples += n;
            i += n;
            ist += (USTIME)(((REAL64)n / record->rate) * UST_SECOND);
        }

	    fclose(file);

		AppendFilenameToList(filename);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID ConvertToBTime(BTIME *btime, USTIME ustime)
{
    INT32 year, doy, hour, minute;
    REAL64 second;

    DecodeUSTime(ustime, &year, &doy, &hour, &minute, &second);

    btime->year = SwapINT16((UINT16)year);
    btime->day = SwapINT16((UINT16)doy);
    btime->hour = (UINT8)hour;
    btime->minute = (UINT8)minute;
    btime->seconds = (UINT8)second;
    btime->align1 = 0;
    btime->fracs = SwapINT16((UINT16)((second - (INT32)second) * 10000.0));

    return;
}

/*---------------------------------------------------------------------*/
static VOID DumpSEEDRecord(SEED_REC *seed)
{
    ASSERT(seed != NULL);

    LogMessage(LOG_DEBUG, "SEED record sequence: %.6s, type: %c\n", seed->hdr.seq, seed->hdr.type);
    LogMessage(LOG_DEBUG, " Station:        '%.5s'\n", seed->hdr.stn);
    LogMessage(LOG_DEBUG, " Location:       '%.2s'\n", seed->hdr.loc);
    LogMessage(LOG_DEBUG, " Channel:        '%.3s'\n", seed->hdr.chn);
    LogMessage(LOG_DEBUG, " Network:        '%.2s'\n", seed->hdr.net);
    LogMessage(LOG_DEBUG, " Time:            %u %03u %02u:%02u:%02u.%04u\n", 
        SwapINT16(seed->hdr.start_time.year),
        SwapINT16(seed->hdr.start_time.day),
        seed->hdr.start_time.hour,
        seed->hdr.start_time.minute,
        seed->hdr.start_time.seconds,
        SwapINT16(seed->hdr.start_time.fracs));
    LogMessage(LOG_DEBUG, " Samples:         %u\n", SwapINT16(seed->hdr.nsamps));
    LogMessage(LOG_DEBUG, " Sampling rate:   factor %d, multiplyer %d\n", 
        SwapINT16(seed->hdr.rate_fac), SwapINT16(seed->hdr.rate_mul));
    LogMessage(LOG_DEBUG, " Activity flags:  0x%02X\n", seed->hdr.act_flags);
    LogMessage(LOG_DEBUG, " I/O flags:       0x%02X\n", seed->hdr.io_flags);
    LogMessage(LOG_DEBUG, " Quality flags:   0x%02X\n", seed->hdr.qual_flags);
    LogMessage(LOG_DEBUG, " Time correction: %d\n", SwapINT32(seed->hdr.time_correct));
    LogMessage(LOG_DEBUG, " Data start:      %u\n", SwapINT16(seed->hdr.data_start));
    LogMessage(LOG_DEBUG, " First block:     %u\n", SwapINT16(seed->hdr.first_blk));

    LogMessage(LOG_DEBUG, " Blockette:       %u\n", SwapINT16(seed->blk_1000.type));
    LogMessage(LOG_DEBUG, "  Next blockette  %u\n", SwapINT16(seed->blk_1000.next));
    LogMessage(LOG_DEBUG, "  Encoding:       %u\n", seed->blk_1000.encoding);
    LogMessage(LOG_DEBUG, "  Byte order:     %u\n", seed->blk_1000.order);
    LogMessage(LOG_DEBUG, "  Length:         2^%u\n", seed->blk_1000.length);

    return;
}
