/* $Id: sac.c,v 1.3 2003/03/22 21:16:35 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	SAC binary output...

-----------------------------------------------------------------------*/

#include "sac.h"

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/
typedef struct _SAC_HEADER {				/* SAC file header (* denotes required value) */
	INT32 npts;								/* * Number of data points */
	REAL32 b;								/* * Beginning value of the independent variable */
	REAL32 e;								/* * Ending value of the independent variable */
	INT32 iftype;							/* * Type of file */
	CHAR leven;								/* * TRUE if data is evenly spaced */
	REAL32 delta;							/* * Nominal increment between samples */
	INT32 idep;								/* Type of dependent variable */
	REAL32 depmin;							/* Max value of dependent variable */
	REAL32 depmax;							/* Min value of dependent variable */
	REAL32 depmen;							/* Mean value of dependent variable */
	INT32 nzyear;							/* GMT year corresponding to zero time in file */
	INT32 nzjday;							/* GMT julian day */
	INT32 nzhour;							/* GMT hour */
	INT32 nzmin;							/* GMT minute */
	INT32 nzsec;							/* GMT second */
	INT32 nzmsec;							/* GMT millisecond */
	REAL32 cmpaz;							/* Azimuth */
	REAL32 cmpinc;							/* Incidence */
	CHAR kcmpnm[8];							/* Component name */
	CHAR knetwk[8];							/* Name of seismic network */
	CHAR kstnm[8];							/* Name of station */
	REAL32 stla;							/* Station latitude (north positive) */
	REAL32 stlo;							/* Station longitude (east positive) */
	REAL32 stel;							/* Station elevation (meters) */
} SAC_HEADER;

/* Module globals -----------------------------------------------------*/
static CHAR name_format[MAX_PATH_LEN + 1];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL OutputSAC(RECORD * record, CHAR *net_name);
static BOOL WriteSACHeader(FILE *file, RECORD * record);
static BOOL WriteSACData(FILE *file, RECORD * record);
static BOOL WriteHeader(SAC_HEADER * sac, FILE *file);
static VOID InitHeader(SAC_HEADER * sac);

/*---------------------------------------------------------------------*/
VOID SACNameFormat(CHAR *format)
{
	strcpy(name_format, format);
	return;
}

/*---------------------------------------------------------------------*/
BOOL SACConvert(RECORD * head, CHAR *net_name)
{
	RECORD *record;

	ASSERT(head != NULL);

	/* Output a SAC file per channel */
	for (record = head; record != NULL; record = record->next) {
		if (net_name != NULL) {
			if (strcmp(record->name.network, net_name) != 0)
				continue;
		}
		if (!OutputSAC(record, net_name))
			return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL OutputSAC(RECORD * record, CHAR *net_name)
{
	CHAR filename[MAX_PATH_LEN + 1];
	FILE *file;

	ASSERT(record != NULL);

	OutputFilename(record, net_name, name_format, filename);
	LogMessage(LOG_INFO, "%s\n", filename);

	if ((file = fopen(filename, "w+b")) == NULL) {
		LogMessage(LOG_ERR, "ERROR: OutputSAC: fopen(%s) failed: %s\n", filename, strerror(errno));
		return FALSE;
	}

	if (!WriteSACHeader(file, record)) {
		fclose(file);
		return FALSE;
	}

	if (!WriteSACData(file, record)) {
		fclose(file);
		return FALSE;
	}

	fclose(file);

	AppendFilenameToList(filename);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL WriteSACHeader(FILE *file, RECORD * record)
{
	INT32 year, doy, hour, minute;
	REAL64 second;
	SAC_HEADER sac;

	ASSERT(file != NULL);
	ASSERT(record != NULL);

	/* Initialize the header */
	InitHeader(&sac);

	/* Fill out the header */
	sac.iftype = 1;
	sac.leven = 1;

	sac.npts = record->n;

	sac.b = 0;
	sac.e = (REAL32)(record->n / record->rate);

	sac.delta = (REAL32)(1.0 / record->rate);

	sac.idep = 5;
	sac.depmin = -8388607.0F;
	sac.depmax = 8388607.0F;
	sac.depmen = 0;

	DecodeUSTime(record->ist, &year, &doy, &hour, &minute, &second);
	sac.nzyear = year;
	sac.nzjday = doy;
	sac.nzhour = hour;
	sac.nzmin = minute;
	sac.nzsec = (INT32)second;
	sac.nzmsec = (INT32)((second - (REAL32)((INT32)second)) * 1000.0);

	if (record->name.network[0] != '\0')
		strncpy(sac.knetwk, record->name.network, 8);
	if (record->name.station[0] != '\0')
		strncpy(sac.kstnm, record->name.station, 8);
	if (record->name.component[0] != '\0')
		strncpy(sac.kcmpnm, record->name.component, 8);

	if (!WriteHeader(&sac, file))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL WriteSACData(FILE *file, RECORD * record)
{
	UINT32 i;
	INT32 *sample;
	REAL32 fsample;

	ASSERT(file != NULL);
	ASSERT(record != NULL);

	sample = record->samples;
	for (i = 0; i < record->n; i++) {
		fsample = SwapREAL32((REAL32)sample[i]);
		if (fwrite(&fsample, 4, 1, file) != 1) {
			LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL WriteHeader(SAC_HEADER * sac, FILE *file)
{
	INT32 i_void;
	UINT32 i_row[5];
	REAL32 f_void, f_row[5];

	i_void = SwapINT32(-12345);
	f_void = SwapREAL32(-12345.0);

	/* Real values */
	/* 1 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* sac->delta, sac->depmin, sac->depmax, -12345.0, -12345.0 ); */
	f_row[0] = SwapREAL32(sac->delta);
	f_row[1] = SwapREAL32(sac->depmin);
	f_row[2] = SwapREAL32(sac->depmax);
	f_row[3] = f_void;
	f_row[4] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 2 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* sac->b, sac->e, -12345.0, -12345.0, -12345.0 ); */
	f_row[0] = SwapREAL32(sac->b);
	f_row[1] = SwapREAL32(sac->e);
	f_row[2] = f_void;
	f_row[3] = f_void;
	f_row[4] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 3 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	f_row[0] = f_void;
	f_row[1] = f_void;
	f_row[2] = f_void;
	f_row[3] = f_void;
	f_row[4] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 4 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 5 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 6 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 7 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, sac->stla, sac->stlo, sac->stel, -12345.0 ); */
	f_row[0] = f_void;
	f_row[1] = SwapREAL32(sac->stla);
	f_row[2] = SwapREAL32(sac->stlo);
	f_row[3] = SwapREAL32(sac->stel);
	f_row[4] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 8 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	f_row[0] = f_void;
	f_row[1] = f_void;
	f_row[2] = f_void;
	f_row[3] = f_void;
	f_row[4] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 9 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 10 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 11 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* 12 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, sac->depmen, -12345.0, -12345.0, -12345.0 ); */
	f_row[1] = SwapREAL32(sac->depmen);
	f_row[2] = SwapREAL32(sac->cmpaz);
	f_row[3] = SwapREAL32(sac->cmpinc);
	fwrite(f_row, 4, 5, file);

	/* 13 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	f_row[1] = f_void;
	f_row[2] = f_void;
	f_row[3] = f_void;
	fwrite(f_row, 4, 5, file);

	/* 14 */
	/* fprintf( file, "%#15.7g%#15.7g%#15.7g%#15.7g%#15.7g\n", */
	/* -12345.0, -12345.0, -12345.0, -12345.0, -12345.0 ); */
	fwrite(f_row, 4, 5, file);

	/* Integer values */
	/* 15 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* sac->nzyear, sac->nzjday, sac->nzhour, sac->nzmin, sac->nzsec ); */
	i_row[0] = SwapINT32(sac->nzyear);
	i_row[1] = SwapINT32(sac->nzjday);
	i_row[2] = SwapINT32(sac->nzhour);
	i_row[3] = SwapINT32(sac->nzmin);
	i_row[4] = SwapINT32(sac->nzsec);
	fwrite(i_row, 4, 5, file);

	/* 16 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* sac->nzmsec, 6, 0, 0, sac->npts ); */
	i_row[0] = SwapINT32(sac->nzmsec);
	i_row[1] = SwapINT32(6);
	i_row[2] = 0x00000000;
	i_row[3] = 0x00000000;
	i_row[4] = SwapINT32(sac->npts);
	fwrite(i_row, 4, 5, file);

	/* 17 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* -12345, -12345, -12345, -12345, -12345 ); */
	i_row[0] = i_void;
	i_row[1] = i_void;
	i_row[2] = i_void;
	i_row[3] = i_void;
	i_row[4] = i_void;
	fwrite(i_row, 4, 5, file);

	/* 18 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* sac->iftype, sac->idep, -12345, -12345, -12345 ); */
	i_row[0] = SwapINT32(sac->iftype);
	i_row[1] = SwapINT32(sac->idep);
	fwrite(i_row, 4, 5, file);

	/* 19 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* -12345, -12345, -12345, -12345, -12345 ); */
	i_row[0] = i_void;
	i_row[1] = i_void;
	fwrite(i_row, 4, 5, file);

	/* 20 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* -12345, -12345, -12345, -12345, -12345 ); */
	fwrite(i_row, 4, 5, file);

	/* 21 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* -12345, -12345, -12345, -12345, -12345 ); */
	fwrite(i_row, 4, 5, file);

	/* 22 */
	/* fprintf( file, "%10d%10d%10d%10d%10d\n", */
	/* sac->leven, -12345, -12345, -12345, -12345 ); */
	i_row[0] = SwapINT32(sac->leven);
	fwrite(i_row, 4, 5, file);

	/* ASCII values */
	/* 23 */
	fprintf(file, "%-8.8s%-16.16s", sac->kstnm, "-12345");
	/* 24 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", "-12345");
	/* 25 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", "-12345");
	/* 26 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", "-12345");
	/* 27 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", "-12345");
	/* 28 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", "-12345");
	/* 29 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", "-12345", "-12345", sac->kcmpnm);
	/* 30 */
	fprintf(file, "%-8.8s%-8.8s%-8.8s", sac->knetwk, "-12345", "-12345");

	if (ferror(file)) {
		LogMessage(LOG_ERR, "ERROR: write error to output file!\n");
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID InitHeader(SAC_HEADER * sac)
{

	ASSERT(sac != NULL);

	sac->npts = -12345;
	sac->b = -12345.0F;
	sac->e = -12345.0F;
	sac->iftype = -12345;
	sac->leven = 1;
	sac->delta = -12345.0F;
	sac->idep = -12345;
	sac->depmin = -12345.0F;
	sac->depmax = -12345.0F;
	sac->depmen = -12345.0F;
	sac->nzyear = -12345;
	sac->nzjday = -12345;
	sac->nzhour = -12345;
	sac->nzmin = -12345;
	sac->nzsec = -12345;
	sac->nzmsec = -12345;
	strcpy(sac->knetwk, "-12345");
	strcpy(sac->kstnm, "-12345");
	strcpy(sac->kcmpnm, "-12345");
	sac->cmpaz = -12345.0F;
	sac->cmpinc = -12345.0F;
	sac->stla = -12345.0F;
	sac->stlo = -12345.0F;
	sac->stel = -12345.0F;

	return;
}
