/* $Id: steim.c,v 1.7 2002/01/03 21:59:04 cvs Exp $
 *
 * Steim 1 compression and decompression routines for SEED 2.0.
 *
 * These routines are based on the work of many others, the compression
 * algorithm implementation is based on the work of Guy Stewart at IRIS.
 */

/* Debug messages */
/*#define DEBUG */

/* Includes ----------------------------------------------------------- */
#include <stdio.h>                          /* Standard C runtime */
#include "miniseed.h"

/* Prototypes --------------------------------------------------------- */

/* Local helper functions, these should not be called externally */
static BOOL SaveChunk(STEIM * rec, DATA_STATE * ds, CHUNK chunk, INT32 c_type);
static VOID FinishRecord(VOID * ptr, DATA_STATE * ds, UINT16 n_samp, INT32 xn);
static INT8 ChunkType(UINT32 flags, INT32 c_ndx);

/* Globals ------------------------------------------------------------ */

/* DFA transition table */
TRANSITION transition[] = {
    {_D1, 0, 0},
    {_D2, 0, 0},
    {_D4_f, 0, 0},
    {_D1_D1, 0, 1},
    {_D2_D2_f, 0, 1},
    {_D4_f, 1, -1},
    {_D2_D2_f, 0, 1},
    {_D2_D2_f, 0, 1},
    {_D4_f, 1, -1},
    {_D1_D1_D1, 0, 2},
    {_D2_D2_f, 1, -1},
    {_D2_D2_f, 1, -1},
    {_D1_D1_D1_D1_f, 0, 3},
    {_D2_D2_f, 2, -1},
    {_D2_D2_f, 2, -1}
};

/* Final state indicators */
BOOL final[] = {
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    TRUE,
    TRUE,
    TRUE
};

/*---------------------------------------------------------------------
 * Encode INT32 samples using Steim 1 compression.
 *
 *    INT16 EncodeSteim1( INT32 *samples, INT16 n_rawsamp, VOID *ptr );
 *
 * Arguments:
 *    samples   = INT32 array of uncompressed samples
 *    n_rawsamp = Number of samples in samples array
 *    ptr       = VOID pointer to a SEED 2.0 data record
 *
 * Returns the actual number of samples Steim 1 encoded in the SEED
 * data record.
 *
 * ptr should point to a 4096 byte block of memory that will contain the
 * SEED output record.  samples should be a INT32 array with at least 3780
 * elements and n_rawsamp should specify the actual number of samples
 * contained in samples starting at samples[0].
 *
 * The algorithm:
 *
 * Each Steim compressed SEED record is composed of a 64 byte fixed data
 * header followed by 63 frames of 64 bytes, each beginning with a 4 byte
 * control header and followed by 15 4 bytes "chunks".  Each of these chunks
 * can contain 4 byte values, 2 INT16 values, or 1 INT32 value, depending on
 * the 2 bit flags contained in the control header.  Generally all binary
 * data in SEED records are in big endian (68000) byte order.
 *
 * The compressor is implemented as a Deterministic Finite Automaton (DFA).
 * The transition table for the DFA is:
 *
 * note: _f signifies a final state.
 * ----------------------------------------------------------
 *                | # of    |                | # of  |
 *                | bytes   |                | DIFS  |
 *                | DIF     |                | to    | DIF
 * Current state  | fits in | New State      | unget | index
 * ---------------+---------+----------------+-------+-------
 * _START_STATE   |  1      | _D1            | 0     |  0
 * _START_STATE   |  2      | _D2            | 0     |  0
 * _START_STATE   |  4      | _D4_f          | 0     |  0
 *          _D1   |  1      | _D1_D1         | 0     |  1
 *          _D1   |  2      | _D2_D2_f       | 0     |  1
 *          _D1   |  4      | _D4_f          | 1     | -1
 *          _D2   |  1      | _D2_D2_f       | 0     |  1
 *          _D2   |  2      | _D2_D2_f       | 0     |  1
 *          _D2   |  4      | _D4_f          | 1     | -1
 *       _D1_D1   |  1      | _D1_D1_D1      | 0     |  2
 *       _D1_D1   |  2      | _D2_D2_f       | 1     | -1
 *       _D1_D1   |  4      | _D2_D2_f       | 1     | -1
 *    _D1_D1_D1   |  1      | _D1_D1_D1_D1_f | 0     |  3
 *    _D1_D1_D1   |  2      | _D2_D2_f       | 2     | -1
 *    _D1_D1_D1   |  4      | _D2_D2_f       | 2     | -1
 * ----------------------------------------------------------
 */

INT16 EncodeSteim1(INT32 * samples, INT16 n_rawsamp, VOID * ptr)
{
    INT8 token;
    UINT16 n_samp;
    INT32 state, i, dif, difs[4], c_type = 0, t_ndx;
    CHUNK chunk;
    DATA_STATE ds;
    DATA_HEADER *hdr;
    STEIM *rec;

    /* Put values in the fixed data header */
    hdr = (DATA_HEADER *) ptr;
/*
    hdr->type = 'D';
    hdr->align1 = ' ';
    hdr->total_blks = 0;
    hdr->data_start = SwapINT16(0x40);
    hdr->first_blk = 0;
*/
    /* Initialize the data state */
    ds.x0 = samples[0];                     /* Save the forward integrating constant */
    ds.xn = 0;                              /* First dif will = x-0 */
    ds.flags = 0x00000000;                  /* Clear the frame flags */
    ds.f_ndx = 0;                           /* Start at first frame, */
    ds.c_ndx = 2;                           /* 3rd chunk, the 1st and 2nd will store */
                                            /* the integrating constants */
    
    /* Initialize the sample counter and record pointer */
    n_samp = 0;
    rec = (STEIM *) ((INT8 *) ptr + 64);

    /* Initialize the DFA state */
    state = _START_STATE;

    /* Start of compression loop, compress until out of samples or the record
       is full (SaveChunk() returns FALSE). */
    for (i = 0; i < n_rawsamp; i++) {

        /* Compute the difference */
        dif = samples[i] - ds.xn;
        ds.xn = samples[i];

        /* Examine the difference */
        if (dif <= 127 && dif >= -128)
            token = 0;
        else if (dif <= 32767 && dif >= -32768)
            token = 1;
        else
            token = 2;

#ifdef DEBUG
        LogMessage(LOG_DEBUG, "state:     %d, i: %ld, sam: %ld, dif: %ld, token: %d\n",
            state, i, samples[i], dif, token);
#endif

        /* Make the transition */
        t_ndx = (state * 3) + token;

        /* Unget samples as needed */
        if (transition[t_ndx].unget) {
            i -= transition[t_ndx].unget;
            ds.xn = samples[i];
        }

        /* Save dif (if needed) in the proper slot in the buffer */
        if (transition[t_ndx].d_ndx >= 0)
            difs[transition[t_ndx].d_ndx] = dif;

        /* Get new state */
        state = transition[t_ndx].new_state;

#ifdef DEBUG
        LogMessage(LOG_DEBUG, "new state: %d, i: %ld, buf: %ld %ld %ld %ld\n",
            state, i, difs[0], difs[1], difs[2], difs[3]);
#endif

        /* Deal with final states */
        if (final[state]) {
            switch (state) {

                /* 1 INT32 difference */
              case _D4_f:
                chunk.l = difs[0];
                c_type = CHUNK_LONG;
                n_samp += 1;
                break;

                /* 2 INT16 differences */
              case _D2_D2_f:
                chunk.w[0] = (INT16) difs[0];
                chunk.w[1] = (INT16) difs[1];
                c_type = CHUNK_WORDS;
                n_samp += 2;
                break;

                /* 4 INT8 differences */
              case _D1_D1_D1_D1_f:
                chunk.b[0] = (INT8) difs[0];
                chunk.b[1] = (INT8) difs[1];
                chunk.b[2] = (INT8) difs[2];
                chunk.b[3] = (INT8) difs[3];
                c_type = CHUNK_BYTES;
                n_samp += 4;
                break;
            }

            /* Save this chunk and see if we're done */
            if (!SaveChunk(rec, &ds, chunk, c_type))
                break;

            /* Reset the state */
            state = _START_STATE;
        }
        /* End of compression loop */
    }

    /* If we ran out of data before running out of room, resolve any non-final
       state */
    if (!final[state]) {
        switch (state) {

            /* 1 INT8 or INT16 difference, save as a INT32 */
          case _D1:
          case _D2:
            chunk.l = difs[0];
            c_type = CHUNK_LONG;
            n_samp += 1;
            SaveChunk(rec, &ds, chunk, c_type);
            break;

            /* 2 INT8 differences, save as INT16S's */
          case _D1_D1:
            chunk.w[0] = (INT16) difs[0];
            chunk.w[1] = (INT16) difs[1];
            c_type = CHUNK_WORDS;
            n_samp += 2;
            SaveChunk(rec, &ds, chunk, c_type);
            break;

            /* 3 INT8 differences, save first 2 as INT16S's, 3rd as a INT32 */
          case _D1_D1_D1:
            chunk.w[0] = (INT16) difs[0];
            chunk.w[1] = (INT16) difs[1];
            c_type = CHUNK_WORDS;
            n_samp += 2;
            if (!SaveChunk(rec, &ds, chunk, c_type))
                break;
            chunk.l = difs[2];
            c_type = CHUNK_LONG;
            n_samp += 1;
            SaveChunk(rec, &ds, chunk, c_type);
            break;
        }
    }

    /* Finish out the record */
    if (n_samp > 0)
        FinishRecord(ptr, &ds, n_samp, samples[n_samp - 1]);

    /* Return the number of samples actually encoded */
    return ((INT16) n_samp);
}

/*---------------------------------------------------------------------
 * Save the current chunk, called by encode_steim().
 * Returns TRUE if there is room for another chunk, FALSE if not.
 */

static BOOL SaveChunk(STEIM * rec, DATA_STATE * ds, CHUNK chunk, INT32 c_type)
{
    INT8 tmp;

    /* Update frame flags */
    ds->flags |= c_type << (2 * (15 - (ds->c_ndx + 1)));

    if (SwappingBytes()) {
        switch (c_type) {
          case CHUNK_WORDS:
            tmp = chunk.b[0];
            chunk.b[0] = chunk.b[1];
            chunk.b[1] = tmp;
            tmp = chunk.b[2];
            chunk.b[2] = chunk.b[3];
            chunk.b[3] = tmp;
            break;
          case CHUNK_LONG:
            tmp = chunk.b[0];
            chunk.b[0] = chunk.b[3];
            chunk.b[3] = tmp;
            tmp = chunk.b[1];
            chunk.b[1] = chunk.b[2];
            chunk.b[2] = tmp;
            break;
        }
    }

    /* Save the chunk */
    rec->frame[ds->f_ndx].chunk[ds->c_ndx].l = chunk.l;

#ifdef DEBUG
    LogMessage(LOG_DEBUG, "final state, saved type %ld chunk in frame %d, chunk %d\n",
        c_type, ds->f_ndx, ds->c_ndx);
#endif

    /* Increment the chunk index */
    ds->c_ndx++;

    /* See if we need a new frame */
    if (ds->c_ndx >= 15) {

        /* Save and reset the flags */
        rec->frame[ds->f_ndx].flags = SwapINT32(ds->flags);
        ds->flags = 0x00000000;

        /* Reset the chunk index */
        ds->c_ndx = 0;

        /* Increment the frame index */
        ds->f_ndx++;

        /* See if we hit the end of the record */
        if (ds->f_ndx >= 63)
            return (FALSE);
    }

    return (TRUE);
}

/*---------------------------------------------------------------------
 * Finish off a SEED record, called by EncodeSteim1()
 */

static VOID FinishRecord(VOID * ptr, DATA_STATE * ds, UINT16 n_samp, INT32 xn)
{
    DATA_HEADER *hdr;
    STEIM *rec;

    /* Update fixed header number of samples */
    hdr = (DATA_HEADER *) ptr;
    hdr->nsamps = SwapINT16(n_samp);

    /* Save integrating constants and flags for last (possibly incomplete)
       frame */
    rec = (STEIM *) ((INT8 *) ptr + 64);
    if(ds->f_ndx < 63)
        rec->frame[ds->f_ndx].flags = SwapINT32(ds->flags);
    rec->frame[0].chunk[0].l = SwapINT32(ds->x0);
    rec->frame[0].chunk[1].l = SwapINT32(ds->xn);

    return;
}

/*---------------------------------------------------------------------
 * Decode INT32 Steim 1 compressed samples in a SEED data record.
 *
 *    INT16 decode_steim( VOID *ptr, INT32 *samples );
 *
 * Arguments:
 *    ptr     = VOID pointer to SEED data record block (4096 bytes)
 *    samples = INT32 array for uncompressed samples
 *
 * Returns the actual number of samples decoded.
 *
 * samples should be a INT32 array containing at least 3780 elements.
 */

INT16 DecodeSteim1(VOID * ptr, INT32 * samples)
{
    BOOL skip_ird;
    INT16 i, n_samp;
    INT32 f_ndx, c_ndx, sum, difs[4], fi_con, ri_con, n_difs = 0;
    UINT32 flags;
    DATA_HEADER *hdr;
    STEIM *rec;

    /* Get the number of samples from header */
    hdr = (DATA_HEADER *) ptr;
    n_samp = SwapINT16(hdr->nsamps);

    /* Get the integration constants */
    rec = (STEIM *) ((INT8 *) ptr + SwapINT16(hdr->data_start));
    fi_con = SwapINT32(rec->frame[0].chunk[0].l);
    ri_con = SwapINT32(rec->frame[0].chunk[1].l);

    /* Init the running sum and save the first sample */
    sum = fi_con;
    n_samp--;
    *samples++ = sum;

    /* Skip the inter-record delta, i.e., the difference between records */
    skip_ird = TRUE;

    /* Process up to 64 frames (usually only 63) */
    for (f_ndx = 0; f_ndx < 64; f_ndx++) {

        /* Get the frame flags */
        flags = SwapINT32(rec->frame[f_ndx].flags);

        /* Process 15 chunks per frame */
        for (c_ndx = 0; c_ndx < 15; c_ndx++) {

            /* Count down samples to 0 */
            if (n_samp <= 0)
                break;

            /* Process a chunk based on type */
            switch (ChunkType(flags, c_ndx)) {

              case CHUNK_NULL:
                continue;

              case CHUNK_BYTES:
                for (i = 0; i < 4; i++)
                    difs[i] = (INT32) rec->frame[f_ndx].chunk[c_ndx].b[i];
                n_difs = 4;
                break;

              case CHUNK_WORDS:
                for (i = 0; i < 2; i++)
                    difs[i] = (INT32) SwapINT16(rec->frame[f_ndx].chunk[c_ndx].w[i]);
                n_difs = 2;
                break;

              case CHUNK_LONG:
                difs[0] = SwapINT32(rec->frame[f_ndx].chunk[c_ndx].l);
                n_difs = 1;
                break;
            }

            /* Save the decoded samples */
            for (i = 0; i < n_difs; i++) {
                if (skip_ird)
                    skip_ird = FALSE;
                else {
                    sum += difs[i];
                    *samples++ = sum;
                    n_samp--;
                    if (n_samp <= 0)
                        break;
                }
            }
        }
    }

    if (n_samp > 0)
        LogMessage(LOG_DEBUG, "WARNING: DecodeSteim1: only partial decompression of sample data, n_samp=%d\n", n_samp);

    if (sum != ri_con)
        LogMessage(LOG_DEBUG, "WARNING: DecodeSteim1: unreconciled reverse integrating constant: %ld, sum: %ld\n",
            ri_con, sum);

    return (SwapINT16(hdr->nsamps));
}

/*---------------------------------------------------------------------
 * Returns the chunk type, called by decode_steim( ).
 */

static INT8 ChunkType(UINT32 flags, INT32 c_ndx)
{

    flags >>= (2 * (15 - (c_ndx + 1)));

    return ((INT8) (flags & 0x03));
}
