/* $Id: config.c,v 1.15 2002/10/03 18:29:06 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Configuration for grfd.

-----------------------------------------------------------------------*/

#include "config.h"
#include "log.h"
#include "mem.h"

/* Module constants ---------------------------------------------------*/
#define MAX_LINE_LENGTH					128
#define MAX_TOKEN_LENGTH				128

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL fatal_error;
static UINT32 line_number, token_number;
static CHAR *config_filespec;
static FILE *config_fptr;
static BOOL opened = FALSE;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL ParseCommandLine(int argc, char *argv[], CMDLINE_ARGS * args);
static VOID CmdLineSetDefaults(CMDLINE_ARGS * args);
static VOID CmdLineHelp(VOID);
static BOOL ParseConfigFile(ARGS * args);
static BOOL ValifyConfiguration(ARGS * args);
static BOOL DefineLogging(CMDLINE_ARGS * args);
static BOOL DefineServer(SVR_ARGS * args);
static BOOL ParseHostList(HOST_LIST ** list);
static BOOL DefineGPS(GPS_ARGS * args);
static BOOL DefineRepository(REPOS_ARGS * args);
static BOOL DefineConnections(GRF_ARGS * args);
static BOOL DefineEndpoints(GRF_ARGS * args);
static BOOL DefineFreeWaveTDMA(GRF_ARGS * args);
static BOOL DefineTDMAMaster(MASTER * master);
static BOOL DefineTDMASlaves(MASTER * master);
static BOOL DefineDigitalStreams(GRF_ARGS * args);
static BOOL DefineDSTPort(DST * dst);
static BOOL DefineDSTChannel(DST_CHANNEL * channel);
static VOID ParseChannelName(GRF_NAME * name, CHAR *token);
static CHAR *GetNextLine(VOID);
static CHAR *GetNextToken(VOID);

/*---------------------------------------------------------------------*/
#include "getopt.c"

/*---------------------------------------------------------------------*/
BOOL Configure(ARGS * args)
{
	CHAR *grf_etc, *home;
	CMDLINE_ARGS cmdline;

	ASSERT(args != NULL); 

    /* Set all command line defaults */
	CmdLineSetDefaults(&cmdline);

    /* Check for GRF_ETC environment variable */
    if((grf_etc = getenv("GRF_ETC")) != NULL) {
		if(grf_etc[0] == '~') {	
			if((home = getenv("HOME")) != NULL) {
			    if(grf_etc[strlen(grf_etc) - 1] == '/' || grf_etc[strlen(grf_etc) - 1] == '\\' )
            		sprintf(cmdline.confile, "%s%s%s", home, &grf_etc[1], DFL_CONF_FILE);
		        else
        		    sprintf(cmdline.confile, "%s%s/%s", home, &grf_etc[1], DFL_CONF_FILE);
			}
		} else {
	        if(grf_etc[strlen(grf_etc) - 1] == '/' || grf_etc[strlen(grf_etc) - 1] == '\\' )
    	        sprintf(cmdline.confile, "%s%s", grf_etc, DFL_CONF_FILE);
        	else
            	sprintf(cmdline.confile, "%s/%s", grf_etc, DFL_CONF_FILE);
		}
    }
    else
        sprintf(cmdline.confile, "%s%s", DFL_GRF_ETC, DFL_CONF_FILE); 

	/* Parse the command line to get the config filename */
	if (!ParseCommandLine(args->main.argc, args->main.argv, &cmdline))
		exit(1);

	/* Set defaults for everything... */
	CmdLineSetDefaults(&args->cmdline);
	GPSSetDefaults(&args->gps);
	GRFSetDefaults(&args->grf);
	ServerSetDefaults(&args->server);

	/* Parse the configuration file */
	strcpy(args->cmdline.confile, cmdline.confile);
	if (!ParseConfigFile(args))
		exit(1);

	/* Fold in command line options overriding config file settings */
	if (cmdline.quiet || cmdline.verbose || cmdline.debug) {
		args->cmdline.quiet = cmdline.quiet;
		args->cmdline.verbose = cmdline.verbose;
		args->cmdline.debug = cmdline.debug;
	}
	if (cmdline.daemon)
		args->cmdline.daemon = TRUE;
	if (cmdline.facility[0] != '\0')
		strcpy(args->cmdline.facility, cmdline.facility);
	if (cmdline.logfile[0] != '\0')
		strcpy(args->cmdline.logfile, cmdline.logfile);

	/* Validate the configuration */
	if (!ValifyConfiguration(args))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL DestroyLists(ARGS * args)
{
	HOST_LIST *hnode, *hlist;

	ASSERT(args != NULL);

	/* Release memory in the various host lists... */
	hlist = args->server.allow;
	while (hlist != NULL) {
		hnode = hlist;
		hlist = hlist->next;
		ReleaseMemory(hnode);
	}
	hlist = args->server.deny;
	while (hlist != NULL) {
		hnode = hlist;
		hlist = hlist->next;
		ReleaseMemory(hnode);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ParseCommandLine(int argc, char *argv[], CMDLINE_ARGS * args)
{
    BOOL confile;
	int i, option;

	ASSERT(args != NULL);
	ASSERT(argv != NULL);

	/* Don't write errors to stderr */
	my_opterr = 0;

	/* Options... */
	while ((option = my_getopt(argc, argv, "hvqdbl:s:")) != -1) {
		switch (option) {
		case 'h':
			CmdLineHelp();
			break;
		case 'v':
			args->verbose = TRUE;
			break;
		case 'q':
			args->quiet = TRUE;
			break;
		case 'd':
			args->debug = TRUE;
			break;
		case 'b':
			args->daemon = TRUE;
			break;
		case 'l':
			if (my_optarg)
				strncpy(args->logfile, my_optarg, MAX_PATH_LEN);
			else
				args->logfile[0] = '\0';
			break;
		case 's':
			if (my_optarg)
				strncpy(args->facility, my_optarg, MAX_PATH_LEN);
			else
				args->facility[0] = '\0';
			break;
		case '?':
			if (isprint(my_optopt))
				fprintf(stderr, "Unknown option `-%c'.\n", my_optopt);
			else
				fprintf(stderr, "Unknown option character `\\x%x'.\n", my_optopt);
			return FALSE;
		default:
			abort();
		}
	}

	/* Arguments... */
    confile = FALSE;
	for (i = my_optind; i < argc; i++) {
		if (confile == FALSE) {
            confile = TRUE;
			strcpy(args->confile, argv[i]);
		}
		else {
			fprintf(stderr, "unrecognized argument: `%s'\n", argv[i]);
			return FALSE;
		}
	}

	/* If we weren't passed a config filename, use the default */
	if (args->confile[0] == '\0')
		strcpy(args->confile, DFL_CONF_FILE);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CmdLineSetDefaults(CMDLINE_ARGS * args)
{
	ASSERT(args != NULL);

	args->verbose = FALSE;
	args->quiet = FALSE;
	args->debug = FALSE;
	args->daemon = FALSE;
	args->pidfile[0] = '\0';
	args->facility[0] = '\0';
	args->logfile[0] = '\0';
	args->confile[0] = '\0';

	return;
}

/*---------------------------------------------------------------------*/
static VOID CmdLineHelp(VOID)
{
#if !defined WIN32
	LogMessage(LOG_NOTICE, "Usage: grfhub [-hvqdb] [-l logfile] [-s facility] [configuration file]\n");
#else
	LogMessage(LOG_NOTICE, "Usage: grfhub [[-hvqdb] [-l logfile] [configuration file]] |\n");
    LogMessage(LOG_NOTICE, "              [-install | -remove | -start | -stop | -status]\n");
#endif
    LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "   -h           Help display.\n");
	LogMessage(LOG_NOTICE, "   -v           Verbose logging.\n");
	LogMessage(LOG_NOTICE, "   -q           Quiet logging.\n");
	LogMessage(LOG_NOTICE, "   -d           Debug logging.\n");
	LogMessage(LOG_NOTICE, "   -b           Run in background. (No) \n");
	LogMessage(LOG_NOTICE, "   -l logfile   Log to logfile. (Off)\n");
#if !defined WIN32
    LogMessage(LOG_NOTICE, "   -s facility  Log to syslog facility, `LOCAL0'-`LOCAL7'. (Off)\n");
#else
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "   -install     Install as a Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -remove      Uninstall Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -start       Start Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -stop        Stop Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -status      Report Win32 service state.\n");
#endif
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "The configuration file defaults to %s%s if not specified.\n", DFL_GRF_ETC, DFL_CONF_FILE);
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "[] = optional, () = default, | = mutually exclusive.\n");

	exit(0);
}

/*---------------------------------------------------------------------*/
static BOOL ParseConfigFile(ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	config_filespec = args->cmdline.confile;
	line_number = token_number = 0;
	fatal_error = FALSE;

	LogMessage(LOG_INFO, "Parsing configuration file: %s\n", config_filespec);

	while ((token = GetNextToken()) != NULL) {
		if (strncasecmp(token, "Daemon", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			/* "Yes", "True", or 1 */
			if (toupper(*token) == 'Y' || toupper(*token) == 'T' || *token == '1')
				args->cmdline.daemon = TRUE;
			/* Anything else is FALSE */
			else
				args->cmdline.daemon = FALSE;
		}
		else if (strncasecmp(token, "PIDFile", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->cmdline.pidfile, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "DateFormat", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			USTSetDateRepresentation(token);
		}
		else if (strncasecmp(token, "Logging", 7) == 0) {
			if (!DefineLogging(&args->cmdline))
				break;
		}
		else if (strncasecmp(token, "Server", 6) == 0) {
			if (!DefineServer(&args->server))
				break;
		}
		else if (strncasecmp(token, "GPS", 3) == 0) {
			if (!DefineGPS(&args->gps))
				break;
		}
		else if (strncasecmp(token, "Connections", 11) == 0) {
			if (!DefineConnections(&args->grf))
				break;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			break;
		}
	}

	if (opened) {
		fclose(config_fptr);
		opened = FALSE;
	}

	if (fatal_error)
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ValifyConfiguration(ARGS * args)
{
	ASSERT(args != NULL);

	/* Decide whether or not things are kosher... */

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineLogging(CMDLINE_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	/* Process statements in scope */
	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Syslog", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strcpy(args->facility, token);
		}
		else if (strncasecmp(token, "File", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strcpy(args->logfile, token);
		}
		else if (strncasecmp(token, "Level", 5) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'Q')
				args->quiet = TRUE;
			else if (toupper(*token) == 'V')
				args->verbose = TRUE;
			else if (toupper(*token) == 'D')
				args->debug = TRUE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineServer(SVR_ARGS * args)
{
	CHAR *token;
	UINT32 unit_id;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "FriendlyName", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			memset(args->name, 0, MAX_FRIENDLY_NAME_LEN + 1);
			strncpy(args->name, token, MAX_FRIENDLY_NAME_LEN);
		}
		else if (strncasecmp(token, "GRFUnitID", 9) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			unit_id = strtoul(token, NULL, 0);
			GRFSetUnitNumber(unit_id);
		}
		else if (strncasecmp(token, "Endpoint", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			ParseEndpoint(&args->listen, token, DFL_PORT);
		}
		else if (strncasecmp(token, "AllowHosts", 10) == 0) {
			if (!ParseHostList(&args->allow))
				return FALSE;
		}
		else if (strncasecmp(token, "DenyHosts", 9) == 0) {
			if (!ParseHostList(&args->deny))
				return FALSE;
		}
		else if (strncasecmp(token, "MaxConnections", 14) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->max_connections = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->queue_depth = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Repository", 10) == 0) {
			if (!DefineRepository(&args->repos))
				return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ParseHostList(HOST_LIST ** list)
{
	CHAR *token, *ptr;
	INT32 i, n;
	UINT32 netmask;
	struct hostent *hptr;
	HOST_LIST *node;

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}

		/* Allocate a new node */
		if (!AllocateMemory((void **)&node, sizeof(HOST_LIST))) {
			LogMessage(LOG_ERR, "ERROR: ParseHostList(): Unable to allocate memory!\n");
			return FALSE;
		}

		/* Get the subnet mask */
		netmask = 0xFFFFFFFF;
		if ((ptr = strchr(token, '/')) != NULL) {
			n = strtoul(ptr + 1, NULL, 0);
			if (n < 32) {
				netmask = 0;
				for (i = 0; i < n; i++) {
					netmask |= (0x00000001 << (32 - i - 1));
				}
			}
			*ptr = '\0';
		}
		node->netmask = htonl(netmask);

		/* Get the host address */
		if (isdigit(token[0])) {
			if ((node->address = inet_addr(token)) == INADDR_NONE) {
				LogMessage(LOG_ERR, "Invalid address string: %s\n", token);
			}
		}
		else if ((hptr = gethostbyname(token)) != NULL) {
			node->address = *(UINT32 *)hptr->h_addr_list[0];
		}
		else {
			LogMessage(LOG_ERR, "Unable to resolve name: %s\n", token);
		}

		/* Link into list at head */
		node->next = (*list);
		(*list) = node;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineGPS(GPS_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return FALSE;
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return FALSE;
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Port", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->device, token, GPS_MAX_DEV_LEN);
		}
		else if (strncasecmp(token, "Baudrate", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->baud = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Parity", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'N')
				args->parity = ASYNC_PARITY_NONE;
			else if (toupper(*token) == 'E')
				args->parity = ASYNC_PARITY_EVEN;
			else if (toupper(*token) == 'O')
				args->parity = ASYNC_PARITY_ODD;
		}
		else if (strncasecmp(token, "Databits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->databits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Stopbits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->stopbits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Latency", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->latency = (INT64)strtol(token, NULL, 0);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineRepository(REPOS_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Path", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->path, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "NameFormat", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->name_fmt, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "RecordLength", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->rec_length = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->queue_depth = strtoul(token, NULL, 0);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineConnections(GRF_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->con_q_depth = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Timeout", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->con_timeout = (USTIME)strtoul(token, NULL, 0) * UST_SECOND;
		}
		else if (strncasecmp(token, "Endpoints", 9) == 0) {
			if (!DefineEndpoints(args))
				return FALSE;
		}
		else if (strncasecmp(token, "DigitalStreams", 14) == 0) {
			if (!DefineDigitalStreams(args))
				return FALSE;
		}
		else if (strncasecmp(token, "FreeWaveTDMA", 12) == 0) {
			if (!DefineFreeWaveTDMA(args))
				return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineEndpoints(GRF_ARGS * args)
{
	CHAR *token;
	CONNECTION *node = NULL;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}

		if (!AllocateMemory((VOID **)&node, sizeof(CONNECTION))) {
			LogMessage(LOG_ERR, "ERROR: DefineEndpoints: Unable to allocate memory!\n");
			return FALSE;
		}

		ParseEndpoint(&node->endpoint, token, DFL_PORT);

		/* Link into list */
		node->next = args->connections;
		args->connections = node;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineFreeWaveTDMA(GRF_ARGS * args)
{
	CHAR *token;
	MASTER *node = NULL;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Master", 6) == 0) {
            if(args->n_masters >= ASYNC_MAX_PORTS) {
		        LogMessage(LOG_ERR, "Too many masters at %s line %lu\n", config_filespec, line_number);
		        fatal_error = TRUE;
		        return (FALSE);
            }
            args->n_masters++;

			if (!AllocateMemory((VOID **)&node, sizeof(MASTER))) {
				LogMessage(LOG_ERR, "ERROR: DefineFreeWaveTDMA: Unable to allocate memory!\n");
				return FALSE;
			}

			MasterSetDefaults(node);
			if (!DefineTDMAMaster(node))
				return FALSE;

			/* Link into list */
			node->next = args->master;
			args->master = node;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineTDMAMaster(MASTER * master)
{
	CHAR *token;

	ASSERT(master != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Device", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(master->args.async.device, token, MAX_ASYNC_DEV_NAME);
		}
		else if (strncasecmp(token, "BaudRate", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			master->args.async.baud_rate = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Parity", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'N')
				master->args.async.parity = ASYNC_PARITY_NONE;
			else if (toupper(*token) == 'E')
				master->args.async.parity = ASYNC_PARITY_EVEN;
			else if (toupper(*token) == 'O')
				master->args.async.parity = ASYNC_PARITY_ODD;
		}
		else if (strncasecmp(token, "DataBits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			master->args.async.data_bits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "StopBits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			master->args.async.stop_bits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "PacketLength", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			master->args.packet_length = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Slaves", 6) == 0) {
            if(!DefineTDMASlaves(master))
                return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineTDMASlaves(MASTER * master)
{
	CHAR *token;

	ASSERT(master != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else {
            if(master->args.n_slaves >= MAX_SLAVES) {
        		LogMessage(LOG_ERR, "Warning: too many slaves defined at %s line %lu\n", config_filespec, line_number);
                continue;
            }

			master->args.slaves[master->args.n_slaves] = strtoul(token, NULL, 0);
            master->args.n_slaves++;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineDigitalStreams(GRF_ARGS * args)
{
	CHAR *token;
	DST *node = NULL;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "DataType", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (strncasecmp(token, "INT32", strlen(token)) == 0)
				args->data_type = GRF_DATA_INT32;
			else if (strncasecmp(token, "INT24", strlen(token)) == 0)
				args->data_type = GRF_DATA_INT24;
			else if (strncasecmp(token, "CM8", strlen(token)) == 0)
				args->data_type = GRF_DATA_CM8;
		}
		else if (strncasecmp(token, "SamplesPerPacket", 14) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->max_samples = (UINT16)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Port", 4) == 0) {
			if (!AllocateMemory((VOID **)&node, sizeof(DST))) {
				LogMessage(LOG_ERR, "ERROR: DefineDigitalStreams: Unable to allocate memory!\n");
				return FALSE;
			}

			DSTSetDefaults(node);
			if (!DefineDSTPort(node))
				return FALSE;

			/* Link into list */
			node->next = args->dst;
			args->dst = node;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineDSTPort(DST * dst)
{
	CHAR *token;
	UINT16 channel;

	ASSERT(dst != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Device", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(dst->params.device, token, GPS_MAX_DEV_LEN);
		}
		else if (strncasecmp(token, "BaudRate", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->params.baud_rate = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Parity", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'N')
				dst->params.parity = ASYNC_PARITY_NONE;
			else if (toupper(*token) == 'E')
				dst->params.parity = ASYNC_PARITY_EVEN;
			else if (toupper(*token) == 'O')
				dst->params.parity = ASYNC_PARITY_ODD;
		}
		else if (strncasecmp(token, "DataBits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->params.data_bits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "StopBits", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->params.stop_bits = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "GRFUnitID", 9) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->unit_id = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "SamplingRate", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->rate = atof(token);
		}
		else if (strncasecmp(token, "ByteOrder", 9) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'L')
				dst->byte_order = DST_BO_LSBMSB;
			else if (toupper(*token) == 'M')
				dst->byte_order = DST_BO_MSBLSB;
		}
		else if (strncasecmp(token, "SyncByte", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			dst->sync = (UINT8)strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "AuxByte", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'T') {
				dst->aux_type = DST_AUX_TIMING;
				dst->aux = 0;
			}
			else {
				dst->aux_type = DST_AUX_FRAME_END;
				dst->aux = (UINT8)strtoul(token, NULL, 0);
			}
		}
		else if (strncasecmp(token, "Channel", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			channel = (UINT16)strtoul(token, NULL, 0);
			if (channel < 1 || channel > DST_MAX_CHANNELS) {
				LogMessage(LOG_ERR, "Channel number out-of-range at %s line %lu: '%s'\n", config_filespec, line_number, token);
				fatal_error = TRUE;
				return FALSE;
			}
			channel -= 1;
			if (!DefineDSTChannel(&dst->channel[channel]))
				return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineDSTChannel(DST_CHANNEL * channel)
{
	CHAR *token;

	ASSERT(channel != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return FALSE;
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return FALSE;
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Name", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			ParseChannelName(&channel->name, token);
			/* Activate this channel */
			channel->active = TRUE;
		}
		else if (strncasecmp(token, "CountsPerVolt", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			channel->counts_per_volt = strtol(token, NULL, 0);
		}
		else if (strncasecmp(token, "DCOffset", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			channel->dc_offset = strtol(token, NULL, 0);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID ParseChannelName(GRF_NAME * name, CHAR *token)
{
	CHAR *ptr;

	ASSERT(name != NULL);
	ASSERT(token != NULL);

	memset(name, 0, sizeof(GRF_NAME));

	/* Component... */
	if ((ptr = strrchr(token, ':')) == NULL) {
		strncpy(name->component, token, GRF_MAX_COMP_LEN);
		return;
	}
	strncpy(name->component, ptr + 1, GRF_MAX_COMP_LEN);
	*ptr = '\0';

	/* Station... */
	if ((ptr = strrchr(token, ':')) == NULL) {
		strncpy(name->station, token, GRF_MAX_STN_LEN);
		return;
	}
	strncpy(name->station, ptr + 1, GRF_MAX_STN_LEN);
	*ptr = '\0';

	/* Network */
	if (token[0] != '\0')
		strncpy(name->network, token, GRF_MAX_NET_LEN);

	return;
}

/*---------------------------------------------------------------------*/
CHAR *GetNextLine(VOID)
{
	static CHAR line_buffer[MAX_LINE_LENGTH + 1];
	BOOL quote;
	CHAR *cptr = NULL;

	if (!opened) {
		if ((config_fptr = fopen(config_filespec, "r")) == NULL) {
			LogMessage(LOG_ERR, "ERROR: fopen(%s) failed: %s\n", config_filespec, strerror(errno));
			fatal_error = TRUE;
			return (NULL);
		}
		line_number = 0;
		opened = TRUE;
	}

	/* Read and cleanup net line... */
	memset(line_buffer, '\0', MAX_LINE_LENGTH);
	while (fgets(line_buffer, MAX_LINE_LENGTH, config_fptr) != NULL) {
		line_number++;
		token_number = 0;

		/* Trim trailing crap (CR, LF, white space, etc...) */
		cptr = line_buffer + MAX_LINE_LENGTH;
		while (!(isalnum(*cptr) || ispunct(*cptr)) && cptr >= line_buffer) {
			*cptr = '\0';
			cptr--;
		}
		if (cptr < line_buffer) {
			cptr = NULL;
			continue;
		}

		/* Look for and dispose of comments anywhere in the line */
		quote = FALSE;
		cptr = line_buffer;
		while (*cptr) {
			if (*cptr == '"') {
				if (quote)
					quote = FALSE;
				else
					quote = TRUE;
			}
			if (!quote && *cptr == '#') {
				do {
					*cptr = '\0';
					if (cptr == line_buffer)
						break;
					cptr--;
				}
				while (!(isalnum(*cptr) || ispunct(*cptr)));
				break;
			}
			cptr++;
		}

		if (line_buffer[0] == '\0') {
			cptr = NULL;
			memset(line_buffer, '\0', MAX_LINE_LENGTH);
			continue;
		}

		/* Trim off leading whitespace */
		cptr = line_buffer;
		while (!(isalnum(*cptr) || ispunct(*cptr)) && *cptr != '\0')
			cptr++;

		/* Success */
		return (cptr);
	}

	/* Check for read error... */
	if (ferror(config_fptr)) {
		LogMessage(LOG_ERR, "ERROR: Reading config file: %s\n", strerror(errno));
		fatal_error = TRUE;
	}

	return (NULL);
}

/*---------------------------------------------------------------------*/
CHAR *GetNextToken(VOID)
{
	static CHAR *cptr = NULL, token[MAX_TOKEN_LENGTH + 1];
	BOOL quote;
	UINT16 i;

	quote = FALSE;

	while (TRUE) {
		if (!cptr) {
			if ((cptr = GetNextLine()) == NULL)
				break;
		}
		else {
			while (*cptr && (isalnum(*cptr) || ispunct(*cptr)))
				cptr++;
		}

		while (*cptr && !isalnum(*cptr) && !ispunct(*cptr))
			cptr++;

		if (*cptr) {
			if (*cptr == '"') {
				cptr++;
				quote = TRUE;
			}
			i = 0;
			while (i < MAX_TOKEN_LENGTH && cptr[i] && cptr[i] != ',' && (isalnum(cptr[i]) || ispunct(cptr[i]) || quote)) {
				if (quote && cptr[i] == '"') {
					cptr += i;
					break;
				}
				token[i] = cptr[i];
				i++;
			}
			token[i] = '\0';
			token_number++;
			return (token);
		}
		else
			cptr = NULL;
	}

	return (cptr);
}

/*---------------------------------------------------------------------*/
VOID DumpConfiguration(ARGS * args)
{
	CHAR string1[64], string2[64];
	HOST_LIST *node;
	CONNECTION *connection;
	DST *dst;
	MASTER *master;
	UINT16 i;

	ASSERT(args != NULL);

	LogMessage(LOG_INFO, "Configuration dump:\n");

	LogMessage(LOG_INFO, "Command line:\n");
	LogMessage(LOG_INFO, " Daemon:          %s\n", (args->cmdline.daemon ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Verbose:         %s\n", (args->cmdline.verbose ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Quiet:           %s\n", (args->cmdline.quiet ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Debug:           %s\n", (args->cmdline.debug ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Facility:        %s\n", (args->cmdline.facility[0] != '\0' ? args->cmdline.facility : "None"));
	LogMessage(LOG_INFO, " Log file:        %s\n", (args->cmdline.logfile[0] != '\0' ? args->cmdline.logfile : "None"));
	LogMessage(LOG_INFO, " Config file:     %s\n", (args->cmdline.confile[0] != '\0' ? args->cmdline.confile : "None"));

	LogMessage(LOG_INFO, "Server:\n");
	LogMessage(LOG_INFO, " Friendly name:   %s\n", args->server.name);
	LogMessage(LOG_INFO, " GRF unit ID:     %u\n", GRFGetUnitNumber());
	LogMessage(LOG_INFO, " Listen at:       %s\n", FormatEndpoint(&args->server.listen, string1));
	LogMessage(LOG_INFO, " Max connections: %u\n", args->server.max_connections);
	LogMessage(LOG_INFO, " Queue depth:     %u\n", args->server.queue_depth);
	for (node = args->server.allow; node != NULL; node = node->next) {
		LogMessage(LOG_INFO, " Allow client:    %s (%s)\n",
			inet_ntop(AF_INET, &node->address, string1, 32), inet_ntop(AF_INET, &node->netmask, string2, 32));
	}
	for (node = args->server.deny; node != NULL; node = node->next) {
		LogMessage(LOG_INFO, " Deny client:     %s (%s)\n",
			inet_ntop(AF_INET, &node->address, string1, 32), inet_ntop(AF_INET, &node->netmask, string2, 32));
	}

	if (args->server.repos.path[0] != '\0' && args->server.repos.rec_length != 0) {
		LogMessage(LOG_INFO, " Repository:\n");
		LogMessage(LOG_INFO, "  Path:           %s\n", args->server.repos.path);
		LogMessage(LOG_INFO, "  Name format:    %s\n", args->server.repos.name_fmt);
		LogMessage(LOG_INFO, "  Record length:  %u seconds\n", (UINT32)args->server.repos.rec_length);
		LogMessage(LOG_INFO, "  Queue depth:    %u\n", (UINT32)args->server.repos.queue_depth);
	}

	if (args->gps.device[0] == '\0')
		LogMessage(LOG_INFO, "GPS:              Not active\n");
	else {
		LogMessage(LOG_INFO, "GPS:\n");
		LogMessage(LOG_INFO, " Device:          %s\n", args->gps.device);
		LogMessage(LOG_INFO, " Baudrate:        %u\n", args->gps.baud);
		if (args->gps.parity == ASYNC_PARITY_NONE)
			LogMessage(LOG_INFO, " Parity:          None\n");
		else if (args->gps.parity == ASYNC_PARITY_EVEN)
			LogMessage(LOG_INFO, " Parity:          Even\n");
		else if (args->gps.parity == ASYNC_PARITY_ODD)
			LogMessage(LOG_INFO, " Parity:          Odd\n");
		LogMessage(LOG_INFO, " Data bits:       %u\n", args->gps.databits);
		LogMessage(LOG_INFO, " Stop bits:       %u\n", args->gps.stopbits);
		LogMessage(LOG_INFO, " Latency:         %d us\n", args->gps.latency);
	}

	LogMessage(LOG_INFO, "Connections:\n");
	LogMessage(LOG_INFO, " Queue depth:     %u\n", args->grf.con_q_depth);
	LogMessage(LOG_INFO, " Socket timeout:  %lf seconds\n", (REAL64)args->grf.con_timeout / (REAL64)UST_SECOND);
	for (connection = args->grf.connections; connection != NULL; connection = connection->next)
		LogMessage(LOG_INFO, " GRF Endpoint:    %s\n", FormatEndpoint(&connection->endpoint, string1));
	if (args->grf.master != NULL)
		LogMessage(LOG_INFO, "FreeWave TDMA Master Radios:\n");
	for (master = args->grf.master; master != NULL; master = master->next) {
		LogMessage(LOG_INFO, " Master device:   %s\n", AsyncFormatParams(string1, 64, &master->args.async));
		LogMessage(LOG_INFO, "  Packet length:  %u\n", master->args.packet_length);
        for(i = 0; i < master->args.n_slaves; i++)
		    LogMessage(LOG_INFO, "  Allowed slave:  %03u-%04u\n", master->args.slaves[i] / 10000, master->args.slaves[i] % 10000);
	}
	if (args->grf.dst != NULL) {
		LogMessage(LOG_INFO, "Digital streams:\n");
		if (args->grf.data_type == GRF_DATA_INT32)
			LogMessage(LOG_INFO, " Data type:      INT32\n");
		else if (args->grf.data_type == GRF_DATA_INT24)
			LogMessage(LOG_INFO, " Data type:      INT24\n");
		else if (args->grf.data_type == GRF_DATA_CM8)
			LogMessage(LOG_INFO, " Data type:      CM8 Compression\n");
		LogMessage(LOG_INFO, " Max samp/packet %u\n", args->grf.max_samples);
	}
	for (dst = args->grf.dst; dst != NULL; dst = dst->next) {
		LogMessage(LOG_INFO, " DST device:     %s\n", AsyncFormatParams(string1, 64, &dst->params));
		LogMessage(LOG_INFO, "  GRF unit ID:   %u\n", dst->unit_id);
		LogMessage(LOG_INFO, "  Sampling rate: %.2lf sps\n", dst->rate);
		LogMessage(LOG_INFO, "  ByteOrder:     %s\n", (dst->byte_order == DST_BO_MSBLSB ? "MSB/LSB" : "LSB/MSB"));
		LogMessage(LOG_INFO, "  Sync byte:     0x%02x\n", dst->sync);
		if (dst->aux_type == DST_AUX_TIMING)
			LogMessage(LOG_INFO, "  Aux byte:      Timing, bit 7 PPS\n");
		else
			LogMessage(LOG_INFO, "  Aux byte:      0x%02x\n", dst->aux);
		for (i = 0; i < DST_MAX_CHANNELS; i++) {
			if (dst->channel[i].active) {
				LogMessage(LOG_INFO, "  Channel %d: %s:%s:%s, volt: %d, dc: %d\n", i + 1,
					dst->channel[i].name.network, dst->channel[i].name.station,
					dst->channel[i].name.component, dst->channel[i].counts_per_volt, dst->channel[i].dc_offset);
			}
		}
	}

	return;
}
