/* $Id: fwtdma.c,v 1.4 2001/06/29 16:04:31 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2001 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    FreeWave TDMA packetized mode...

-----------------------------------------------------------------------*/

#include "fwtdma.h"
#include "log.h"

/* Module constants ---------------------------------------------------*/
#define RX_BUFFER_SIZE			256

/* Module types -------------------------------------------------------*/
typedef enum {
	TDMA_IDLE,
	TDMA_ID1,
	TDMA_ID2,
	TDMA_ID3,
	TDMA_LENGTH,
	TDMA_PAYLOAD
} TDMA_STATE;

typedef enum {
	GRF_IDLE,
	GRF_SIGN_G,
	GRF_SIGN_R,
	GRF_SIGN_F,
	GRF_VER,
	GRF_LENGTH,
	GRF_PAYLOAD
} GRF_STATE;

typedef struct TAG_TDMA_PORT {
	MUTEX mutex;
	TDMA_STATE state;
	UINT32 address;
	UINT8 length;
	UINT8 count;
	UINT8 packet[TDMA_MAX_SIZE];
} TDMA_PORT;

typedef struct TAG_TDMA_PACKET {
	UINT8 delimiter;						/* TDMA packet delimiter byte */
	UINT8 addr1;							/* Radio address MSB */
	UINT8 addr2;
	UINT8 addr3;							/* Radio address LSB */
	UINT8 length;							/* Payload length (TDMA packet length - 5) */
	UINT8 payload[TDMA_MAX_PAYLOAD];		/* Packet payload */
} TDMA_PACKET;

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;
static TDMA_PORT port[ASYNC_MAX_PORTS];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static VOID Initialize(VOID);

/*---------------------------------------------------------------------*/
BOOL TDMAReceive(H_ASYNC handle, UINT32 *address, VOID *data, size_t *length)
{
	UINT8 byte;
	size_t n;

	ASSERT(address != NULL);
	ASSERT(data != NULL);
	ASSERT(length != NULL);

	if (!initialized)
		Initialize();

	/* Consume and decode bytes until we have a complete TDMA packet or we timeout... */

	*length = 0;

	while (TRUE) {

		n = 1;
		if (!AsyncRead(handle, &byte, &n))
			return FALSE;
		if (n == 0)
			break;

		if (port[handle].state == TDMA_IDLE) {
			if (byte == TDMA_DELIMITER) {
				/* Start decoding a TDMA packet... */
				port[handle].state = TDMA_ID1;
				port[handle].length = 0;
				port[handle].count = 0;
			}
		}
		else if (port[handle].state == TDMA_ID1) {
			if (byte == 0) {
				/* This is a command packet... */
				port[handle].address = 0xFF000000;
			}
			else {
				/* This is data... */
				port[handle].address = (UINT32)byte << 16;
			}
			port[handle].state = TDMA_ID2;
		}
		else if (port[handle].state == TDMA_ID2) {
			if (port[handle].address == 0xFF000000)
				port[handle].length = (size_t)byte;
			else
				port[handle].address |= (UINT32)byte << 8;
			port[handle].state = TDMA_ID3;
		}
		else if (port[handle].state == TDMA_ID3) {
			if (port[handle].address == 0xFF000000)
				port[handle].state = TDMA_PAYLOAD;
			else {
				port[handle].address |= (UINT32)byte;
				port[handle].state = TDMA_LENGTH;
			}
		}
		else if (port[handle].state == TDMA_LENGTH) {
			port[handle].length = (size_t)byte;
			port[handle].state = TDMA_PAYLOAD;
		}
		else {
			if (byte == TDMA_DELIMITER) {
				port[handle].packet[port[handle].count++] = byte;
				*address = port[handle].address;
				*length = port[handle].count;
				memcpy(data, port[handle].packet, *length);
				LogMessage(LOG_ERR, "TDMAReceive: Framing error, missing bytes: %u (%u:%u)\n",
					port[handle].length - port[handle].count + 1, port[handle].length, port[handle].count);
				port[handle].state = TDMA_ID1;
				port[handle].length = 0;
				port[handle].count = 0;
				return TRUE;
			}
			port[handle].packet[port[handle].count++] = byte;
			if (port[handle].count >= port[handle].length) {
				*address = port[handle].address;
				*length = port[handle].count;
				memcpy(data, port[handle].packet, *length);
				port[handle].state = TDMA_IDLE;
				return TRUE;
			}
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL TDMATransmit(H_ASYNC handle, UINT32 address, VOID *data, size_t length)
{
	UINT8 *bytes;
	TDMA_PACKET packet;
	size_t i, n;

	ASSERT(data != NULL);

	if (!initialized)
		Initialize();

	if (length > TDMA_MAX_PAYLOAD)
		return FALSE;

	packet.delimiter = TDMA_DELIMITER;
	packet.addr1 = (UINT8)((address & 0x00FF0000) >> 16);
	packet.addr2 = (UINT8)((address & 0x0000FF00) >> 8);
	packet.addr3 = (UINT8)(address & 0x000000FF);
	packet.length = (UINT8)length;

	bytes = (UINT8 *)data;
	for (i = 0; i < packet.length; i++)
		packet.payload[i] = bytes[i];

	n = packet.length + 5;

	MUTEX_LOCK(&port[handle].mutex);

	if (!AsyncWrite(handle, &packet, &n)) {
		MUTEX_UNLOCK(&port[handle].mutex);
		return FALSE;
	}
	if (!AsyncDrain(handle)) {
		MUTEX_UNLOCK(&port[handle].mutex);
		return FALSE;
	}

	MUTEX_UNLOCK(&port[handle].mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL EncodeGRFPacket(GRF_PACKET * packet, UINT8 *data, size_t *n)
{
	UINT16 crc;
	size_t i, j, length;

	ASSERT(packet != NULL);
	ASSERT(data != NULL);
	ASSERT(n != NULL);

	length = *n;
	*n = 0;

	if (packet->hdr.length > GRF_MAX_PACKET_LEN - 2)
		return FALSE;

	/* Compute CRC of raw data and tack on end */
	crc = ComputeCRC16(packet->serial.bytes, packet->hdr.length);
	packet->serial.bytes[packet->hdr.length] = (UINT8)((crc & 0xFF00) >> 8);
	packet->serial.bytes[packet->hdr.length + 1] = (UINT8)(crc & 0x00FF);

	/* Copy and perform transparency encoding... */
	j = 0;
	for (i = 0; i < (size_t)packet->hdr.length + 2; i++) {
		if (packet->serial.bytes[i] == TDMA_DELIMITER || packet->serial.bytes[i] == TDMA_ESCAPE) {
			if (j + 2 >= length)
				return FALSE;
			data[j++] = TDMA_ESCAPE;
			data[j++] = packet->serial.bytes[i] - 2;
		}
		else {
			if (j + 1 >= length)
				return FALSE;
			data[j++] = packet->serial.bytes[i];
		}
	}
	*n = j;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL DetectGRFPacket(UNIT * unit, GRF_PACKET * packet)
{
	UINT16 crc, length, count;
	size_t index, i;
	GRF_STATE state;

	ASSERT(packet != NULL);

  restart:

	crc = 0;
	count = length = 0;
	state = GRF_IDLE;
	packet->hdr.length = 0;

	index = unit->head;
	while (index != unit->tail) {
		switch (state) {
		case GRF_IDLE:
			if (unit->queue[index] == 'G') {
				state = GRF_SIGN_G;
				crc = UpdateCRC16(0, unit->queue[index]);
			}
			else {
				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;
				goto restart;
			}
			break;
		case GRF_SIGN_G:
			if (unit->queue[index] == 'R') {
				state = GRF_SIGN_R;
				crc = UpdateCRC16(crc, unit->queue[index]);
			}
			else {
				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;
				goto restart;
			}
			break;
		case GRF_SIGN_R:
			if (unit->queue[index] == 'F') {
				state = GRF_SIGN_F;
				crc = UpdateCRC16(crc, unit->queue[index]);
			}
			else {
				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;
				goto restart;
			}
			break;
		case GRF_SIGN_F:
			if (unit->queue[index] == 1) {
				state = GRF_VER;
				crc = UpdateCRC16(crc, unit->queue[index]);
			}
			else {
				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;
				goto restart;
			}
			break;
		case GRF_VER:
			crc = UpdateCRC16(crc, unit->queue[index]);
			length = ((UINT16)unit->queue[index]) << 8;
			state = GRF_LENGTH;
			break;
		case GRF_LENGTH:
			crc = UpdateCRC16(crc, unit->queue[index]);
			length |= ((UINT16)unit->queue[index]);
			if (length > GRF_MAX_PACKET_LEN) {
				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;
				goto restart;
			}
			count = 6;
			state = GRF_PAYLOAD;
			break;
		case GRF_PAYLOAD:
			crc = UpdateCRC16(crc, unit->queue[index]);
			count++;
			if (count >= (size_t)length + 2) {
				/* We have the full packet with attached CRC */
				if (crc != 0) {
					LogMessage(LOG_DEBUG, "DetectGRFPacket: CRC failure, length: %u\n", length);

					/* CRC failed, increment the head and try again */
					state = GRF_IDLE;
					unit->head++;
					if (unit->head >= GRF_MAX_PACKET_LEN)
						unit->head = 0;
					goto restart;
				}

				index = unit->head;
				for (i = 0; i < length; i++) {
					packet->serial.bytes[i] = unit->queue[index];
					index++;
					if (index >= GRF_MAX_PACKET_LEN)
						index = 0;
				}

				state = GRF_IDLE;
				unit->head++;
				if (unit->head >= GRF_MAX_PACKET_LEN)
					unit->head = 0;

				return TRUE;
			}
			break;
		}

		index++;
		if (index >= GRF_MAX_PACKET_LEN)
			index = 0;
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static VOID Initialize(VOID)
{
	size_t i;

	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		memset(&port[i], 0, sizeof(TDMA_PORT));
		MUTEX_INIT(&port[i].mutex);
	}

	initialized = TRUE;

	return;
}
