/* $Id: gps.c,v 1.7 2002/10/02 03:18:09 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000-2001 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	This is the GPS thread.  It consumes NMEA 0183 messages and maintains
	the delta t between system time and UTC.  It also maintains the last
	reported position.

	The thread recognizes only the GPRMC (Recommended Minimum Specific
	GPS/Transit Data) message that must be supported by all NMEA 0183
	devices.

-----------------------------------------------------------------------*/
//#define GPS_UNIT_TEST

#include <platform.h>

#include "sock.h"
#include <grf.h>
#include "gps.h"
#include "log.h"
#include "threads.h"

/* Module constants ---------------------------------------------------*/
//#define AVG_DELTA         /* Uncomment to average the delta */
#define AVG_CONSTANT		100

#define MAX_MSG_LEN			128

#define NMEA_START			'$'
#define NMEA_RMC			"$GPRMC"

/* Module types -------------------------------------------------------*/
typedef enum TAG_GPS_STATE {
	GPS_STOPPED,
	GPS_CLOSED,
	GPS_UNLOCKED,
	GPS_LOCKED
} GPS_STATE;

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;
static THREAD thread_id;
static BOOL stop = FALSE;

static struct {
	MUTEX mutex;
	GPS_STATE state;
	H_ASYNC handle;
	ASYNC_PARAMS params;
	USTIME latency;

	USTIME delta;
	REAL64 latitude;
	REAL64 longitude;

	BOOL first_lock;
} gps;

static GRF_PACKET info;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC GPSThread(void *argument);
static VOID GPSInit(VOID);
static BOOL GPSOpen(VOID);
static VOID GPSClose(VOID);
static BOOL GPSReadMessage(CHAR *message);
static BOOL GPSCheckMessage(CHAR *message);
static VOID GPSProcessMessage(CHAR *message);
static CHAR *GPSNextToken(CHAR *string);
static VOID SetSystemClock(USTIME ustime);

/*---------------------------------------------------------------------*/
VOID GPSSetDefaults(GPS_ARGS * args)
{
	ASSERT(args != NULL);

	memset(args->device, 0, GPS_MAX_DEV_LEN);
	args->baud = GPS_DFL_BAUD;
	args->parity = GPS_DFL_PARITY;
	args->databits = GPS_DFL_DATABITS;
	args->stopbits = GPS_DFL_STOPBITS;
	args->latency = 0;

	return;
}

/*---------------------------------------------------------------------*/
BOOL GPSThreadStart(GPS_ARGS * args, THREAD_LIST * threads)
{
	ASSERT(args != NULL);

	if (!initialized)
		GPSInit();

	/* If we received parameters... */
	if (args != NULL) {
		MUTEX_LOCK(&gps.mutex);

		strcpy(gps.params.device, args->device);
		gps.params.baud_rate = args->baud;
		gps.params.parity = args->parity;
		gps.params.data_bits = args->databits;
		gps.params.stop_bits = args->stopbits;

		gps.latency = args->latency;

		gps.first_lock = TRUE;

		MUTEX_UNLOCK(&gps.mutex);
	}

	/* Don't start if not activated */
	if (gps.params.device[0] == '\0')
		return TRUE;

	/* Don't start more than one thread */
	MUTEX_LOCK(&gps.mutex);
	if (gps.state != GPS_STOPPED) {
		MUTEX_UNLOCK(&gps.mutex);
		return FALSE;
	}
	MUTEX_UNLOCK(&gps.mutex);

	stop = FALSE;
	if (!THREAD_CREATE(&thread_id, GPSThread, NULL)) {
		LogMessage(LOG_ERR, "ERROR: GPSThreadStart: THREAD_CREATE(GPSThread) failed: %u %s\n",
			THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}
	if (!ThreadRegister(threads, thread_id, "GPSThread"))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL GPSThreadStop(BOOL join)
{
	if (stop)
		return FALSE;

	stop = TRUE;
	if (join)
		THREAD_JOIN(thread_id);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL GPSIsLocked(VOID)
{
	BOOL locked;

	if (!initialized)
		GPSInit();

	MUTEX_LOCK(&gps.mutex);
	locked = (gps.state == GPS_LOCKED);
	MUTEX_UNLOCK(&gps.mutex);

	return locked;
}

/*---------------------------------------------------------------------*/
USTIME GPSGetDelta(BOOL *locked)
{
	USTIME delta;

	if (!initialized)
		GPSInit();

	MUTEX_LOCK(&gps.mutex);
	delta = gps.delta;
	if (locked != NULL)
		*locked = (gps.state == GPS_LOCKED);
	MUTEX_UNLOCK(&gps.mutex);

	return delta;
}

/*---------------------------------------------------------------------*/
BOOL GPSGetPosition(REAL64 *latitude, REAL64 *longitude)
{
	BOOL result;

	if (!initialized)
		GPSInit();

	ASSERT(latitude != NULL);
	ASSERT(longitude != NULL);

	result = FALSE;

	MUTEX_LOCK(&gps.mutex);
	if (gps.latitude > GPS_VOID_POSITION + 0.1) {
		*latitude = gps.latitude;
		*longitude = gps.longitude;
		result = TRUE;
	}
	MUTEX_UNLOCK(&gps.mutex);

	return result;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC GPSThread(void *argument)
{
	CHAR message[MAX_MSG_LEN + 1];

	MUTEX_LOCK(&gps.mutex);
	gps.state = GPS_CLOSED;
	MUTEX_UNLOCK(&gps.mutex);

	LogMessage(LOG_NOTICE, "GPSThread (tid:%d, pid:%d) start\n", THREAD_SELF(), getpid());

	/* Open the device */
	if (!GPSOpen())
		stop = TRUE;

	/* Process messages */
	while (!ShutDownRequested() && !stop) {
		if (GPSReadMessage(message)) {
			if (!GPSCheckMessage(message))
				LogMessage(LOG_INFO, "GPS message checksum failure!\n");
			else
				GPSProcessMessage(message);
		}
	}

	/* Close the device */
	GPSClose();

	MUTEX_LOCK(&gps.mutex);
	gps.state = GPS_STOPPED;
	MUTEX_UNLOCK(&gps.mutex);

	LogMessage(LOG_NOTICE, "GPSThread (tid:%d, pid:%d) exit\n", THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GPSCheckMessage(CHAR *message)
{
	CHAR *ptr;
	UINT8 computed;
	int checksum;
	size_t length;

	ASSERT(message != NULL);

	/* If it doesn't start with '$'... */
	if (message[0] != NMEA_START)
		return FALSE;

	/* If '*' followed by the two hex digit checksum isn't on the end... */
	length = strlen(message);
	if (length < 9)
		return FALSE;
	ptr = &message[length - 3];
	if (*ptr != '*')
		return FALSE;

	/* Trim off the '*' and the checksum */
	*ptr = '\0';
	ptr++;

	/* Decode the checksum... */
	sscanf(ptr, "%X", &checksum);

	/* Compute the checksum... */
	/* The NMEA 0183 checksum is the exclusive or of all characters excluding the '$' and '*' */
	computed = 0;
	ptr = &message[1];
	while (*ptr != '\0')
		computed ^= (UINT8)*ptr++;

	/* If it doesn't match... */
	if (computed != (UINT8)checksum)
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID GPSInit(VOID)
{
	if (initialized)
		return;

	MUTEX_INIT(&gps.mutex);
	MUTEX_LOCK(&gps.mutex);

	initialized = TRUE;

	gps.state = GPS_STOPPED;

	AsyncDefaultParams(&gps.params);
	gps.params.flow_ctrl = ASYNC_FLOW_NONE;

	gps.params.baud_rate = GPS_DFL_BAUD;
	gps.params.parity = GPS_DFL_PARITY;
	gps.params.data_bits = GPS_DFL_DATABITS;
	gps.params.stop_bits = GPS_DFL_STOPBITS;
	gps.latency = 0;

	gps.delta = 0;
	gps.latitude = GPS_VOID_POSITION;
	gps.longitude = GPS_VOID_POSITION;

	MUTEX_UNLOCK(&gps.mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL GPSOpen(VOID)
{
	CHAR string[128];

	ASSERT(initialized);

	MUTEX_LOCK(&gps.mutex);

	if (gps.state != GPS_CLOSED) {
		MUTEX_UNLOCK(&gps.mutex);
		GPSClose();
		MUTEX_LOCK(&gps.mutex);
	}

	/* Open the port */
	if ((gps.handle = AsyncOpen(&gps.params)) == ASYNC_INVALID) {
		MUTEX_UNLOCK(&gps.mutex);
		LogMessage(LOG_ERR, "GPSOpen: AsyncOpen: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(gps.handle)));
		return FALSE;
	}

	LogMessage(LOG_NOTICE, "GPS device opened: %s\n", AsyncFormatParams(string, 128, &gps.params));

	gps.state = GPS_UNLOCKED;

	MUTEX_UNLOCK(&gps.mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID GPSClose(VOID)
{
	ASSERT(initialized);

	MUTEX_LOCK(&gps.mutex);

	if (gps.state != GPS_CLOSED) {
		AsyncClose(gps.handle);
		gps.state = GPS_CLOSED;
	}

	MUTEX_UNLOCK(&gps.mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL GPSReadMessage(CHAR *message)
{
	CHAR byte, string[128];
	size_t n;
	static size_t i = 0;

	/* Read bytes and extract messages */
	while (TRUE) {
		/* Read a single byte */
		n = 1;
		if (!AsyncRead(gps.handle, &byte, &n)) {
			LogMessage(LOG_ERR, "GPSReadMessage: AsyncRead: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(gps.handle)));
			stop = TRUE;
			return FALSE;
		}

		if (n == 0) {
			/* Read timed out */
			return FALSE;
		}

		/* Start of message? */
		if (i == 0) {
			if (byte == NMEA_START) {
				message[i++] = byte;
			}
		}
		else {
			/* End of message */
			if (byte == 0x0A || byte == 0x0D) {
				message[i] = '\0';
				i = 0;
				return TRUE;
			}
			/* Within message */
			message[i++] = byte;
		}

		/* Truncate long messages */
		if (i >= MAX_MSG_LEN) {
			message[i] = '\0';
			i = 0;
			return TRUE;
		}
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static VOID GPSProcessMessage(CHAR *message)
{
	CHAR string[32], *ptr;
	INT32 year, month, day, hour, minute, latd, lond;
	REAL64 second, latm, lonm, latitude, longitude;
	USTIME ustime;
	USTIME delta, average;
	GPS_STATE state;
	static INT32 last_minute = 0, last_hour = 0;
	static BOOL marking = TRUE;

#if 0
	static USTIME last_delta = 0;
	static USTIME last_average = 0;
#endif

	ASSERT(message != NULL);

	//LogMessage(LOG_DEBUG, "GPS message: %s\n", message);

	/* We only want NMEA GPRMC messages */
	if (strncmp(NMEA_RMC, message, strlen(NMEA_RMC)) != 0) {
		LogMessage(LOG_DEBUG, "GPSThread: GPSProcessMessage: tossing non-RMC message: %s\n", message);
		return;
	}

	/* Decode time */
	if ((ptr = GPSNextToken(message)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	sscanf(ptr, "%2d%2d%lf", &hour, &minute, &second);

	if (marking) {
		AsyncSetLine(gps.handle, ASYNC_DEASSERT_DTR);
		marking = FALSE;
	}
	else if (minute != last_minute) {
		AsyncSetLine(gps.handle, ASYNC_ASSERT_DTR);
		marking = TRUE;
	}
	last_minute = minute;
	/* Status */
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	if (*ptr != 'A') {
		/* Bail out if not locked */
		MUTEX_LOCK(&gps.mutex);
		state = gps.state;
		gps.state = GPS_UNLOCKED;
		delta = gps.delta;
		MUTEX_UNLOCK(&gps.mutex);
		if (state != GPS_UNLOCKED) {
			ustime = SystemUSTime() + delta;
			LogMessage(LOG_NOTICE, "GPSThread: GPS unlocked: %s\n", FormatUSTime(string, 32, UST_DATE_TIME, ustime));
			GRFInfoPacket(&info, "GPS unlocked: %s", FormatUSTime(string, 32, UST_DATE_TIME, ustime));
		}
		return;
	}

	/* Latitude */
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	sscanf(ptr, "%2d%lf", &latd, &latm);
	latitude = (REAL64)latd + (latm / 60.0);
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	if (*ptr == 'S')
		latitude = -latitude;

	/* Longitude */
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	sscanf(ptr, "%3d%lf", &lond, &lonm);
	longitude = (REAL64)lond + (lonm / 60.0);
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	if (*ptr == 'W')
		longitude = -longitude;

	/* Skip speed and heading */
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}

	/* Date */
	if ((ptr = GPSNextToken(ptr)) == NULL) {
		LogMessage(LOG_ERR, "GPSThread: GPSProcessMessage: premature end of message: %s\n", message);
		return;
	}
	sscanf(ptr, "%2d%2d%2d", &day, &month, &year);

	/* Sanity check */
	if (year < 0 || year > 99 ||
        month < 1 || month > 12 ||
		day < 1 || day > 31 || 
        hour < 0 || hour > 23 || 
        minute < 0 || minute > 59 || 
        second < 0.0 || second > 60.0) {
		LogMessage(LOG_ERR, "GPSThread: failed decode of RMC message: %s\n", message);
		return;
	}
	year += 2000;

	ustime = EncodeUSTime(year, DayOfYear(year, month, day), hour, minute, second);
	ustime += gps.latency;

	/* Try to set the system clock on first lock... */
	if (gps.state == GPS_LOCKED && gps.first_lock) {
		SetSystemClock(ustime);
		gps.first_lock = FALSE;
	}

	delta = ustime - SystemUSTime();

#if defined AVG_DELTA
	/* Fast attack running average */
	MUTEX_LOCK(&gps.mutex);
	average = gps.delta;
	MUTEX_UNLOCK(&gps.mutex);
	if (average == 0)
		average = delta;
	else
		average += ((delta - average) / AVG_CONSTANT);
#else
	average = delta;
#endif

	MUTEX_LOCK(&gps.mutex);
	gps.delta = average;
	gps.latitude = latitude;
	gps.longitude = longitude;
	state = gps.state;
	gps.state = GPS_LOCKED;
	MUTEX_UNLOCK(&gps.mutex);

	if (state != GPS_LOCKED) {
		LogMessage(LOG_NOTICE, "GPSThread: GPS locked: %s\n", FormatUSTime(string, 32, UST_DATE_TIME, ustime));
		GRFInfoPacket(&info, "GPS locked: %s", FormatUSTime(string, 32, UST_DATE_TIME, ustime));
	}

	if (hour != last_hour) {
		last_hour = hour;
		MUTEX_LOCK(&gps.mutex);
		delta = gps.delta;
		MUTEX_UNLOCK(&gps.mutex);
		ustime = SystemUSTime() + delta;
		GPSGetPosition(&latitude, &longitude);
		LogMessage(LOG_NOTICE, "GPSThread: Position: %+.6lf %+.6lf\n", latitude, longitude);
		GRFInfoPacket(&info, "GPS position at %s, lat:%+.6lf, long:%+.6lf",
			FormatUSTime(string, 32, UST_DATE_TIME, ustime), latitude, longitude);
	}

#if 0
	LogMessage(LOG_DEBUG, "GPSThread: %s %+.6lf\n", FormatUSTime(string, 32, UST_DATE_TIME, ustime),
		(REAL64)(delta) / (REAL64)UST_SECOND);
#endif

	return;
}

/*---------------------------------------------------------------------*/
static VOID SetSystemClock(USTIME ustime)
{
#if !defined WIN32
	struct timeval tv;

	tv.tv_sec = ustime / UST_SECOND;
	tv.tv_usec = ustime % UST_SECOND;

	if (settimeofday(&tv, NULL) == 0)
		LogMessage(LOG_NOTICE, "GPSThread: system clock updated!\n");
	else
		LogMessage(LOG_NOTICE, "GPSThread: system clock update failed!\n");
#endif

	return;
}

/*---------------------------------------------------------------------*/
static CHAR *GPSNextToken(CHAR *string)
{
	CHAR *ptr;

	ASSERT(string != NULL);

	ptr = string;
	while (*ptr != ',' && *ptr != '\0')
		ptr++;

	ptr++;
	if (*ptr == '\0')
		ptr = NULL;

	return ptr;
}

/*---------------------------------------------------------------------*/
#if defined GPS_UNIT_TEST
int main(int argc, char *argv[])
{
	CHAR command[32], string[32];
	BOOL locked;
	REAL64 lat, lon;
	USTIME delta;
	GPS_ARGS args;
	THREAD_LIST threads;

	ThreadListInit(&threads);

	LogOpen(argv[0], "", "./gps.log");
	LogSetLevel(LOG_FIL | LOG_CON, LOG_DEBUG);

	LogMessage(LOG_INFO, "MainThread start (0x%x)\n", THREAD_SELF());

	GPSSetDefaults(&args);

	//strcpy(args.device, "COM2");
	strcpy(args.device, "/dev/ttyS0");
	args.baud = 4800;
	args.parity = ASYNC_PARITY_NONE;
	args.databits = 8;
	args.stopbits = 1;

	lat = lon = GPS_VOID_POSITION;
	while (gets(command) != NULL) {
		if (command[0] == 't') {
			delta = GPSGetDelta(&locked);
			LogMessage(LOG_NOTICE, "UTC: %s (%s: %d us)\n",
				FormatUSTime(string, 32, UST_DATE_TIME_USEC, SystemUSTime() + delta), (locked ? "locked" : "unlocked"), delta);
		}
		else if (command[0] == 'p') {
			GPSGetPosition(&lat, &lon);
			LogMessage(LOG_NOTICE, "GPS position: %+.6lf %+.6lf\n", lat, lon);
		}
		else if (command[0] == 'l') {
			LogMessage(LOG_NOTICE, "GPS is %s\n", (GPSIsLocked()? "locked" : "unlocked"));
		}
		else if (command[0] == 's')
			GPSThreadStart(&args, &threads);
		else if (command[0] == 'S')
			GPSThreadStop(TRUE);
		else if (command[0] == ' ')
			ThreadLogAll(&threads, LOG_NOTICE);
	}

	/* Shut down and wait for all threads... */
	RequestShutDown(TRUE);
	ThreadJoinAll(&threads);

	LogMessage(LOG_INFO, "MainThread exit (0x%x)\n", THREAD_SELF());
	LogClose();

	return 0;
}

#endif
