/* $Id: grfthread.c,v 1.8 2001/10/15 15:29:44 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    GRFThread...

-----------------------------------------------------------------------*/

#include "grfthread.h"
#include "mem.h"
#include "log.h"

/* Module constants ---------------------------------------------------*/
#define CHECK_INTERVAL	(UST_MSECOND * 200)
#define DFL_Q_DEPTH		32

/* Module types -------------------------------------------------------*/
typedef struct TAG_CLIENT_QUEUES {
	struct TAG_CLIENT_QUEUES *next;
	UINT32 host;
	UINT32 process;
	H_QUEUE queue;
} CLIENT_QUEUES;

/* Module globals -----------------------------------------------------*/
static MUTEX mutex;
static BOOL initialized = FALSE;
static BOOL running = FALSE;
static BOOL stop = FALSE;

static THREAD thread_id;
static THREAD_LIST threads;

static GRF_ARGS grf;

static MUTEX list_mutex;
static CLIENT_QUEUES *cli_list = NULL;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
/* This routine is visble to the outside world because GRFInfoPacket() in
   grf.c needs to see it */
BOOL HandlePacket(GRF_PACKET * packet);

static THREAD_FUNC GRFThread(void *argument);
static BOOL GetPacket(GRF_PACKET ** packet);
static BOOL CreateConnectionQueues(VOID);
static BOOL DestroyConnectionQueues(VOID);
static BOOL CreateConnections(VOID);
static BOOL KillConnections(VOID);
static VOID DestroyClientList(VOID);
static VOID GRFInit(VOID);
static BOOL StopRequested(VOID);

/*---------------------------------------------------------------------*/
VOID GRFSetDefaults(GRF_ARGS * args)
{
	ASSERT(args != NULL);

	if (!initialized)
		GRFInit();

	args->data_type = GRF_DATA_INT32;
	args->max_samples = GRF_MAX_INT32S;

	args->con_q_depth = DFL_Q_DEPTH;
	args->con_timeout = GRF_DFL_TIMEOUT;

	args->connections = NULL;
	args->dst = NULL;
    args->n_masters = 0;
    args->master = NULL;

	return;
}

/*---------------------------------------------------------------------*/
BOOL GRFThreadStart(GRF_ARGS * args, THREAD_LIST * threads)
{

	ASSERT(args != NULL);
	ASSERT(threads != NULL);

	if (!initialized)
		GRFInit();

	MUTEX_LOCK(&mutex);

	/* Only start one thread */
	if (running) {
		LogMessage(LOG_ERR, "ERROR: GRFThreadStart(), thread is already running!\n");
		MUTEX_UNLOCK(&mutex);
		return FALSE;
	}

	/* Save a local copy of current args */
	memcpy(&grf, args, sizeof(GRF_ARGS));

	MUTEX_UNLOCK(&mutex);

	/* Fire up the thread */
	if (!THREAD_CREATE(&thread_id, GRFThread, NULL)) {
		LogMessage(LOG_ERR, "ERROR: GRFThreadStart: THREAD_CREATE(GRFThread) failed: %u %s\n",
			THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}
	if (!ThreadRegister(threads, thread_id, "GRFThread"))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL GRFThreadStop(BOOL join)
{
	if (!initialized)
		GRFInit();

	MUTEX_LOCK(&mutex);
	stop = TRUE;
	MUTEX_UNLOCK(&mutex);

	if (join)
		THREAD_JOIN(thread_id);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL GRFThreadIsRunning(VOID)
{
	BOOL result;

	if (!initialized)
		GRFInit();

	MUTEX_LOCK(&mutex);
	result = running;
	MUTEX_UNLOCK(&mutex);

	return result;
}

/*---------------------------------------------------------------------*/
BOOL GRFRegisterClient(UINT32 host, UINT32 process, H_QUEUE queue)
{
	CHAR string[MAX_HOST_LEN + 1];
	CLIENT_QUEUES *node;

	if (!initialized)
		GRFInit();

	if (!AllocateMemory((VOID **)&node, sizeof(CLIENT_QUEUES))) {
		LogMessage(LOG_ERR, "ERROR: GRFRegisterClient: Unable to allocate memory!\n");
		return FALSE;
	}

	node->host = host;
	node->process = process;
	node->queue = queue;

	LogMessage(LOG_INFO, "GRFRegisterClient: %s\n", FormatClient(host, process, string));

	MUTEX_LOCK(&list_mutex);
	node->next = cli_list;
	cli_list = node;
	MUTEX_UNLOCK(&list_mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL GRFUnregisterClient(UINT32 host, UINT32 process)
{
	CHAR string[MAX_HOST_LEN + 1];
	CLIENT_QUEUES *node, *prev;
	UINT16 result;
	VOID *ptr = NULL;

	if (!initialized)
		GRFInit();

	MUTEX_LOCK(&list_mutex);

	/* Find and unlink the node */
	prev = NULL;
	for (node = cli_list; node != NULL; node = node->next) {
		if (node->host == host && node->process == process) {
			if (prev == NULL)
				cli_list = node->next;
			else
				prev->next = node->next;
			break;
		}
		prev = node;
	}

	MUTEX_UNLOCK(&list_mutex);

	/* If we found it, destroy it */
	if (node != NULL) {
		LogMessage(LOG_INFO, "GRFUnregisterClient: %s\n", FormatClient(node->host, node->process, string));

		/* Flush the queue */
		while ((result = QueueGet(node->queue, &ptr, FALSE)) != Q_EMPTY) {
			if (result != Q_NO_ERROR) {
				LogMessage(LOG_ERR, "ERROR: GRFThread: UnregisterClient: QueueGet: %s\n", QueueErrorString(result));
				return FALSE;
			}
			if ((result = QueuePut(grf.empties, ptr, FALSE)) != Q_NO_ERROR) {
				LogMessage(LOG_ERR, "ERROR: GRFThread: UnregisterClient: QueuePut(empties): %s\n", QueueErrorString(result));
				return FALSE;
			}
		}
		ReleaseMemory(node);
		return TRUE;
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC GRFThread(void *argument)
{
	GRF_PACKET *packet;

	MUTEX_LOCK(&mutex);
	stop = FALSE;
	running = TRUE;
	MUTEX_UNLOCK(&mutex);

	LogMessage(LOG_NOTICE, "GRFThread (tid:%d, pid:%d) start\n", THREAD_SELF(), getpid());

	if (!CreateConnectionQueues())
		RequestShutDown(TRUE);

	if (!CreateConnections())
		RequestShutDown(TRUE);

	while (!ShutDownRequested() && !StopRequested()) {
		if (GetPacket(&packet)) {
			if (!HandlePacket(packet))
				RequestShutDown(TRUE);
		}
	}

	if (!KillConnections())
		RequestShutDown(TRUE);

	if (!DestroyConnectionQueues())
		RequestShutDown(TRUE);

	DestroyClientList();

	MUTEX_LOCK(&mutex);
	running = FALSE;
	MUTEX_UNLOCK(&mutex);

	LogMessage(LOG_NOTICE, "GRFThread (tid:%d, pid:%d) exit\n", THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GetPacket(GRF_PACKET ** packet)
{
	UINT16 result;

	ASSERT(packet != NULL);

	if ((result = QueueGet(grf.con_input, (VOID **)packet, FALSE)) != Q_NO_ERROR) {
		if (result != Q_EMPTY) {
			LogMessage(LOG_ERR, "ERROR: GRFThread: GetPacket: QueueGet(input): %s\n", QueueErrorString(result));
		}
		USSleep(CHECK_INTERVAL);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL HandlePacket(GRF_PACKET * packet)
{
	CHAR string[32];
	UINT16 result;
	CLIENT_QUEUES *node;
	GRF_PACKET *copy;

	ASSERT(packet != NULL);

	/* Load packet into all active client queues... */
	MUTEX_LOCK(&list_mutex);
	for (node = cli_list; node != NULL; node = node->next) {
		/* Get an empty slot */
		if ((result = QueueGet(grf.empties, (VOID **)&copy, FALSE)) != Q_NO_ERROR) {
			if (result != Q_EMPTY) {
				LogMessage(LOG_ERR, "ERROR: GRFThread: HandlePacket: QueueGet(empties): %s\n", QueueErrorString(result));
			}
			MUTEX_UNLOCK(&list_mutex);
			return FALSE;
		}

		ASSERT(IsValidBasePointer(copy));
		memcpy(copy, packet, sizeof(GRF_PACKET));

		/* Put onto the client queue */
		if ((result = QueuePut(node->queue, (VOID *)copy, FALSE)) != Q_NO_ERROR) {
			if (result != Q_FULL) {
				LogMessage(LOG_ERR, "ERROR: GRFThread: HandlePacket: QueuePut(client:%s): %s\n",
					FormatClient(node->host, node->process, string), QueueErrorString(result));
				MUTEX_UNLOCK(&list_mutex);
				return FALSE;
			}
			else {
				/* Warn if queue is full and drop the packet */
				LogMessage(LOG_INFO, "GRFThread: Client %s queue full!\n", FormatClient(node->host, node->process, string));
				/* Put it back into the empties queue */
				if ((result = QueuePut(grf.empties, copy, FALSE)) != Q_NO_ERROR) {
					LogMessage(LOG_ERR, "ERROR: GRFThread: HandlePacket: QueueGet(empties): %s\n", QueueErrorString(result));
					MUTEX_UNLOCK(&list_mutex);
					return FALSE;
				}
			}
		}

	}
	MUTEX_UNLOCK(&list_mutex);

	if ((result = QueuePut(grf.con_empties, (VOID *)packet, FALSE)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: GRFThread: HandlePacket: QueuePut(empties): %s\n", QueueErrorString(result));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL CreateConnectionQueues(VOID)
{
	UINT16 result;
	VOID *ptr;
	size_t i;

	if ((result = QueueCreate(&grf.con_empties, grf.con_q_depth)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: GRFThread: CreateConnectionQueues: QueueCreate(empty): %s\n", QueueErrorString(result));
		return FALSE;
	}

	if ((result = QueueCreate(&grf.con_input, grf.con_q_depth)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: GRFThread: CreateConnectionQueues: QueueCreate(input): %s\n", QueueErrorString(result));
		return FALSE;
	}

	/* Allocate slot memory */
	for (i = 0; i < grf.con_q_depth; i++) {
		if (!AllocateMemory(&ptr, sizeof(GRF_PACKET))) {
			LogMessage(LOG_ERR, "ERROR: GRFThread: CreateConnectionQueue: Can't allocate memory!\n");
			return FALSE;
		}
		if ((result = QueuePut(grf.con_empties, ptr, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: GRFThread: CreateConnectionQueue: QueuePut: %s\n", QueueErrorString(result));
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DestroyConnectionQueues(VOID)
{
	UINT16 result;
	VOID *ptr;

	while ((result = QueueGet(grf.con_empties, &ptr, FALSE)) != Q_EMPTY) {
		if (result != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: GRFThread: DestroyConnectionQueues: QueueGet(empty): %s\n", QueueErrorString(result));
			return FALSE;
		}
		ReleaseMemory(ptr);
	}
	if (QueueDestroy(grf.con_empties) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: GRFThread: DestroyConnectionQueues: QueueDestroy(empty: %s\n", QueueErrorString(result));
		return FALSE;
	}

	while ((result = QueueGet(grf.con_input, &ptr, FALSE)) != Q_EMPTY) {
		if (result != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: GRFThread: DestroyConnectionQueues: QueueGet(input): %s\n", QueueErrorString(result));
			return FALSE;
		}
		ReleaseMemory(ptr);
	}
	if (QueueDestroy(grf.con_input) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: GRFThread: DestroyConnectionQueues: QueueDestroy(input): %s\n", QueueErrorString(result));
		return FALSE;
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static BOOL CreateConnections(VOID)
{
	CONNECTION *connection;
	DST *dst;
	MASTER *master;

	/* Create connections to upstream servers... */
	for (connection = grf.connections; connection != NULL; connection = connection->next) {
		connection->empties = grf.con_empties;
		connection->input = grf.con_input;
		MUTEX_INIT(&connection->mutex);
		connection->running = FALSE;
		connection->stop = FALSE;
		connection->socket = INVALID_SOCKET;
		connection->timeout = grf.con_timeout;

		if (!ConnectionThreadStart(connection))
			return FALSE;
	}

	/* Create digital stream connections... */
	for (dst = grf.dst; dst != NULL; dst = dst->next) {
		dst->empties = grf.con_empties;
		dst->input = grf.con_input;
		dst->data_type = grf.data_type;
		dst->max_samples = grf.max_samples;

		if (!DSTThreadStart(dst))
			return FALSE;
	}

	/* Create FreeWave TDMA master connections... */
	for (master = grf.master; master != NULL; master = master->next) {
		master->empties = grf.con_empties;
		master->input = grf.con_input;

		if (!MasterThreadStart(master))
			return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL KillConnections(VOID)
{
	BOOL running;
	CONNECTION *connection;
	DST *dst;
	MASTER *master;

	/* Tear down all server connections */
	while (TRUE) {
		/* Get the first connection... */
		connection = grf.connections;
		if (connection == NULL)
			break;

		/* If thread is running... */
		MUTEX_LOCK(&connection->mutex);
		running = connection->running;
		MUTEX_UNLOCK(&connection->mutex);
		if (running) {
			/* Tell it to stop... */
			MUTEX_LOCK(&connection->mutex);
			connection->stop = TRUE;
			MUTEX_UNLOCK(&connection->mutex);
			/* Wait for it to stop... */
			THREAD_JOIN(connection->thread_id);
		}

		/* Destroy resources... */
		MUTEX_DESTROY(&connection->mutex);
		grf.connections = connection->next;
		ReleaseMemory((VOID *)connection);
	}

	/* Tear down all DST connections */
	while (TRUE) {
		/* Get the first connection... */
		dst = grf.dst;
		if (dst == NULL)
			break;

		/* If thread is running... */
		MUTEX_LOCK(&dst->mutex);
		running = dst->running;
		MUTEX_UNLOCK(&dst->mutex);
		if (running) {
			/* Tell it to stop... */
			MUTEX_LOCK(&dst->mutex);
			dst->stop = TRUE;
			MUTEX_UNLOCK(&dst->mutex);
			/* Wait for it to stop... */
			THREAD_JOIN(dst->thread_id);
		}

		/* Destroy resources... */
		MUTEX_DESTROY(&dst->mutex);
		grf.dst = dst->next;
		ReleaseMemory((VOID *)dst);
	}

	/* Tear down all FreeWave connections */
	while (TRUE) {
		/* Get the first connection... */
		master = grf.master;
		if (master == NULL)
			break;

		/* If thread is running... */
		MUTEX_LOCK(&master->mutex);
		running = master->running;
		MUTEX_UNLOCK(&master->mutex);
		if (running) {
			/* Tell it to stop... */
			MUTEX_LOCK(&master->mutex);
			master->stop = TRUE;
			MUTEX_UNLOCK(&master->mutex);
			/* Wait for it to stop... */
			THREAD_JOIN(master->thread_id);
		}

		/* Destroy resources... */
		MUTEX_DESTROY(&master->mutex);
		grf.master = master->next;
		ReleaseMemory((VOID *)master);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID DestroyClientList(VOID)
{

	/* Wait for clients to unregister... */
	while (TRUE) {
		MUTEX_LOCK(&list_mutex);
		if (cli_list == NULL) {
			MUTEX_UNLOCK(&list_mutex);
			break;
		}
		MUTEX_UNLOCK(&list_mutex);
		USSleep(UST_MSECOND * 200);
	}

	return;
}

/*---------------------------------------------------------------------*/
static VOID GRFInit(VOID)
{
	ASSERT(!initialized);

	initialized = TRUE;

	MUTEX_INIT(&mutex);
	MUTEX_INIT(&list_mutex);

	ThreadListInit(&threads);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(VOID)
{
	BOOL request;

	MUTEX_LOCK(&mutex);
	request = stop;
	MUTEX_UNLOCK(&mutex);

	return request;
}
