/* $Id: main.c,v 1.12 2002/10/03 18:29:06 cvs Exp $ */
/*-----------------------------------------------------------------------

	GRFHub main module...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include "main.h"
#include "mem.h"
#include "queues.h"
#include "sock.h"
#include "config.h"
#include "log.h"
#include "threads.h"
#include "win32svc.h"

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL hupped = FALSE;

/* Module prototypes --------------------------------------------------*/
static BOOL CreateEmptiesQueue(ARGS * args);
static BOOL DestroyEmptiesQueue(ARGS * args);
static BOOL DaemonInit(VOID);
static BOOL LockProcess(CHAR *lockfile, UINT32 pid);
static BOOL UpdateProcessLock(CHAR *lockfile, UINT32 pid);
static BOOL UnlockProcess(CHAR *lockfile);
static BOOL SetSignalHandlers(VOID);
static VOID CatchSignal(int sig);

/*---------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	MAIN_ARGS args;

#if defined WIN32
    int i;
    VOID *errorString;
    SERVICE_TABLE_ENTRY ServiceTable[] = {
        { SERVICE_NAME, ServiceMain },
        { NULL,         NULL }
    };

    for (i = 0; i < argc; i++) {
        if (stricmp(argv[i], "-install") == 0) 
            return (ServiceInstall() ? 0 : 1);
        else if (stricmp(argv[i], "-remove") == 0) 
            return (ServiceRemove() ? 0 : 1);
        else if (stricmp(argv[i], "-start") == 0) 
            return (ServiceStart() ? 0 : 1);
        else if (stricmp(argv[i], "-stop") == 0) 
            return (ServiceStop() ? 0 : 1);
        else if (stricmp(argv[i], "-status") == 0) 
            return (ServiceStatus() ? 0 : 1);
        else if (stricmp(argv[i], "-service") == 0) {
            if (!StartServiceCtrlDispatcher(ServiceTable)) {
		        FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER 
                        | FORMAT_MESSAGE_FROM_SYSTEM 
                        | FORMAT_MESSAGE_IGNORE_INSERTS,
			            NULL, 
                        GetLastError(), 
                        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
                        (LPTSTR)&errorString, 
                        0, 
                        NULL
                );
		        ServiceDebugMessage("ERROR: StartServiceCtrlDispatcher(): %s", errorString);
		        LocalFree(errorString);
                return 1;
            }
            return 0;
        }
    }
#endif

	args.argc = argc;
	args.argv = argv;

	LogMessage(LOG_NOTICE, "GRFHub version %s (%s %s), pid:%d\n\n", VERSION, __DATE__, __TIME__, getpid());

	MainThread(&args);

    return 0;
}

/*--------------------------------------------------------------------- */
static BOOL DaemonInit(VOID)
{
#if !defined WIN32
	pid_t pid;

	/* Become a background process */
	LogMessage(LOG_NOTICE, "MainThread: pid %u is becoming a daemon...\n", getpid());

	if ((pid = fork()) < 0) {
		LogMessage(LOG_ERR, "MainThread: fork() failed: %s\n", strerror(errno));
		return FALSE;
	}
	else if (pid != 0)
		exit(0);							/* Parent goes bye-bye */

	/* Child continues... */
	setsid();								/* Become session leader */
	chdir("/");								/* Make the root dir current */
	umask(02);								/* rw-rw-r-- */
#endif
	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL LockProcess(CHAR *lockfile, UINT32 pid)
{
#if !defined WIN32
	CHAR string[32];
	int fd = 0;

	if (lockfile[0] == '\0')
		return TRUE;

	snprintf(string, sizeof(string), "%u\n", pid);

	//LogMessage(LOG_DEBUG, "LockProcess: pid %u attempting lock: %s\n", pid, lockfile);
	fd = open(lockfile, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd < 0) {
		if (errno != EEXIST) {
			LogMessage(LOG_ERR, "ERROR: LockProcess: open(%s) failed: %s\n", lockfile, strerror(errno));
			return FALSE;
		}
		LogMessage(LOG_ERR, "LockProcess: pid %u did not obtain lock: %s\n", pid, lockfile);
		LogMessage(LOG_ERR, "ERROR: Could not lock process. Already running?\n");
		return FALSE;
	}

	if (ftruncate(fd, 0) < 0) {
		LogMessage(LOG_ERR, "ERROR: LockProcess: ftruncate(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}
	if (write(fd, string, strlen(string)) < 0) {
		LogMessage(LOG_ERR, "ERROR: LockProcess: write(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}
	if (close(fd) < 0) {
		LogMessage(LOG_ERR, "ERROR: LockProcess: close(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}

	//LogMessage(LOG_DEBUG, "LockProcess: pid %u obtained lock: %s\n", pid, lockfile);
#endif
	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL UpdateProcessLock(CHAR *lockfile, UINT32 pid)
{
#if !defined WIN32
	CHAR string[32];
	int fd = 0;

	snprintf(string, sizeof(string), "%u\n", pid);

	if (lockfile[0] == '\0')
		return TRUE;

	//LogMessage(LOG_DEBUG, "UpdateProcessLock: pid %u attempting to update lock: %s\n", pid, lockfile);
	fd = open(lockfile, O_RDWR);
	if (fd < 0) {
		LogMessage(LOG_ERR, "ERROR: UpdateProcessLock: open(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}
	if (ftruncate(fd, 0) < 0) {
		LogMessage(LOG_ERR, "ERROR: UpdateProcessLock: ftruncate(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}
	if (write(fd, string, strlen(string)) < 0) {
		LogMessage(LOG_ERR, "ERROR: UpdateProcessLock: write(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}
	if (close(fd) < 0) {
		LogMessage(LOG_ERR, "ERROR: UpdateProcessLock: close(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}

	//LogMessage(LOG_DEBUG, "UpdateProcessLock: pid %u updated lock: %s\n", pid, lockfile);

#endif
	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL UnlockProcess(CHAR *lockfile)
{
#if !defined WIN32
	if (lockfile[0] == '\0')
		return TRUE;

	if (unlink(lockfile) < 0) {
		LogMessage(LOG_ERR, "ERROR: UnlockProcess: unlink(%s) failed: %s\n", lockfile, strerror(errno));
		return FALSE;
	}

	//LogMessage(LOG_DEBUG, "UnlockProcess: removed lock: %s\n", lockfile);
#endif
	return TRUE;
}

/*--------------------------------------------------------------------- */
THREAD_FUNC MainThread(VOID *argument)
{
	ARGS args;
	THREAD_LIST threads;

    ASSERT(argument != NULL);

	ThreadListInit(&threads);

	if (!SetSignalHandlers())
		exit(1);

	if (!SocketsInit())
		exit(1);

	args.main.argc = ((MAIN_ARGS *) argument)->argc;
	args.main.argv = ((MAIN_ARGS *) argument)->argv;

  restart:
	if (!Configure(&args))
		exit(1);

	if (!LockProcess(args.cmdline.pidfile, getpid()))
		exit(1);

	if (!LogOpen(args.main.argv[0], args.cmdline.facility, args.cmdline.logfile))
		exit(1);
	if (args.cmdline.quiet)
		LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_WARNING);
	else if (args.cmdline.verbose)
		LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_INFO);
	else if (args.cmdline.debug)
		LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_DEBUG);

#if !defined WIN32
	/* Become a daemon process... */
	if (args.cmdline.daemon) {
		if (!DaemonInit())
			exit(1);
		LogSetLevel(LOG_CON, LOG_EMERG);
		if (!UpdateProcessLock(args.cmdline.pidfile, getpid()))
			exit(1);
	}
#endif

    if (IsService()) {
      	LogMessage(LOG_NOTICE, "GRFHub version %s (%s %s), starting as a Win32 service...\n", 
                VERSION, __DATE__, __TIME__);
    }

	LogMessage(LOG_NOTICE, "MainThread (tid:%d, pid:%d) start\n", THREAD_SELF(), getpid());

	DumpConfiguration(&args);

	RequestShutDown(FALSE);

	if (!CreateEmptiesQueue(&args))
		exit(1);

	/* Start threads... */
	if (!GPSThreadStart(&args.gps, &threads))
		exit(1);
	if (!GRFThreadStart(&args.grf, &threads))
		exit(1);
	if (!ServerThreadStart(&args.server, &threads))
		exit(1);

	/* Check for shutdown and HUP's */
	while (!ShutDownRequested()) {
		USSleep(UST_SECOND);
        LogFlush();
	}

	/* Wait for all the threads we started here to shutdown... */
	ThreadJoinAll(&threads);

	if (!DestroyEmptiesQueue(&args))
		exit(1);

	if (!DestroyLists(&args))
		exit(1);

	DumpMemoryInfo(LOG_DEBUG);

	LogMessage(LOG_NOTICE, "MainThread (tid:%d, pid:%d) exit\n", THREAD_SELF(), getpid());
	LogClose();
	UnlockProcess(args.cmdline.pidfile);

	if (hupped) {
		LogMessage(LOG_NOTICE, "Restart requested...\n");
		hupped = FALSE;
		goto restart;
	}

	exit(0);
}

/*--------------------------------------------------------------------- */
static BOOL CreateEmptiesQueue(ARGS * args)
{
	UINT16 result;
	size_t i, slots, n_queues;
	H_QUEUE queue;
	VOID *ptr;

	ASSERT(args != NULL);

	/* Init the queuing subsystem */
	n_queues = args->server.max_connections + 6;
	if ((result = QueuesOpen(n_queues)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: MainThread: CreateEmptiesQueue: QueuesOpen: %s", QueueErrorString(result));
		return FALSE;
	}

	/* Create the empties queue */
	slots = (args->server.max_connections * args->server.queue_depth) + 1;
	if ((result = QueueCreate(&queue, slots)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: MainThread: CreateEmptiesQueue: QueueCreate: %s\n", QueueErrorString(result));
		return FALSE;
	}

	/* Allocate slot memory */
	for (i = 0; i < slots; i++) {
		if (!AllocateMemory(&ptr, sizeof(GRF_PACKET))) {
			LogMessage(LOG_ERR, "ERROR: MainThread: CreateEmptiesQueue: Can't allocate memory!\n");
			return FALSE;
		}
		if ((result = QueuePut(queue, ptr, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: MainThread: CreateEmptiesQueue: QueuePut: %s\n", QueueErrorString(result));
			return FALSE;
		}
	}

	args->server.empties = queue;
	args->grf.empties = queue;

	LogMessage(LOG_DEBUG, "MainThread: Created empties queue\n");

	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL DestroyEmptiesQueue(ARGS * args)
{
	UINT16 result;
	H_QUEUE queue;
	VOID *ptr = NULL;

	ASSERT(args != NULL);

	queue = args->server.empties;

	/* Empty the queue... */
	while ((result = QueueGet(queue, &ptr, FALSE)) != Q_EMPTY) {
		if (result != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: MainThread: DestroyEmptiesQueue: QueueGet: %s\n", QueueErrorString(result));
			return FALSE;
		}
		ReleaseMemory(ptr);
	}

	/* Destroy the queue */
	if (QueueDestroy(queue) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: MainThread: DestroyEmptiesQueue: QueueDestroy: %s\n", QueueErrorString(result));
		return FALSE;
	}

	/* Shut down the queuing subsystem */
	if ((result = QueuesClose()) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: MainThread: DestroyEmptiesQueue: QueuesClose: %s", QueueErrorString(result));
		return FALSE;
	}

	LogMessage(LOG_DEBUG, "MainThread: Destroyed empties queue\n");

	return TRUE;
}

/*--------------------------------------------------------------------- */
VOID HexDump(UINT32 loglevel, VOID *ptr, size_t length)
{
	CHAR string[128];
	UINT8 *data;
	size_t i, j, n;

	if (ptr == NULL)
		return;

/*                        0123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789 */
	LogMessage(loglevel, "%08X  00 01 02 03  04 05 06 07  08 09 0A 0B  0C 0D 0E 0F  0123456789ABCDEF\n", length);
	LogMessage(loglevel, "--------  --------------------------------------------------  ----------------\n");
	data = (UINT8 *)ptr;
	for (n = 0; n < length; n += 16) {
		j = sprintf(string, "%08x ", n);
		for (i = n; i < n + 16; i++) {
			if (i % 4 == 0)
				j += sprintf(string + j, " ");
			if (i < length)
				j += sprintf(string + j, "%02x ", data[i]);

			else
				j += sprintf(string + j, "   ");
		}
		j += sprintf(string + j, " ");
		for (i = n; i < n + 16; i++) {
			if (i < length) {
				if (isprint(data[i]))
					j += sprintf(string + j, "%c", data[i]);

				else
					j += sprintf(string + j, ".");
			}
		}
		LogMessage(loglevel, "%s\n", string);
	}
	return;
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandlers(VOID)
{
	if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
		perror("signal(SIGTERM)");
		return FALSE;
	}
	if (signal(SIGINT, CatchSignal) == SIG_ERR) {
		perror("signal(SIGINT)");
		return FALSE;
	}
#if !defined WIN32
	if (signal(SIGHUP, CatchSignal) == SIG_ERR) {
		perror("signal(SIGHUP)");
		return FALSE;
	}
	/* Ignore the SIGPIPE signal */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal(SIGPIPE)");
		return FALSE;
	}
#endif

	return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{
	switch (sig) {
	case SIGTERM:
		//fprintf(stderr, "\rCaught SIGTERM! \n");
		RequestShutDown(TRUE);
		break;
	case SIGINT:
		//fprintf(stderr, "\rCaught SIGINT! \n");
		RequestShutDown(TRUE);
		break;
#if !defined WIN32
	case SIGHUP:
		//fprintf(stderr, "\rCaught SIGHUP! \n");
		if (!hupped)
			hupped = TRUE;
		RequestShutDown(TRUE);
		break;
#endif
	default:
		fprintf(stderr, "\rCaught unexpected signal %d, ignored! \n", sig);
		break;
	}

	signal(sig, CatchSignal);
}
