/* $Id: master.c,v 1.7 2001/07/12 15:37:56 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2001 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Master TDMA mode routines...

-----------------------------------------------------------------------*/

#include "main.h"

/* Module constants ---------------------------------------------------*/
#define RX_BUFFER_SIZE 256

/* Module types -------------------------------------------------------*/
typedef enum {
	GRF_IDLE,
	GRF_SIGN_G,
	GRF_SIGN_R,
	GRF_SIGN_F,
	GRF_VER,
	GRF_LENGTH,
	GRF_PAYLOAD
} GRF_STATE;

typedef struct TAG_TDMA_MASTER {
	MUTEX mutex;
	size_t packet_length;
	struct {
		UINT8 buffer[RX_BUFFER_SIZE];
		size_t n_units;
		UNIT unit[MAX_SLAVES];
	} rx;
	struct {
		USTIMER outbound;					/* To measure outbound packet rate */
		REAL64 mean_outbound;
		UINT8 buffer[GRF_MAX_PACKET_LEN * 2];
		UINT16 sequence;
		USTIME timeout;
		GRF_PACKET ack;
	} tx;
} TDMA_MASTER;

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;
static TDMA_MASTER master[ASYNC_MAX_PORTS];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL Acknowledge(H_ASYNC handle, GRF_PACKET * packet, UINT32 address);
static VOID Initialize(VOID);

/*---------------------------------------------------------------------*/
VOID TDMAMasterSetParams(H_ASYNC handle, MASTER_ARGS * args)
{
    size_t i;
	UNIT *unit;

	if (!initialized)
		Initialize();

	master[handle].packet_length = args->packet_length;
	master[handle].tx.timeout = args->timeout;

    for(i = 0; i < args->n_slaves; i++) {
		unit = &master[handle].rx.unit[i];
		master[handle].rx.n_units++;

		unit->address = args->slaves[i];
		unit->escape = FALSE;
		unit->head = unit->tail = 0;
		unit->sequence = 0;
		unit->started = FALSE;
    }

	return;
}

/*---------------------------------------------------------------------*/
BOOL TDMAMasterWrite(H_ASYNC handle, GRF_PACKET * packet, UINT32 address)
{
	CHAR string[128];
	UINT8 *ptr;
	REAL64 elapsed;
	size_t n, remaining, length;

	ASSERT(packet != NULL);

	if (!initialized)
		Initialize();

//  MUTEX_LOCK(&master[handle].mutex);

	packet->hdr.sequence = master[handle].tx.sequence;
	master[handle].tx.sequence++;

	SerializePacket(packet);

	n = GRF_MAX_PACKET_LEN * 2;
	if (!EncodeGRFPacket(packet, master[handle].tx.buffer, &n)) {
//      MUTEX_UNLOCK(&master[handle].mutex);
		LogMessage(LOG_ERR, "TDMAMasterWrite: EncodeGRFPacket failed, encode buffer too small!\n");
		return FALSE;
	}

	elapsed = (REAL64)USTimerElapsed(&master[handle].tx.outbound) / (REAL64)UST_SECOND;
	StartUSTimer(&master[handle].tx.outbound, 0);
	if (master[handle].tx.mean_outbound < 0.0)
		master[handle].tx.mean_outbound = elapsed;
	else
		master[handle].tx.mean_outbound += (elapsed - master[handle].tx.mean_outbound) / RATE_AVG_CONSTANT;
	LogMessage(LOG_DEBUG, "Tx %u:%05u to %u, delta: %.3lf (%.3lf)\n",
		packet->hdr.unit, packet->hdr.sequence, address, elapsed, master[handle].tx.mean_outbound);

	ptr = master[handle].tx.buffer;
	remaining = n;
	while (remaining > 0) {
		if (remaining > master[handle].packet_length)
			length = master[handle].packet_length;
		else
			length = remaining;

		if (!TDMATransmit(handle, address, ptr, length)) {
//          MUTEX_UNLOCK(&master[handle].mutex);
			LogMessage(LOG_ERR, "TDMAMasterWrite: AsyncWrite failed: %s\n",
				AsyncErrorString(string, 128, AsyncErrorCode(handle)));
			return FALSE;
		}

		ptr += length;
		remaining -= length;
	}

//  MUTEX_UNLOCK(&master[handle].mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL TDMAMasterRead(H_ASYNC handle, GRF_PACKET * packet, UINT32 *address)
{
	CHAR string[128];
	size_t i, n;
	UINT32 rx_addr;
	REAL64 elapsed;
	UNIT *unit;

	ASSERT(packet != NULL);
	ASSERT(address != NULL);

	if (!initialized)
		Initialize();

	*address = TDMA_VOID_ADDR;

	while (TRUE) {
		n = RX_BUFFER_SIZE;
		if (!TDMAReceive(handle, &rx_addr, master[handle].rx.buffer, &n)) {
			LogMessage(LOG_ERR, "TDMAMasterRead: TDMAReceive: AsyncRead failed: %s\n",
				AsyncErrorString(string, 128, AsyncErrorCode(handle)));
			return FALSE;
		}
		if (n == 0)
			break;

#if 0
		LogMessage(LOG_DEBUG, "TDMA Rx %u bytes from %u\n", n, rx_addr);
		HexDump(LOG_DEBUG, master[handle].rx.buffer, n);
#endif

		/* Find this unit based on rx_addr */
		unit = NULL;
		for (i = 0; i < master[handle].rx.n_units; i++) {
			if (master[handle].rx.unit[i].address == rx_addr) {
				unit = &master[handle].rx.unit[i];
				break;
			}
		}
        if (unit == NULL) {
    		LogMessage(LOG_DEBUG, "Ignoring %u byte TDMA packet from %u\n", n, rx_addr);
            break;
        }

		/* Decode into this units byte queue... */
		i = 0;
		while (i < n) {
			if (unit->escape) {
				unit->escape = FALSE;
				unit->queue[unit->tail++] = master[handle].rx.buffer[i] + 2;
			}
			else if (master[handle].rx.buffer[i] == TDMA_ESCAPE)
				unit->escape = TRUE;
			else
				unit->queue[unit->tail++] = master[handle].rx.buffer[i];

			if (unit->tail >= GRF_MAX_PACKET_LEN)
				unit->tail = 0;
			i++;
		}

		if (DetectGRFPacket(unit, packet)) {
			DeserializePacket(packet);

			if (unit->started) {
				/* This one we're waiting for... */
				if (packet->hdr.sequence == unit->sequence) {
					Acknowledge(handle, packet, rx_addr);
					unit->sequence = packet->hdr.sequence + 1;
					*address = unit->address;
				}
				/* Old packet */
				else if (GRF_SEQ_GTEQ(packet->hdr.sequence, unit->sequence - TX_Q_DEPTH) &&
					GRF_SEQ_LT(packet->hdr.sequence, unit->sequence)) {
					LogMessage(LOG_DEBUG, "Old packet %u:%05u from %u, expected %u\n",
						packet->hdr.unit, packet->hdr.sequence, unit->address, unit->sequence);
					Acknowledge(handle, packet, rx_addr);
					break;
				}
				/* Inside our sequence space... */
				else if (GRF_SEQ_GT(packet->hdr.sequence, unit->sequence) &&
					GRF_SEQ_LTEQ(packet->hdr.sequence, unit->sequence + TX_Q_DEPTH)) {
					LogMessage(LOG_DEBUG, "Premature packet %u:%05u from %u, expected %u\n",
						packet->hdr.unit, packet->hdr.sequence, unit->address, unit->sequence);
					Acknowledge(handle, packet, rx_addr);
					break;
				}
				/* Outside of our sequence space */
				else {
					LogMessage(LOG_DEBUG, "Invalid packet %u:%05u from %u, expected %u, restart!\n",
						packet->hdr.unit, packet->hdr.sequence, rx_addr, unit->sequence);
					Acknowledge(handle, packet, rx_addr);
					unit->sequence = packet->hdr.sequence + 1;
					break;
				}
			}
			else {
				Acknowledge(handle, packet, rx_addr);
				*address = unit->address;
				unit->sequence = packet->hdr.sequence + 1;
				unit->started = TRUE;
			}

			elapsed = (REAL64)USTimerElapsed(&unit->inbound) / (REAL64)UST_SECOND;
			StartUSTimer(&unit->inbound, 0);
			if (unit->mean_inbound < 0.0)
				unit->mean_inbound = elapsed;
			else
				unit->mean_inbound += (elapsed - unit->mean_inbound) / RATE_AVG_CONSTANT;
			LogMessage(LOG_DEBUG, "Rx %u:%05u, %u, from %u, delta: %.3lf (%.3lf)\n",
				packet->hdr.unit, packet->hdr.sequence, packet->hdr.length, rx_addr, elapsed, unit->mean_inbound);

			break;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL Acknowledge(H_ASYNC handle, GRF_PACKET * packet, UINT32 address)
{
	CHAR string[128];
	UINT8 *ptr;
	size_t n, remaining, length;

	ASSERT(packet != NULL);

//  MUTEX_LOCK(&master[handle].mutex);

//  LogMessage(LOG_DEBUG, "Acking %u:%u to slave %u\n", packet->hdr.unit,
//      packet->hdr.sequence, address);

	address = TDMA_BROADCAST;

	memcpy(&master[handle].tx.ack.hdr, &packet->hdr, sizeof(GRF_HDR));
	master[handle].tx.ack.hdr.type = GRF_TYPE_ACK;
	master[handle].tx.ack.hdr.length = GRF_HDR_SIZE;

	SerializePacket(&master[handle].tx.ack);

	n = GRF_MAX_PACKET_LEN * 2;
	if (!EncodeGRFPacket(&master[handle].tx.ack, master[handle].tx.buffer, &n)) {
//      MUTEX_UNLOCK(&master[handle].mutex);
		LogMessage(LOG_ERR, "TDMAMaster: Acknowledge: EncodeGRFPacket failed, encode buffer too small!\n");
		return FALSE;
	}

	ptr = master[handle].tx.buffer;
	remaining = n;
	while (remaining > 0) {
		if (remaining > master[handle].packet_length)
			length = master[handle].packet_length;
		else
			length = remaining;

		if (!TDMATransmit(handle, address, ptr, length)) {
//          MUTEX_UNLOCK(&master[handle].mutex);
			LogMessage(LOG_ERR, "TDMAMaster: Acknowledge: AsyncWrite failed: %s\n",
				AsyncErrorString(string, 128, AsyncErrorCode(handle)));
			return FALSE;
		}

		ptr += length;
		remaining -= length;
	}

//  MUTEX_UNLOCK(&master[handle].mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID Initialize(VOID)
{
	size_t i, j;

	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		memset(&master[i], 0, sizeof(TDMA_MASTER));
		MUTEX_INIT(&master[i].mutex);
		master[i].packet_length = 100;
		master[i].tx.timeout = UST_SECOND * 10;
		StartUSTimer(&master[i].tx.outbound, 0);
		master[i].tx.mean_outbound = -1;

		for (j = 0; j < MAX_SLAVES; j++) {
			StartUSTimer(&master[i].rx.unit[j].inbound, 0);
			master[i].rx.unit[j].mean_inbound = -1.0;
		}
	}

	initialized = TRUE;

	return;
}
