/* $Id: master_thread.c,v 1.5 2001/07/12 15:37:56 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Service TDMA master radio connections...

-----------------------------------------------------------------------*/

#include "master_thread.h"
#include "fwtdma.h"
#include "grfthread.h"

/* Module constants ---------------------------------------------------*/
#define CHECK_INTERVAL	(UST_MSECOND * 200)
#define RECONNECT_INTERVAL (UST_SECOND * 5)

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC MasterThread(void *argument);
static BOOL GetPacket(MASTER * master, GRF_PACKET * packet);
static VOID RequestStop(MASTER * master);
static BOOL StopRequested(MASTER * master);

/*---------------------------------------------------------------------*/
VOID MasterSetDefaults(MASTER * master)
{

	ASSERT(master != NULL);

	memset(master, 0, sizeof(MASTER));

	MUTEX_INIT(&master->mutex);
	MUTEX_LOCK(&master->mutex);

	master->running = FALSE;
	AsyncDefaultParams(&master->args.async);
    master->args.async.baud_rate = 115200;
	master->args.packet_length = 32;
	master->args.timeout = UST_SECOND * 10;
    master->args.n_slaves = 0;

	MUTEX_UNLOCK(&master->mutex);

	return;
}

/*---------------------------------------------------------------------*/
BOOL MasterThreadStart(MASTER * master)
{
	CHAR string[128];

	ASSERT(master != NULL);

	MUTEX_LOCK(&master->mutex);

	if (master->running) {
		MUTEX_UNLOCK(&master->mutex);
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: MasterThread(%s) is already running!\n", master->args.async.device);
		return FALSE;
	}
	if (master->args.n_slaves == 0) {
		MUTEX_UNLOCK(&master->mutex);
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: MasterThread(%s) has no slaves!\n", master->args.async.device);
		return FALSE;
	}

	master->stop = FALSE;

	LogMessage(LOG_DEBUG, "MasterThreadStart: opening device: %s\n", AsyncFormatParams(string, 128, &master->args.async));
	if ((master->handle = AsyncOpen(&master->args.async)) == ASYNC_INVALID) {
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: AsyncOpen: %s\n",
			AsyncErrorString(string, 128, AsyncErrorCode(master->handle)));
		return FALSE;
	}
	TDMAMasterSetParams(master->handle, &master->args);
	if (!AsyncSetLine(master->handle, ASYNC_ASSERT_DTR)) {
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: AsyncSetLine(ASSERT_DTR): %s\n",
			AsyncErrorString(string, 128, AsyncErrorCode(master->handle)));
		AsyncClose(master->handle);
		return FALSE;
	}
	USSleep(UST_SECOND);
	if (!AsyncFlush(master->handle, ASYNC_FLUSH_BOTH)) {
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: AsyncFlush(both): %s\n",
			AsyncErrorString(string, 128, AsyncErrorCode(master->handle)));
		AsyncClose(master->handle);
		return FALSE;
	}

	MUTEX_UNLOCK(&master->mutex);

	if (!THREAD_CREATE(&master->thread_id, MasterThread, master)) {
		LogMessage(LOG_ERR, "ERROR: MasterThreadStart: THREAD_CREATE(%s) failed: %u %s\n",
			master->args.async.device, THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC MasterThread(void *argument)
{
	UINT16 result;
	MASTER *master;
	GRF_PACKET *packet;

	ASSERT(argument != NULL);

	master = (MASTER *) argument;

	MUTEX_LOCK(&master->mutex);
	master->running = TRUE;
	MUTEX_UNLOCK(&master->mutex);

	LogMessage(LOG_NOTICE, "MasterThread %s (tid:%d, pid:%d) start\n", master->args.async.device, THREAD_SELF(), getpid());

	packet = NULL;
	while (!ShutDownRequested() && !StopRequested(master)) {
		/* Get an empty packet... */
		if (packet == NULL) {
			if ((result = QueueGet(master->empties, (VOID **)&packet, FALSE)) != Q_NO_ERROR) {
				if (result != Q_EMPTY) {
					LogMessage(LOG_ERR, "ERROR: MasterThread(%s): QueueGet(empties): %s\n",
						master->args.async.device, QueueErrorString(result));
				}
				USSleep(CHECK_INTERVAL);
				continue;
			}
		}

		/* Get a packet from our server */
		if (GetPacket(master, packet)) {
			/* Timed out? */
			if (packet->hdr.length == 0) {
				USSleep(CHECK_INTERVAL);
				continue;
			}
			/* Got a packet */
#if 0
			LogMessage(LOG_DEBUG, "%s -> %05u %s:%s:%s %s\n", master->args.async.device,
				packet->hdr.sequence,
				packet->payload_hdr.data.name.network,
				packet->payload_hdr.data.name.station,
				packet->payload_hdr.data.name.component,
				FormatUSTime(string, 32, UST_DOY_TIME_USEC, packet->payload_hdr.data.timing.time +
					packet->payload_hdr.data.timing.time_correction));
#endif

			if ((result = QueuePut(master->input, packet, FALSE)) != Q_NO_ERROR) {
				LogMessage(LOG_ERR, "ERROR: MasterThread(%s): PutPacket: QueuePut(input): %s\n",
					master->args.async.device, QueueErrorString(result));
				RequestShutDown(TRUE);
			}
			packet = NULL;
		}
	}

	/* Put the packet back into the empties queue */
	if (packet != NULL) {
		if ((result = QueuePut(master->empties, packet, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: MasterThread(%s): PutPacket: QueuePut(empties): %s\n",
				master->args.async.device, QueueErrorString(result));
			RequestShutDown(TRUE);
		}
	}

	AsyncClose(master->handle);

	/* Destroy our data structures */
	MUTEX_LOCK(&master->mutex);
	master->running = FALSE;
	MUTEX_UNLOCK(&master->mutex);

	LogMessage(LOG_NOTICE, "MasterThread %s (tid:%d, pid:%d) exit\n", master->args.async.device, THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GetPacket(MASTER * master, GRF_PACKET * packet)
{
	UINT32 address;

	ASSERT(master != NULL);
	ASSERT(packet != NULL);

	if (!TDMAMasterRead(master->handle, packet, &address))
		RequestStop(master);

	if (address == TDMA_VOID_ADDR)
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID RequestStop(MASTER * master)
{
	ASSERT(master != NULL);

	MUTEX_LOCK(&master->mutex);
	master->stop = TRUE;
	MUTEX_UNLOCK(&master->mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(MASTER * master)
{
	BOOL stop;

	ASSERT(master != NULL);

	MUTEX_LOCK(&master->mutex);
	stop = master->stop;
	MUTEX_UNLOCK(&master->mutex);

	return stop;
}
