/* $Id: queues.c,v 1.2 2001/06/29 16:04:31 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Generic thread safe pointer queues module.

-----------------------------------------------------------------------*/

#include "queues.h"
#include "mem.h"

/* Module constants ---------------------------------------------------*/

/* Uncomment and compile to perform unit testing */
//#define QUEUES_UNIT_TEST

/* Module types -------------------------------------------------------*/
typedef struct TAG_QUEUE {
	BOOL active;							/* In use flag */
	SEMAPHORE n_occupied;					/* Number of occupied slots */
	SEMAPHORE n_empty;						/* Number of empty slots */
	INT32 used;								/* Number of packets in queue */
	INT32 unused;							/* Number of packets available */
	INT32 low;								/* Min packets since check */
	INT32 high;								/* Max packets since check */
	INT32 head;								/* Head index */
	INT32 tail;								/* Tail index */
	INT32 n_slots;							/* Depth of queue */
	VOID **slots;							/* Array of void pointers */
} QUEUE;

/* Module globals -----------------------------------------------------*/
static BOOL queues_initialized = FALSE;

static struct TAG_QUEUES {
	MUTEX mutex;							/* Mutual exclusion */
	UINT32 n_queues;						/* Number of allocated queue structs */
	QUEUE *queue;							/* Queue structs */
} queues;

/* Number of errors defined in queues.h */
#define Q_ERRORS                8

static CHAR *error_string[Q_ERRORS + 1] = {
	"No queue error occurred!",
	"Queue is empty!",
	"Queue is full!",
	"Queues module is already initialized!",
	"Queues module not initialized!",
	"No more queue handles available!",
	"Invalid queue handle!",
	"Unable to allocate memory!",
	"Invalid queue error code!"
};

/* Module prototypes --------------------------------------------------*/
static BOOL IsValidHandle(H_QUEUE handle);

/*--------------------------------------------------------------------- */
UINT16 QueuesOpen(size_t n_queues)
{
	UINT16 i;
	QUEUE *queue;

	if (queues_initialized)
		return Q_ALREADY_INIT;

	/* Allocate queue structures */
	queues.n_queues = n_queues;
	if (!AllocateMemory((VOID **)&queues.queue, sizeof(QUEUE) * n_queues))
		return Q_NO_MEMORY;

	MUTEX_INIT(&queues.mutex);
	MUTEX_LOCK(&queues.mutex);

	/* Init mutex on each queue */
	for (i = 0; i < queues.n_queues; i++) {
		queue = &queues.queue[i];
		queue->active = FALSE;
	}

	queues_initialized = TRUE;

	MUTEX_UNLOCK(&queues.mutex);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
UINT16 QueuesClose(VOID)
{
	UINT16 result;
	H_QUEUE i;
	QUEUE *queue;

	if (!queues_initialized)
		return Q_NO_INIT;

	/* Destroy any active queues */
	for (i = 0; i < queues.n_queues; i++) {
		queue = &queues.queue[i];
		if (queue->active) {
			if ((result = QueueDestroy(i)) != Q_NO_ERROR)
				return result;
		}
	}

	MUTEX_LOCK(&queues.mutex);

	queues_initialized = FALSE;
	NoteMemoryRef(queues.queue);
	ReleaseMemory(queues.queue);

	MUTEX_UNLOCK(&queues.mutex);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
UINT16 QueueCreate(H_QUEUE * handle, size_t n_slots)
{
	UINT16 result;
	H_QUEUE i;
	QUEUE *queue;

	ASSERT(handle != NULL);

	if (!queues_initialized)
		return Q_NO_INIT;

	result = Q_NO_HANDLE;

	MUTEX_LOCK(&queues.mutex);

	/* Init first inactive queue... */
	for (i = 0; i < queues.n_queues; i++) {
		queue = &queues.queue[i];
		if (!queue->active) {

			/* Slot pointer storage, an array of void pointers */
			if (!AllocateMemory((VOID **)(&queue->slots), n_slots * sizeof(VOID *))) {
				result = Q_NO_MEMORY;
				break;
			}
			queue->n_slots = n_slots;

			SEM_INIT(&queue->n_occupied, 0, n_slots);
			SEM_INIT(&queue->n_empty, n_slots, n_slots);
			queue->used = 0;
			queue->unused = n_slots;
			queue->low = 0;
			queue->high = 0;
			queue->head = 0;
			queue->tail = 0;
			queue->active = TRUE;

			*handle = i;
			result = Q_NO_ERROR;
			break;
		}
	}

	MUTEX_UNLOCK(&queues.mutex);

	return result;
}

/*--------------------------------------------------------------------- */
UINT16 QueueDestroy(H_QUEUE handle)
{
	QUEUE *queue;

	if (!queues_initialized)
		return Q_NO_INIT;
	if (!IsValidHandle(handle))
		return Q_BAD_HANDLE;

	queue = &queues.queue[handle];
	MUTEX_LOCK(&queues.mutex);

	/* If still active after lock, deactivate */
	if (queue->active) {
		queue->active = FALSE;
		SEM_DESTROY(&queue->n_occupied);
		SEM_DESTROY(&queue->n_empty);
		NoteMemoryRef(queue->slots);
		ReleaseMemory(queue->slots);
	}

	MUTEX_UNLOCK(&queues.mutex);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
UINT16 QueuePut(H_QUEUE handle, VOID *ptr, BOOL block)
{
	QUEUE *queue;

	if (!queues_initialized)
		return Q_NO_INIT;
	if (!IsValidHandle(handle))
		return Q_BAD_HANDLE;

	queue = &queues.queue[handle];

	if (block) {
		/* Wait for at least one empty slot */
		SEM_WAIT(&queue->n_empty);
	}
	else {
		/* See if there's an empty slot */
		if (SEM_TRYWAIT(&queue->n_empty) != 0)
			return Q_FULL;
	}

	/* Lock the queue */
	MUTEX_LOCK(&queues.mutex);

	/* Sanity checks and stats */
	ASSERT(queue->used < queue->n_slots);
	queue->used++;
	ASSERT(queue->unused > 0);
	queue->unused--;

	if (queue->high < queue->used)
		queue->high = queue->used;

	/* Put pointer in tail slot */
	*(queue->slots + queue->tail) = ptr;
	queue->tail++;
	if (queue->tail >= queue->n_slots)
		queue->tail = 0;

	/* Unlock the queue */
	MUTEX_UNLOCK(&queues.mutex);

	/* One more occupied slot */
	SEM_POST(&queue->n_occupied);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
UINT16 QueueGet(H_QUEUE handle, VOID **ptr, BOOL block)
{
	QUEUE *queue;

	if (!queues_initialized)
		return Q_NO_INIT;
	if (!IsValidHandle(handle))
		return Q_BAD_HANDLE;

	queue = &queues.queue[handle];

	if (block) {
		/* Wait for at least one occupied slot */
		SEM_WAIT(&queue->n_occupied);
	}
	else {
		/* See if there's an occupied slot */
		if (SEM_TRYWAIT(&queue->n_occupied) != 0)
			return Q_EMPTY;
	}

	/* Lock the queue */
	MUTEX_LOCK(&queues.mutex);

	/* Sanity checks and stats */
	ASSERT(queue->used > 0);
	queue->used--;
	ASSERT(queue->unused < queue->n_slots);
	queue->unused++;

	if (queue->low > queue->used)
		queue->low = queue->used;

	/* Get pointer from head slot */
	*ptr = *(queue->slots + queue->head);
	queue->head++;
	if (queue->head >= queue->n_slots)
		queue->head = 0;

	/* Unlock the queue */
	MUTEX_UNLOCK(&queues.mutex);

	/* One more empty slot */
	SEM_POST(&queue->n_empty);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
CHAR *QueueErrorString(UINT16 error)
{

	if (error > Q_ERRORS)
		error = Q_ERRORS;

	return error_string[error];
}

/*--------------------------------------------------------------------- */
UINT16 QueueStat(H_QUEUE handle, QUEUE_STAT stat, INT32 *value)
{
	QUEUE *queue;

	if (!queues_initialized)
		return Q_NO_INIT;
	if (!IsValidHandle(handle))
		return Q_BAD_HANDLE;

	queue = &queues.queue[handle];
	MUTEX_LOCK(&queues.mutex);

	switch (stat) {
	case Q_USED:
		*value = queue->used;
		break;
	case Q_UNUSED:
		*value = queue->unused;
		break;
	case Q_LOW:
		*value = queue->low;
		break;
	case Q_HIGH:
		*value = queue->high;
		break;
	case Q_RESET:
		queue->low = queue->used;
		queue->high = queue->used;
		*value = queue->used;
		break;
	}

	MUTEX_UNLOCK(&queues.mutex);

	return Q_NO_ERROR;
}

/*--------------------------------------------------------------------- */
static BOOL IsValidHandle(H_QUEUE handle)
{

	if (handle >= queues.n_queues)
		return FALSE;
	if (!queues.queue[handle].active)
		return FALSE;

	return TRUE;
}

/*--------------------------------------------------------------------- */
/*--------------------------------------------------------------------- */
#if defined QUEUES_UNIT_TEST

#include "main.h"

#define DEPTH    8
#define N_QUEUES 8

BOOL putblock = FALSE;
BOOL getblock = FALSE;

static THREAD_FUNC GetThread(VOID *argument);
VOID CatchSignal(int sig);
BOOL SetSignalHandler(VOID);

H_QUEUE queue;

/*--------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
	UINT16 result;
	THREAD thread_id;

	if (!SetSignalHandler())
		exit(1);

	LogOpen("./queues_unit_test.log");

	LogMessage(LOG_INFO, "MainThread starting, id=0x%X\n", THREAD_SELF());

	SocketsInit();

	/* Init the queues system... */
	if ((result = QueuesOpen(N_QUEUES)) != Q_NO_ERROR) {
		LogMessage(LOG_ERROR, "ERROR: QueuesOpen(): %s", QueueErrorString(result));
		exit(1);
	}

	/* Create a queue... */
	if ((result = QueueCreate(&queue, DEPTH)) != Q_NO_ERROR) {
		LogMessage(LOG_ERROR, "ERROR: QueueCreate(): %s", QueueErrorString(result));
		exit(1);
	}

	/* Fire up the Get thread... */
	if (!THREAD_CREATE(&thread_id, GetThread, NULL)) {
		LogMessage(LOG_ERROR, "ERROR: THREAD_CREATE: GetThread: %s\n", strerror(errno));
		exit(1);
	}

	/* Fire up the Console thread... */
	if (!ConsoleInitialize(12345))
		exit(1);

	/* The console thread will callback HandleKey() as keys are pressed */

	/* Wait for the threads to exit... */
	THREAD_JOIN(thread_id);
	THREAD_JOIN(ctrl.threads.console_tid);

	/* This call is unnecessary in this case as QueuesClose will destroy all
	   active queues */
	if ((result = QueueDestroy(queue)) != Q_NO_ERROR) {
		LogMessage(LOG_ERROR, "ERROR: QueueCreate(): %s", QueueErrorString(result));
		exit(1);
	}

	if ((result = QueuesClose()) != Q_NO_ERROR) {
		LogMessage(LOG_ERROR, "ERROR: QueuesClose(): %s", QueueErrorString(result));
		exit(1);
	}

	/* Will assert if anything is fishy */
	CheckMemoryRefs();

	LogMessage(LOG_INFO, "MainThread exiting, id=0x%X\n", THREAD_SELF());
	exit(0);
}

/*--------------------------------------------------------------------- */
/* Callback from console thread with key presses */
VOID HandleKey(CHAR key)
{
	UINT16 result, *ptr;
	INT32 value, q_unused, q_used, q_low, q_high;

	/* 1 key gets queue status */
	if (key == '1') {
		if ((result = QueueStat(queue, Q_USED, &q_used)) != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: MainThread: QueueStat: Used: %s\n", QueueErrorString(result));
			shut_down = TRUE;
		}
		if ((result = QueueStat(queue, Q_UNUSED, &q_unused)) != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: MainThread: QueueStat: Unused: %s\n", QueueErrorString(result));
			shut_down = TRUE;
		}
		if ((result = QueueStat(queue, Q_LOW, &q_low)) != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: MainThread: QueueStat: Low: %s\n", QueueErrorString(result));
			shut_down = TRUE;
		}
		if ((result = QueueStat(queue, Q_HIGH, &q_high)) != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: MainThread: QueueStat: High: %s\n", QueueErrorString(result));
			shut_down = TRUE;
		}
		LogMessage(LOG_DEFAULT, "PutThread: queue stats: unused=%d, used=%d, low=%d, high=%d\n", q_unused, q_used, q_low,
			q_high);
		if ((result = QueueStat(queue, Q_RESET, &value)) != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: MainThread: QueueStat: Reset: %s\n", QueueErrorString(result));
			shut_down = TRUE;
		}
	}
	else if (key == '2') {
		putblock = TRUE;
		LogMessage(LOG_INFO, "PutThread: QueuePut: Blocking\n");
	}
	else if (key == '3') {
		putblock = FALSE;
		LogMessage(LOG_INFO, "PutThread: QueuePut: Nonblocking\n");
	}
	/* Everything else gets sent to the Get thread... */
	else {
		/* Memory will be released at other end of queue  */
		if (!AllocateMemory((VOID **)&ptr, sizeof(UINT16))) {
			LogMessage(LOG_ERROR, "ERROR: PutThread: AllocateMemory() failed!\n");
			shut_down = TRUE;
			return;
		}

		*ptr = (UINT16)key;
		LogMessage(LOG_INFO, "PutThread: QueuePut: 0x%X (%c)\n", *ptr, (*ptr >= 32 && *ptr <= 126 ? (CHAR)*ptr : '.'));

		result = QueuePut(queue, ptr, putblock);
		if (result == Q_FULL) {
			LogMessage(LOG_ERROR, "PutThread: QueuePut: %s\n", QueueErrorString(result));
			NoteMemoryRef(ptr);
			ReleaseMemory(ptr);
		}
		else if (result != Q_NO_ERROR) {
			LogMessage(LOG_ERROR, "ERROR: PutThread: QueuePut: %s\n", QueueErrorString(result));
			shut_down = TRUE;
			return;
		}
	}

	return;
}

/*--------------------------------------------------------------------- */
static THREAD_FUNC GetThread(VOID *argument)
{
	UINT16 result, key, *ptr;

	LogMessage(LOG_INFO, "GetThread starting, id=0x%X\n", THREAD_SELF());

	ptr = NULL;

	while (!shut_down) {
		key = 0;
		result = QueueGet(queue, (VOID **)&ptr, getblock);
		if (result == Q_NO_ERROR) {
			LogMessage(LOG_INFO, "GetThread: QueueGet: 0x%X (%c)\n", *ptr, (*ptr >= 32 && *ptr <= 126 ? (CHAR)*ptr : '.'));
			key = *ptr;
			/* Memory was allocated in the other thread */
			NoteMemoryRef(ptr);
			ReleaseMemory(ptr);
		}
		else if (result == Q_EMPTY) {
			LogMessage(LOG_DEBUG, "GetThread: QueueGet: %s\n", QueueErrorString(result));
			USSleep(UST_MSECOND * 200);
		}
		else {
			LogMessage(LOG_ERROR, "ERROR: GetThread: QueueGet: %s\n", QueueErrorString(result));
			key = 0;
			shut_down = TRUE;
		}

		if (key == '4') {
			getblock = TRUE;
			LogMessage(LOG_INFO, "GetThread: QueueGet: Blocking\n");
		}
		else if (key == '5') {
			getblock = FALSE;;
			LogMessage(LOG_INFO, "GetThread: QueueGet: Nonblocking\n");
		}
		else if (key == ' ') {
			LogMessage(LOG_INFO, "GetThread: sleeping for 5 seconds\n");
			USSleep(UST_SECOND * 5);
		}
	}

	LogMessage(LOG_INFO, "GetThread exiting, id=0x%X\n", THREAD_SELF());

	THREAD_EXIT((VOID *)0);
}

#endif
