/* $Id: main.c,v 1.20 2006/04/30 16:24:38 cvs Exp $ */
/*-----------------------------------------------------------------------

	GRF packet logging utility main module...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include "main.h"
#include "log.h"
#include "mem.h" 

/* Module constants ---------------------------------------------------*/
#define MAX_SAMPLES  1024

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL quit = FALSE;
static INT32 samples[MAX_SAMPLES];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL GRFLog(MAIN_ARGS * args);
static VOID CheckTiming(MAIN_ARGS * args, GRF_UNIT * unit, GRF_PACKET * packet);
static BOOL ReadPacket(MAIN_ARGS * args, GRF_PACKET * packet);
static BOOL LogPacket(MAIN_ARGS * args, GRF_PACKET * packet);
static BOOL OpenInput(MAIN_ARGS * args);
static VOID CloseInput(MAIN_ARGS * args);
static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args);
static VOID CommandLineHelp(VOID);
static BOOL SetSignalHandlers(VOID);
static VOID CatchSignal(int sig);

/*---------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	MAIN_ARGS args;

	if (!SetSignalHandlers())
		abort();

	if (!SocketsInit())
		abort();

	LogMessage(LOG_NOTICE, "GRFLog version %s (%s %s), pid:%d\n", VERSION, __DATE__, __TIME__, getpid());

	if (!ParseCommandLine(argc, argv, &args))
		exit(1);

	if (!LogOpen(argv[0], NULL, args.output_spec))
		exit(1);
	LogSetLevel(LOG_CON | LOG_SYS | LOG_FIL, LOG_DEBUG);

	if (args.input_spec[0] == '\0') {
		LogMessage(LOG_ERR, "\nERROR: No input source specified!\n");
		CommandLineHelp();
	}

	while (!QuitRequested()) {
		if (OpenInput(&args))
			break;
		USSleep(UST_SECOND * 5);
		CloseInput(&args);
	}

	if (!GRFLog(&args)) {
		CloseInput(&args);
		LogClose();
		exit(1);
	}

	CloseInput(&args);
	LogClose();

	return 0;
}

/*--------------------------------------------------------------------- */
static BOOL GRFLog(MAIN_ARGS *args)
{
	BOOL ok;
	UINT32 t_packets, t_seq_errors, t_time_tears;
	GRF_PACKET *packet;
	GRF_UNIT *list = NULL, *unit;
	REAL64 elapsed;
	USTIMER timer;
	FILE *fptr = NULL;

	if (!AllocateMemory((VOID **)&packet, GRF_MAX_PACKET_LEN)) {
		LogMessage(LOG_ERR, "ERROR: GRFLog: AllocateMemory failed!\n");
		return FALSE;
	}

	if (args->capture_spec[0] != '\0') {
		if ((fptr = fopen(args->capture_spec, "w+b")) == NULL) {
			LogMessage(LOG_ERR, "ERROR: GRFLog(): fopen(%s) failed: %s\n", 
					args->capture_spec, strerror(errno));
			args->capture_spec[0] = '\0';
		}
	}

	StartUSTimer(&timer, 0);
	ok = TRUE;
	while (!QuitRequested()) {
		if (ReadPacket(args, packet)) {
			if (packet->hdr.length == 0)
				continue;

			if (fptr != NULL) {
				if (fwrite(packet->serial.bytes, (size_t)packet->hdr.length, 1, fptr) != 1) {
					LogMessage(LOG_ERR, "ERROR: GRFLog: fwrite(%s) failed: %s\n",
						args->capture_spec, strerror(errno));
					fclose(fptr);
					fptr = NULL;
					args->capture_spec[0] = '\0';
				}
			}

			/* Check sequence numbers */
			for (unit = list; unit != NULL; unit = unit->next) {
				if (packet->hdr.unit == unit->unit_id)
					break;
			}
			if (unit == NULL) {
				if (!AllocateMemory((VOID **)&unit, sizeof(GRF_UNIT))) {
					LogMessage(LOG_ERR, "ERROR: Unable to allocate memory!\n");
					ok = FALSE;
					break;
				}
				unit->unit_id = packet->hdr.unit;
				unit->channel = 0;
				unit->sequence = packet->hdr.sequence;
				unit->count = 0;
				unit->seq_errors = 0;
				unit->time_tears = 0;
				unit->expect_time = 0;
				unit->next = list;
				list = unit;
			}
			if (packet->hdr.sequence != unit->sequence) {
				LogMessage(LOG_ERR, "ERROR: sequence break on unit %u, got:%u, expected: %u\n",
					unit->unit_id, packet->hdr.sequence, unit->sequence);
				unit->seq_errors++;
			}
			unit->sequence = packet->hdr.sequence + 1;
			unit->count++;

			if (args->pps_comp[0] != '\0') {
				CheckTiming(args, unit, packet);
			}
			else {
				if (!LogPacket(args, packet)) {
					ok = FALSE;
					break;
				}
			}
		}
		/* If we're reading a socket... */
		else if (args->file == NULL) {
			while (!QuitRequested()) {
				CloseInput(args);
				USSleep(UST_SECOND * 5);
				if (OpenInput(args))
					break;
			}
		}
		else {
			break;
		}
	}

	elapsed = (REAL64)USTimerElapsed(&timer) / (REAL64)UST_SECOND;

	if (fptr != NULL)
		fclose(fptr);

	ReleaseMemory(packet);

	/* Release sequence check list */
	t_packets = t_seq_errors = t_time_tears = 0;
	while (list != NULL) {
		unit = list;
		list = list->next;
		if (args->pps_comp[0] != '\0') {
			LogMessage(LOG_NOTICE, "Unit %u packets: %u @ %.2lf/sec, sequence breaks: %u, time tears: %u\n",
				unit->unit_id, unit->count, (REAL64)unit->count / elapsed, unit->seq_errors, unit->time_tears);
		} else {
			LogMessage(LOG_NOTICE, "Unit %u packets: %u @ %.2lf/sec, sequence breaks: %u\n",
				unit->unit_id, unit->count, (REAL64)unit->count / elapsed, unit->seq_errors);
		}
		t_packets += unit->count;
		t_seq_errors += unit->seq_errors;
		t_time_tears += unit->time_tears;
		ReleaseMemory(unit);
	}
	if (args->pps_comp[0] != '\0') {
		LogMessage(LOG_NOTICE, "Total packets: %u @ %.2lf/sec, sequence breaks: %u, time tears: %u\n",
			t_packets, (REAL64)t_packets / elapsed, t_seq_errors, t_time_tears);
	} else {
		LogMessage(LOG_NOTICE, "Total packets: %u @ %.2lf/sec, sequence breaks: %u\n",
			t_packets, (REAL64)t_packets / elapsed, t_seq_errors);
	}

	DumpMemoryInfo(LOG_NOTICE);

	return ok;
}

/*--------------------------------------------------------------------- */
static VOID CheckTiming(MAIN_ARGS *args, GRF_UNIT *unit, GRF_PACKET *packet)
{
	CHAR string[32];
	REAL64 rate;
	USTIME time, half_period;

	ASSERT(args != NULL);
	ASSERT(unit != NULL);
	ASSERT(packet != NULL);

	if (packet->hdr.type == GRF_TYPE_INFO) {
		if (args->kill_info)
			return;
		if (packet->hdr.unit != 0)
			LogMessage(LOG_NOTICE, "Unit %d: %s\n", packet->hdr.unit, packet->serial.info.message);
		else
			LogMessage(LOG_NOTICE, "%s\n", packet->serial.info.message);
	}
	if (packet->hdr.type != GRF_TYPE_DATA || args->info_only)
		return;

	if (args->pps_stn[0] != '*' && args->pps_stn[0] != '\0' &&
		strcmp(packet->payload_hdr.data.name.station, args->pps_stn) != 0) return;
	if (args->pps_comp[0] != '*' && strcmp(packet->payload_hdr.data.name.component, args->pps_comp) != 0)
		return;
	if (args->pps_comp[0] != '*' && args->pps_stn[0] != '*' && args->pps_stn[0] != '\0')
		strcpy(args->pps_stn, packet->payload_hdr.data.name.station);

	if (unit->channel == 0)
		unit->channel = packet->payload_hdr.data.channel;
	else if (unit->channel != packet->payload_hdr.data.channel)
		return;

	time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
	rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;
	if (rate > 0.0)
		half_period = (USTIME)(((1.0 / rate) * 1.0) * (REAL64)UST_SECOND);
	else
		return;

	if (unit->expect_time != 0) {
		LogMessage(LOG_NOTICE, "%s:%s:%s %u:%u:%05u %s (%s) %.6lf (%s) %+.6lf\n",
			packet->payload_hdr.data.name.network,
			packet->payload_hdr.data.name.station,
			packet->payload_hdr.data.name.component,
			packet->hdr.unit,
			packet->payload_hdr.data.channel,
			packet->hdr.sequence,
			FormatUSTime(string, 32, UST_DATE_TIME_USEC, time),
			GRFDescribeQuality(packet->payload_hdr.data.timing.time_quality),
			rate,
			GRFDescribeQuality(packet->payload_hdr.data.timing.rate_quality),
			(REAL64)(time - unit->expect_time) / (REAL64)UST_SECOND);

		if (time < unit->expect_time - half_period || time > unit->expect_time + half_period) {
			unit->time_tears++;
			LogMessage(LOG_ERR, "ERROR: time tear on unit %u, %+.6lf\n",
				packet->hdr.unit, (REAL64)(time - unit->expect_time) / (REAL64)UST_SECOND);
		}
	}

	unit->expect_time = time + (USTIME)(((REAL64)packet->payload_hdr.data.n / rate) * (REAL64)UST_SECOND);

	return;
}

/*--------------------------------------------------------------------- */
static BOOL ReadPacket(MAIN_ARGS * args, GRF_PACKET * packet)
{

	ASSERT(args != NULL);
	ASSERT(packet != NULL);

	/* Read from file? */
	if (args->file != NULL) {
		if (!GRFReadFile(args->file, packet))
			return FALSE;
		/* End-of-file? */
		if (packet->hdr.type == GRF_TYPE_NONE && packet->hdr.length == 0) {
			RequestQuit(TRUE);
			return FALSE;
		}
	}
	else {
		if (!GRFReadSocket(args->socket, packet, args->read_timeout)) {
			return FALSE;
		}
	}

	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL LogPacket(MAIN_ARGS *args, GRF_PACKET *packet)
{
	CHAR string[32], string2[32], data_desc[32];
	UINT16 i, bytes;
	INT32 min, max;
	REAL64 sample, sum, mean, sigma;
	USTIME time;
	REAL64 rate;

	ASSERT(args != NULL);
	ASSERT(packet != NULL);


	if (args->verbose) {
		if (args->info_only && packet->hdr.type != GRF_TYPE_INFO)
			return TRUE;
		LogMessage(LOG_INFO, "GRF common header:\n");
		LogMessage(LOG_INFO, " Signature:       %.3s\n", packet->hdr.signature);
		LogMessage(LOG_INFO, " Version:         %u\n", packet->hdr.version);
		LogMessage(LOG_INFO, " Length:          %u\n", packet->hdr.length);
		LogMessage(LOG_INFO, " Sequence number: %u\n", packet->hdr.sequence);
		LogMessage(LOG_INFO, " Unit ID:         %u\n", packet->hdr.unit);
		LogMessage(LOG_INFO, " Type:            %u - %s\n", packet->hdr.type, GRFDescribeType(packet->hdr.type));
	}

	if (packet->hdr.type == GRF_TYPE_INFO) {
		if (args->kill_info)
			return TRUE;
		if (packet->hdr.unit != 0)
			LogMessage(LOG_NOTICE, "Unit %d: %s\n", packet->hdr.unit, packet->serial.info.message);
		else
			LogMessage(LOG_NOTICE, "%s\n", packet->serial.info.message);
		return TRUE;
	}
	if (args->info_only)
		return TRUE;

	if (packet->hdr.type == GRF_TYPE_DATA) {
		time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
		rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;

		if (args->verbose) {
			LogMessage(LOG_INFO, " GRF data header:\n");
			LogMessage(LOG_INFO, "  Channel:        %u\n", packet->payload_hdr.data.channel);
			LogMessage(LOG_INFO, "  Network:        %s\n", packet->payload_hdr.data.name.network);
			LogMessage(LOG_INFO, "  Station:        %s\n", packet->payload_hdr.data.name.station);
			LogMessage(LOG_INFO, "  Component:      %s\n", packet->payload_hdr.data.name.component);
			LogMessage(LOG_INFO, "  Time:           %s\n",
				FormatUSTime(string, 32, UST_DATE_TIME_MSEC, packet->payload_hdr.data.timing.time));
			LogMessage(LOG_INFO, "  Corrected time: %s quality: %s\n",
				FormatUSTime(string, 32, UST_DATE_TIME_MSEC, time),
				GRFDescribeQuality(packet->payload_hdr.data.timing.time_quality));
			LogMessage(LOG_INFO, "  Rate:           %lf sps\n", packet->payload_hdr.data.timing.rate);
			LogMessage(LOG_INFO, "  Corrected rate: %lf quality: %s\n", rate,
				GRFDescribeQuality(packet->payload_hdr.data.timing.rate_quality));
			LogMessage(LOG_INFO, "  Counts/volt:    %d\n", packet->payload_hdr.data.counts_per_volt);
			LogMessage(LOG_INFO, "  Data type:      %u - %s\n", packet->payload_hdr.data.type,
				GRFDescribeDataType(packet->payload_hdr.data.type));
			LogMessage(LOG_INFO, "  Data samples:   %u\n", packet->payload_hdr.data.n);
		}
		else {
			if (packet->payload_hdr.data.type == GRF_DATA_CM8) {
				bytes = packet->hdr.length - GRF_HDR_SIZE - GRF_DATA_HDR_SIZE - 2;
				sprintf(data_desc, "CM8 %u %.2lf:1", packet->payload_hdr.data.n,
					(REAL64)(packet->payload_hdr.data.n * 4) / (REAL64)bytes);
			}
			else if (packet->payload_hdr.data.type == GRF_DATA_INT24)
				sprintf(data_desc, "INT24 %u", packet->payload_hdr.data.n);
			else
				sprintf(data_desc, "INT32 %u", packet->payload_hdr.data.n);
			LogMessage(LOG_NOTICE, "%s:%s:%s %u:%u:%05u %s/%s (%s) %s\n",
				packet->payload_hdr.data.name.network,
				packet->payload_hdr.data.name.station,
				packet->payload_hdr.data.name.component,
				packet->hdr.unit,
				packet->payload_hdr.data.channel,
				packet->hdr.sequence,
				FormatUSTime(string, 32, UST_DATE_TIME_MSEC, time),
				FormatUSTime(string2, 32, UST_TIME_MSEC, time + (USTIME)((packet->payload_hdr.data.n / rate) * (REAL64)UST_SECOND)), 
				GRFDescribeQuality(packet->payload_hdr.data.timing.time_quality), 
				data_desc);
		}

		if (GRFDecodeSamples(packet, samples, MAX_SAMPLES)) {
			if (args->verbose && !args->avgs && !args->nodata)
				LogMessage(LOG_INFO, " Sample data:\n");
			if (args->avgs && packet->payload_hdr.data.n > 1) {
				max = min = samples[0];
				sum = 0.0;
				for (i = 0; i < packet->payload_hdr.data.n; i++) {
					if (min > samples[i])
						min = samples[i];
					if (max < samples[i])
						max = samples[i];
					sum += (REAL64)samples[i];
				}
				mean = sum / (REAL64)packet->payload_hdr.data.n;
				sum = 0.0;
				for (i = 0; i < packet->payload_hdr.data.n; i++) {
					sample = (REAL64)samples[i] - mean;
					sum += (sample * sample);
				}
				sigma = sqrt(sum / (REAL64)(packet->payload_hdr.data.n - 1));
				LogMessage(LOG_NOTICE, " Min: %d, Max: %d, Range: %d, Mean: %.2lf, Sigma: %.2lf\n",
					min, max, max - min + 1, mean, sigma);
			}
			else if (!args->nodata) {
				for (i = 0; i < packet->payload_hdr.data.n; i++) {
					LogMessage(LOG_NOTICE, "%+9d", samples[i]);
					if ((i + 1) % 8 == 0)
						LogMessage(LOG_NOTICE, "\n");
				}
			}
			if (!args->nodata) {
				if (i % 8 != 0 && !args->avgs)
					LogMessage(LOG_NOTICE, "\n");
			}
		}
	}
	else {
		LogMessage(LOG_ERR, "Unimplemented packet type %u from %u, length %u\n", packet->hdr.type,
			packet->hdr.unit, packet->hdr.length);
	}

	return TRUE;
}

/*--------------------------------------------------------------------- */
static BOOL OpenInput(MAIN_ARGS * args)
{
	BOOL result;
	CHAR string[128];
	GRF_PACKET *packet;

	ASSERT(args != NULL);

	/* Is it a file or a socket */
	if (args->endpoint.sin_addr.s_addr != 0) {
		if ((args->socket = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
			LogMessage(LOG_ERR, "ERROR: OpenInput: socket: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
			return FALSE;
		}
		args->endpoint.sin_family = AF_INET;
		if (connect(args->socket, (struct sockaddr *)&args->endpoint, sizeof(ENDPOINT)) != 0) {
			LogMessage(LOG_ERR, "ERROR: OpenInput: connect(%s): %s\n",
				FormatEndpoint(&args->endpoint, string), SOCK_ERRSTRING(SOCK_ERRNO()));
			SOCK_CLOSE(args->socket);
			args->socket = INVALID_SOCKET;
			return FALSE;
		}
		if (!AllocateMemory((VOID **)&packet, GRF_MAX_PACKET_LEN)) {
			LogMessage(LOG_ERR, "ERROR: GRFLog: AllocateMemory failed!\n");
			SOCK_CLOSE(args->socket);
			args->socket = INVALID_SOCKET;
			return FALSE;
		}
		if ((result = GRFConnect(args->socket, &args->endpoint, GRF_ATTRIB_DATA, args->read_timeout, packet)) == FALSE) {
			SOCK_CLOSE(args->socket);
			args->socket = INVALID_SOCKET;
		}
		ReleaseMemory(packet);
		return result;
	}
	else {
		if ((args->file = fopen(args->input_spec, "rb")) == NULL) {
			LogMessage(LOG_ERR, "ERROR: OpenInput: fopen(%s) failed: %s\n", args->input_spec, strerror(errno));
			return FALSE;
		}
	}

	return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID CloseInput(MAIN_ARGS * args)
{
	static GRF_PACKET packet;

	ASSERT(args != NULL);

	if (args->file != NULL)
		fclose(args->file);
	else if (args->socket != INVALID_SOCKET) {
		GRFDisconnect(args->socket, &packet);
		SOCK_CLOSE(args->socket);
		args->socket = INVALID_SOCKET;
	}

	return;
}

/*--------------------------------------------------------------------- */
#include "getopt.c"

static BOOL ParseCommandLine(int argc, char *argv[], MAIN_ARGS * args)
{
	CHAR string[32], *ptr;
	int i, option;

	ASSERT(args != NULL);
	ASSERT(argv != NULL);

	/* Set defaults */
	args->input_spec[0] = '\0';
	args->output_spec[0] = '\0';
	args->capture_spec[0] = '\0';
	args->terse = FALSE;
	args->verbose = FALSE;
	args->debug = FALSE;
	args->read_timeout = GRF_DFL_TIMEOUT;
	args->nodata = FALSE;
	args->avgs = FALSE;
	args->info_only = FALSE;
	args->kill_info = FALSE;
	memset(args->pps_stn, 0, GRF_MAX_COMP_LEN);
	memset(args->pps_comp, 0, GRF_MAX_COMP_LEN);

	memset(&args->endpoint, 0, sizeof(ENDPOINT));
	args->file = NULL;
	args->socket = INVALID_SOCKET;
	args->output = NULL;

	/* Don't write errors to stderr */
	my_opterr = 0;

	/* Options... */
	while ((option = my_getopt(argc, argv, "hvtdanixp:s:f:o:r:c:")) != -1) {
		switch (option) {
		case 'h':
			CommandLineHelp();
			break;
		case 'v':
			args->verbose = TRUE;
			break;
		case 't':
			args->terse = TRUE;
			break;
		case 'd':
			args->debug = TRUE;
			break;
		case 'n':
			args->nodata = TRUE;
			break;
		case 'a':
			args->avgs = TRUE;
			break;
		case 'i':
			args->info_only = TRUE;
			break;
		case 'x':
			args->kill_info = TRUE;
			break;
		case 'p':
			if (my_optarg) {
				if ((ptr = strchr(my_optarg, ':')) == NULL)
					strncpy(args->pps_comp, my_optarg, GRF_MAX_COMP_LEN);
				else {
					strncpy(args->pps_comp, ptr + 1, GRF_MAX_COMP_LEN);
					*ptr = '\0';
					strncpy(args->pps_stn, my_optarg, GRF_MAX_STN_LEN);
				}
			}
			else {
				LogMessage(LOG_ERR, "ERROR: no component identifier specified: '-p'\n");
				return FALSE;
			}
			break;
		case 's':
			if (my_optarg) {
				if (!ParseEndpoint(&args->endpoint, my_optarg, DFL_PORT))
					return FALSE;
				sprintf(args->input_spec, "%s (%s)", my_optarg, FormatEndpoint(&args->endpoint, string));
			}
			else {
				LogMessage(LOG_ERR, "ERROR: No endpoint specified: '-s'\n");
				return FALSE;
			}
			break;
		case 'r':
			if (my_optarg) {
				args->read_timeout = (USTIME)strtoul(my_optarg, NULL, 0) * UST_SECOND;
			}
			break;
		case 'f':
			if (my_optarg)
				strncpy(args->input_spec, my_optarg, MAX_PATH_LEN);
			else {
				LogMessage(LOG_ERR, "ERROR: No input file specified: '-f'\n");
				return FALSE;
			}
			break;
		case 'o':
			if (my_optarg)
				strncpy(args->output_spec, my_optarg, MAX_PATH_LEN);
			else {
				LogMessage(LOG_ERR, "ERROR: No output file specified: '-o'\n");
				return FALSE;
			}
			break;
		case 'c':
			if (my_optarg)
				strncpy(args->capture_spec, my_optarg, MAX_PATH_LEN);
			else {
				LogMessage(LOG_ERR, "ERROR: No capture file specified: '-c'\n");
				return FALSE;
			}
			break;
		case '?':
			if (isprint(my_optopt))
				fprintf(stderr, "Unknown option '-%c'.\n", my_optopt);
			else
				fprintf(stderr, "Unknown option character '\\x%x'.\n", my_optopt);
			return FALSE;
		default:
			abort();
		}
	}

	/* Arguments... */
	for (i = my_optind; i < argc; i++) {
        if (args->input_spec[0] == '\0') {
            strcpy(args->input_spec, argv[i]);
        }
        else {
            fprintf(stderr, "unrecognized argument: `%s'\n", argv[i]);
            return FALSE;
        }
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CommandLineHelp(VOID)
{
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "Usage: grflog [-htvixna] [-p [station:]component] [-r seconds]\n");
	LogMessage(LOG_NOTICE, "              [-o output_file] [-c capture_file]\n");
	LogMessage(LOG_NOTICE, "              -s ip_address[:port] | [-f] input_file \n");
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "   -h          Help display.\n");
	LogMessage(LOG_NOTICE, "   -t          Terse output, single line headers.\n");
	LogMessage(LOG_NOTICE, "   -v          Verbose output, fully annotated headers.\n");
	LogMessage(LOG_NOTICE, "   -i          Log only GRF information messages.\n");
	LogMessage(LOG_NOTICE, "   -x          Exclude GRF information messages.\n");
	LogMessage(LOG_NOTICE, "   -n          No sample data, log headers only.\n");
	LogMessage(LOG_NOTICE, "   -a          Analyze data packets and log statistics.\n");
	LogMessage(LOG_NOTICE, "   -p channel  Analyze timing on specified station:component.\n");
    LogMessage(LOG_NOTICE, "   -r seconds  Set socket read timeout. (%u seconds)\n", (UINT32)(GRF_DFL_TIMEOUT / UST_SECOND));
	LogMessage(LOG_NOTICE, "   -o file     Output file. (stdout)\n");
	LogMessage(LOG_NOTICE, "   -c file     GRF packet capture file. (none)\n");
	LogMessage(LOG_NOTICE, "   -s socket   Socket to connect to for input data,\n");
	LogMessage(LOG_NOTICE, "                port number defaults to %u if not specified.\n", DFL_PORT);
	LogMessage(LOG_NOTICE, "   [-f] file   GRF packet image file to read for input data.\n");
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "[] = optional, () = default, | = mutually exclusive.\n");

	exit(0);
}

/*--------------------------------------------------------------------- */
BOOL QuitRequested(VOID)
{
	return quit;
}

/*--------------------------------------------------------------------- */
VOID RequestQuit(BOOL request)
{
	quit = request;
	return;
}

/*--------------------------------------------------------------------- */
VOID HexDump(UINT32 loglevel, VOID *ptr, size_t length)
{
	CHAR string[128];
	UINT8 *data;
	size_t i, j, n;

	ASSERT(ptr != NULL);

	if (length == 0)
		return;

/*                        0123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789 */
	LogMessage(loglevel, "%08X  00 01 02 03  04 05 06 07  08 09 0A 0B  0C 0D 0E 0F  0123456789ABCDEF\n", length);
	LogMessage(loglevel, "--------  --------------------------------------------------  ----------------\n");
	data = (UINT8 *)ptr;
	for (n = 0; n < length; n += 16) {
		j = sprintf(string, "%08x ", n);
		for (i = n; i < n + 16; i++) {
			if (i % 4 == 0)
				j += sprintf(string + j, " ");
			if (i < length)
				j += sprintf(string + j, "%02x ", data[i]);

			else
				j += sprintf(string + j, "   ");
		}
		j += sprintf(string + j, " ");
		for (i = n; i < n + 16; i++) {
			if (i < length) {
				if (isprint(data[i]))
					j += sprintf(string + j, "%c", data[i]);

				else
					j += sprintf(string + j, ".");
			}
		}
		LogMessage(loglevel, "%s\n", string);
	}
	return;
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandlers(VOID)
{
	if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
		perror("signal(SIGTERM)");
		return FALSE;
	}
	if (signal(SIGINT, CatchSignal) == SIG_ERR) {
		perror("signal(SIGINT)");
		return FALSE;
	}
#if !defined WIN32
	/* Ignore the SIGPIPE signal */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal(SIGPIPE)");
		return FALSE;
	}
#endif

	return TRUE;
}

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{
	switch (sig) {
	case SIGTERM:
		//fprintf(stderr, "\rCaught SIGTERM! \n");
		RequestQuit(TRUE);
		break;
	case SIGINT:
		//fprintf(stderr, "\rCaught SIGINT! \n");
		RequestQuit(TRUE);
		break;
	default:
		fprintf(stderr, "\rCaught unexpected signal %d, ignored! \n", sig);
		break;
	}

	signal(sig, CatchSignal);
}
