/* $Id: clients.c,v 1.5 2006/04/30 16:24:38 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Client connections...

-----------------------------------------------------------------------*/

#include "clients.h"
#include "server.h"
#include "grfthread.h"

/* Module constants ---------------------------------------------------*/
#define CHECK_INTERVAL	(UST_MSECOND * 200)

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC ClientThread(void *argument);
static BOOL GetPacket(CLIENT * client, GRF_PACKET ** packet);
static BOOL PutPacket(CLIENT * client, GRF_PACKET * packet);
static VOID RequestStop(CLIENT * client);
static BOOL ClientRequestedStop(CLIENT * client);
static BOOL StopRequested(CLIENT * client);
static BOOL Suspended(CLIENT * client);

/*---------------------------------------------------------------------*/
BOOL ClientThreadStart(CLIENT * client)
{
	CHAR string[32];

	ASSERT(client != NULL);

	/* 123456789a123456789b12345678 */
	/* Client:123.123.123.123:12345 */
	sprintf(string, "Client:%s", FormatEndpoint(&client->endpoint, string));

	MUTEX_LOCK(&client->mutex);
	client->stop = FALSE;
	client->suspended = FALSE;
	client->resume = FALSE;
	MUTEX_UNLOCK(&client->mutex);

	if (!THREAD_CREATE(&client->thread_id, ClientThread, client)) {
		LogMessage(LOG_ERR, "ERROR: ClientThreadStart: THREAD_CREATE(%s) failed: %u %s\n",
			string, THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL ClientThreadStop(CLIENT * client)
{
	THREAD thread_id;

	ASSERT(client != NULL);

	MUTEX_LOCK(&client->mutex);
	if (!client->stop)
		client->stop = TRUE;
	thread_id = client->thread_id;
	MUTEX_UNLOCK(&client->mutex);

	//THREAD_JOIN(thread_id);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC ClientThread(void *argument)
{
	CHAR description[32];
	UINT16 result;
	INT32 unused;
	UINT32 id;
	CLIENT *client;
	GRF_PACKET *packet;

	ASSERT(argument != NULL);

	client = (CLIENT *) argument;

	MUTEX_LOCK(&client->mutex);
	client->running = TRUE;
	MUTEX_UNLOCK(&client->mutex);

	LogMessage(LOG_NOTICE, "ClientThread %s(%u), tid:%d, pid:%d, start\n",
		FormatEndpoint(&client->endpoint, description), client->process_id, THREAD_SELF(), getpid());

	client->type = CLI_TYPE_DATA;
	LogMessage(LOG_DEBUG, "ClientThread %s(%u): connection timeout: %f\n",
		description, client->process_id, (REAL32)client->timeout / (REAL32)UST_SECOND);

	/* Register for data */
	LogMessage(LOG_DEBUG, "ClientThread %s(%u): registering for data\n", description, client->process_id);
	if (!GRFRegisterClient(client->endpoint.sin_addr.s_addr, client->process_id, client->queue))
		RequestStop(client);

	while (!ShutDownRequested() && !StopRequested(client)) {
		if (!Suspended(client)) {
			/* Get a packet */
			if (GetPacket(client, &packet)) {
				/* Write to client */
				if (!PutPacket(client, packet)) {
					MUTEX_LOCK(&client->mutex);
					client->suspended = TRUE;
					MUTEX_UNLOCK(&client->mutex);
					SOCK_CLOSE(client->socket);
					client->socket = INVALID_SOCKET;
					LogMessage(LOG_DEBUG, "ClientThread %s(%u): Suspended!\n", description, client->process_id);
				}
			}
		}
		else {
			/* Are we resuming? */
			MUTEX_LOCK(&client->mutex);
			if (client->resume) {
				/* No longer suspended... */
				client->suspended = FALSE;
				MUTEX_UNLOCK(&client->mutex);
				LogMessage(LOG_NOTICE, "ClientThread %s(%u): Resuming connection!\n",
					FormatEndpoint(&client->endpoint, description), client->process_id);
				continue;
			}
			MUTEX_UNLOCK(&client->mutex);

			/* Check if queue is getting full... */
			if ((result = QueueStat(client->queue, Q_UNUSED, &unused)) != Q_NO_ERROR) {
				LogMessage(LOG_ERR, "ERROR: ClientThread %s(%u): QueueStat: %s\n",
					description, client->process_id, QueueErrorString(result));
				RequestShutDown(TRUE);
			}
			if (unused <= 2) {
				MUTEX_LOCK(&client->mutex);
				client->suspended = FALSE;
				client->resume = FALSE;
				MUTEX_UNLOCK(&client->mutex);
				RequestStop(client);
				LogMessage(LOG_DEBUG, "ClientThread %s(%u): Queue is full, shutting down!\n", description, client->process_id);
			}
			USSleep(UST_MSECOND * 200);
		}
	}

	/* Unregister for data and flush queue */
	LogMessage(LOG_DEBUG, "ClientThread %s(%u): unregistering for data\n", description, client->process_id);
	GRFUnregisterClient(client->endpoint.sin_addr.s_addr, client->process_id);

	/* If we're suspended, we need to put the current packet back */
	if (Suspended(client)) {
		if ((result = QueuePut(client->empties, packet, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: ClientThread: QueuePut(empties): %s\n", QueueErrorString(result));
		}
	}

	/* Close the connection */
	LogMessage(LOG_NOTICE, "ClientThread %s(%u): disconnecting from client\n", description, client->process_id);
	SOCK_CLOSE(client->socket);
	client->socket = INVALID_SOCKET;

	/* Destroy our data structures */
	MUTEX_LOCK(&client->mutex);
	id = client->process_id;
	client->running = FALSE;
	MUTEX_UNLOCK(&client->mutex);

	DestroyClient(client);

	/* Decrement the connection counter in the server */
	DecrementConnections();

	LogMessage(LOG_NOTICE, "ClientThread %s(%u), tid:%d, pid:%d, exit\n",
		description, id, THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GetPacket(CLIENT * client, GRF_PACKET ** packet)
{
	UINT16 result;

	ASSERT(client != NULL);
	ASSERT(packet != NULL);

	/* If we're resuming a connection, use the last packet we got */
	MUTEX_LOCK(&client->mutex);
	if (client->resume) {
		client->resume = FALSE;
		MUTEX_UNLOCK(&client->mutex);
		return TRUE;
	}
	MUTEX_UNLOCK(&client->mutex);

	if ((result = QueueGet(client->queue, (VOID **)packet, FALSE)) != Q_NO_ERROR) {
		if (result != Q_EMPTY) {
			LogMessage(LOG_ERR, "ERROR: ClientThread: GetPacket: QueueGet: %s\n", QueueErrorString(result));
			RequestStop(client);
			return FALSE;
		}
		USSleep(CHECK_INTERVAL);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL PutPacket(CLIENT * client, GRF_PACKET * packet)
{
	UINT16 result;

	ASSERT(client != NULL);
	ASSERT(packet != NULL);

	if (!GRFWriteSocket(client->socket, packet, client->timeout))
		return FALSE;

	if ((result = QueuePut(client->empties, packet, FALSE)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: ClientThread: PutPacket: QueuePut(empties): %s\n", QueueErrorString(result));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID RequestStop(CLIENT * client)
{
	ASSERT(client != NULL);

	MUTEX_LOCK(&client->mutex);
	if (!client->stop)
		client->stop = TRUE;
	MUTEX_UNLOCK(&client->mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(CLIENT * client)
{
	BOOL stop;

	ASSERT(client != NULL);

	MUTEX_LOCK(&client->mutex);
	stop = client->stop;
	MUTEX_UNLOCK(&client->mutex);

	if (stop)
		return stop;

	/* Check if the client has sent a stop request */
	return ClientRequestedStop(client);
}

/*---------------------------------------------------------------------*/
static BOOL ClientRequestedStop(CLIENT * client)
{
	CHAR string[32];
	static GRF_PACKET packet;

	ASSERT(client != NULL);

	if (client->socket == INVALID_SOCKET)
		return FALSE;

	/* Read with no blocking */
	GRFReadSocket(client->socket, &packet, 1);
//	if (!GRFReadSocket(client->socket, &packet, 1))
//		return TRUE;

	/* Does the client want to disconnect? */
	if (packet.hdr.type == GRF_TYPE_DISCON) {
		LogMessage(LOG_DEBUG, "ClientThread %s: client requests disconnect!\n",
			FormatClient(client->endpoint.sin_addr.s_addr, client->process_id, string));
		return TRUE;
	}
	/* Pass info packets off to digital I/O to catch message events */
	else if (packet.hdr.type == GRF_TYPE_INFO) {
		LogMessage(LOG_DEBUG, "ClientThread %s: received info packet: '%s'\n",
			FormatClient(client->endpoint.sin_addr.s_addr, client->process_id, string),
			packet.serial.info.message);
		return FALSE; 
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static BOOL Suspended(CLIENT * client)
{
	BOOL suspended;

	ASSERT(client != NULL);

	MUTEX_LOCK(&client->mutex);
	suspended = client->suspended;
	MUTEX_UNLOCK(&client->mutex);

	return suspended;
}
