/* $Id: config.c,v 1.10 2002/10/03 18:29:06 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Configuration for grftrig.

-----------------------------------------------------------------------*/

#include "config.h"
#include "log.h"
#include "mem.h"

/* Module constants ---------------------------------------------------*/
#define MAX_LINE_LENGTH					128
#define MAX_TOKEN_LENGTH				128

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL fatal_error;
static UINT32 line_number, token_number;
static CHAR *config_filespec;
static FILE *config_fptr;
static BOOL opened = FALSE;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL ParseCommandLine(int argc, char *argv[], CMDLINE_ARGS * args);
static VOID CmdLineSetDefaults(CMDLINE_ARGS * args);
static VOID CmdLineHelp(VOID);
static BOOL ParseConfigFile(ARGS * args);
static BOOL ValifyConfiguration(ARGS * args);
static BOOL DefineLogging(CMDLINE_ARGS * args);
static BOOL DefineServer(SVR_ARGS * args);
static BOOL ParseHostList(HOST_LIST ** list);
static BOOL DefineRepository(REPOS_ARGS * args);
static BOOL DefineConnections(GRF_ARGS * args);
static BOOL DefineEndpoints(GRF_ARGS * args);
static BOOL DefineNetwork(GRF_ARGS * args);
static BOOL DefineVoters(NETWORK * net);
static BOOL DefineAmplitudeTrigger(VOTER * voter);
static BOOL DefineSTALTATrigger(VOTER * voter);
static BOOL DefineChannels(NETWORK * net);
static VOID ParseChannelName(GRF_NAME * name, CHAR *token);
static CHAR *GetNextLine(VOID);
static CHAR *GetNextToken(VOID);

/*---------------------------------------------------------------------*/
#include "getopt.c"

/*---------------------------------------------------------------------*/
BOOL Configure(ARGS * args)
{
	CHAR *grf_etc, *home;
	CMDLINE_ARGS cmdline;

	ASSERT(args != NULL);

    /* Set all command line defaults */
	CmdLineSetDefaults(&cmdline);

    /* Check for GRF_ETC environment variable */
    if((grf_etc = getenv("GRF_ETC")) != NULL) {
		if(grf_etc[0] == '~') {	
			if((home = getenv("HOME")) != NULL) {
			    if(grf_etc[strlen(grf_etc) - 1] == '/' || grf_etc[strlen(grf_etc) - 1] == '\\' )
            		sprintf(cmdline.confile, "%s%s%s", home, &grf_etc[1], DFL_CONF_FILE);
		        else
        		    sprintf(cmdline.confile, "%s%s/%s", home, &grf_etc[1], DFL_CONF_FILE);
			}
		} else {
	        if(grf_etc[strlen(grf_etc) - 1] == '/' || grf_etc[strlen(grf_etc) - 1] == '\\' )
    	        sprintf(cmdline.confile, "%s%s", grf_etc, DFL_CONF_FILE);
        	else
            	sprintf(cmdline.confile, "%s/%s", grf_etc, DFL_CONF_FILE);
		}
    }
    else
        sprintf(cmdline.confile, "%s%s", DFL_GRF_ETC, DFL_CONF_FILE);

	/* Parse the command line to get the config filename */
	if (!ParseCommandLine(args->main.argc, args->main.argv, &cmdline))
		exit(1);

	/* Set defaults for everything... */
	CmdLineSetDefaults(&args->cmdline);
	GRFSetDefaults(&args->grf);
	ServerSetDefaults(&args->server);

	/* Parse the configuration file */
	strcpy(args->cmdline.confile, cmdline.confile);
	if (!ParseConfigFile(args))
		exit(1);

	/* Fold in command line options overriding config file settings */
	if (cmdline.quiet || cmdline.verbose || cmdline.debug) {
		args->cmdline.quiet = cmdline.quiet;
		args->cmdline.verbose = cmdline.verbose;
		args->cmdline.debug = cmdline.debug;
	}
	if (cmdline.daemon)
		args->cmdline.daemon = TRUE;
	if (cmdline.facility[0] != '\0')
		strcpy(args->cmdline.facility, cmdline.facility);
	if (cmdline.logfile[0] != '\0')
		strcpy(args->cmdline.logfile, cmdline.logfile);

	/* Validate the configuration */
	if (!ValifyConfiguration(args))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL DestroyNetworks(ARGS * args)
{
	NETWORK *net;
	VOTER *voter;
	CHANNEL *chan;
	PL_NODE *head, *tail, *node;
	VOID *ptr;

	net = args->grf.networks;
	while (net != NULL) {
		voter = net->voters;
		while (voter != NULL) {
			ptr = voter;
			voter = voter->next;
			ReleaseMemory(ptr);
		}
		chan = net->channels;
		while (chan != NULL) {
			head = &chan->packets.head;
			tail = &chan->packets.tail;
			node = head->next;
			while (node != tail) {
				node->prev->next = node->next;
				node->next->prev = node->prev;
				ReleaseMemory((VOID *)node);
				node = head->next;
			}
			ptr = chan;
			chan = chan->next;
			ReleaseMemory(ptr);
		}
		ptr = net;
		net = net->next;
		ReleaseMemory(ptr);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL DestroyHostLists(ARGS * args)
{
	HOST_LIST *node, *list;

	ASSERT(args != NULL);

	/* Release memory in the various host lists... */

	list = args->server.allow;
	while (list != NULL) {
		node = list;
		list = list->next;
		ReleaseMemory(node);
	}
	list = args->server.deny;
	while (list != NULL) {
		node = list;
		list = list->next;
		ReleaseMemory(node);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ParseCommandLine(int argc, char *argv[], CMDLINE_ARGS * args)
{
    BOOL confile;
	int i, option;

	ASSERT(args != NULL);
	ASSERT(argv != NULL);

	/* Don't write errors to stderr */
	my_opterr = 0;

	/* Options... */
	while ((option = my_getopt(argc, argv, "hvqdbl:s:")) != -1) {
		switch (option) {
		case 'h':
			CmdLineHelp();
			break;
		case 'v':
			args->verbose = TRUE;
			break;
		case 'q':
			args->quiet = TRUE;
			break;
		case 'd':
			args->debug = TRUE;
			break;
		case 'b':
			args->daemon = TRUE;
			break;
		case 'l':
			if (my_optarg)
				strncpy(args->logfile, my_optarg, MAX_PATH_LEN);
			else
				args->logfile[0] = '\0';
			break;
		case 's':
			if (my_optarg)
				strncpy(args->facility, my_optarg, MAX_PATH_LEN);
			else
				args->facility[0] = '\0';
			break;
		case '?':
			if (isprint(my_optopt))
				fprintf(stderr, "Unknown option `-%c'.\n", my_optopt);
			else
				fprintf(stderr, "Unknown option character `\\x%x'.\n", my_optopt);
			return FALSE;
		default:
			abort();
		}
	}

	/* Arguments... */
    confile = FALSE;
	for (i = my_optind; i < argc; i++) {
		if (confile == FALSE) {
            confile = TRUE;
			strcpy(args->confile, argv[i]);
		}
		else {
			fprintf(stderr, "unrecognized argument: `%s'\n", argv[i]);
			return FALSE;
		}
	}

	/* If we weren't passed a config filename, use the default */
	if (args->confile[0] == '\0')
		strcpy(args->confile, DFL_CONF_FILE);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID CmdLineSetDefaults(CMDLINE_ARGS * args)
{
	ASSERT(args != NULL);

	args->verbose = FALSE;
	args->quiet = FALSE;
	args->debug = FALSE;
	args->daemon = FALSE;
	args->pidfile[0] = '\0';
	args->facility[0] = '\0';
	args->logfile[0] = '\0';
	args->confile[0] = '\0';

	return;
}

/*---------------------------------------------------------------------*/
static VOID CmdLineHelp(VOID)
{
#if !defined WIN32
	LogMessage(LOG_NOTICE, "Usage: grftrig [-hvqdb] [-l logfile] [-s facility] [configuration file]\n");
#else
	LogMessage(LOG_NOTICE, "Usage: grftrig [[-hvqdb] [-l logfile] [configuration file]] |\n");
    LogMessage(LOG_NOTICE, "               [-install | -remove | -start | -stop | -status]\n");
#endif
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "   -h           Help display.\n");
	LogMessage(LOG_NOTICE, "   -v           Verbose logging.\n");
	LogMessage(LOG_NOTICE, "   -q           Quiet logging.\n");
	LogMessage(LOG_NOTICE, "   -d           Debug logging.\n");
	LogMessage(LOG_NOTICE, "   -b           Run in background. (No) \n");
	LogMessage(LOG_NOTICE, "   -l logfile   Log to logfile. (Off)\n");
#if !defined WIN32
    LogMessage(LOG_NOTICE, "   -s facility  Log to syslog facility, `LOCAL0'-`LOCAL7'. (Off)\n");
#else
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "   -install     Install as a Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -remove      Uninstall Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -start       Start Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -stop        Stop Win32 service.\n");
    LogMessage(LOG_NOTICE, "   -status      Report Win32 service state.\n");
#endif
	LogMessage(LOG_NOTICE, "\n");
	LogMessage(LOG_NOTICE, "The configuration file defaults to %s%s if not specified.\n", DFL_GRF_ETC, DFL_CONF_FILE);
	LogMessage(LOG_NOTICE, "\n");
    LogMessage(LOG_NOTICE, "[] = optional, () = default, | = mutually exclusive.\n");

	exit(0);
}

/*---------------------------------------------------------------------*/
static BOOL ParseConfigFile(ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	config_filespec = args->cmdline.confile;
	line_number = token_number = 0;
	fatal_error = FALSE;

	LogMessage(LOG_INFO, "Parsing configuration file: %s\n", config_filespec);

	while ((token = GetNextToken()) != NULL) {
		if (strncasecmp(token, "Daemon", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			/* "Yes", "True", or 1 */
			if (toupper(*token) == 'Y' || toupper(*token) == 'T' || *token == '1')
				args->cmdline.daemon = TRUE;
			/* Anything else is FALSE */
			else
				args->cmdline.daemon = FALSE;
		}
		else if (strncasecmp(token, "PIDFile", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->cmdline.pidfile, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "DateFormat", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			USTSetDateRepresentation(token);
		}
		else if (strncasecmp(token, "Logging", 7) == 0) {
			if (!DefineLogging(&args->cmdline))
				break;
		}
		else if (strncasecmp(token, "Server", 6) == 0) {
			if (!DefineServer(&args->server))
				break;
		}
		else if (strncasecmp(token, "Connections", 11) == 0) {
			if (!DefineConnections(&args->grf))
				break;
		}
		else if (strncasecmp(token, "Network", 7) == 0) {
			if (!DefineNetwork(&args->grf))
				break;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			break;
		}
	}

	if (opened) {
		fclose(config_fptr);
		opened = FALSE;
	}

	if (fatal_error)
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ValifyConfiguration(ARGS * args)
{
	BOOL result;
	NETWORK *net;

	ASSERT(args != NULL);

	result = TRUE;

	/* At least one network must be defined... */
	if (args->grf.networks == NULL) {
		LogMessage(LOG_ERR, "At least one network must be defined!\n");
		result = FALSE;
	}

	/* Each network must contain at least one channel and one voter */
	for (net = args->grf.networks; net != NULL; net = net->next) {
		if (net->voters == NULL) {
			LogMessage(LOG_ERR, "Network '%s' has no voting channel triggers!\n", net->name);
			result = FALSE;
		}
		if (net->channels == NULL) {
			LogMessage(LOG_ERR, "Network '%s' has no data channels!\n", net->name);
			result = FALSE;
		}
	}

	return result;
}

/*---------------------------------------------------------------------*/
static BOOL DefineLogging(CMDLINE_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	/* Process statements in scope */
	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Syslog", 6) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strcpy(args->facility, token);
		}
		else if (strncasecmp(token, "File", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strcpy(args->logfile, token);
		}
		else if (strncasecmp(token, "Level", 5) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			if (toupper(*token) == 'Q')
				args->quiet = TRUE;
			else if (toupper(*token) == 'V')
				args->verbose = TRUE;
			else if (toupper(*token) == 'D')
				args->debug = TRUE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineServer(SVR_ARGS * args)
{
	CHAR *token;
	UINT32 unit_id;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "FriendlyName", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			memset(args->name, 0, MAX_FRIENDLY_NAME_LEN + 1);
			strncpy(args->name, token, MAX_FRIENDLY_NAME_LEN);
		}
		else if (strncasecmp(token, "GRFUnitID", 9) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			unit_id = strtoul(token, NULL, 0);
			GRFSetUnitNumber(unit_id);
		}
		else if (strncasecmp(token, "Endpoint", 8) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			ParseEndpoint(&args->listen, token, DFL_PORT);
		}
		else if (strncasecmp(token, "AllowHosts", 10) == 0) {
			if (!ParseHostList(&args->allow))
				return FALSE;
		}
		else if (strncasecmp(token, "DenyHosts", 9) == 0) {
			if (!ParseHostList(&args->deny))
				return FALSE;
		}
		else if (strncasecmp(token, "MaxConnections", 14) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->max_connections = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->queue_depth = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Repository", 10) == 0) {
			if (!DefineRepository(&args->repos))
				return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ParseHostList(HOST_LIST ** list)
{
	CHAR *token, *ptr;
	INT32 i, n;
	UINT32 netmask;
	struct hostent *hptr;
	HOST_LIST *node;

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}

		/* Allocate a new node */
		if (!AllocateMemory((void **)&node, sizeof(HOST_LIST))) {
			LogMessage(LOG_ERR, "ERROR: ParseHostList(): Unable to allocate memory!\n");
			return FALSE;
		}

		/* Get the subnet mask */
		netmask = 0xFFFFFFFF;
		if ((ptr = strchr(token, '/')) != NULL) {
			n = strtoul(ptr + 1, NULL, 0);
			if (n < 32) {
				netmask = 0;
				for (i = 0; i < n; i++) {
					netmask |= (0x00000001 << (32 - i - 1));
				}
			}
			*ptr = '\0';
		}
		node->netmask = htonl(netmask);

		/* Get the host address */
		if (isdigit(token[0])) {
			if ((node->address = inet_addr(token)) == INADDR_NONE) {
				LogMessage(LOG_ERR, "Invalid address string: %s\n", token);
			}
		}
		else if ((hptr = gethostbyname(token)) != NULL) {
			node->address = *(UINT32 *)hptr->h_addr_list[0];
		}
		else {
			LogMessage(LOG_ERR, "Unable to resolve name: %s\n", token);
		}

		/* Link into list at head */
		node->next = (*list);
		(*list) = node;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineRepository(REPOS_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Path", 4) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->path, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "NameFormat", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			strncpy(args->name_fmt, token, MAX_PATH_LEN);
		}
		else if (strncasecmp(token, "RecordLength", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->rec_length = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->queue_depth = strtoul(token, NULL, 0);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineConnections(GRF_ARGS * args)
{
	CHAR *token;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "QueueDepth", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->con_q_depth = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "Timeout", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			args->con_timeout = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "Endpoints", 9) == 0) {
			if (!DefineEndpoints(args))
				return FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineEndpoints(GRF_ARGS * args)
{
	CHAR *token;
	CONNECTION *node = NULL;

	ASSERT(args != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}

		if (!AllocateMemory((VOID **)&node, sizeof(CONNECTION))) {
			LogMessage(LOG_ERR, "ERROR: DefineEndpoints: Unable to allocate memory!\n");
			fatal_error = TRUE;
			return FALSE;
		}

		ParseEndpoint(&node->endpoint, token, DFL_PORT);

		/* Link into list */
		node->next = args->connections;
		args->connections = node;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineNetwork(GRF_ARGS * args)
{
	CHAR *token;
	NETWORK *net;
	static UINT32 count = 0;

	ASSERT(args != NULL);

	if (!AllocateMemory((VOID **)&net, sizeof(NETWORK))) {
		LogMessage(LOG_ERR, "ERROR: DefineNetwork: Unable to allocate memory!\n");
		fatal_error = TRUE;
		return FALSE;
	}
	SetNetworkDefaults(net);

	/* Insert and head of net list */
	net->next = args->networks;
	args->networks = net;

	/* Find network name or scoping token */
	count++;
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token == '{') {
		sprintf(net->name, "Net%u", count);
	}
	else {
		strncpy(net->name, token, GRF_MAX_NET_LEN);
		if ((token = GetNextToken()) == NULL)
			return (FALSE);
		if (*token != '{') {
			LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Voters", 6) == 0) {
			if (!DefineVoters(net))
				return FALSE;
		}
		else if (strncasecmp(token, "Channels", 8) == 0) {
			if (!DefineChannels(net))
				return FALSE;
		}
		else if (strncasecmp(token, "TriggerVotes", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->trig_votes = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "DetriggerVotes", 14) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->detrig_votes = strtoul(token, NULL, 0);
		}
		else if (strncasecmp(token, "CoincidenceWindow", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->window = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "PreEventSeconds", 15) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->pre_event = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "PostEventSeconds", 16) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->post_event = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "MinEventSeconds", 15) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->min_event = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "MaxEventSeconds", 15) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			net->max_event = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineVoters(NETWORK * net)
{
	CHAR *token;
	VOTER *voter;

	ASSERT(net != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Trigger", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;

			if (!AllocateMemory((VOID **)&voter, sizeof(VOTER))) {
				LogMessage(LOG_ERR, "ERROR: DefineVoters: Unable to allocate memory!\n");
				fatal_error = TRUE;
				return FALSE;
			}
			memset(voter, 0, sizeof(voter));

			/* Insert and head of channel list */
			voter->next = net->voters;
			net->voters = voter;

			if (strncasecmp(token, "STA/LTA", 7) == 0) {
				if (!DefineSTALTATrigger(voter))
					return FALSE;
			}
			else if (strncasecmp(token, "Amplitude", 9) == 0) {
				if (!DefineAmplitudeTrigger(voter))
					return FALSE;
			}
			else {
				LogMessage(LOG_ERR, "Unrecognized trigger type at %s line %lu: '%s'\n", config_filespec, line_number, token);
				fatal_error = TRUE;
				return (FALSE);
			}
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineAmplitudeTrigger(VOTER * voter)
{
	CHAR *token;

	ASSERT(voter != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	SetTriggerDefaults(&voter->trigger, TRIG_AMP);

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Channel", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			ParseChannelName(&voter->name, token);
		}
		else if (strncasecmp(token, "MeanSeconds", 11) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.amp.mean_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "TriggerAmplitude", 16) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.amp.trig_amp = strtol(token, NULL, 0);
		}
		else if (strncasecmp(token, "TriggerSeconds", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.amp.trig_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "DetriggerAmplitude", 16) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.amp.detrig_amp = strtol(token, NULL, 0);
		}
		else if (strncasecmp(token, "DetriggerSeconds", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.amp.detrig_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineSTALTATrigger(VOTER * voter)
{
	CHAR *token;

	ASSERT(voter != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	SetTriggerDefaults(&voter->trigger, TRIG_STALTA);

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}
		else if (strncasecmp(token, "Channel", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			ParseChannelName(&voter->name, token);
		}
		else if (strncasecmp(token, "MeanSeconds", 11) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.stalta.mean_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "STASeconds", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.stalta.sta_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "LTASeconds", 10) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.stalta.lta_sec = (USTIME)(atof(token) * (REAL64)UST_SECOND);
		}
		else if (strncasecmp(token, "TriggerRatio", 12) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.stalta.trig_ratio = (REAL32)atof(token);
		}
		else if (strncasecmp(token, "DetriggerRatio", 14) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			voter->trigger.prm.stalta.detrig_ratio = (REAL32)atof(token);
		}
		else if (strncasecmp(token, "LTAHold", 7) == 0) {
			if ((token = GetNextToken()) == NULL)
				break;
			/* "Yes", "True", or 1 */
			if (toupper(*token) == 'Y' || toupper(*token) == 'T' || *token == '1')
				voter->trigger.prm.stalta.lta_hold = TRUE;
			else
				voter->trigger.prm.stalta.lta_hold = FALSE;
		}
		else {
			LogMessage(LOG_ERR, "Out of scope or unrecognized statement at %s line %lu: '%s'\n",
				config_filespec, line_number, token);
			fatal_error = TRUE;
			return (FALSE);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL DefineChannels(NETWORK * net)
{
	CHAR *token;
	CHANNEL *chan;

	ASSERT(net != NULL);

	/* Find scoping token */
	if ((token = GetNextToken()) == NULL)
		return (FALSE);
	if (*token != '{') {
		LogMessage(LOG_ERR, "Expected '{' at %s line %lu: '%s'\n", config_filespec, line_number, token);
		fatal_error = TRUE;
		return (FALSE);
	}

	while ((token = GetNextToken()) != NULL) {
		if (*token == '}') {
			break;
		}

		if (!AllocateMemory((VOID **)&chan, sizeof(CHANNEL))) {
			LogMessage(LOG_ERR, "ERROR: DefineChannels: Unable to allocate memory!\n");
			fatal_error = TRUE;
			return FALSE;
		}
		SetChannelDefaults(chan);

		/* Insert and head of channel list */
		chan->next = net->channels;
		net->channels = chan;

		ParseChannelName(&chan->name, token);
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID ParseChannelName(GRF_NAME * name, CHAR *token)
{
	CHAR *ptr;

	ASSERT(name != NULL);
	ASSERT(token != NULL);

	strcpy(name->network, "*");
	strcpy(name->station, "*");
	strcpy(name->component, "*");

	/* Component... */
	if ((ptr = strrchr(token, ':')) == NULL) {
		strncpy(name->component, token, GRF_MAX_COMP_LEN);
		return;
	}
	strncpy(name->component, ptr + 1, GRF_MAX_COMP_LEN);
	*ptr = '\0';

	/* Station... */
	if ((ptr = strrchr(token, ':')) == NULL) {
		strncpy(name->station, token, GRF_MAX_STN_LEN);
		return;
	}
	strncpy(name->station, ptr + 1, GRF_MAX_STN_LEN);
	*ptr = '\0';

	/* Network */
	if (token[0] != '\0')
		strncpy(name->network, token, GRF_MAX_NET_LEN);

	return;
}

/*---------------------------------------------------------------------*/
CHAR *GetNextLine(VOID)
{
	static CHAR line_buffer[MAX_LINE_LENGTH + 1];
	BOOL quote;
	CHAR *cptr = NULL;

	if (!opened) {
		if ((config_fptr = fopen(config_filespec, "r")) == NULL) {
			LogMessage(LOG_ERR, "ERROR: fopen(%s) failed: %s\n", config_filespec, strerror(errno));
			fatal_error = TRUE;
			return (NULL);
		}
		line_number = 0;
		opened = TRUE;
	}

	/* Read and cleanup net line... */
	memset(line_buffer, '\0', MAX_LINE_LENGTH);
	while (fgets(line_buffer, MAX_LINE_LENGTH, config_fptr) != NULL) {
		line_number++;
		token_number = 0;

		/* Trim trailing crap (CR, LF, white space, etc...) */
		cptr = line_buffer + MAX_LINE_LENGTH;
		while (!(isalnum(*cptr) || ispunct(*cptr)) && cptr >= line_buffer) {
			*cptr = '\0';
			cptr--;
		}
		if (cptr < line_buffer) {
			cptr = NULL;
			continue;
		}

		/* Look for and dispose of comments anywhere in the line */
		quote = FALSE;
		cptr = line_buffer;
		while (*cptr) {
			if (*cptr == '"') {
				if (quote)
					quote = FALSE;
				else
					quote = TRUE;
			}
			if (!quote && *cptr == '#') {
				do {
					*cptr = '\0';
					if (cptr == line_buffer)
						break;
					cptr--;
				}
				while (!(isalnum(*cptr) || ispunct(*cptr)));
				break;
			}
			cptr++;
		}

		if (line_buffer[0] == '\0') {
			cptr = NULL;
			memset(line_buffer, '\0', MAX_LINE_LENGTH);
			continue;
		}

		/* Trim off leading whitespace */
		cptr = line_buffer;
		while (!(isalnum(*cptr) || ispunct(*cptr)) && *cptr != '\0')
			cptr++;

		/* Success */
		return (cptr);
	}

	/* Check for read error... */
	if (ferror(config_fptr)) {
		LogMessage(LOG_ERR, "ERROR: Reading config file: %s\n", strerror(errno));
		fatal_error = TRUE;
	}

	return (NULL);
}

/*---------------------------------------------------------------------*/
CHAR *GetNextToken(VOID)
{
	static CHAR *cptr = NULL, token[MAX_TOKEN_LENGTH + 1];
	BOOL quote;
	UINT16 i;

	quote = FALSE;

	while (TRUE) {
		if (!cptr) {
			if ((cptr = GetNextLine()) == NULL)
				break;
		}
		else {
			while (*cptr && (isalnum(*cptr) || ispunct(*cptr)))
				cptr++;
		}

		while (*cptr && !isalnum(*cptr) && !ispunct(*cptr))
			cptr++;

		if (*cptr) {
			if (*cptr == '"') {
				cptr++;
				quote = TRUE;
			}
			i = 0;
			while (i < MAX_TOKEN_LENGTH && cptr[i] && cptr[i] != ',' && (isalnum(cptr[i]) || ispunct(cptr[i]) || quote)) {
				if (quote && cptr[i] == '"') {
					cptr += i;
					break;
				}
				token[i] = cptr[i];
				i++;
			}
			token[i] = '\0';
			token_number++;
			return (token);
		}
		else
			cptr = NULL;
	}

	return (cptr);
}

/*---------------------------------------------------------------------*/
VOID DumpConfiguration(ARGS * args)
{
	CHAR string1[32], string2[32];
	HOST_LIST *node;
	CONNECTION *connection;
	NETWORK *net;
	CHANNEL *chan;
	VOTER *voter;

	ASSERT(args != NULL);

	LogMessage(LOG_INFO, "Configuration dump:\n");

	LogMessage(LOG_INFO, "Command line:\n");
	LogMessage(LOG_INFO, " Daemon:                  %s\n", (args->cmdline.daemon ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Verbose:                 %s\n", (args->cmdline.verbose ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Quiet:                   %s\n", (args->cmdline.quiet ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Debug:                   %s\n", (args->cmdline.debug ? "Yes" : "No"));
	LogMessage(LOG_INFO, " Facility                 %s\n",
		(args->cmdline.facility[0] != '\0' ? args->cmdline.facility : "None"));
	LogMessage(LOG_INFO, " Log file                 %s\n",
		(args->cmdline.logfile[0] != '\0' ? args->cmdline.logfile : "None"));
	LogMessage(LOG_INFO, " Config file              %s\n",
		(args->cmdline.confile[0] != '\0' ? args->cmdline.confile : "None"));

	LogMessage(LOG_INFO, "Server:\n");
	LogMessage(LOG_INFO, " Listen at:               %s\n", FormatEndpoint(&args->server.listen, string1));
	LogMessage(LOG_INFO, " Max connections          %u\n", args->server.max_connections);
	LogMessage(LOG_INFO, " Queue depth              %u\n", args->server.queue_depth);

	LogMessage(LOG_INFO, " Allow hosts:\n");
	for (node = args->server.allow; node != NULL; node = node->next) {
		LogMessage(LOG_INFO, "  %s (%s)\n",
			inet_ntop(AF_INET, &node->address, string1, 32), inet_ntop(AF_INET, &node->netmask, string2, 32));
	}
	LogMessage(LOG_INFO, " Deny hosts:\n");
	for (node = args->server.deny; node != NULL; node = node->next) {
		LogMessage(LOG_INFO, "  %s (%s)\n",
			inet_ntop(AF_INET, &node->address, string1, 32), inet_ntop(AF_INET, &node->netmask, string2, 32));
	}

	if (args->server.repos.path[0] != '\0' && args->server.repos.rec_length != 0) {
		LogMessage(LOG_INFO, " Repository:\n");
		LogMessage(LOG_INFO, "  Path:                   %s\n", args->server.repos.path);
		LogMessage(LOG_INFO, "  Name format:            %s\n", args->server.repos.name_fmt);
		LogMessage(LOG_INFO, "  Record length:          %u seconds\n", (UINT32)args->server.repos.rec_length);
		LogMessage(LOG_INFO, "  Queue depth:            %u\n", (UINT32)args->server.repos.queue_depth);
	}

	LogMessage(LOG_INFO, "Connections:\n");
	LogMessage(LOG_INFO, " Queue depth              %u\n", args->grf.con_q_depth);
	LogMessage(LOG_INFO, " Timeout                  %lf seconds\n", (REAL64)args->grf.con_timeout / (REAL64)UST_SECOND);
	LogMessage(LOG_INFO, " Endpoints:\n");
	for (connection = args->grf.connections; connection != NULL; connection = connection->next)
		LogMessage(LOG_INFO, "  %s\n", FormatEndpoint(&connection->endpoint, string1));

	LogMessage(LOG_INFO, "Networks:\n");
	for (net = args->grf.networks; net != NULL; net = net->next) {
		LogMessage(LOG_INFO, " %s:\n", net->name);
		LogMessage(LOG_INFO, "  Voters:\n");
		for (voter = net->voters; voter != NULL; voter = voter->next) {
			LogMessage(LOG_INFO, "   Channel:               %s:%s:%s\n",
				voter->name.network, voter->name.station, voter->name.component);
			if (voter->trigger.type == TRIG_AMP) {
				LogMessage(LOG_INFO, "    Trigger:              Amplitude\n");
				LogMessage(LOG_INFO, "     Mean removal:        %.2lf seconds\n",
					(REAL64)voter->trigger.prm.amp.mean_sec / (REAL64)UST_SECOND);
				LogMessage(LOG_INFO, "     Trigger amplitude:   %d counts\n", voter->trigger.prm.amp.trig_amp);
				LogMessage(LOG_INFO, "     Trigger seconds:     %.2lf seconds\n",
					(REAL64)voter->trigger.prm.amp.trig_sec / (REAL64)UST_SECOND);
				LogMessage(LOG_INFO, "     Detrigger amplitude: %d counts\n", voter->trigger.prm.amp.detrig_amp);
				LogMessage(LOG_INFO, "     Detrigger seconds:   %.2lf seconds\n",
					(REAL64)voter->trigger.prm.amp.detrig_sec / (REAL64)UST_SECOND);
			}
			else if (voter->trigger.type == TRIG_STALTA) {
				LogMessage(LOG_INFO, "    Trigger:              STA/LTA\n");
				LogMessage(LOG_INFO, "     Mean removal:        %.2lf seconds\n",
					(REAL64)voter->trigger.prm.stalta.mean_sec / (REAL64)UST_SECOND);
				LogMessage(LOG_INFO, "     STA length:          %.2lf seconds\n",
					(REAL64)voter->trigger.prm.stalta.sta_sec / (REAL64)UST_SECOND);
				LogMessage(LOG_INFO, "     LTA length:          %.2lf seconds\n",
					(REAL64)voter->trigger.prm.stalta.lta_sec / (REAL64)UST_SECOND);
				LogMessage(LOG_INFO, "     Trigger ratio:       %.2f\n", voter->trigger.prm.stalta.trig_ratio);
				LogMessage(LOG_INFO, "     Detrigger ratio:     %.2f\n", voter->trigger.prm.stalta.detrig_ratio);
				LogMessage(LOG_INFO, "     LTA hold:            %s\n", (voter->trigger.prm.stalta.lta_hold ? "Yes" : "No"));
			}
		}
		LogMessage(LOG_INFO, "  Trigger votes:          %u\n", net->trig_votes);
		LogMessage(LOG_INFO, "  Detrigger votes:        %u\n", net->detrig_votes);
		LogMessage(LOG_INFO, "  Coincidence window:     %.2lf seconds\n", (REAL64)net->window / (REAL64)UST_SECOND);
		LogMessage(LOG_INFO, "  Pre-event length:       %.2lf seconds\n", (REAL64)net->pre_event / (REAL64)UST_SECOND);
		LogMessage(LOG_INFO, "  Post-event length:      %.2lf seconds\n", (REAL64)net->post_event / (REAL64)UST_SECOND);
		LogMessage(LOG_INFO, "  Minimum event length:   %.2lf seconds\n", (REAL64)net->min_event / (REAL64)UST_SECOND);
		LogMessage(LOG_INFO, "  Maximum event length:   %.2lf seconds\n", (REAL64)net->max_event / (REAL64)UST_SECOND);
		LogMessage(LOG_INFO, "  Channels:\n");
		for (chan = net->channels; chan != NULL; chan = chan->next)
			LogMessage(LOG_INFO, "   %s:%s:%s\n", chan->name.network, chan->name.station, chan->name.component);

	}

	return;
}
