/* $Id: connection.c,v 1.5 2001/06/29 16:04:32 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Server connections...

-----------------------------------------------------------------------*/

#include "connection.h"
#include "grfthread.h"

/* Module constants ---------------------------------------------------*/
#define CHECK_INTERVAL	(UST_MSECOND * 200)
#define RECONNECT_INTERVAL (UST_SECOND * 5)

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static GRF_PACKET dpacket;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC ConnectionThread(void *argument);
static BOOL GetPacket(CONNECTION * connection, GRF_PACKET * packet);
static VOID RequestStop(CONNECTION * connection);
static BOOL StopRequested(CONNECTION * connection);

/*---------------------------------------------------------------------*/
BOOL ConnectionThreadStart(CONNECTION * connection)
{
	CHAR string[32];

	ASSERT(connection != NULL);

	if (!THREAD_CREATE(&connection->thread_id, ConnectionThread, connection)) {
		LogMessage(LOG_ERR, "ERROR: ConnectionThreadStart: THREAD_CREATE(%s) failed: %u %s\n",
			FormatEndpoint(&connection->endpoint, string), THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC ConnectionThread(void *argument)
{
	CHAR description[32];
	UINT16 result;
	CONNECTION *connection;
	GRF_PACKET *packet;
	GRF_UNIT *list = NULL, *unit;

	ASSERT(argument != NULL);

	connection = (CONNECTION *) argument;

	MUTEX_LOCK(&connection->mutex);
	connection->running = TRUE;
	MUTEX_UNLOCK(&connection->mutex);

	LogMessage(LOG_NOTICE, "ConnectionThread %s (tid:%d, pid:%d) start\n",
		FormatEndpoint(&connection->endpoint, description), THREAD_SELF(), getpid());

	packet = NULL;
	while (!ShutDownRequested() && !StopRequested(connection)) {
		/* Get an empty packet... */
		if (packet == NULL) {
			if ((result = QueueGet(connection->empties, (VOID **)&packet, FALSE)) != Q_NO_ERROR) {
				if (result != Q_EMPTY) {
					LogMessage(LOG_ERR, "ERROR: ConnectionThread(%s): QueueGet(empties): %s\n",
						description, QueueErrorString(result));
				}
				USSleep(CHECK_INTERVAL);
				continue;
			}
		}

		/* Get a packet from our server */
		if (GetPacket(connection, packet)) {
			/* Timed out? */
			if (packet->hdr.length == 0) {
				USSleep(CHECK_INTERVAL);
				continue;
			}
			/* Got a packet */
#if 0
			LogMessage(LOG_DEBUG, "%s -> %05u %s:%s:%s %s\n", description,
				packet->hdr.sequence,
				packet->payload_hdr.data.name.network,
				packet->payload_hdr.data.name.station,
				packet->payload_hdr.data.name.component,
				FormatUSTime(string, 32, UST_DOY_TIME_USEC, packet->payload_hdr.data.timing.time +
					packet->payload_hdr.data.timing.time_correction));
#endif

			/* Check sequence numbers */
			for (unit = list; unit != NULL; unit = unit->next) {
				if (packet->hdr.unit == unit->unit_id)
					break;
			}
			if (unit == NULL) {
				if (!AllocateMemory((VOID **)&unit, sizeof(GRF_UNIT))) {
					LogMessage(LOG_ERR, "ERROR: Unable to allocate memory!\n");
					break;
				}
				unit->unit_id = packet->hdr.unit;
				unit->sequence = packet->hdr.sequence;
				unit->next = list;
				list = unit;
			}
			if (packet->hdr.sequence != unit->sequence) {
				LogMessage(LOG_ERR, "ConnectionThread(%s): sequence break on unit %u, got:%u, expected: %u\n",
					description, unit->unit_id, packet->hdr.sequence, unit->sequence);
			}
			unit->sequence = packet->hdr.sequence + 1;

			if ((result = QueuePut(connection->input, packet, FALSE)) != Q_NO_ERROR) {
				LogMessage(LOG_ERR, "ERROR: ConnectionThread(%s): PutPacket: QueuePut(input): %s\n",
					description, QueueErrorString(result));
				SOCK_CLOSE(connection->socket);
				connection->socket = INVALID_SOCKET;
				RequestStop(connection);
				RequestShutDown(TRUE);
			}
			packet = NULL;
		}
	}

	if (connection->socket != INVALID_SOCKET) {
		GRFDisconnect(connection->socket, &dpacket);
		SOCK_CLOSE(connection->socket);
		connection->socket = INVALID_SOCKET;
	}

	/* Put the packet back into the empties queue */
	if (packet != NULL) {
		if ((result = QueuePut(connection->input, packet, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: ConnectionThread(%s): PutPacket: QueuePut(input): %s\n",
				description, QueueErrorString(result));
			RequestShutDown(TRUE);
		}
	}

	/* Release sequence check list */
	while (list != NULL) {
		unit = list;
		list = list->next;
		ReleaseMemory(unit);
	}

	/* Destroy our data structures */
	MUTEX_LOCK(&connection->mutex);
	connection->running = FALSE;
	MUTEX_UNLOCK(&connection->mutex);

	LogMessage(LOG_NOTICE, "ConnectionThread %s (tid:%d, pid:%d) exit\n", description, THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GetPacket(CONNECTION * connection, GRF_PACKET * packet)
{
	CHAR string[32];

	ASSERT(connection != NULL);
	ASSERT(packet != NULL);

	/* If socket is not connected... */
	if (connection->socket == INVALID_SOCKET) {
		if ((connection->socket = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
			LogMessage(LOG_ERR, "ERROR: ConnectionThread(%s): socket: %s\n",
				FormatEndpoint(&connection->endpoint, string), SOCK_ERRSTRING(SOCK_ERRNO()));
			RequestShutDown(TRUE);
			return FALSE;
		}
		connection->endpoint.sin_family = AF_INET;
		if (connect(connection->socket, (struct sockaddr *)&connection->endpoint, sizeof(ENDPOINT)) != 0) {
			LogMessage(LOG_ERR, "ERROR: ConnectionThread(%s): connect: %s\n",
				FormatEndpoint(&connection->endpoint, string), SOCK_ERRSTRING(SOCK_ERRNO()));
			SOCK_CLOSE(connection->socket);
			connection->socket = INVALID_SOCKET;
			USSleep(RECONNECT_INTERVAL);
			return FALSE;
		}
		if (!GRFConnect(connection->socket, &connection->endpoint, GRF_ATTRIB_DATA, GRF_DFL_TIMEOUT, packet)) {
			SOCK_CLOSE(connection->socket);
			connection->socket = INVALID_SOCKET;
			USSleep(RECONNECT_INTERVAL);
			return FALSE;
		}
	}

	/* Read a packet */
	if (!GRFReadSocket(connection->socket, packet, connection->timeout)) {
		SOCK_CLOSE(connection->socket);
		connection->socket = INVALID_SOCKET;
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID RequestStop(CONNECTION * connection)
{
	ASSERT(connection != NULL);

	MUTEX_LOCK(&connection->mutex);
	connection->stop = TRUE;
	MUTEX_UNLOCK(&connection->mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(CONNECTION * connection)
{
	BOOL stop;

	ASSERT(connection != NULL);

	MUTEX_LOCK(&connection->mutex);
	stop = connection->stop;
	MUTEX_UNLOCK(&connection->mutex);

	return stop;
}
