/* $Id: log.c,v 1.6 2002/10/02 20:44:33 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Logging functions.

-----------------------------------------------------------------------*/

#include "log.h"
#include "win32svc.h"

#include <ustime.h>

/* Module constants ---------------------------------------------------*/
#define MAX_PROGNAME	32

/* Module types -------------------------------------------------------*/
typedef struct TAG_LOGGING {
	MUTEX mutex;
	CHAR progname[MAX_PROGNAME + 1];
	UINT32 console_level;
	UINT32 file_level;
	UINT32 syslog_level;
	FILE *fptr;
	BOOL syslog;
	int facility;
} LOGGING;

/* Module globals -----------------------------------------------------*/
static LOGGING logging;
static BOOL logging_initialized = FALSE;

/* Module prototypes --------------------------------------------------*/
VOID LogClose(VOID);

/* Console output routine defined in console.c */
extern VOID ConsoleOutput(CHAR *format, ...);

/*---------------------------------------------------------------------*/
BOOL LogOpen(CHAR *progname, CHAR *facility, CHAR *filespec)
{

	ASSERT(progname != NULL);

	if (!logging_initialized) {
		MUTEX_INIT(&logging.mutex);
		logging_initialized = TRUE;
	}

	MUTEX_LOCK(&logging.mutex);

	logging.console_level = LOG_NOTICE;
	logging.file_level = LOG_NOTICE;
	logging.syslog_level = LOG_NOTICE;

	logging.fptr = NULL;
	logging.syslog = FALSE;
	strncpy(logging.progname, progname, MAX_PROGNAME);

	logging_initialized = TRUE;

#if defined HAVE_SYSLOG
	if (facility != NULL) {
		logging.facility = -1;
		if (strncasecmp(facility, "local0", 6) == 0)
			logging.facility = LOG_LOCAL0;
		else if (strncasecmp(facility, "local1", 6) == 0)
			logging.facility = LOG_LOCAL1;
		else if (strncasecmp(facility, "local2", 6) == 0)
			logging.facility = LOG_LOCAL2;
		else if (strncasecmp(facility, "local3", 6) == 0)
			logging.facility = LOG_LOCAL3;
		else if (strncasecmp(facility, "local4", 6) == 0)
			logging.facility = LOG_LOCAL4;
		else if (strncasecmp(facility, "local5", 6) == 0)
			logging.facility = LOG_LOCAL5;
		else if (strncasecmp(facility, "local6", 6) == 0)
			logging.facility = LOG_LOCAL6;
		else if (strncasecmp(facility, "local6", 6) == 0)
			logging.facility = LOG_LOCAL7;

		if (logging.facility != -1) {
			logging.syslog = TRUE;
			openlog(logging.progname, LOG_PID, logging.facility);
		}
	}
#endif

	if (filespec != NULL) {
		/* Try to open the log file... */
		if (filespec[0] != '\0') {
			if ((logging.fptr = fopen(filespec, "at")) == NULL) {
				MUTEX_UNLOCK(&logging.mutex);
				LogMessage(LOG_ERR, "ERROR: LogInit(): Unable to open log file: %s: %s\n", filespec, strerror(errno));
				return FALSE;
			}
			atexit(LogClose);
		}
	}

	MUTEX_UNLOCK(&logging.mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
VOID LogClose(VOID)
{

	if (!logging_initialized)
		return;

	MUTEX_LOCK(&logging.mutex);
#if defined HAVE_SYSLOG
	if (logging.syslog)
		closelog();
#endif
	if (logging.fptr != NULL) {
		fclose(logging.fptr);
	}

	logging_initialized = FALSE;

	MUTEX_UNLOCK(&logging.mutex);
	MUTEX_DESTROY(&logging.mutex);

	return;
}

/*---------------------------------------------------------------------*/
VOID LogSetLevel(UINT8 which, UINT32 level)
{

	if (!logging_initialized)
		return;

	MUTEX_LOCK(&logging.mutex);

	if (which & LOG_CON)
		logging.console_level = level;
	if (which & LOG_FIL)
		logging.file_level = level;
	if (which & LOG_SYS)
		logging.syslog_level = level;

	MUTEX_UNLOCK(&logging.mutex);

	return;
}

/*---------------------------------------------------------------------*/
UINT32 LogGetLevel(UINT8 which)
{
	UINT32 level;

	if (!logging_initialized)
		return 0;

	level = 0;

	MUTEX_LOCK(&logging.mutex);

	switch (which) {
	case LOG_CON:
		level = logging.console_level;
		break;
	case LOG_FIL:
		level = logging.file_level;
		break;
	case LOG_SYS:
		level = logging.syslog_level;
		break;
	}

	MUTEX_UNLOCK(&logging.mutex);

	return level;
}

/*---------------------------------------------------------------------*/
VOID LogFlush(VOID)
{

	if (!logging_initialized)
		return;

	MUTEX_LOCK(&logging.mutex);

	if (logging.fptr != NULL)
		fflush(logging.fptr);

	MUTEX_UNLOCK(&logging.mutex);

	return;
}

/*---------------------------------------------------------------------*/
VOID LogMessage(UINT32 level, CHAR *format, ...)
{
	CHAR string[MAX_MESSAGE_LENGTH + 1], timestring[32];
	va_list marker;

	ASSERT(format != NULL);

	if (!logging_initialized) {
		va_start(marker, format);
		vfprintf(stderr, format, marker);
		va_end(marker);
		return;
	}

	MUTEX_LOCK(&logging.mutex);

	va_start(marker, format);
	vsnprintf(string, MAX_MESSAGE_LENGTH, format, marker);
	va_end(marker);
	string[MAX_MESSAGE_LENGTH] = '\0';

#if defined HAVE_SYSLOG
	if (logging.syslog && (level <= logging.syslog_level))
		syslog(level, "%s", string);
#endif

	if (level <= logging.file_level) {
		/* Write to log, if open... */
		if (logging.fptr != NULL)
			fprintf(logging.fptr, "%s: %s", 
					FormatUSTime(timestring, 32, UST_DATE_TIME_MSEC, SystemUSTime()), string);
	}

	if (level <= logging.console_level) {
		/* Write to console... */
#if defined WIN32 && ! defined UNIT_TEST
		if (IsService()) 
            ServiceDebugMessage("%s: %s: %s", SERVICE_NAME, 
					FormatUSTime(timestring, 32, UST_DATE_TIME_MSEC, SystemUSTime()), string);
        else
#endif				
		    printf("%s: %s", FormatUSTime(timestring, 32, UST_TIME_MSEC, SystemUSTime()), string);
	}

	MUTEX_UNLOCK(&logging.mutex);

	return;
}
