/* $Id: repository.c,v 1.5 2002/01/06 19:04:10 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Repository routines...

-----------------------------------------------------------------------*/

#include "server.h"
#include "grfthread.h"
#include "log.h"

/* Module constants ---------------------------------------------------*/
#define CHECK_INTERVAL	(UST_MSECOND * 200)

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static MUTEX mutex;
static BOOL initialized = FALSE;
static BOOL running = FALSE;
static BOOL stop = FALSE;
static THREAD thread_id;
static REPOS_ARGS repos;
static H_QUEUE empties;

/* This is a quick kludge... */
static CHAR filespec[MAX_PATH_LEN + 1], newspec[MAX_PATH_LEN + 1];
static FILE *file = NULL;
static USTIME start_time = 0;
static USTIME end_time = 0;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL StopRequested(VOID);
static THREAD_FUNC ReposThread(void *argument);
static BOOL GetPacket(GRF_PACKET ** packet, H_QUEUE queue);
static BOOL PutPacket(GRF_PACKET * packet);
static CHAR *CreateFileSpec(USTIME start, USTIME end, UINT32 unit, CHAR *path, CHAR *string);

/*---------------------------------------------------------------------*/
VOID ReposSetDefaults(REPOS_ARGS * args)
{
	ASSERT(args != NULL);

	memset(args->path, 0, MAX_PATH_LEN + 1);
    strcpy(args->name_fmt, DFL_NAME_FMT);
	args->rec_length = 0;
	args->queue_depth = DFL_QUEUE_DEPTH;

	return;
}

/*---------------------------------------------------------------------*/
BOOL ReposThreadStart(REPOS_ARGS * args, H_QUEUE empties_queue)
{
	ASSERT(args != NULL);

	if (!initialized) {
		MUTEX_INIT(&mutex);
		initialized = TRUE;
	}

	if (args->path[0] == '\0' || args->rec_length == 0 || ReposThreadIsRunning())
		return TRUE;

	MUTEX_LOCK(&mutex);
	strcpy(repos.path, args->path);
    strcpy(repos.name_fmt, args->name_fmt);
	repos.rec_length = args->rec_length;
	repos.queue_depth = args->queue_depth;
	empties = empties_queue;
	stop = FALSE;
	MUTEX_UNLOCK(&mutex);

	/* Fire up the thread */
	if (!THREAD_CREATE(&thread_id, ReposThread, NULL)) {
		LogMessage(LOG_ERR, "ERROR: ReposThreadStart: THREAD_CREATE() failed: %u %s\n", THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL ReposThreadStop(BOOL join)
{

	MUTEX_LOCK(&mutex);
	stop = TRUE;
	MUTEX_UNLOCK(&mutex);

	if (join)
		THREAD_JOIN(thread_id);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL ReposThreadIsRunning(VOID)
{
	BOOL result;

	MUTEX_LOCK(&mutex);
	result = running;
	MUTEX_UNLOCK(&mutex);

	return result;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(VOID)
{
	BOOL result;

	MUTEX_LOCK(&mutex);
	result = stop;
	MUTEX_UNLOCK(&mutex);

	return result;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC ReposThread(void *argument)
{
	UINT16 result;
	UINT32 mypid;
	ENDPOINT endpoint;
	GRF_PACKET *packet;
	H_QUEUE queue;

	MUTEX_LOCK(&mutex);
	running = TRUE;
	MUTEX_UNLOCK(&mutex);

	LogMessage(LOG_NOTICE, "RepositoryThread (tid:%d, pid:%d) start\n", THREAD_SELF(), getpid());

	/* Create our queue */
	if ((result = QueueCreate(&queue, repos.queue_depth)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: RepositoryThread: QueueCreate: %s\n", QueueErrorString(result));
		ReposThreadStop(FALSE);
	}

	/* Register for data */
	mypid = getpid();
	if (!ShutDownRequested() && !StopRequested()) {
		LogMessage(LOG_DEBUG, "RepositoryThread registering for data\n");
		memset(&endpoint, 0, sizeof(ENDPOINT));
		if (!GRFRegisterClient(endpoint.sin_addr.s_addr, mypid, queue))
			ReposThreadStop(FALSE);
	}

	while (!ShutDownRequested() && !StopRequested()) {
		/* Get a packet */
		if (GetPacket(&packet, queue)) {
			/* Write to client */
			if (!PutPacket(packet)) {
				ReposThreadStop(FALSE);
			}
		}
	}

	/* Unregister for data and flush queue */
	LogMessage(LOG_DEBUG, "RepositoryThread unregistering for data\n");
	GRFUnregisterClient(endpoint.sin_addr.s_addr, mypid);

	/* Destroy our queue */
	if ((result = QueueGet(queue, (VOID **)&packet, FALSE)) != Q_EMPTY) {
		LogMessage(LOG_ERR, "ERROR: RepositoryThread: QueueGet: %s\n", QueueErrorString(result));
	}
	if ((result = QueueDestroy(queue)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: RepositoryThread: QueueDestory: %s\n", QueueErrorString(result));
	}

	/* Close any open file */
	if (file != NULL) {
		LogMessage(LOG_DEBUG, "RepositoryThread: Close: %s\n", filespec);
		fclose(file);
		/* Rename the file */
		CreateFileSpec(start_time, end_time, GRFGetUnitNumber(), repos.path, newspec);
		if (strcmp(filespec, newspec) != 0) {
			LogMessage(LOG_DEBUG, "RepositoryThread: New name: %s\n", newspec);
		    remove(newspec);
			rename(filespec, newspec);
		}
		file = NULL;
	}

	MUTEX_LOCK(&mutex);
	running = FALSE;
	MUTEX_UNLOCK(&mutex);

	LogMessage(LOG_NOTICE, "RepositoryThread (tid:%d, pid:%d) exit\n", THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL GetPacket(GRF_PACKET ** packet, H_QUEUE queue)
{
	UINT16 result;

	ASSERT(packet != NULL);

	if ((result = QueueGet(queue, (VOID **)packet, FALSE)) != Q_NO_ERROR) {
		if (result != Q_EMPTY) {
			LogMessage(LOG_ERR, "ERROR: RepositoryThread: GetPacket: QueueGet: %s\n", QueueErrorString(result));
			ReposThreadStop(FALSE);
			return FALSE;
		}
		USSleep(CHECK_INTERVAL);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL PutPacket(GRF_PACKET * packet)
{
	BOOL ok;
    CHAR *ptr;
	UINT16 result;
	USTIME time = 0;
	REAL64 rate;

	ASSERT(IsValidBasePointer(packet));
	ASSERT(IsValidMemoryRange(packet->serial.bytes, packet->hdr.length));

	ok = TRUE;
    if (packet->hdr.type == GRF_TYPE_INFO) {
        /* Is this a network trigger message? */
        if (strncasecmp(packet->serial.info.message, "Network trigger:", 16) == 0) {
            /* Close if we're open */
            if (file != NULL) {
		        LogMessage(LOG_DEBUG, "RepositoryThread: Close due to trigger message: %s\n", filespec);
		        fclose(file);
		        /* Rename the file */
		        CreateFileSpec(start_time, end_time, GRFGetUnitNumber(), repos.path, newspec);
		        if (strcmp(filespec, newspec) != 0) {
			        LogMessage(LOG_DEBUG, "RepositoryThread: New name: %s\n", newspec);
		            remove(newspec);
			        rename(filespec, newspec);
		        }
		        file = NULL;
            }

            /* Get trigger time */
            if ((ptr = strrchr(packet->serial.info.message, ' ')) == NULL) 
                return TRUE;
            time = ParseUSTimeDOY(ptr);
        } else if (strncasecmp(packet->serial.info.message, "Network detrigger:", 18) == 0
                || strncasecmp(packet->serial.info.message, "Channel trigger:", 16) == 0
                || strncasecmp(packet->serial.info.message, "Channel detrigger:", 18) == 0) {
            /* Get (de)trigger time and store */
            if ((ptr = strrchr(packet->serial.info.message, ' ')) == NULL) 
                return TRUE;
            time = ParseUSTimeDOY(ptr);
        } else {
            /* Store packet only if we're open */
            if (file == NULL) 
                return TRUE;
        }
    } else if (packet->hdr.type == GRF_TYPE_DATA) {
	    /* Compute corrected time and rate */
	    time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
	    rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;
	    /* Update end time */
	    end_time = time + (USTIME)((REAL64)(packet->payload_hdr.data.n - 1) / rate);
    } else {
        return TRUE;
    }

    /* Is file open and it's time to close? */
	if (file != NULL && packet->hdr.type == GRF_TYPE_DATA &&
            packet->payload_hdr.data.timing.time_quality >= GRF_TIME_POOR &&
            ((time > start_time + (repos.rec_length * UST_SECOND)) ||
            (time < start_time - (repos.rec_length * UST_SECOND)))) {
		LogMessage(LOG_DEBUG, "RepositoryThread: Close due to record length: %s\n", filespec);
		fclose(file);
		/* Rename the file */
		CreateFileSpec(start_time, end_time, GRFGetUnitNumber(), repos.path, newspec);
		if (strcmp(filespec, newspec) != 0) {
			LogMessage(LOG_DEBUG, "RepositoryThread: New name: %s\n", newspec);
		    remove(newspec);
			rename(filespec, newspec);
		}
		file = NULL;
	}

	/* Is file not open? */
	if (file == NULL) {
		CreateFileSpec(time, 0, GRFGetUnitNumber(), repos.path, filespec);
		LogMessage(LOG_DEBUG, "RepositoryThread: Open: %s\n", filespec);
		if ((file = fopen(filespec, "w+b")) == NULL) {
			LogMessage(LOG_ERR, "ERROR: RepositoryThread: fopen(%s) failed: %s\n", filespec, strerror(errno));
			ReposThreadStop(FALSE);
			ok = FALSE;
		}
		start_time = time;
	}

	if (!GRFWriteFile(file, packet)) {
		ReposThreadStop(FALSE);
		ok = FALSE;
	}

	if ((result = QueuePut(empties, (VOID *)packet, FALSE)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: RepositoryThread: PutPacket: QueuePut(empties): %s\n", QueueErrorString(result));
		ReposThreadStop(FALSE);
		ok = FALSE;
	}

    /* If this is a detrigger message, close */
    if (packet->hdr.type == GRF_TYPE_INFO) {
        if (strncasecmp(packet->serial.info.message, "Network detrigger:", 18) == 0) {
		    LogMessage(LOG_DEBUG, "RepositoryThread: Close due to detrigger message: %s\n", filespec);
		    fclose(file);
		    /* Rename the file */
		    CreateFileSpec(start_time, end_time, GRFGetUnitNumber(), repos.path, newspec);
		    if (strcmp(filespec, newspec) != 0) {
			    LogMessage(LOG_INFO, "RepositoryThread: Event file: %s\n", newspec);
		        remove(newspec);
			    rename(filespec, newspec);
		    }
		    file = NULL;
        }
    }

	return ok;
}

/*---------------------------------------------------------------------*/
/*
Filename format specifiers:

    %Y  - four digit year
    %y  - two digit year
    %M	- two digit month
    %D  - two digit day of month
    %d  - three digit day of year
    %h	- two digit hour of day
    %m  - two digit minute of hour
    %s  - two digit integer second of minute
    %ms - three digit integer millisecond of second
    %us - six digit integer microsecond of second

    %l  - record length as integer seconds (variable length)
    %lm - three digit integer millisecond of record length
    %lu - six digit integer microsecond of record length

    %u  - GRF unit number (variable length)

    %%	- literal '%'
*/
static CHAR *CreateFileSpec(USTIME start, USTIME end, UINT32 unit, CHAR *path, CHAR *string)
{
    BOOL escape;
    CHAR *ptr;
    INT32 year, month, day, doy, hour, minute;
    REAL64 second;
	USTIME length;
    size_t i;

    ASSERT(path != NULL);
    ASSERT(string != NULL);

	length = 0;
	if(end > start)
		length = end - start;

    DecodeUSTime(start, &year, &doy, &hour, &minute, &second);
    MonthDay(&month, &day, year, doy);

    ptr = repos.name_fmt;
	if(string[strlen(string) - 1] == '/' || string[strlen(string) - 1] == '\\')
	    i = sprintf(string, "%s", path);
	else
	    i = sprintf(string, "%s/", path);
    escape = FALSE;

    while (*ptr) {

        if (*ptr == '%') {
            if (escape) {
                escape = FALSE;
            }
            else {
                escape = TRUE;
                ptr++;
                continue;
            }
        }

        if (escape) {
            if (*ptr == 'Y')
                i += sprintf(&string[i], "%04d", year);
            else if (*ptr == 'y')
                i += sprintf(&string[i], "%02d", year % 100);
            else if (*ptr == 'M')
                i += sprintf(&string[i], "%02d", month);
            else if (*ptr == 'D')
                i += sprintf(&string[i], "%02d", day);
            else if (*ptr == 'd')
                i += sprintf(&string[i], "%03d", doy);
            else if (*ptr == 'h')
                i += sprintf(&string[i], "%02d", hour);
            else if (*ptr == 'm') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&string[i], "%03d", (INT32) ((second - (INT32) second) * 1000));
                }
                else {
                    i += sprintf(&string[i], "%02d", minute);
                }
            }
            else if (*ptr == 's')
                i += sprintf(&string[i], "%02d", (INT32) second);
            else if (*ptr == 'u') {
                if (*(ptr + 1) == 's') {
                    ptr++;
                    i += sprintf(&string[i], "%06d", (INT32) ((second - (INT32) second) * 1000000));
                }
                else {
                    i += sprintf(&string[i], "%d", unit);
                }
            }
            else if (*ptr == 'l') {
                if (*(ptr + 1) == 'm') {
                    ptr++;
                    i += sprintf(&string[i], "%03u", (UINT32) ((length % UST_SECOND) / UST_MSECOND));
                }
                else if (*(ptr + 1) == 'u') {
                    ptr++;
                    i += sprintf(&string[i], "%06u", (UINT32) (length % UST_SECOND));
                }
                else {
                    i += sprintf(&string[i], "%u", (UINT32) (length / UST_SECOND));
                }
            }
            else
                i += sprintf(&string[i], "%c", *ptr);
        }
        else {
            i += sprintf(&string[i], "%c", *ptr);
        }

        ptr++;
        escape = FALSE;
    }

    return string;
}
