/* $Id: server.c,v 1.4 2003/04/07 20:05:57 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

    Server thread...

-----------------------------------------------------------------------*/

#include "server.h"
#include "log.h"
#include "mem.h"
#include "queues.h"

/* Module constants ---------------------------------------------------*/
#define BACKLOG 			10				/* Listen backlog */
#define SELECT_TIMEOUT      (UST_MSECOND * 500);	/* select timeout in usecs */

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;

static struct {
	MUTEX mutex;
	THREAD thread_id;
	BOOL running;
	BOOL stop;
	UINT32 connections;
	SVR_ARGS args;
	CLIENT head;							/* Client list */
	CLIENT tail;
	GRF_PACKET packet;
} svr;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC ServerThread(void *argument);
static BOOL HandleConnection(SOCKET connected_fd, ENDPOINT * endpoint);
static BOOL ConnectResponse(SOCKET fd, UINT8 code, CHAR *message);
static BOOL IsAllowedHost(ENDPOINT * endpoint, HOST_LIST * allow, HOST_LIST * deny);
static BOOL IsInHostList(ENDPOINT * endpoint, HOST_LIST * list);
static BOOL ResumeClient(SOCKET socket, ENDPOINT * endpoint);
static BOOL CreateClient(SOCKET socket, ENDPOINT * endpoint);
static VOID KillClients(VOID);
static VOID DumpClients(UINT8 log_level);
static VOID ServerInit(VOID);
static BOOL StopRequested(VOID);

/*---------------------------------------------------------------------*/
VOID ServerSetDefaults(SVR_ARGS * args)
{
	ASSERT(args != NULL);

	if (!initialized)
		ServerInit();

	strcpy(args->name, DFL_FRIENDLY_NAME);

	memset(&args->listen, 0, sizeof(ENDPOINT));
	args->listen.sin_family = AF_INET;
	args->listen.sin_addr.s_addr = htonl(DFL_ADDRESS);
	args->listen.sin_port = htons(DFL_PORT);

	args->max_connections = DFL_CONNECTIONS;
	args->queue_depth = DFL_QUEUE_DEPTH;

	args->allow = NULL;
	args->deny = NULL;

	ReposSetDefaults(&args->repos);

	return;
}

/*---------------------------------------------------------------------*/
BOOL ServerThreadStart(SVR_ARGS * args, THREAD_LIST * threads)
{
	ASSERT(args != NULL);

	if (!initialized)
		ServerInit();

	MUTEX_LOCK(&svr.mutex);
	if (svr.running) {
		MUTEX_UNLOCK(&svr.mutex);
		return FALSE;
	}

	memcpy(&svr.args, args, sizeof(SVR_ARGS));

	MUTEX_UNLOCK(&svr.mutex);

	svr.stop = FALSE;
	if (!THREAD_CREATE(&svr.thread_id, ServerThread, NULL)) {
		LogMessage(LOG_ERR, "ERROR: ServerThreadStart: THREAD_CREATE(ServerThread) failed: %u %s\n",
			THREAD_ERRNO, strerror(THREAD_ERRNO));
		return FALSE;
	}
	if (!ThreadRegister(threads, svr.thread_id, "ServerThread"))
		return FALSE;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL ServerThreadStop(BOOL join)
{
	if (svr.stop || !svr.running)
		return FALSE;

	svr.stop = TRUE;

	if (join)
		THREAD_JOIN(svr.thread_id);

	return TRUE;
}

/*---------------------------------------------------------------------*/
VOID DecrementConnections(VOID)
{
	MUTEX_LOCK(&svr.mutex);
	if (svr.connections > 0)
		svr.connections--;
	MUTEX_UNLOCK(&svr.mutex);

	return;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC ServerThread(void *argument)
{
	int optval, maxfd, result;
	fd_set input_set;
	struct timeval timeout;
	SOCKET listen_fd, connected_fd;
	ENDPOINT endpoint;
	int length;

	LogMessage(LOG_NOTICE, "ServerThread (tid:%d, pid:%d) start\n", THREAD_SELF(), getpid());

	MUTEX_LOCK(&svr.mutex);
	svr.running = TRUE;
	MUTEX_UNLOCK(&svr.mutex);

	/* Start the repository thread (if on) */
	if (!ReposThreadStart(&svr.args.repos, svr.args.empties))
		RequestShutDown(TRUE);

	/* Create and bind the socket */
	if ((listen_fd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: socket: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
		RequestShutDown(TRUE);
	}
	optval = 1;
	if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (CHAR *)&optval, sizeof(optval)) != 0) {
		LogMessage(LOG_INFO, "ERROR: ServerThread: setsockopt(SO_REUSEADDR): %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
		RequestShutDown(TRUE);
	}
	if (bind(listen_fd, (struct sockaddr *)&svr.args.listen, sizeof(ENDPOINT)) != 0) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: bind: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
		RequestShutDown(TRUE);
	}

	/* Listen for connections */
	if (listen(listen_fd, BACKLOG) != 0) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: listen: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
		RequestShutDown(TRUE);
	}

	/* Accept connections... */
	while (!ShutDownRequested() && !StopRequested()) {

		/* Is a connection pending? */
		FD_ZERO(&input_set);
		FD_SET(listen_fd, &input_set);
		maxfd = listen_fd + 1;
		timeout.tv_sec = 0;
		timeout.tv_usec = SELECT_TIMEOUT;
		if ((result = select(maxfd, &input_set, NULL, NULL, &timeout)) == 0) {
			continue;
		}
		else if (result < 0) {
			if (errno == EINTR)
				continue;
			LogMessage(LOG_ERR, "ERROR: ServerThread: select failed: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
			RequestShutDown(TRUE);
		}

		/* Accept it... */
		length = sizeof(ENDPOINT);
		if ((connected_fd = accept(listen_fd, (struct sockaddr *)&endpoint, &length)) == INVALID_SOCKET) {
			LogMessage(LOG_ERR, "ERROR: ServerThread: accept: %s\n", SOCK_ERRSTRING(SOCK_ERRNO()));
			RequestShutDown(TRUE);
			break;
		}

		/* Handle it... */
		if (!HandleConnection(connected_fd, &endpoint)) {
			RequestShutDown(TRUE);
			break;
		}
	}

	/* Destroy listening socket */
	SOCK_CLOSE(listen_fd);

	/* Shutdown all client connections */
	KillClients();

	/* Shut down the repository thread... */
	if (ReposThreadIsRunning())
		ReposThreadStop(TRUE);

	MUTEX_LOCK(&svr.mutex);
	svr.running = FALSE;
	MUTEX_UNLOCK(&svr.mutex);

	LogMessage(LOG_NOTICE, "ServerThread (tid:%d, pid:%d) exit\n", THREAD_SELF(), getpid());

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static BOOL HandleConnection(SOCKET connected_fd, ENDPOINT * endpoint)
{
	CHAR string[32];

	ASSERT(endpoint != NULL);

	DumpClients(LOG_DEBUG);

	/* Read a packet... */
	if (!GRFReadSocket(connected_fd, &svr.packet, GRF_DFL_TIMEOUT)) {
		SOCK_CLOSE(connected_fd);
		return TRUE;
	}

	/* What does the client want? */
	if (svr.packet.hdr.type != GRF_TYPE_CON_REQ) {
		SOCK_CLOSE(connected_fd);
		LogMessage(LOG_NOTICE, "ServerThread: Unrecognized connection request from %s\n", FormatEndpoint(endpoint, string));
		return TRUE;
	}
	/* Command connection... */
	if (svr.packet.payload_hdr.connect.attributes & GRF_ATTRIB_CMD) {
		if (!IsAllowedHost(endpoint, svr.args.allow, svr.args.deny)) {
			ConnectResponse(connected_fd, GRF_TYPE_CON_NAK, "Permission denied!");
			SOCK_CLOSE(connected_fd);
			LogMessage(LOG_NOTICE, "ServerThread: denying console connection: %s\n", FormatEndpoint(endpoint, string));
			return TRUE;
		}
		ConnectResponse(connected_fd, GRF_TYPE_CON_NAK, "Console is not available!");
		SOCK_CLOSE(connected_fd);
		LogMessage(LOG_NOTICE, "ServerThread: console not available: %s\n", FormatEndpoint(endpoint, string));
		return TRUE;
	}
	/* Data connection... */
	else if (svr.packet.payload_hdr.connect.attributes & GRF_ATTRIB_DATA) {
		if (!IsAllowedHost(endpoint, svr.args.allow, svr.args.deny)) {
			ConnectResponse(connected_fd, GRF_TYPE_CON_NAK, "Permission denied!");
			SOCK_CLOSE(connected_fd);
			LogMessage(LOG_NOTICE, "ServerThread: denying client connection: %s(%u)\n",
				FormatEndpoint(endpoint, string), svr.packet.payload_hdr.connect.process_id);
			return TRUE;
		}

		/* Is this a reconnect? */
		if (ResumeClient(connected_fd, endpoint)) {
			LogMessage(LOG_NOTICE, "ServerThread: reestablishing client connection: %s(%u)\n",
				FormatEndpoint(endpoint, string), svr.packet.payload_hdr.connect.process_id);
			return TRUE;
		}

		MUTEX_LOCK(&svr.mutex);
		if (svr.connections < svr.args.max_connections)
			svr.connections++;
		else {
			MUTEX_UNLOCK(&svr.mutex);
			ConnectResponse(connected_fd, GRF_TYPE_CON_NAK, "Too many connections!");
			SOCK_CLOSE(connected_fd);
			LogMessage(LOG_NOTICE, "ServerThread: too many connections: %s(%u)\n",
				FormatEndpoint(endpoint, string), svr.packet.payload_hdr.connect.process_id);
			return TRUE;
		}
		MUTEX_UNLOCK(&svr.mutex);

		LogMessage(LOG_NOTICE, "ServerThread: accepting client connection: %s(%u)\n",
			FormatEndpoint(endpoint, string), svr.packet.payload_hdr.connect.process_id);
		if (!ConnectResponse(connected_fd, GRF_TYPE_CON_ACK, svr.args.name)) {
			SOCK_CLOSE(connected_fd);
			return TRUE;
		}
		if (!CreateClient(connected_fd, endpoint)) {
			RequestShutDown(TRUE);
		}
	}
	else {
		SOCK_CLOSE(connected_fd);
		LogMessage(LOG_NOTICE, "ServerThread: Unrecognized connection attribute requested: 0x%08X %s(%u)\n",
			svr.packet.payload_hdr.connect.attributes, FormatEndpoint(endpoint, string),
			svr.packet.payload_hdr.connect.process_id);
		return TRUE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL ConnectResponse(SOCKET fd, UINT8 code, CHAR *message)
{

	ASSERT(message != NULL);

	svr.packet.hdr.type = code;
	svr.packet.hdr.unit = GRFGetUnitNumber();

	sprintf(svr.packet.serial.connect.message, message);

	return GRFWriteSocket(fd, &svr.packet, GRF_DFL_TIMEOUT);
}

/*---------------------------------------------------------------------*/
static BOOL IsAllowedHost(ENDPOINT * endpoint, HOST_LIST * allow, HOST_LIST * deny)
{
	BOOL allowed;

	ASSERT(endpoint != NULL);

	/* If both lists are empty, allow all */
	if (allow == NULL && deny == NULL)
		return TRUE;

	allowed = FALSE;

	/* Is it in the allow list or is the allow list empty? */
	if (IsInHostList(endpoint, allow))
		allowed = TRUE;

	/* Is the deny list empty? */
	if (deny != NULL) {
		/* Is it in the deny list? */
		if (IsInHostList(endpoint, deny))
			allowed = FALSE;
	}

	return allowed;
}

/*---------------------------------------------------------------------*/
static BOOL IsInHostList(ENDPOINT * endpoint, HOST_LIST * list)
{
	BOOL found;
	HOST_LIST *host;

	ASSERT(endpoint != NULL);

	/* If the list is empty, it's found */
	if (list == NULL)
		return TRUE;

	/* See if it's in the list */
	found = FALSE;
	for (host = list; host != NULL; host = host->next) {
		if ((host->address & host->netmask) == (endpoint->sin_addr.s_addr & host->netmask)) {
			found = TRUE;
			break;
		}
	}

	return found;
}

/*---------------------------------------------------------------------*/
static VOID KillClients(VOID)
{
	CLIENT *client;

	LogMessage(LOG_DEBUG, "ServerThread: Killing all clients\n");

	MUTEX_LOCK(&svr.mutex);
	client = svr.head.next;
	while (client->next != client) {
		ClientThreadStop(client);
		client = client->next;
	}
	MUTEX_UNLOCK(&svr.mutex);

	return;
}

/*---------------------------------------------------------------------*/
static VOID DumpClients(UINT8 log_level)
{
	CHAR description[32], state[128];
	CLIENT *client;
	INT32 i, count;

	count = 0;

	MUTEX_LOCK(&svr.mutex);
	client = svr.head.next;
	while (client->next != client) {
		i = 0;
		if (client->running) 
			i += sprintf(&state[i],"running");
		if (client->suspended) {
			if (i > 0)
				i += sprintf(&state[i],", ");
			i += sprintf(&state[i],"suspended");
		}
		if (client->resume) {
			if (i > 0)
				i += sprintf(&state[i],", ");
			i += sprintf(&state[i],"resuming");
		}
		if (client->stop) {
			if (i > 0)
				i += sprintf(&state[i],", ");
			i += sprintf(&state[i],"stopping");
		}

		if (i == 0) {
			LogMessage(LOG_ERR, "ERROR: ServerThread: Client #%d %s(%d), Thread is not running!\n",
					++count, FormatEndpoint(&client->endpoint, description), 
					client->process_id, state);
		} else {
			LogMessage(log_level, "ServerThread: Client #%d %s(%d), %s\n",
					++count, FormatEndpoint(&client->endpoint, description), 
					client->process_id, state);
		}

		client = client->next;
	}
	MUTEX_UNLOCK(&svr.mutex);

	return;
}

/*---------------------------------------------------------------------*/
static BOOL ResumeClient(SOCKET socket, ENDPOINT * endpoint)
{
	CHAR string[32];
	BOOL found;
	CLIENT *client;

	ASSERT(endpoint != NULL);

	/* Look through connected clients... */
	found = FALSE;
	MUTEX_LOCK(&svr.mutex);
	client = svr.head.next;
	while (client->next != client) {
		/* Is this the same client? */
		MUTEX_LOCK(&client->mutex);
		if (IsSameHost(&client->endpoint, endpoint) && 
				client->process_id == svr.packet.payload_hdr.connect.process_id &&
				client->running) {

			found = TRUE;

			MUTEX_UNLOCK(&client->mutex);
			break;
		}
		MUTEX_UNLOCK(&client->mutex);
		client = client->next;
	}
	MUTEX_UNLOCK(&svr.mutex);

	if (found) {
		MUTEX_LOCK(&client->mutex);
		if (!client->suspended) {
			MUTEX_UNLOCK(&client->mutex);
			ConnectResponse(socket, GRF_TYPE_CON_NAK, "Active connection already exists!");
			SOCK_CLOSE(socket);
			LogMessage(LOG_NOTICE, "ServerThread: active connection already exists: %s(%u)\n",
				FormatEndpoint(endpoint, string), svr.packet.payload_hdr.connect.process_id);
			return TRUE;
		}
		MUTEX_UNLOCK(&client->mutex);

		if (!ConnectResponse(socket, GRF_TYPE_CON_ACK, svr.args.name)) {
			SOCK_CLOSE(socket);
			return TRUE;
		}

		/* Resume this connection on this new socket... */
		MUTEX_LOCK(&client->mutex);
		client->socket = socket;
		memcpy(&client->endpoint, endpoint, sizeof(ENDPOINT));
		client->resume = TRUE;
		MUTEX_UNLOCK(&client->mutex);

		return TRUE;
	}

	return FALSE;
}

/*---------------------------------------------------------------------*/
static BOOL CreateClient(SOCKET socket, ENDPOINT * endpoint)
{
	UINT16 result;
	CLIENT *client;

	ASSERT(endpoint != NULL);

	/* Allocate node memory */
	if (!AllocateMemory((VOID **)&client, sizeof(CLIENT))) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: Can't allocate memory!\n");
		return FALSE;
	}

	/* Set up the client structure */
	MUTEX_INIT(&client->mutex);
	client->socket = socket;
	memcpy(&client->endpoint, endpoint, sizeof(ENDPOINT));
	client->process_id = svr.packet.payload_hdr.connect.process_id;
	client->timeout = svr.packet.payload_hdr.connect.time_out;
	client->empties = svr.args.empties;
	if ((result = QueueCreate(&client->queue, svr.args.queue_depth)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: QueueCreate: %s\n", QueueErrorString(result));
		return FALSE;
	}

	/* Link into client list */
	MUTEX_LOCK(&svr.mutex);
	client->prev = &svr.head;
	client->next = svr.head.next;
	client->next->prev = client;
	client->prev->next = client;
	MUTEX_UNLOCK(&svr.mutex);

	/* Start the client thread */
	ClientThreadStart(client);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL DestroyClient(CLIENT * client)
{
	UINT16 result;
	VOID *ptr = NULL;

	ASSERT(client != NULL);

	ASSERT(!client->running);

	/* Unlink */
	MUTEX_LOCK(&svr.mutex);
	client->prev->next = client->next;
	client->next->prev = client->prev;
	MUTEX_UNLOCK(&svr.mutex);

	MUTEX_DESTROY(&client->mutex);

	while ((result = QueueGet(client->queue, &ptr, FALSE)) != Q_EMPTY) {
		if (result != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: ServerThread: DestroyClient: QueueGet: %s\n", QueueErrorString(result));
		}
		if ((result = QueuePut(client->empties, &ptr, FALSE)) != Q_NO_ERROR) {
			LogMessage(LOG_ERR, "ERROR: ServerThread: DestroyClient: QueuePut: %s\n", QueueErrorString(result));
		}
	}
	if ((result = QueueDestroy(client->queue)) != Q_NO_ERROR) {
		LogMessage(LOG_ERR, "ERROR: ServerThread: DestroyClient: QueueDestory: %s\n", QueueErrorString(result));
		return FALSE;
	}

	ReleaseMemory(client);

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL StopRequested(VOID)
{
	BOOL stop;

	MUTEX_LOCK(&svr.mutex);
	stop = svr.stop;
	MUTEX_UNLOCK(&svr.mutex);

	return stop;
}

/*---------------------------------------------------------------------*/
static VOID ServerInit(VOID)
{
	ASSERT(!initialized);

	MUTEX_INIT(&svr.mutex);
	MUTEX_LOCK(&svr.mutex);

	initialized = TRUE;

	svr.running = FALSE;
	svr.stop = FALSE;
	svr.connections = 0;

	/* Init the client list */
	svr.head.prev = &svr.head;
	svr.head.next = &svr.tail;
	svr.tail.prev = &svr.head;
	svr.tail.next = &svr.tail;

	MUTEX_UNLOCK(&svr.mutex);

	return;
}
