/* $Id: trigger.c,v 1.22 2002/11/06 08:27:37 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Network and channel triggering...

-----------------------------------------------------------------------*/

#include "trigger.h"
#include "mem.h"
#include "log.h"

/* Module constants ---------------------------------------------------*/
#define MAX_SAMPLES		1024

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static GRF_PACKET info;						/* Storage for info packets */
static INT32 sample_buffer[MAX_SAMPLES];

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL StorePacket(NETWORK * net, CHANNEL * chan, GRF_PACKET * packet);
static BOOL RecordPackets(NETWORK * net);
static BOOL TrimPacket(PL_NODE * node, USTIME start, USTIME end);
static VOID PurgePackets(NETWORK * net);

//static VOID DumpPacketList(CHANNEL *chan);
static UINT8 NetworkTrigger(NETWORK * net);
static USTIME GetNextEarliestTime(NETWORK * net, USTIME time);

//static USTIME GetDetriggerTime(NETWORK *net);
static VOID UpdateNetworkTime(NETWORK * net);
static BOOL ChannelTrigger(NETWORK * net, VOTER * voter, GRF_PACKET * packet);
static BOOL AmplitudeTrigger(VOTER * voter, INT32 sample);
static BOOL StaLtaTrigger(VOTER * voter, INT32 sample);

/* Defined in grfthread.c */
extern BOOL DistributePacketUpstream(GRF_PACKET * packet);
extern BOOL DistributePacket(GRF_PACKET * packet);

/*---------------------------------------------------------------------*/
VOID SetNetworkDefaults(NETWORK * net)
{
	ASSERT(net != NULL);

	net->next = NULL;
	net->voters = NULL;
	net->channels = NULL;

	net->trig_votes = 1;
	net->detrig_votes = 0;
	net->window = UST_SECOND * 1;

	net->pre_event = UST_SECOND * 5;
	net->post_event = UST_SECOND * 5;
	net->min_event = UST_SECOND * 1;
	net->max_event = UST_SECOND * 120;

	net->time = VOID_USTIME;
	net->state = STA_UNINITIALIZED;
	net->trig_time = VOID_USTIME;
	net->detrig_time = VOID_USTIME;

	net->recording = FALSE;
	net->msg_trigger = FALSE;

	return;
}

/*---------------------------------------------------------------------*/
VOID SetChannelDefaults(CHANNEL * chan)
{
	PL_NODE *head, *tail;

	ASSERT(chan != NULL);

	memset(chan, 0, sizeof(CHANNEL));

	chan->rate = 0.0;

	/* Init the packet list */
	head = &chan->packets.head;
	tail = &chan->packets.tail;

	head->prev = head;
	head->next = tail;
	tail->next = tail;
	tail->prev = head;

	return;
}

/*---------------------------------------------------------------------*/
VOID SetTriggerDefaults(TRIGGER * trigger, UINT8 type)
{
	ASSERT(trigger != NULL);

	trigger->type = type;
	trigger->state = STA_UNINITIALIZED;
	trigger->time = VOID_USTIME;
	trigger->trig_time = VOID_USTIME;
	trigger->detrig_time = VOID_USTIME;

	if (type == TRIG_AMP) {
		trigger->prm.amp.mean_sec = UST_SECOND * 30;
		trigger->prm.amp.trig_amp = 50000;
		trigger->prm.amp.trig_sec = UST_MSECOND * 250;
		trigger->prm.amp.detrig_amp = 5000;
		trigger->prm.amp.detrig_sec = UST_MSECOND * 250;
	}
	else if (type == TRIG_STALTA) {
		trigger->prm.stalta.mean_sec = UST_SECOND * 30;
		trigger->prm.stalta.sta_sec = UST_MSECOND * 100;
		trigger->prm.stalta.lta_sec = UST_SECOND * 10;
		trigger->prm.stalta.trig_ratio = 8.0;
		trigger->prm.stalta.detrig_ratio = 2.0;
		trigger->prm.stalta.lta_hold = TRUE;
	}
	else {
		LogMessage(LOG_ERR, "SetTriggerDefaults: unknown trigger type: %u\n", type);
		memset(&trigger, 0, sizeof(TRIGGER));
	}

	return;
}

/*---------------------------------------------------------------------*/
BOOL HandlePacket(NETWORK * networks, GRF_PACKET * packet)
{
	BOOL trig = FALSE;
	CHAR *ptr, string[32];
	USTIME time;
	NETWORK *net;
	VOTER *voter;
	CHANNEL *chan;

	ASSERT(packet != NULL);

	/* Store packet in every network... */
	for (net = networks; net != NULL; net = net->next) {
		for (chan = net->channels; chan != NULL; chan = chan->next) {
			if (packet->hdr.type != GRF_TYPE_DATA) {
				if (!StorePacket(net, chan, packet))
					return FALSE;
				break;
			}
			else if (GRFCompareName(&packet->payload_hdr.data.name, &chan->name)) {
				if (!StorePacket(net, chan, packet))
					return FALSE;
			}
		}
	}

	if (packet->hdr.type == GRF_TYPE_INFO) {
		if (strncmp(packet->serial.info.message, "GRFTrig trigger", 15) == 0) {
			ptr = &packet->serial.info.message[16];
			trig = TRUE;
		}
		else if (strncmp(packet->serial.info.message, "GRFTrig detrigger", 17) == 0) {
			ptr = &packet->serial.info.message[18];
			trig = FALSE;
		}
		else
			return TRUE;

		for (net = networks; net != NULL; net = net->next) {
			if (strncmp(net->name, ptr, strlen(net->name)) == 0) {
				while (*ptr && *ptr != ' ')	
					ptr++;
				time = ParseUSTimeDOY(ptr);

				LogMessage(LOG_DEBUG, "Received %s message: %s %s\n",
					(trig ? "trigger" : "detrigger"), net->name,
					FormatUSTime(string, 31, UST_DATE_TIME_MSEC, time));

				if (trig) {
					if (net->state == STA_UNTRIGGERED) {
						net->msg_trigger = TRUE;
						net->trig_time = time;
						net->detrig_time = VOID_USTIME;
						net->state = STA_TRIGGER;
					}
				}
				else {
					net->detrig_time = time;
					//net->state = STA_DETRIGGER;
				}
				break;
			}
		}

		return TRUE;
	}
	else if (packet->hdr.type != GRF_TYPE_DATA)
		return TRUE;

	/* Look at each network */
	for (net = networks; net != NULL; net = net->next) {

		/* Is this station a voter? */
		for (voter = net->voters; voter != NULL; voter = voter->next) {
			if (GRFCompareName(&packet->payload_hdr.data.name, &voter->name)) {
				/* Run the trigger against this data */
				if (ChannelTrigger(net, voter, packet))
					break;
			}
		}

		UpdateNetworkTime(net);

		if (NetworkTrigger(net) == STA_TRIGGER) {
			net->recording = TRUE;
		}

		if (net->recording) {
			if (!RecordPackets(net))
				return FALSE;
		}
		else {
			PurgePackets(net);
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL RecordPackets(NETWORK * net)
{
	BOOL done;
	CHAR string[32], string2[32];
	CHANNEL *chan;
	PL_NODE *head, *tail, *node;
	USTIME start, end;

	ASSERT(net != NULL);

	if (net->trig_time == VOID_USTIME)
		return TRUE;

	start = net->trig_time - net->pre_event;
	if (net->detrig_time != VOID_USTIME) {
		end = net->detrig_time + net->post_event;
		if (end - start < net->min_event)
			end = start + net->min_event;
	}
	else if (net->msg_trigger && net->state == STA_TRIGGERED) {
		end = net->trig_time - net->pre_event + net->max_event;
		net->detrig_time = end - net->post_event;
		if (end - start < net->min_event)
			end = start + net->min_event;
	}
	else
		end = VOID_USTIME;

	done = TRUE;
	for (chan = net->channels; chan != NULL; chan = chan->next) {
		if (!chan->active)
			continue;
		head = &chan->packets.head;
		tail = &chan->packets.tail;
		node = head->next;
		while (node != tail) {

			/* Distribute non-data packets... */
			if (node->packet.hdr.type != GRF_TYPE_DATA) {
				if (!DistributePacket(&node->packet))
					return FALSE;
			}
			/* Start time is in packet... */
			else if (start >= node->ist && start <= node->lst) {
				LogMessage(LOG_DEBUG, "1st: %s:%s:%s %s %s\n",
					node->packet.payload_hdr.data.name.network,
					node->packet.payload_hdr.data.name.station,
					node->packet.payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, node->ist),
					FormatUSTime(string2, 32, UST_DATE_TIME_MSEC, node->lst));
				/* Trim it up... */
				if (TrimPacket(node, start, end)) {
					if (!DistributePacket(&node->packet))
						return FALSE;
				}
			}
			/* Haven't detriggered or packet is entirely inside event */
			else if (node->ist >= start && (end == VOID_USTIME || node->lst <= end)) {
				LogMessage(LOG_DEBUG, "%s:%s:%s %s %s\n",
					node->packet.payload_hdr.data.name.network,
					node->packet.payload_hdr.data.name.station,
					node->packet.payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, node->ist),
					FormatUSTime(string2, 32, UST_DATE_TIME_MSEC, node->lst));
				if (!DistributePacket(&node->packet))
					return FALSE;
			}
			/* End time is in packet... */
			else if (end != VOID_USTIME && node->ist <= end && node->lst >= end) {
				LogMessage(LOG_DEBUG, "Last: %s:%s:%s %s %s\n",
					node->packet.payload_hdr.data.name.network,
					node->packet.payload_hdr.data.name.station,
					node->packet.payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, node->ist),
					FormatUSTime(string2, 32, UST_DATE_TIME_MSEC, node->lst));
				/* Trim it up... */
				if (TrimPacket(node, start, end)) {
					if (!DistributePacket(&node->packet))
						return FALSE;
				}
			}
			else {
				LogMessage(LOG_DEBUG, "Toss: %s:%s:%s %s %s\n",
					node->packet.payload_hdr.data.name.network,
					node->packet.payload_hdr.data.name.station,
					node->packet.payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, node->ist),
					FormatUSTime(string2, 32, UST_DATE_TIME_MSEC, node->lst));
			}

			node->prev->next = node->next;
			node->next->prev = node->prev;
			ReleaseMemory((VOID *)node);
			node = head->next;
		}

		if (end == VOID_USTIME || chan->latest < end)
			done = FALSE;
	}

	if (done) {
		LogMessage(LOG_INFO, "%s\n", GRFInfoPacket(&info, "Network detrigger: %s %s", net->name,
				FormatUSTime(string, 32, UST_DATE_TIME_MSEC, net->detrig_time)));
		if (!net->msg_trigger )
			DistributePacketUpstream(&info);
		else
			net->state = STA_UNTRIGGERED;

		if (!DistributePacket(&info))
			return FALSE;

		net->msg_trigger = FALSE;
		net->recording = FALSE;
		net->trig_time = VOID_USTIME;
		net->detrig_time = VOID_USTIME;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL TrimPacket(PL_NODE * node, USTIME start, USTIME end)
{
	CHAR string[32], string2[32];
	INT32 *samples;
	UINT16 first, last;

	ASSERT(node != NULL);

	/* If start time is in packet... */
	if (start >= node->ist && start <= node->lst) {
		first = (UINT16)((start - node->ist) / node->period);
		if (end >= node->ist && end <= node->lst)
			last = (UINT16)((end - node->ist) / node->period);
		else
			last = node->packet.payload_hdr.data.n - 1;
	}
	/* or end time is in packet... */
	else if (end >= node->ist && end <= node->lst) {
		last = (UINT16)((end - node->ist) / node->period);
		if (start >= node->ist && start <= node->lst)
			first = (UINT16)((start - node->ist) / node->period);
		else
			first = 0;
	}
	/* otherwise, error... */
	else {
		return FALSE;
	}

	/* Sanity check... */
	ASSERT(first <= node->packet.payload_hdr.data.n - 1);
	ASSERT(last <= node->packet.payload_hdr.data.n - 1);

	/* Would this be zero length? */
	if (last - first == 0)
		return FALSE;

	/* Manipulate the samples... */
	samples = sample_buffer;
	if (!GRFDecodeSamples(&node->packet, samples, MAX_SAMPLES))
		return FALSE;
	if (!GRFEncodeSamples(&node->packet, samples + first, (UINT16)(last - first + 1)))
		return FALSE;

	/* Adjust the timestamp */
	if (first > 0)
		node->packet.payload_hdr.data.timing.time += node->period * (USTIME)first;

	/* Update node times... */
	node->ist = node->packet.payload_hdr.data.timing.time + node->packet.payload_hdr.data.timing.time_correction;
	node->lst = node->ist + (node->period * (node->packet.payload_hdr.data.n - 1));
	LogMessage(LOG_DEBUG, "Trim: %s:%s:%s %s %s\n",
		node->packet.payload_hdr.data.name.network,
		node->packet.payload_hdr.data.name.station,
		node->packet.payload_hdr.data.name.component,
		FormatUSTime(string, 32, UST_DATE_TIME_MSEC, node->ist), FormatUSTime(string2, 32, UST_DATE_TIME_MSEC, node->lst));

	return TRUE;
}

/*---------------------------------------------------------------------*/
static BOOL StorePacket(NETWORK * net, CHANNEL * chan, GRF_PACKET * packet)
{
	PL_NODE *tail, *node;

	ASSERT(chan != NULL);
	ASSERT(packet != NULL);

	if (!chan->active) {
		chan->active = TRUE;
		chan->unit = packet->hdr.unit;
		chan->channel = packet->payload_hdr.data.channel;
	}

	/* Create a packet list node */
	if (!AllocateMemory((VOID **)&node, sizeof(PL_NODE))) {
		LogMessage(LOG_ERR, "ERROR: StorePacket: unable to allocate memory!\n");
		return FALSE;
	}

	/* Compute IST, LST, and sampling period in us */
	node->ist = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
	node->period = (USTIME)((1.0 / (packet->payload_hdr.data.timing.rate +
				packet->payload_hdr.data.timing.rate_correction)) * (REAL64)UST_SECOND);
	node->lst = node->ist + (node->period * (packet->payload_hdr.data.n - 1));

	chan->latest = node->lst;

	/* Copy packet into node */
	memcpy(&node->packet, packet, sizeof(GRF_PACKET));

	/* If this is a data packet, change network name */
	if (packet->hdr.type == GRF_TYPE_DATA)
		strcpy(node->packet.payload_hdr.data.name.network, net->name);

	/* Link into list in front of tail */
	tail = &chan->packets.tail;

	node->prev = tail->prev;
	node->next = tail;
	node->prev->next = node;
	tail->prev = node;

	return TRUE;
}

/*---------------------------------------------------------------------*/
static VOID PurgePackets(NETWORK * net)
{
	CHANNEL *chan;
	PL_NODE *head, *tail, *node;

	ASSERT(net != NULL);

	/* Destroy all packets whose endtime is earlier than net->time - pre_event */
	for (chan = net->channels; chan != NULL; chan = chan->next) {
		head = &chan->packets.head;
		tail = &chan->packets.tail;
		node = head->next;
		while (node != tail) {
			if (node->lst >= net->time - net->pre_event)
				break;
			node->prev->next = node->next;
			node->next->prev = node->prev;
			ReleaseMemory((VOID *)node);
			node = head->next;
		}
	}

	return;
}

/*---------------------------------------------------------------------*/
#if 0
static VOID DumpPacketList(CHANNEL * chan)
{
	CHAR string1[32], string2[32];
	PL_NODE *head, *tail, *node;
	UINT32 count;

	ASSERT(chan != NULL);

	LogMessage(LOG_DEBUG, "%s:%s:%s packet list:\n", chan->name.network, chan->name.station, chan->name.component);

	count = 0;
	head = &chan->packets.head;
	tail = &chan->packets.tail;
	for (node = head->next; node != tail; node = node->next)
		LogMessage(LOG_DEBUG, "	%04u %s %s\n", ++count,
			FormatUSTime(string1, 32, UST_DATE_TIME_USEC, node->ist), FormatUSTime(string2, 32, UST_DATE_TIME_USEC, node->lst));

	return;
}
#endif

/*---------------------------------------------------------------------*/
static UINT8 NetworkTrigger(NETWORK * net)
{
	CHAR string[32];
	VOTER *voter;
	UINT32 count;
	USTIME time, last, first;

	ASSERT(net != NULL);

	if (net->state == STA_UNINITIALIZED) {
		net->trig_time = VOID_USTIME;
		net->detrig_time = VOID_USTIME;
		net->state = STA_UNTRIGGERED;
	}

	if (net->msg_trigger) {
		if (net->state == STA_TRIGGER) {
			count = 0;
			LogMessage(LOG_INFO, "%s\n", GRFInfoPacket(&info, "Network trigger: %s(%u) %s",
					net->name, count, FormatUSTime(string, 32, UST_DATE_TIME_MSEC, net->trig_time)));
			if (!DistributePacket(&info))
				return FALSE;
			net->state = STA_TRIGGERED;
			net->recording = TRUE;
		}
		else if (net->state == STA_DETRIGGER)
			net->state = STA_UNTRIGGERED;

		return net->state;
	}

	switch (net->state) {
	  case STA_DETRIGGER:
	  case STA_UNTRIGGERED:
		net->state = STA_UNTRIGGERED;
		/* Don't retrigger inside of post-event recording */
		if (net->recording) {
			LogMessage(LOG_DEBUG, "NetworkTrigger(%s): no retrigger during post event recording!\n", net->name);
			break;
		}
		count = 0;
		time = VOID_USTIME;
		last = VOID_USTIME;
		first = VOID_USTIME;
		/* Find earliest trigger time */
		first = VOID_USTIME;
		for (voter = net->voters; voter != NULL; voter = voter->next) {
			if (voter->trigger.trig_time != VOID_USTIME) {
				/* Toss times earlier than the net time */
				if (voter->trigger.trig_time < net->time - net->window)
					voter->trigger.trig_time = VOID_USTIME;
				else if (first == VOID_USTIME)
					first = voter->trigger.trig_time;
				else if (voter->trigger.trig_time < first)
					first = voter->trigger.trig_time;
			}
		}

		if (first == VOID_USTIME)
			break;

		LogMessage(LOG_DEBUG, "NetworkTrigger(%s): earliest channel: %s\n", net->name,
			FormatUSTime(string, 32, UST_DATE_TIME_USEC, first));

		/* Count, triggers within window */
		count = 0;
		for (voter = net->voters; voter != NULL; voter = voter->next) {
			if (voter->trigger.trig_time != VOID_USTIME) {
				if (voter->trigger.trig_time >= first && voter->trigger.trig_time <= first + net->window) {
					count++;
					LogMessage(LOG_DEBUG, "NetworkTrigger(%s): yea channel %s:%s %s\n",
						net->name, voter->name.station, voter->name.component,
						FormatUSTime(string, 32, UST_DATE_TIME_USEC, voter->trigger.trig_time));
				}
				else {
					LogMessage(LOG_DEBUG, "NetworkTrigger(%s): nay channel %s:%s %s\n",
						net->name, voter->name.station, voter->name.component,
						FormatUSTime(string, 32, UST_DATE_TIME_USEC, voter->trigger.trig_time));
				}
			}
			else {
				LogMessage(LOG_DEBUG, "NetworkTrigger(%s): nay channel %s:%s %s\n",
					net->name, voter->name.station, voter->name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_USEC, voter->trigger.trig_time));
			}
		}

		/* Are there enough? */
		if (count >= net->trig_votes) {
			/* Declare event */
			net->trig_time = first;
			net->detrig_time = VOID_USTIME;
			net->state = STA_TRIGGER;
			LogMessage(LOG_INFO, "%s\n", GRFInfoPacket(&info, "Network trigger: %s(%u) %s",
					net->name, count, FormatUSTime(string, 32, UST_DATE_TIME_MSEC, net->trig_time)));
			DistributePacketUpstream(&info);
			if (!DistributePacket(&info))
				return FALSE;
			break;
		}
		break;
	  case STA_TRIGGER:
	  case STA_TRIGGERED:
		net->state = STA_TRIGGERED;
		count = 0;
		time = VOID_USTIME;
		/* Count triggered voters and get latest (detrigger) time */
		for (voter = net->voters; voter != NULL; voter = voter->next) {
			/* Get latest detrigger time */
			if (voter->trigger.detrig_time != VOID_USTIME) {
				if (time == VOID_USTIME || time < voter->trigger.detrig_time)
					time = voter->trigger.detrig_time;
			}
			/* Count up triggered voters */
			if (voter->trigger.trig_time != VOID_USTIME && voter->trigger.detrig_time == VOID_USTIME) {
				count++;
			}
		}
		/* Detrigger? */
		if (count <= net->detrig_votes) {
			ASSERT(time != VOID_USTIME);
			net->detrig_time = time;
			net->state = STA_DETRIGGER;
			/* Clear triggered voters */
			for (voter = net->voters; voter != NULL; voter = voter->next) {
				if (voter->trigger.trig_time != VOID_USTIME)
					voter->trigger.trig_time = VOID_USTIME;
			}
		}
		break;
	}

	return net->state;
}

/*---------------------------------------------------------------------*/
static USTIME GetNextEarliestTime(NETWORK * net, USTIME last)
{
	CHAR string[32], string1[32];
	VOTER *voter;
	USTIME time;

	ASSERT(net != NULL);

	/* Find earliest time later than last */
	time = VOID_USTIME;
	for (voter = net->voters; voter != NULL; voter = voter->next) {
		LogMessage(LOG_DEBUG, "Voter %s:%s:%s, state: %u, trig: %s, detrig: %s\n",
			net->name, voter->name.station, voter->name.component, voter->trigger.state,
			FormatUSTime(string, 32, UST_DATE_TIME_USEC, voter->trigger.trig_time),
			FormatUSTime(string1, 32, UST_DATE_TIME_USEC, voter->trigger.detrig_time));

		if (voter->trigger.trig_time != VOID_USTIME) {
			/* Toss times earlier than the net time */
			if (voter->trigger.trig_time < net->time - net->window)
				voter->trigger.trig_time = VOID_USTIME;
			else if ((time == VOID_USTIME || voter->trigger.trig_time < time) &&
				(last == VOID_USTIME || voter->trigger.trig_time > last))
				time = voter->trigger.trig_time;
		}
	}

	return time;
}

/*---------------------------------------------------------------------*/
static VOID UpdateNetworkTime(NETWORK * net)
{
	VOTER *voter;

	ASSERT(net != NULL);

	if (net->msg_trigger)
		return;

	/* Get earliest time among voters... */
	for (voter = net->voters; voter != NULL; voter = voter->next) {
		if (net->time == VOID_USTIME || net->time < voter->trigger.time)
			net->time = voter->trigger.time;
	}

	return;
}

/*---------------------------------------------------------------------*/
static BOOL ChannelTrigger(NETWORK * net, VOTER * voter, GRF_PACKET * packet)
{
	BOOL chan_event;
	CHAR string[32];
	USTIME time;
	REAL64 rate;
	INT32 sample;
	UINT32 i;
	struct TAG_AMP *amp;
	struct TAG_STALTA *stalta;

	ASSERT(voter != NULL);
	ASSERT(packet != NULL);

	/* Initial sample time... */
	time = packet->payload_hdr.data.timing.time + packet->payload_hdr.data.timing.time_correction;
	voter->trigger.time = time;

	/* Compute sampling period */
	rate = packet->payload_hdr.data.timing.rate + packet->payload_hdr.data.timing.rate_correction;
	voter->trigger.period = (USTIME)((1.0 / rate) * (REAL64)UST_SECOND);

	/* Update the various trigger time constants... */
	if (voter->trigger.type == TRIG_AMP) {
		amp = &voter->trigger.prm.amp;
		amp->mean_n = (UINT32)(((REAL64)amp->mean_sec / (REAL64)UST_SECOND) * rate);
		amp->trig_n = (UINT32)(((REAL64)amp->trig_sec / (REAL64)UST_SECOND) * rate);
		amp->detrig_n = (UINT32)(((REAL64)amp->detrig_sec / (REAL64)UST_SECOND) * rate);
		amp->max_n = (UINT32)(((REAL64)net->max_event / (REAL64)UST_SECOND) * rate);
	}
	else if (voter->trigger.type == TRIG_STALTA) {
		stalta = &voter->trigger.prm.stalta;
		stalta->mean_n = (UINT32)(((REAL64)stalta->mean_sec / (REAL64)UST_SECOND) * rate);
		stalta->sta_n = (UINT32)(((REAL64)stalta->sta_sec / (REAL64)UST_SECOND) * rate);
		if (stalta->sta_n < 1)
			stalta->sta_n = 1;
		stalta->lta_n = (UINT32)(((REAL64)stalta->lta_sec / (REAL64)UST_SECOND) * rate);
		if (stalta->lta_n < 1)
			stalta->lta_n = 1;
		stalta->max_n = (UINT32)(((REAL64)net->max_event / (REAL64)UST_SECOND) * rate);
	}

	if (!GRFDecodeSamples(packet, sample_buffer, MAX_SAMPLES))
		return FALSE;

	chan_event = FALSE;
	for (i = 0; i < packet->payload_hdr.data.n; i++) {
		/* Get next sample */
		sample = sample_buffer[i];
		/* Trigger... */
		if (voter->trigger.type == TRIG_AMP)
			AmplitudeTrigger(voter, sample);
		else if (voter->trigger.type == TRIG_STALTA)
			StaLtaTrigger(voter, sample);

		if (voter->trigger.state == STA_TRIGGER) {
			voter->trigger.trig_time = time + (voter->trigger.period * i);
			voter->trigger.detrig_time = VOID_USTIME;
			chan_event = TRUE;
			LogMessage(LOG_INFO, "%s\n", GRFInfoPacket(&info, "Channel trigger: %s:%s:%s %s",
					net->name,
					packet->payload_hdr.data.name.station,
					packet->payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, voter->trigger.trig_time)));
			DistributePacketUpstream(&info);
			if (!DistributePacket(&info))
				return FALSE;
		}
		else if (voter->trigger.state == STA_DETRIGGER) {
			voter->trigger.detrig_time = time + (voter->trigger.period * i);
			chan_event = TRUE;
			LogMessage(LOG_INFO, "%s\n", GRFInfoPacket(&info, "Channel detrigger: %s:%s:%s %s",
					net->name,
					packet->payload_hdr.data.name.station,
					packet->payload_hdr.data.name.component,
					FormatUSTime(string, 32, UST_DATE_TIME_MSEC, voter->trigger.detrig_time)));
			DistributePacketUpstream(&info);
			if (!DistributePacket(&info))
				return FALSE;
		}
	}

	return chan_event;
}

/*---------------------------------------------------------------------*/
static BOOL AmplitudeTrigger(VOTER * voter, INT32 sample)
{
	INT32 x;
	struct TAG_AMP *prm;

	ASSERT(voter != NULL);

	prm = &voter->trigger.prm.amp;

	if (voter->trigger.state == STA_UNINITIALIZED) {
		prm->mean = (REAL32)sample;
		prm->holdoff = prm->mean_n * 2;
		prm->n = 0;
		voter->trigger.state = STA_UNTRIGGERED;
		return voter->trigger.state;
	}
	else if (prm->mean_n > 0) {
		/* Compute running mean */
		prm->mean += (((REAL32)sample - prm->mean) / (REAL32)prm->mean_n);
		/* Demeaned and rectified sample */
		x = abs(sample - (INT32)(prm->mean + 0.5));
	}
	else {
		/* Rectified sample */
		x = abs(sample);
	}

	switch (voter->trigger.state) {
	  case STA_DETRIGGER:
	  case STA_UNTRIGGERED:
		voter->trigger.state = STA_UNTRIGGERED;
		/* Are we holding off... */
		if (prm->holdoff > 0) {
			prm->holdoff--;
			break;
		}
		/* Amplitude over threshold? */
		if (x > prm->trig_amp) {
			/* For long enough? */
			prm->count++;
			if (prm->count >= prm->trig_n) {
				/* Trigger... */
				voter->trigger.state = STA_TRIGGER;
				prm->count = 0;
				prm->n = 0;
			}
		}
		else {
			prm->count = 0;
		}
		break;
	  case STA_TRIGGER:
	  case STA_TRIGGERED:
		voter->trigger.state = STA_TRIGGERED;
		/* Force detrigger? */
		prm->n++;
		if (prm->n >= prm->max_n) {
			voter->trigger.state = STA_DETRIGGER;
			LogMessage(LOG_INFO, "%s:%s:%s forced detrigger: amplitude:%d\n",
				voter->name.network, voter->name.station, voter->name.component, x);
			prm->holdoff = prm->mean_n;
			prm->count = 0;
			prm->n = 0;
		}
		/* Amplitude under threshold? */
		if (x <= prm->detrig_amp) {
			/* For long enough */
			prm->count++;
			if (prm->count >= prm->detrig_n) {
				/* Detrigger... */
				voter->trigger.state = STA_DETRIGGER;
				prm->holdoff = prm->mean_n;
				prm->count = 0;
				prm->n = 0;
			}
		}
		else {
			prm->count = 0;
		}
		break;
	}

	return voter->trigger.state;
}

/*---------------------------------------------------------------------*/
static BOOL StaLtaTrigger(VOTER * voter, INT32 sample)
{
	REAL32 x;
	struct TAG_STALTA *prm;

	ASSERT(voter != NULL);

	prm = &voter->trigger.prm.stalta;

	if (voter->trigger.state == STA_UNINITIALIZED) {
		prm->mean = (REAL32)sample;
		prm->sta = 0;
		prm->lta = 0;
		prm->holdoff = prm->lta_n * 2;
		prm->n = 0;
		voter->trigger.state = STA_UNTRIGGERED;
		return voter->trigger.state;
	}
	else if (prm->mean_n > 0) {
		prm->mean += (((REAL32)sample - prm->mean) / (REAL32)prm->mean_n);
		x = (REAL32)fabs((REAL32)sample - prm->mean);
	}
	else
		x = (REAL32)abs(sample);

	switch (voter->trigger.state) {
	  case STA_DETRIGGER:
	  case STA_UNTRIGGERED:
		voter->trigger.state = STA_UNTRIGGERED;
		/* Compute STA and LTA */
		prm->sta += (x - prm->sta) / (REAL32)prm->sta_n;
		prm->lta += (x - prm->lta) / (REAL32)prm->lta_n;
		/* Are we holding off... */
		if (prm->holdoff > 0) {
			prm->holdoff--;
			break;
		}
		/* Compute the ratio, guard against div by 0 */
		if (prm->lta == 0.0)
			return voter->trigger.state;
		prm->ratio = prm->sta / prm->lta;
		if (prm->ratio >= prm->trig_ratio) {
			voter->trigger.state = STA_TRIGGER;
			prm->n = 0;
		}
		break;
	  case STA_TRIGGER:
	  case STA_TRIGGERED:
		voter->trigger.state = STA_TRIGGERED;
		/* Force detrigger? */
		prm->n++;
		if (prm->n >= prm->max_n) {
			LogMessage(LOG_INFO, "%s:%s:%s forced detrigger: sta:%.2f, lta:%.2f, ratio:%.2f\n",
				voter->name.network, voter->name.station, voter->name.component, prm->sta, prm->lta, prm->ratio);
			voter->trigger.state = STA_DETRIGGER;
			prm->holdoff = prm->lta_n;
			prm->n = 0;
		}
		/* Compute STA */
		prm->sta += (x - prm->sta) / (REAL32)prm->sta_n;
		/* If LTA hold is not on, compute LTA */
		if (!prm->lta_hold)
			prm->lta += (x - prm->lta) / (REAL32)prm->lta_n;
		/* Compute the ratio, guard against div by 0 */
		if (prm->lta == 0.0)
			return voter->trigger.state;
		prm->ratio = prm->sta / prm->lta;
		if (prm->ratio <= prm->detrig_ratio) {
			voter->trigger.state = STA_DETRIGGER;
			/* Don't retrigger for LTA length */
			prm->holdoff = prm->lta_n;
			prm->n = 0;
		}
		break;
	}

	return voter->trigger.state;
}
