/* $Id: async.h,v 1.10 2004/02/04 03:25:26 cvs Exp $ */
/*-----------------------------------------------------------------------

	General, platform independent, async serial I/O library...

    Copyright (c) 2000-2004 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/
#if !defined _ASYNC_H_INCLUDED_
#define _ASYNC_H_INCLUDED_

#include <platform.h>

/* Constants ----------------------------------------------------------*/
#define ASYNC_MAX_PORTS        	32			/* Maxmimum number of active ports */

#define ASYNC_FLUSH_INPUT		1			/* Use with AsyncFlush to specify queue(s) */
#define ASYNC_FLUSH_OUTPUT      2
#define ASYNC_FLUSH_BOTH        3

#define ASYNC_INVALID			-1			/* Invalid handle value */
#define MAX_ASYNC_DEV_NAME		64			/* Max length of device name */

/* Error codes... */
#define AE_NONE					0			/* No error */
#define AE_INVALID				1			/* Invalid handle */
#define AE_TOO_MANY				2			/* Out of handles */
#define AE_PENDING				3
#define AE_SYSTEM				4			/* System defined error */
#define AE_UNKNOWN				5			/* Undefined or unknown error */

#define XON						0x11
#define XOFF					0x13

/* Types --------------------------------------------------------------*/
typedef INT32 H_ASYNC;						/* Handle type */

/* Controllable outbound line conditions... */
typedef enum TAG_ASYNC_ACTION {				/* May be or'ed together */
	ASYNC_ASSERT_RTS = 1,					/* Request to send, really 'ready to receive' */
	ASYNC_DEASSERT_RTS = 2,
	ASYNC_ASSERT_DTR = 4,					/* Data terminal ready */
	ASYNC_DEASSERT_DTR = 8
} ASYNC_ACTION;

/* Testable and waitable inbound line conditions... */
typedef enum TAG_ASYNC_TEST {				/* May be or'ed together */
	ASYNC_CTS_ASSERTED = 1,					/* Cleared to send */
	ASYNC_CTS_DEASSERTED = 2,
	ASYNC_DCD_ASSERTED = 4,					/* Data carrier detected, aka RLSD */
	ASYNC_DCD_DEASSERTED = 8,
	ASYNC_DSR_ASSERTED = 16,				/* Data set ready */
	ASYNC_DSR_DEASSERTED = 32,
	ASYNC_RI_ASSERTED = 64,					/* Ring indicator */
	ASYNC_RI_DEASSERTED = 128
} ASYNC_TEST;

/* Waitable events... */
typedef enum TAG_ASYNC_EVENT {				/* May be or'ed together */
	ASYNC_EVENT_CTS = 1,					/* CTS line state change */
	ASYNC_EVENT_DSR = 2,					/* DSR line state change */
	ASYNC_EVENT_DCD = 4,					/* DCD line state change */
	ASYNC_EVENT_RI = 8,						/* RI line state change */
	ASYNC_EVENT_RX = 16,					/* A byte has been received */
	ASYNC_EVENT_TX_EMPTY = 32				/* TX buffer is empty */
} ASYNC_EVENT;

/* Parameters... */
typedef enum TAG_ASYNC_PARITY {
	ASYNC_PARITY_NONE,						/* No parity */
	ASYNC_PARITY_EVEN,						/* Even parity */
	ASYNC_PARITY_ODD						/* Odd parity */
} ASYNC_PARITY;
typedef enum TAG_ASYNC_FLOWCTRL {			/* May be or'ed together */
	ASYNC_FLOW_NONE = 1,					/* No flow control */
	ASYNC_FLOW_RTSCTS = 2,					/* Hardware flow control */
	ASYNC_FLOW_XONXOFF = 4					/* Software flow control */
} ASYNC_FLOWCTRL;
typedef struct TAG_ASYNC_PARAMS {
	CHAR device[MAX_ASYNC_DEV_NAME + 1];	/* Device name */
	UINT32 baud_rate;						/* Baudrate as integer, 300 - 115200 */
	ASYNC_PARITY parity;					/* Use enum above */
	UINT8 data_bits;						/* Data bits as integer, 7 or 8 */
	UINT8 stop_bits;						/* Stop bits as integer, 1 or 2 */
	ASYNC_FLOWCTRL flow_ctrl;				/* Use enum above */
	UINT32 read_timeout;					/* Read timeout in mseconds */
	UINT32 write_timeout;					/* Write timeout in mseconds */
	UINT32 read_buffer;						/* Read buffer size in bytes */
	UINT32 write_buffer;					/* Write buffer size in bytes */
} ASYNC_PARAMS;

/* API function prototypes --------------------------------------------*/

/* Parameter helpers... */
VOID AsyncParseParams(CHAR *string, ASYNC_PARAMS * params);
VOID AsyncDefaultParams(ASYNC_PARAMS * params);
CHAR *AsyncFormatParams(CHAR *string, size_t n, ASYNC_PARAMS * params);

/* Device open and close */
H_ASYNC AsyncOpen(ASYNC_PARAMS * params);
BOOL AsyncClose(H_ASYNC handle);

/* Read and write... */
BOOL AsyncRead(H_ASYNC handle, VOID *data, size_t *n);
BOOL AsyncWrite(H_ASYNC handle, VOID *data, size_t *n);

/* Cancel I/O and flush buffers... */
BOOL AsyncFlush(H_ASYNC handle, INT32 queue);
BOOL AsyncDrain(H_ASYNC handle);

/* Test and control interface lines, -1 on error, timeout is milliseconds... */
BOOL AsyncSetLine(H_ASYNC handle, ASYNC_ACTION action);
INT32 AsyncTestLine(H_ASYNC handle, ASYNC_TEST * test, BOOL any);
INT32 AsyncWaitForLine(H_ASYNC handle, ASYNC_TEST * test, UINT32 timeout);
INT32 AsyncWaitForEvent(H_ASYNC handle, ASYNC_EVENT * event, UINT32 timeout);

/* Report errors... */
INT32 AsyncErrorCode(H_ASYNC handle);
CHAR *AsyncErrorString(CHAR *string, size_t n, INT32 error_code);

#endif
