/* $Id: linux.c,v 1.8 2002/02/19 17:34:09 cvs Exp $ */
/*-----------------------------------------------------------------------

	General, platform independent, thread-safe, async serial I/O library.
	Linux implementation...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include <async.h>

#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/ioctl.h>

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/
typedef struct TAG_PORT {
	MUTEX mutex;
	BOOL active;
	INT32 error;
	ASYNC_PARAMS params;
	int fd;
} PORT;

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;
static MUTEX mutex;
static INT32 error = AE_NONE;
static PORT ports[ASYNC_MAX_PORTS];

#define N_ERRORS 6							/* Number of error codes defined */
static const CHAR *error_strings[N_ERRORS] = {
	"No error",
	"Invalid handle",
	"Too many ports open",
	"An asynchronous read is pending",
	"System defined error",
	"Undefined or unknown error code"
};

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL IsValidHandle(H_ASYNC handle);
static UINT32 TranslateBaud(UINT32 baud);
static VOID Initialize(VOID);

/*--------------------------------------------------------------------- */
VOID AsyncParseParams(CHAR *string, ASYNC_PARAMS * params)
{
	CHAR delimiters[] = ",:;- \t\n";
	CHAR *token;
	INT32 count;

	ASSERT(params != NULL);
	if (string == NULL)
		return;

	/* Parses strings of the form: "device,baud,parity,databits,stopbits" */

	count = 0;
	token = strtok(string, delimiters);

	while (token != NULL) {
		if (count == 0)
			strcpy(params->device, token);
		else if (count == 1)
			params->baud_rate = (UINT32)atoi(token);
		else if (count == 2) {
			if (tolower(*token) == 'o')
				params->parity = ASYNC_PARITY_ODD;
			else if (tolower(*token) == 'e')
				params->parity = ASYNC_PARITY_EVEN;
			else
				params->parity = ASYNC_PARITY_NONE;
		}
		else if (count == 3)
			params->data_bits = (UINT8)atoi(token);
		else if (count == 4)
			params->stop_bits = (UINT8)atoi(token);

		count++;
		token = strtok(NULL, delimiters);
	}

	return;
}

/*---------------------------------------------------------------------*/
VOID AsyncDefaultParams(ASYNC_PARAMS * params)
{

	ASSERT(params != NULL);

	strcpy(params->device, "/dev/ttyS0");
	params->baud_rate = 9600;
	params->parity = ASYNC_PARITY_NONE;
	params->data_bits = 8;
	params->stop_bits = 1;
	params->flow_ctrl = ASYNC_FLOW_RTSCTS;
	params->read_buffer = 0;
	params->write_buffer = 0;
	params->read_timeout = 1000;
	params->write_timeout = 1000;

	return;
}

/*---------------------------------------------------------------------*/
CHAR *AsyncFormatParams(CHAR *string, size_t n, ASYNC_PARAMS * params)
{
	CHAR parity;

	ASSERT(string != NULL);
	ASSERT(params != NULL);

	if (params->parity == ASYNC_PARITY_EVEN)
		parity = 'e';
	else if (params->parity == ASYNC_PARITY_ODD)
		parity = 'o';
	else
		parity = 'n';

	snprintf(string, n, "%s,%u,%c,%u,%u", params->device, params->baud_rate, parity, params->data_bits, params->stop_bits);

	return string;
}

/*---------------------------------------------------------------------*/
H_ASYNC AsyncOpen(ASYNC_PARAMS * params)
{
	H_ASYNC i, handle;
	UINT32 baudrate;
	struct termios attributes;

	ASSERT(params != NULL);

	if (!initialized)
		Initialize();

	/* Get a handle */
	handle = ASYNC_INVALID;
	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		MUTEX_LOCK(&ports[i].mutex);
		if (!ports[i].active) {
			handle = i;
			break;
		}
		MUTEX_UNLOCK(&ports[i].mutex);
	}
	if (handle == ASYNC_INVALID) {
		MUTEX_LOCK(&mutex);
		error = AE_TOO_MANY;
		MUTEX_UNLOCK(&mutex);
		return FALSE;
	}

	/* Mutex is always locked if we get to here */

	memcpy(&ports[handle].params, params, sizeof(ASYNC_PARAMS));

	/* Open the device for read/write, we're not the controlling terminal */
	if ((ports[handle].fd = open(ports[handle].params.device, O_RDWR | O_NOCTTY | O_NDELAY)) == -1) {
		MUTEX_UNLOCK(&ports[handle].mutex);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Use POSIX terminal interface to set port attributes */

	/* Get the current attributes... */
	tcgetattr(ports[handle].fd, &attributes);

	/* Raw mode (non-canonical) */
	cfmakeraw(&attributes);

	/* Baud rate */
	baudrate = TranslateBaud(ports[handle].params.baud_rate);
	cfsetispeed(&attributes, baudrate);
	cfsetospeed(&attributes, baudrate);

	/* Enable receiver and set local mode */
	attributes.c_cflag |= (CLOCAL | CREAD);

	/* Parity */
	if (ports[handle].params.parity == ASYNC_PARITY_NONE)
		attributes.c_cflag &= ~PARENB;
	else if (ports[handle].params.parity == ASYNC_PARITY_EVEN) {
		attributes.c_cflag |= PARENB;
		attributes.c_cflag &= ~PARODD;
	}
	else if (ports[handle].params.parity == ASYNC_PARITY_ODD) {
		attributes.c_cflag |= PARENB;
		attributes.c_cflag |= PARODD;
	}

	/* Data bits */
	attributes.c_cflag &= ~CSIZE;
	if (ports[handle].params.data_bits == 7)
		attributes.c_cflag |= CS7;
	else
		attributes.c_cflag |= CS8;

	/* Stop bits */
	if (ports[handle].params.stop_bits == 1)
		attributes.c_cflag &= ~CSTOPB;
	else
		attributes.c_cflag |= CSTOPB;

	/* Hardware flow control */
	if (ports[handle].params.flow_ctrl & ASYNC_FLOW_RTSCTS)
		attributes.c_cflag |= CRTSCTS;
	else
		attributes.c_cflag &= ~CRTSCTS;

	/* Software flow control */
	if (ports[handle].params.flow_ctrl & ASYNC_FLOW_XONXOFF)
		attributes.c_iflag |= (IXON | IXOFF | IXANY);
	else
		attributes.c_iflag &= ~(IXON | IXOFF | IXANY);

	/* Non-blocking reads returning whatever is available... */
	attributes.c_cc[VMIN] = 0;
	attributes.c_cc[VTIME] = 0;

	/* Set these attributes now and flush I/O buffers */
	tcsetattr(ports[handle].fd, TCSANOW | TCSAFLUSH, &attributes);

	ports[handle].active = TRUE;
	MUTEX_UNLOCK(&ports[handle].mutex);

	return handle;
}

/*---------------------------------------------------------------------*/
BOOL AsyncClose(H_ASYNC handle)
{

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	MUTEX_LOCK(&ports[handle].mutex);

	close(ports[handle].fd);
	ports[handle].active = FALSE;

	MUTEX_UNLOCK(&ports[handle].mutex);

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncRead(H_ASYNC handle, VOID *data, size_t *n)
{
	UINT8 *ptr;
	fd_set input_set;
	struct timeval timeout;
	int maxfd, result;
	size_t remaining;
	ssize_t this_read;

	ASSERT(data != NULL);
	ASSERT(n != NULL);

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	remaining = *n;
	*n = 0;
	ptr = (UINT8 *)data;

	while (remaining > 0) {
		/* Use select to test for readability... */
		FD_ZERO(&input_set);
		FD_SET(ports[handle].fd, &input_set);
		maxfd = ports[handle].fd + 1;
		timeout.tv_sec = ports[handle].params.read_timeout / 1000;
		timeout.tv_usec = (ports[handle].params.read_timeout % 1000) * 1000;
		result = select(maxfd, &input_set, NULL, NULL, &timeout);
		if (result < 0) {
			/* Were we interrupted? */
			if (errno == EINTR) {
				/* Treat like timeout */
				return TRUE;
			}
			/* A system error has occured! */
			MUTEX_LOCK(&ports[handle].mutex);
			ports[handle].error = AE_SYSTEM;
			MUTEX_UNLOCK(&ports[handle].mutex);
			return FALSE;
		}
		else if (result == 0) {
			/* We timed-out */
			return TRUE;
		}

		/* Something's ready, read it... */
		MUTEX_LOCK(&ports[handle].mutex);
		if ((this_read = read(ports[handle].fd, ptr + *n, remaining)) < 0) {
			ports[handle].error = AE_SYSTEM;
			MUTEX_UNLOCK(&ports[handle].mutex);
			return FALSE;
		}
		MUTEX_UNLOCK(&ports[handle].mutex);
		(*n) += this_read;
		remaining -= this_read;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncWrite(H_ASYNC handle, VOID *data, size_t *n)
{
	UINT8 *ptr;
	fd_set output_set;
	struct timeval timeout;
	int maxfd, result;
	size_t written, remaining;
	ssize_t this_write;

	ASSERT(data != NULL);
	ASSERT(n != NULL);

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	remaining = *n;
	*n = written = 0;
	ptr = (UINT8 *)data;

	while (remaining > 0) {
		/* Use select to test for writability... */
		FD_ZERO(&output_set);
		FD_SET(ports[handle].fd, &output_set);
		maxfd = ports[handle].fd + 1;
		timeout.tv_sec = ports[handle].params.write_timeout / 1000;
		timeout.tv_usec = (ports[handle].params.write_timeout % 1000) * 1000;
		result = select(maxfd, NULL, &output_set, NULL, &timeout);
		if (result < 0) {
			/* Were we interrupted? */
			if (errno == EINTR) {
				/* Treat like timeout */
				return TRUE;
			}
			/* A system error has occured! */
			MUTEX_LOCK(&ports[handle].mutex);
			ports[handle].error = AE_SYSTEM;
			MUTEX_UNLOCK(&ports[handle].mutex);
			return FALSE;
		}
		else if (result == 0) {
			/* We timed-out */
			return TRUE;
		}
		MUTEX_LOCK(&ports[handle].mutex);
		if ((this_write = write(ports[handle].fd, ptr + written, remaining)) < 0) {
			ports[handle].error = AE_SYSTEM;
			MUTEX_UNLOCK(&ports[handle].mutex);
			return FALSE;
		}
		MUTEX_UNLOCK(&ports[handle].mutex);
		written += this_write;
		*n = written;
		remaining -= this_write;
	}

#if 0
	if (tcdrain(ports[handle].fd) < 0) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}
#endif

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncDrain(H_ASYNC handle)
{
	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	if (tcdrain(ports[handle].fd)) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncFlush(H_ASYNC handle, INT32 queue)
{
	int queue_selector;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	queue_selector = 0;
	if (queue == ASYNC_FLUSH_INPUT)
		queue_selector = TCIFLUSH;
	else if (queue == ASYNC_FLUSH_OUTPUT)
		queue_selector = TCOFLUSH;
	else
		queue_selector = TCIOFLUSH;

	if (tcflush(ports[handle].fd, queue_selector)) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncSetLine(H_ASYNC handle, ASYNC_ACTION action)
{
	int state;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	if (ioctl(ports[handle].fd, TIOCMGET, &state) != 0) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}

	if (ports[handle].params.flow_ctrl != ASYNC_FLOW_RTSCTS) {
		if (action == ASYNC_ASSERT_RTS)
			state = state | TIOCM_RTS;
		else if (action == ASYNC_DEASSERT_RTS)
			state = state & ~TIOCM_RTS;
	}

	if (action == ASYNC_ASSERT_DTR)
		state = state | TIOCM_DTR;
	else if (action == ASYNC_DEASSERT_DTR)
		state = state & ~TIOCM_DTR;

	if (ioctl(ports[handle].fd, TIOCMSET, &state) != 0) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
INT32 AsyncTestLine(H_ASYNC handle, ASYNC_TEST * test, BOOL any)
{
	int modem_lines;
	ASYNC_TEST result;

	ASSERT(test != NULL);

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	if (ioctl(ports[handle].fd, TIOCMGET, &modem_lines) != 0) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return FALSE;
	}

	/* Test for any specified line state */
	result = 0;
	if ((*test & ASYNC_CTS_ASSERTED) && (modem_lines & TIOCM_CTS))
		result |= ASYNC_CTS_ASSERTED;
	else if ((*test & ASYNC_CTS_DEASSERTED) && !(modem_lines & TIOCM_CTS))
		result |= ASYNC_CTS_DEASSERTED;

	if ((*test & ASYNC_DCD_ASSERTED) && (modem_lines & TIOCM_CD))
		result |= ASYNC_DCD_ASSERTED;
	else if ((*test & ASYNC_DCD_DEASSERTED) && !(modem_lines & TIOCM_CD))
		result |= ASYNC_DCD_DEASSERTED;

	if ((*test & ASYNC_DSR_ASSERTED) && (modem_lines & TIOCM_DSR))
		result |= ASYNC_DSR_ASSERTED;
	else if ((*test & ASYNC_DSR_DEASSERTED) && !(modem_lines & TIOCM_DSR))
		result |= ASYNC_DSR_DEASSERTED;

	if ((*test & ASYNC_RI_ASSERTED) && (modem_lines & TIOCM_RI))
		result |= ASYNC_RI_ASSERTED;
	else if ((*test & ASYNC_RI_DEASSERTED) && !(modem_lines & TIOCM_RI))
		result |= ASYNC_RI_DEASSERTED;

	if (any) {
		/* If any matched... */
		if (result != 0) {
			*test = result;
			return TRUE;
		}
	}
	else {
		/* If all matched... */
		if (result == *test)
			return TRUE;
	}

	*test = result;
	return FALSE;
}

/*---------------------------------------------------------------------*/
INT32 AsyncErrorCode(H_ASYNC handle)
{
	INT32 code;

	code = AE_INVALID;

	if (!initialized)
		return code;

	MUTEX_LOCK(&mutex);
	if (error != AE_NONE) {
		code = error;
		error = AE_NONE;
	}
	MUTEX_UNLOCK(&mutex);

	if (code == AE_INVALID && IsValidHandle(handle)) {
		MUTEX_LOCK(&ports[handle].mutex);
		code = ports[handle].error;
		ports[handle].error = AE_NONE;
		MUTEX_UNLOCK(&ports[handle].mutex);
	}

	return code;
}

/*---------------------------------------------------------------------*/
CHAR *AsyncErrorString(CHAR *string, size_t n, INT32 error_code)
{

	ASSERT(string != NULL);

	if (error_code == AE_SYSTEM)
		strncpy(string, strerror(errno), n);
	else if (error_code >= AE_NONE && error_code <= N_ERRORS)
		strncpy(string, error_strings[error_code], n);
	else
		strncpy(string, error_strings[N_ERRORS - 1], n);

	return string;
}

/*---------------------------------------------------------------------*/
/* Module local helpers... */
/*---------------------------------------------------------------------*/
static BOOL IsValidHandle(H_ASYNC handle)
{

	if (handle < 0 || handle > ASYNC_MAX_PORTS) {
		error = AE_INVALID;
		return FALSE;
	}

	MUTEX_LOCK(&ports[handle].mutex);

	if (ports[handle].active) {
		//ports[handle].error = AE_NONE;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return TRUE;
	}

	MUTEX_UNLOCK(&ports[handle].mutex);

	return FALSE;
}

/*---------------------------------------------------------------------*/
static UINT32 TranslateBaud(UINT32 baud)
{
	if (baud >= 115200)
		return B115200;
	else if (baud >= 57600)
		return B57600;
	else if (baud >= 38400)
		return B38400;
	else if (baud >= 19200)
		return B19200;
	else if (baud >= 9600)
		return B9600;
	else if (baud >= 4800)
		return B4800;
	else if (baud >= 2400)
		return B2400;
	else if (baud >= 1200)
		return B1200;
	else
		return B300;
}

/*---------------------------------------------------------------------*/
static VOID Initialize(VOID)
{
	H_ASYNC i;

	if (initialized)
		return;

	MUTEX_INIT(&mutex);
	MUTEX_LOCK(&mutex);

	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		MUTEX_INIT(&ports[i].mutex);
		ports[i].active = FALSE;
		ports[i].error = AE_NONE;
	}

	error = AE_NONE;
	initialized = TRUE;

	MUTEX_UNLOCK(&mutex);

	return;
}
