/* $Id: test.c,v 1.12 2002/02/19 17:34:09 cvs Exp $ */
/*-----------------------------------------------------------------------
    Copyright (C) 2000 - Robert Banfill - All rights reserved.
-------------------------------------------------------------------------

	Testing the async library... 

    Connect a loopback plug to a port and run.  The program accepts a
    standard port parameter string as its only argument.  If no 
    argument is provided, the library defaults are used:
    
      /dev/ttyS0,9600,n,8,1 or COM1:9600,n,8,1
    
    The main thread will create a read and a write thread and then 
    periodically toggle the DTR line.  It then tests the DCD line which 
    should reflect the state of DTR if the loopback plug is wired right.
    The write thread writes chunks of WRITE_SIZE sequencial bytes to the
    the port.  The read thread reads READ_SIZE chunks of bytes and 
    sequence check them.

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include <platform.h>
#include <async.h>
#include <ustime.h>

/* Module constants ---------------------------------------------------*/
#define WRITE_SIZE  20
#define READ_SIZE   25

/* Module types -------------------------------------------------------*/

/* Module globals -----------------------------------------------------*/
static BOOL stop = FALSE;
static REAL64 bits;

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static THREAD_FUNC ReadThread(VOID *argument);
static THREAD_FUNC WriteThread(VOID *argument);
static BOOL SetSignalHandler(VOID);
static VOID CatchSignal(int sig);

/*---------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	CHAR string[128];
	BOOL toggle;
	INT32 result;
	ASYNC_PARAMS params;
	ASYNC_TEST test;
	H_ASYNC handle;
	THREAD tid[2];

	if (!SetSignalHandler())
		exit(1);

	AsyncDefaultParams(&params);
	AsyncParseParams(argv[1], &params);
	params.flow_ctrl = ASYNC_FLOW_RTSCTS;

	/* Bits per byte with current line settings */
	bits = 1.0 + (REAL64)(params.data_bits + params.stop_bits);

	printf("Opening device: %s\n", AsyncFormatParams(string, 128, &params));
	if ((handle = AsyncOpen(&params)) == ASYNC_INVALID) {
		printf("ERROR: AsyncOpen: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
		exit(1);
	}

	/* Start threads... */
	if (!THREAD_CREATE(&tid[0], ReadThread, &handle)) {
		printf("THREAD_CREATE: ReadThread: %s\n", strerror(errno));
		stop = TRUE;
	}
	if (!THREAD_CREATE(&tid[1], WriteThread, &handle)) {
		printf("THREAD_CREATE: WriteThread: %s\n", strerror(errno));
		stop = TRUE;
	}

	/* Test the line actions */
	toggle = TRUE;
	while (!stop) {
		USSleep(UST_SECOND);
		continue;

		printf("AsyncSetLine: %ssserting DTR\n", (toggle ? "A" : "Dea"));
		if (!AsyncSetLine(handle, (toggle ? ASYNC_ASSERT_DTR : ASYNC_DEASSERT_DTR)))
			printf("ERROR: AsyncSetLine(): %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));

		test = ASYNC_DCD_ASSERTED | ASYNC_DSR_ASSERTED | ASYNC_RI_ASSERTED;
		if ((result = AsyncTestLine(handle, &test, TRUE)) < 0)
			printf("ERROR: AsyncTestLine: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
		printf("AsyncTestLine returns %s, DCD: %ssserted, DSR: %ssserted, RI: %ssserted\n",
			(result ? "true" : "false"),
			(test & ASYNC_DCD_ASSERTED ? "a" : "dea"),
			(test & ASYNC_DSR_ASSERTED ? "a" : "dea"), (test & ASYNC_RI_ASSERTED ? "a" : "dea"));

		USSleep(UST_MSECOND * 2500);
		toggle = !toggle;
	}

	/* Wait for threads to exit */
	THREAD_JOIN(tid[1]);
	THREAD_JOIN(tid[0]);

	if (!AsyncFlush(handle, ASYNC_FLUSH_BOTH)) {
		printf("ERROR: AsyncFlush: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
		exit(1);
	}

	if (!AsyncClose(handle)) {
		printf("ERROR: AsyncClose: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
		exit(1);
	}

	return 0;
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC ReadThread(VOID *argument)
{
	CHAR string[128];
	UINT8 byte, buffer[READ_SIZE];
	size_t i, n;
	UINT32 count;
	INT64 elapsed;
	H_ASYNC handle;
	USTIMER timer;

	printf("ReadThread started...\n");

	handle = *((H_ASYNC *) argument);

	StartUSTimer(&timer, 0);

	count = 0;
	byte = 0;
	while (!stop) {
		n = READ_SIZE;
		if (!AsyncRead(handle, &buffer, &n)) {
			printf("ERROR: AsyncRead: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
			stop = TRUE;
		}

		for (i = 0; i < n; i++) {
			if (buffer[i] != byte)
				printf("Got %u, expected %u\n", buffer[i], byte);
			byte = buffer[i] + 1;
		}
		count += n;

		/* Exercise flow control */
#if 0
		if (n > 0)
			USSleep(UST_MSECOND * 10);
#endif
	}

	elapsed = USTimerElapsed(&timer);

	printf("Read %u bytes at %.1lf bps\n", count, ((REAL64)count * bits) / ((REAL64)elapsed / (REAL64)UST_SECOND));

	THREAD_EXIT((VOID *)0);
}

/*---------------------------------------------------------------------*/
static THREAD_FUNC WriteThread(VOID *argument)
{
	CHAR string[128];
	UINT8 byte, buffer[WRITE_SIZE];
	size_t i, n;
	UINT32 count;
	INT64 elapsed;
	H_ASYNC handle;
	USTIMER timer;

	printf("WriteThread started...\n");

	handle = *((H_ASYNC *) argument);

	StartUSTimer(&timer, 0);

	count = 0;
	byte = 0;
	while (!stop) {
		for (i = 0; i < WRITE_SIZE; i++)
			buffer[i] = byte++;
		n = WRITE_SIZE;
		if (!AsyncWrite(handle, buffer, &n)) {
			printf("ERROR: AsyncWrite: %s\n", AsyncErrorString(string, 128, AsyncErrorCode(handle)));
			stop = TRUE;
		}
		count += n;
	}

	elapsed = USTimerElapsed(&timer);

	printf("Wrote %u bytes at %.1lf bps\n", count, ((REAL64)count * bits) / ((REAL64)elapsed / (REAL64)UST_SECOND));

	THREAD_EXIT((VOID *)0);
}

/*--------------------------------------------------------------------- */
static VOID CatchSignal(int sig)
{
	switch (sig) {
	case SIGTERM:
		fprintf(stderr, "Caught SIGTERM! \n");
		stop = TRUE;
		break;
	case SIGINT:
		fprintf(stderr, "Caught SIGINT! \n");
		stop = TRUE;
		break;
	default:
		fprintf(stderr, "Caught unexpected signal %d, ignored! \n", sig);
		break;
	}

	signal(sig, CatchSignal);
}

/*--------------------------------------------------------------------- */
static BOOL SetSignalHandler(VOID)
{
	if (signal(SIGTERM, CatchSignal) == SIG_ERR) {
		perror("signal(SIGTERM)");
		return FALSE;
	}
	if (signal(SIGINT, CatchSignal) == SIG_ERR) {
		perror("signal(SIGINT)");
		return FALSE;
	}

	return TRUE;
}
