/* $Id: win32.c,v 1.6 2002/02/19 17:34:09 cvs Exp $ */
/*-----------------------------------------------------------------------

	General, platform independent, thread-safe, async serial I/O library.
	Win32 implementation...

	Copyright (C) 2000-2002 - Robert Banfill - All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (src/license.txt) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

-----------------------------------------------------------------------*/

#include <async.h>

/* Module constants ---------------------------------------------------*/

/* Module types -------------------------------------------------------*/
typedef struct TAG_PORT {
	MUTEX mutex;
	BOOL active;
	INT32 error;
	ASYNC_PARAMS params;
	HANDLE handle;
	OVERLAPPED ol_write;
	OVERLAPPED ol_read;
} PORT;

/* Module globals -----------------------------------------------------*/
static BOOL initialized = FALSE;
static MUTEX mutex;
static INT32 error = AE_NONE;
static PORT ports[ASYNC_MAX_PORTS];

#define N_ERRORS 6							/* Number of error codes defined */
static const CHAR *error_strings[N_ERRORS] = {
	"No error",
	"Invalid handle",
	"Too many ports open",
	"An asynchronous read is pending",
	"System defined error",
	"Undefined or unknown error code"
};

/* Module macros ------------------------------------------------------*/

/* Module prototypes --------------------------------------------------*/
static BOOL IsValidHandle(H_ASYNC handle);
static VOID Initialize(VOID);

/*--------------------------------------------------------------------- */
VOID AsyncParseParams(CHAR *string, ASYNC_PARAMS * params)
{
	CHAR delimiters[] = ",:;- \t\n";
	CHAR *token;
	INT32 count;

	ASSERT(params != NULL);
	if (string == NULL)
		return;

	/* Parses strings of the form: "device,baud,parity,databits,stopbits" */

	count = 0;
	token = strtok(string, delimiters);

	while (token != NULL) {
		if (count == 0)
			strcpy(params->device, token);
		else if (count == 1)
			params->baud_rate = (UINT32)atoi(token);
		else if (count == 2) {
			if (tolower(*token) == 'o')
				params->parity = ASYNC_PARITY_ODD;
			else if (tolower(*token) == 'e')
				params->parity = ASYNC_PARITY_EVEN;
			else
				params->parity = ASYNC_PARITY_NONE;
		}
		else if (count == 3)
			params->data_bits = (UINT8)atoi(token);
		else if (count == 4)
			params->stop_bits = (UINT8)atoi(token);

		count++;
		token = strtok(NULL, delimiters);
	}

	return;
}

/*---------------------------------------------------------------------*/
VOID AsyncDefaultParams(ASYNC_PARAMS * params)
{

	ASSERT(params != NULL);

	strcpy(params->device, "COM1");
	params->baud_rate = 9600;
	params->parity = ASYNC_PARITY_NONE;
	params->data_bits = 8;
	params->stop_bits = 1;
	params->flow_ctrl = ASYNC_FLOW_RTSCTS;
	params->read_buffer = 4096;
	params->write_buffer = 4096;
	params->read_timeout = 1000;
	params->write_timeout = 1000;

	return;
}

/*---------------------------------------------------------------------*/
CHAR *AsyncFormatParams(CHAR *string, size_t n, ASYNC_PARAMS * params)
{
	CHAR parity;

	ASSERT(string != NULL);
	ASSERT(params != NULL);

	if (params->parity == ASYNC_PARITY_EVEN)
		parity = 'e';
	else if (params->parity == ASYNC_PARITY_ODD)
		parity = 'o';
	else
		parity = 'n';

	snprintf(string, n, "%s:%u,%c,%u,%u", params->device, params->baud_rate, parity, params->data_bits, params->stop_bits);

	return string;
}

/*---------------------------------------------------------------------*/
H_ASYNC AsyncOpen(ASYNC_PARAMS * params)
{
	H_ASYNC i, handle;
	PORT *port;
	DCB dcb;
	COMMTIMEOUTS cto;

	ASSERT(params != NULL);

	if (!initialized)
		Initialize();

	/* Get a handle */
	handle = ASYNC_INVALID;
	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		port = &ports[i];
		MUTEX_LOCK(&port->mutex);
		if (!port->active) {
			handle = i;
			break;
		}
		MUTEX_UNLOCK(&port->mutex);
	}
	if (handle == ASYNC_INVALID) {
		MUTEX_LOCK(&mutex);
		error = AE_TOO_MANY;
		MUTEX_UNLOCK(&mutex);
		return FALSE;
	}

	/* Mutex is always locked if we get to here */

	memcpy(&port->params, params, sizeof(ASYNC_PARAMS));

	/* Open the device */
	if ((port->handle = CreateFile(port->params.device, GENERIC_READ | GENERIC_WRITE, 0, NULL,
				OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL)) == INVALID_HANDLE_VALUE) {
		MUTEX_UNLOCK(&port->mutex);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Setup buffers */
	if (!SetupComm(port->handle, port->params.read_buffer, port->params.write_buffer)) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Get the current state  of the port */
	dcb.DCBlength = sizeof(DCB);
	if (!GetCommState(port->handle, &dcb)) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Line settings */
	dcb.BaudRate = port->params.baud_rate;
	dcb.ByteSize = port->params.data_bits;
	if (port->params.parity == ASYNC_PARITY_ODD)
		dcb.Parity = 1;
	else if (port->params.parity == ASYNC_PARITY_EVEN)
		dcb.Parity = 2;
	else
		dcb.Parity = 0;
	if (port->params.stop_bits == 2)
		dcb.StopBits = 2;
	else
		dcb.StopBits = 0;

	dcb.fNull = FALSE;
	dcb.fBinary = TRUE;
	dcb.fAbortOnError = FALSE;
	if (port->params.parity == ASYNC_PARITY_NONE)
		dcb.fParity = FALSE;
	else
		dcb.fParity = FALSE;

	/* Turn off all flow-control */
	dcb.fInX = FALSE;
	dcb.fOutX = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;

	/* XON/XOFF */
	if (port->params.flow_ctrl & ASYNC_FLOW_XONXOFF) {
		dcb.fOutX = TRUE;
		dcb.fInX = TRUE;
		dcb.XonChar = XON;
		dcb.XoffChar = XOFF;
	}
	/* RTS/CTS */
	if (port->params.flow_ctrl & ASYNC_FLOW_RTSCTS) {
		dcb.fOutxCtsFlow = TRUE;
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
	}

	/* Off threshold at 3/4 buffer */
	dcb.XoffLim = (port->params.read_buffer / 4) * 3;
	/* On threshold at 1/4 buffer */
	dcb.XonLim = port->params.read_buffer / 4;

	if (!SetCommState(port->handle, &dcb)) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Set timeouts... */
	cto.ReadIntervalTimeout = MAXDWORD;
	cto.ReadTotalTimeoutMultiplier = 0;
	cto.ReadTotalTimeoutConstant = port->params.read_timeout;
	cto.WriteTotalTimeoutMultiplier = 0;
	cto.WriteTotalTimeoutConstant = port->params.write_timeout;

	if (!SetCommTimeouts(port->handle, &cto)) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	/* Setup overlapped read and write events */
	memset(&port->ol_write, 0, sizeof(OVERLAPPED));
	if ((port->ol_write.hEvent = CreateEvent(NULL, TRUE, TRUE, NULL)) == NULL) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}
	memset(&port->ol_read, 0, sizeof(OVERLAPPED));
	port->ol_read.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (port->ol_read.hEvent == NULL) {
		MUTEX_UNLOCK(&port->mutex);
		AsyncClose(handle);
		MUTEX_LOCK(&mutex);
		error = AE_SYSTEM;
		MUTEX_UNLOCK(&mutex);
		return ASYNC_INVALID;
	}

	port->active = TRUE;
	MUTEX_UNLOCK(&port->mutex);

	return handle;
}

/*---------------------------------------------------------------------*/
BOOL AsyncClose(H_ASYNC handle)
{
	BOOL result;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	result = TRUE;

	MUTEX_LOCK(&ports[handle].mutex);

	if (!CloseHandle(ports[handle].handle))
		result = FALSE;
	if (!CloseHandle(ports[handle].ol_read.hEvent))
		result = FALSE;
	if (!CloseHandle(ports[handle].ol_write.hEvent))
		result = FALSE;
	ports[handle].active = FALSE;

	MUTEX_UNLOCK(&ports[handle].mutex);

	return result;
}

/*---------------------------------------------------------------------*/
BOOL AsyncRead(H_ASYNC handle, VOID *data, size_t *n)
{
	UINT32 last_error;

	ASSERT(data != NULL);
	ASSERT(n != NULL);

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	/* There should never be a read pending! */
	if (HasOverlappedIoCompleted(&ports[handle].ol_read) != TRUE) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_PENDING;
		MUTEX_LOCK(&ports[handle].mutex);
		return FALSE;
	}

	/* Read... */
	if (!ReadFile(ports[handle].handle, data, *n, n, &ports[handle].ol_read)) {
		last_error = GetLastError();
		/* Is our read pending? */
		if (last_error == ERROR_IO_PENDING) {
			/* Block until it's finished... */
			if (!GetOverlappedResult(ports[handle].handle, &ports[handle].ol_read, n, TRUE)) {
				MUTEX_LOCK(&ports[handle].mutex);
				ports[handle].error = AE_SYSTEM;
				MUTEX_LOCK(&ports[handle].mutex);
				return FALSE;
			}
		}
		else {
			SetLastError(last_error);
			MUTEX_LOCK(&ports[handle].mutex);
			ports[handle].error = AE_SYSTEM;
			MUTEX_LOCK(&ports[handle].mutex);
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncWrite(H_ASYNC handle, VOID *data, size_t *n)
{
	UINT32 last_error;
	size_t pending, written;

	ASSERT(data != NULL);
	ASSERT(n != NULL);

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	/* Block until any pending write in finished... */
	pending = 0;
	if (!GetOverlappedResult(ports[handle].handle, &ports[handle].ol_write, &pending, TRUE)) {
		MUTEX_LOCK(&ports[handle].mutex);
		ports[handle].error = AE_SYSTEM;
		MUTEX_LOCK(&ports[handle].mutex);
		return FALSE;
	}

	/* Write... */
	written = 0;
	if (!WriteFile(ports[handle].handle, data, *n, &written, &ports[handle].ol_write)) {
		last_error = GetLastError();
		if (last_error != ERROR_IO_PENDING) {
			SetLastError(last_error);
			MUTEX_LOCK(&ports[handle].mutex);
			ports[handle].error = AE_SYSTEM;
			MUTEX_LOCK(&ports[handle].mutex);
			return FALSE;
		}
	}

	*n = pending + written;

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncDrain(H_ASYNC handle)
{

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	if (!FlushFileBuffers(ports[handle].handle)) {
		ports[handle].error = AE_SYSTEM;
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncFlush(H_ASYNC handle, INT32 queue)
{
	UINT32 action;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	action = 0;
	if (queue & ASYNC_FLUSH_INPUT)
		action |= PURGE_RXABORT | PURGE_RXCLEAR;
	if (queue & ASYNC_FLUSH_OUTPUT)
		action |= PURGE_TXABORT | PURGE_TXCLEAR;

	if (!PurgeComm(ports[handle].handle, action)) {
		ports[handle].error = AE_SYSTEM;
		return FALSE;
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
BOOL AsyncSetLine(H_ASYNC handle, ASYNC_ACTION action)
{
	UINT32 func;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	if (ports[handle].params.flow_ctrl != ASYNC_FLOW_RTSCTS) {
		func = 0;
		if (action & ASYNC_ASSERT_RTS)
			func = SETRTS;
		else if (action & ASYNC_DEASSERT_RTS)
			func = CLRRTS;

		if (func) {
			if (!EscapeCommFunction(ports[handle].handle, func)) {
				ports[handle].error = AE_SYSTEM;
				return FALSE;
			}
		}
	}

	func = 0;
	if (action & ASYNC_ASSERT_DTR)
		func = SETDTR;
	else if (action & ASYNC_DEASSERT_DTR)
		func = CLRDTR;

	if (func) {
		if (!EscapeCommFunction(ports[handle].handle, func)) {
			ports[handle].error = AE_SYSTEM;
			return FALSE;
		}
	}

	return TRUE;
}

/*---------------------------------------------------------------------*/
INT32 AsyncTestLine(H_ASYNC handle, ASYNC_TEST * test, BOOL any)
{
	UINT32 modem_lines;
	ASYNC_TEST result;

	ASSERT(test != NULL);

	if (!initialized || !IsValidHandle(handle))
		return -1;

	if (!GetCommModemStatus(ports[handle].handle, &modem_lines)) {
		ports[handle].error = AE_SYSTEM;
		return -1;
	}

	/* Test for any specified line state */
	result = 0;
	if ((*test & ASYNC_CTS_ASSERTED) && (modem_lines & MS_CTS_ON))
		result |= ASYNC_CTS_ASSERTED;
	else if ((*test & ASYNC_CTS_DEASSERTED) && !(modem_lines & MS_CTS_ON))
		result |= ASYNC_CTS_DEASSERTED;

	if ((*test & ASYNC_DCD_ASSERTED) && (modem_lines & MS_RLSD_ON))
		result |= ASYNC_DCD_ASSERTED;
	else if ((*test & ASYNC_DCD_DEASSERTED) && !(modem_lines & MS_RLSD_ON))
		result |= ASYNC_DCD_DEASSERTED;

	if ((*test & ASYNC_DSR_ASSERTED) && (modem_lines & MS_DSR_ON))
		result |= ASYNC_DSR_ASSERTED;
	else if ((*test & ASYNC_DSR_DEASSERTED) && !(modem_lines & MS_DSR_ON))
		result |= ASYNC_DSR_DEASSERTED;

	if ((*test & ASYNC_RI_ASSERTED) && (modem_lines & MS_RING_ON))
		result |= ASYNC_RI_ASSERTED;
	else if ((*test & ASYNC_RI_DEASSERTED) && !(modem_lines & MS_RING_ON))
		result |= ASYNC_RI_DEASSERTED;

	if (any) {
		/* If any matched... */
		if (result != 0) {
			*test = result;
			return TRUE;
		}
	}
	else {
		/* If all matched... */
		if (result == *test)
			return TRUE;
	}

	*test = result;
	return FALSE;
}

/*---------------------------------------------------------------------*/
INT32 AsyncWaitForLine(H_ASYNC handle, ASYNC_TEST * test, UINT32 timeout)
{
	INT32 result;
	ASYNC_TEST saved;
	ASYNC_EVENT event;

	if (!initialized || !IsValidHandle(handle))
		return FALSE;

	saved = *test;
	if ((result = AsyncTestLine(handle, test, TRUE)) < 0)
		return -1;
	else if (result > 0)
		return TRUE;

	event = 0;
	if (saved & ASYNC_CTS_ASSERTED || saved & ASYNC_CTS_DEASSERTED)
		event |= ASYNC_EVENT_CTS;
	if (saved & ASYNC_DCD_ASSERTED || saved & ASYNC_DCD_DEASSERTED)
		event |= ASYNC_EVENT_DCD;
	if (saved & ASYNC_DSR_ASSERTED || saved & ASYNC_DSR_DEASSERTED)
		event |= ASYNC_EVENT_DSR;
	if (saved & ASYNC_RI_ASSERTED || saved & ASYNC_RI_DEASSERTED)
		event |= ASYNC_EVENT_RI;

	if ((result = AsyncWaitForEvent(handle, &event, timeout)) < 0)
		return -1;
	else if (result == 0)
		return FALSE;

	*test = saved;
	if ((result = AsyncTestLine(handle, test, TRUE)) < 0)
		return -1;
	else if (result > 0)
		return TRUE;

	return FALSE;
}

/*---------------------------------------------------------------------*/
INT32 AsyncWaitForEvent(H_ASYNC handle, ASYNC_EVENT * event, UINT32 timeout)
{
	BOOL timed_out;
	UINT32 event_mask, last_error, result, dummy;
	OVERLAPPED overlapped;

	if (!initialized || !IsValidHandle(handle))
		return -1;

	memset(&overlapped, 0, sizeof(OVERLAPPED));
	overlapped.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (overlapped.hEvent == NULL) {
		ports[handle].error = AE_SYSTEM;
		return -1;
	}

	event_mask = 0;
	if (*event & ASYNC_EVENT_DCD)
		event_mask |= EV_RLSD;
	if (*event & ASYNC_EVENT_CTS)
		event_mask |= EV_CTS;
	if (*event & ASYNC_EVENT_DSR)
		event_mask |= EV_DSR;
	if (*event & ASYNC_EVENT_RI)
		event_mask |= EV_RING;
	if (*event & ASYNC_EVENT_RX)
		event_mask |= EV_RXCHAR;
	if (*event & ASYNC_EVENT_TX_EMPTY)
		event_mask |= EV_TXEMPTY;
	if (!SetCommMask(ports[handle].handle, event_mask)) {
		ports[handle].error = AE_SYSTEM;
		return -1;
	}

	timed_out = FALSE;
	if (!WaitCommEvent(ports[handle].handle, &event_mask, &overlapped)) {
		last_error = GetLastError();
		if (last_error != ERROR_IO_PENDING)
			goto error;

		last_error = 0;
		result = WaitForSingleObject(overlapped.hEvent, timeout);
		if (result == WAIT_FAILED)
			goto error;
		else if (result != WAIT_OBJECT_0)
			timed_out = TRUE;
	}
	if (!timed_out && last_error == ERROR_IO_PENDING) {
		if (!GetOverlappedResult(ports[handle].handle, &overlapped, &dummy, FALSE))
			goto error;
	}

	if (!CloseHandle(overlapped.hEvent))
		return -1;

	*event = 0;

	if (timed_out)
		return FALSE;

	if (event_mask & EV_RLSD)
		*event |= ASYNC_EVENT_DCD;
	if (event_mask & EV_CTS)
		*event |= ASYNC_EVENT_CTS;
	if (event_mask & EV_DSR)
		*event |= ASYNC_EVENT_DSR;
	if (event_mask & EV_RING)
		*event |= ASYNC_EVENT_RI;
	if (event_mask & EV_RXCHAR)
		*event |= ASYNC_EVENT_RX;
	if (event_mask & EV_TXEMPTY)
		*event |= ASYNC_EVENT_TX_EMPTY;

	return TRUE;

  error:
	CloseHandle(overlapped.hEvent);
	if (last_error != 0)
		SetLastError(last_error);

	return -1;
}

/*---------------------------------------------------------------------*/
INT32 AsyncErrorCode(H_ASYNC handle)
{
	INT32 code;

	code = AE_INVALID;

	if (!initialized)
		return code;

	MUTEX_LOCK(&mutex);
	if (error != AE_NONE) {
		code = error;
		error = AE_NONE;
	}
	MUTEX_UNLOCK(&mutex);

	if (code == AE_INVALID && IsValidHandle(handle)) {
		MUTEX_LOCK(&ports[handle].mutex);
		code = ports[handle].error;
		ports[handle].error = AE_NONE;
		MUTEX_UNLOCK(&ports[handle].mutex);
	}

	return code;
}

/*---------------------------------------------------------------------*/
CHAR *AsyncErrorString(CHAR *string, size_t n, INT32 error_code)
{
	VOID *sys_err_str;

	if (error_code == AE_SYSTEM) {
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) & sys_err_str, 0, NULL);
		strncpy(string, sys_err_str, n);
		LocalFree(sys_err_str);
	}
	else if (error_code >= AE_NONE && error_code <= N_ERRORS)
		strncpy(string, error_strings[error_code], n);
	else
		strncpy(string, error_strings[N_ERRORS - 1], n);

	return string;
}

/*---------------------------------------------------------------------*/
/* Module local helpers... */
/*---------------------------------------------------------------------*/
static BOOL IsValidHandle(H_ASYNC handle)
{

	if (handle < 0 || handle > ASYNC_MAX_PORTS) {
		error = AE_INVALID;
		return FALSE;
	}

	MUTEX_LOCK(&ports[handle].mutex);

	if (ports[handle].active) {
//      ports[handle].error = AE_NONE;
		MUTEX_UNLOCK(&ports[handle].mutex);
		return TRUE;
	}

	MUTEX_UNLOCK(&ports[handle].mutex);

	return FALSE;
}

/*---------------------------------------------------------------------*/
static VOID Initialize(VOID)
{
	H_ASYNC i;

	if (initialized)
		return;

	MUTEX_INIT(&mutex);
	MUTEX_LOCK(&mutex);

	for (i = 0; i < ASYNC_MAX_PORTS; i++) {
		MUTEX_INIT(&ports[i].mutex);
		ports[i].active = FALSE;
		ports[i].error = AE_NONE;
	}

	error = AE_NONE;
	initialized = TRUE;

	MUTEX_UNLOCK(&mutex);

	return;
}
